/* sysALib.s - Intel Agilex system-dependent routines */

/*
 * Copyright (c) 2022 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION
This module contains system-dependent routines written in assembly
language. It contains the entry code, sysInit(), for VxWorks images
that start running from RAM, such as 'vxWorks'.  These images are
loaded into memory by some external program (e.g., a boot ROM) and then
started.  The routine sysInit() must come first in the text segment.
Its job is to perform the minimal setup needed to call the generic C
routine usrInit().

sysInit() masks interrupts in the processor and the interrupt
controller and sets the initial stack pointer.  Other hardware and
device initialisation is performed later in the sysHwInit routine in
sysLib.c.

*/

#define _ASMLANGUAGE
#include <vxWorks.h>
#include <vsbConfig.h>
#include <asm.h>
#include <regs.h>
#include <arch/arm/intArmLib.h>
#include "prjParams.h"

/* defines */

#define CNTHCTL_EL2_EL1PCTEN            (0x1)
#define CNTHCTL_EL2_EL1PCEN             (0x1 << 1)

#define HCR_EL2_RW                      (0x1 << 31)
#define HCR_EL2_HCD                     (0x1 << 29)

#define SCR_EL3_NS                      (0x1)
#define SCR_EL3_RES1                    (0x3 << 4)
#define SCR_EL3_SMD                     (0x1 << 7)
#define SCR_EL3_HCE                     (0x1 << 8)
#define SCR_EL3_RW                      (0x1 << 10)

#define CPTR_EL2_RES1                   (0x33ff)
#define SCTLR_EL2_RES1                  (0x30C50830)

#define SPSR_EL3_M_EL2h                 (0x9)
#define SPSR_EL3_F                      (0x1 << 6)
#define SPSR_EL3_I                      (0x1 << 7)
#define SPSR_EL3_A                      (0x1 << 8)
#define SPSR_EL3_E                      (0x1 << 9)

#define SCTLR_EL1_RES1                  (0x30D00800)

#define SPSR_EL2_M_EL1h                 (0x5)
#define SPSR_EL2_F                      (0x1 << 6)
#define SPSR_EL2_I                      (0x1 << 7)
#define SPSR_EL2_A                      (0x1 << 8)
#define SPSR_EL2_E                      (0x1 << 9)

#define CPACR_EL1_FPEN_NO_INST_TRAPPED  (0x3 << 20)

#define SCR_EL3_DEFAULT         (SCR_EL3_NS  | SCR_EL3_RES1 | SCR_EL3_SMD | \
                                 SCR_EL3_HCE | SCR_EL3_RW)

#define SPSR_EL3_DEFAULT        (SPSR_EL3_E | SPSR_EL3_A | SPSR_EL3_I | \
                                 SPSR_EL3_F | SPSR_EL3_M_EL2h)

#define SPSR_EL2_DEFAULT        (SPSR_EL2_E | SPSR_EL2_A | SPSR_EL2_I | \
                                 SPSR_EL2_F | SPSR_EL2_M_EL1h)

        /* exports */

        FUNC_EXPORT (sysInit)
#ifdef _WRS_CONFIG_LP64
        FUNC_EXPORT (armWarmReboot)
#endif /* _WRS_CONFIG_LP64 */

        /* externals */

        FUNC_IMPORT (usrInit)

#ifdef INCLUDE_STANDALONE_DTB

        .data

#define TO_STRING(exp)                  #exp
#define DTS_ASM_FILE(file)              TO_STRING (file)

#include DTS_ASM_FILE (DTB_ASM)

#endif /* INCLUDE_STANDALONE_DTB */

#if defined (__llvm__)
        .section .text.entry, "ax"
#elif defined (__GNUC__)
        .section .text.entry, "ax"
#else
#error "TOOL not supported!"
#endif

#ifndef _WRS_CONFIG_LP64
        .code 32
#endif /* _WRS_CONFIG_LP64 */

/*******************************************************************************
*
* sysInit - start after boot
*
* This routine is the system start-up entry point for VxWorks in RAM, the
* first code executed after booting.  It disables interrupts, sets up
* the stack, and jumps to the C routine usrInit() in usrConfig.c.
*
* The initial stack is set to grow down from the address of sysInit().  This
* stack is used only by usrInit() and is never used again.  Memory for the
* stack must be accounted for when determining the system load address.
*
* NOTE: This routine should not be called by the user.
*
* SYNOPSIS
* \ss
* VOID sysInit
*     (
*     void * pDtb   /@ physical pointer for device-tree in RAM  @/
*     )
* \se
*
* NOTE:
*   1. This routine should not be called by the user;
*   2. <pOf> is unused because bootng from Open Firmware is unsupported now;
*   3. The initial stack is placed in .data section, and mapped with the entire
*   image vxWorks.
*
* RETURNS: N/A
*
* sysInit ()              /@ THIS IS NOT A CALLABLE ROUTINE @/
*
*/

FUNC_LABEL (sysInit)
#ifdef _WRS_CONFIG_LP64

        /* save X0 */

        MOV     X20, X0

        /* read current exception level */

        MRS	X1, CurrentEL
        CMP     X1, 0xC
        B.EQ    el3_mode
        CMP     X1, 0x8
        B.EQ    el2_mode
        CMP     X1, 0x4
        B.EQ    el1_mode

el3_mode:

        /* EL3 */

        MOV     X1, #(SCR_EL3_DEFAULT)  /* SCR */
        MSR     SCR_EL3, X1
        MSR     CPTR_EL3, XZR           /* CPTR_EL3 */

        MOV     X1, #(SPSR_EL3_DEFAULT)
        MSR     SPSR_EL3, X1            /* SPSR_EL3 */

        MRS     X1, VBAR_EL3
        MSR     VBAR_EL2, X1            /* VBAR_EL2 */

        ADR     X0, el2_mode
        MOV     LR, X0
        MSR     ELR_EL3, LR
        ERET

el2_mode:

        /* EL2 */

        /* initilize MPID/MPIDR registers */

        MRS     X1, MIDR_EL1
        MSR     VPIDR_EL2, X1           /* VPIDR_EL2 */

        MRS     X1, MPIDR_EL1
        MSR     VMPIDR_EL2, X1          /* VMPIDR_EL2 */

        /* initialize Generic Timers */

        MRS     X1, CNTHCTL_EL2
        ORR     X1, X1, #(CNTHCTL_EL2_EL1PCTEN | CNTHCTL_EL2_EL1PCEN)
        MSR     CNTHCTL_EL2, X1         /* CNTHCTL_EL2 */
        MSR     CNTVOFF_EL2, XZR        /* CNTVOFF_EL2 */

        MOV     X1, #(SPSR_EL2_DEFAULT)
        MSR     SPSR_EL2, X1            /* SPSR_EL2 */

        MOV     X1, #(CPTR_EL2_RES1)
        MSR     CPTR_EL2, X1            /* CPTR_EL2 */

        LDR     X1, =SCTLR_EL2_RES1
        MSR     SCTLR_EL2, X1           /* SCTLR_EL2 */

        MRS     X0, SCTLR_EL1
        MOV     X1, #SCTLR_C | SCTLR_I
        AND     X0, X0, X1

        LDR     X1, =SCTLR_EL1_RES1
        ORR     X1, X1, X0
        MSR     SCTLR_EL1, X1           /* SCTLR_EL1 */

        MSR     HSTR_EL2, XZR           /* HSTR_EL2 */
        MOV     X1, #(CPACR_EL1_FPEN_NO_INST_TRAPPED)
        MSR     CPACR_EL1, X1           /* CPACR_EL1 */

        LDR     X1, =(HCR_EL2_RW | HCR_EL2_HCD)
        MSR     HCR_EL2, X1             /* HCR_EL2 */

        MRS     X1, VBAR_EL2
        MSR     VBAR_EL1, X1            /* VBAR_EL1 */

        ADR     X0, el1_mode
        MOV     LR, X0
        MSR     ELR_EL2, LR
        ERET                            /* to EL1 */

#define ICC_SRE_EL1_SRE                 (0x1)

el1_mode:

        /* EL1 */

        /* enable ICC_SRE_EL1_SRE */

        MRS     X1, ICC_SRE_EL1
        MOV     X2, #ICC_SRE_EL1_SRE
        ORR     X1, X1, X2
        MSR     ICC_SRE_EL1, X1
        ISB     SY

        /* set initial stack pointer so stack grows down from start of code */

        ADR     X1, sysInit
        MOV     SP, X1
        MOV     X29, #0
        LDR     X1, =sysInit
        ADR     X2, sysInit
        SUB     X2, X2, X1

#ifdef _WRS_CONFIG_SMP
        MRS     X1, MPIDR_EL1
        ANDS    X1, X1, #0xf
        BNE     2f
#endif

#ifdef INCLUDE_STANDALONE_DTB
        LDR     X0, =dt_blob_start
#else
        MOV     X0, X20
#endif /* INCLUDE_STANDALONE_DTB */

        LDR     X1, =gpDtbInit
        ADD     X1, X1, X2
        STR     X0, [X1]        /* set start of device tree blob */
2:
        MOV     X0, #2      /* BOOT_COLD */

armWarmReboot:

        CBNZ    X0, 3f      /* X2 includes invalid value if warm reboot */
        LDR     X1, =RAM_LOW_ADRS
        ADR     X2, sysInit
        SUB     X2, X2, X1

3:
        MOV     X19, X0
        LDR     X1, =vxCpuInit
        ADD     X1, X1, X2
        BLR     X1

        LDR     X1, =sysInit
        ADR     X2, sysInit
        SUB     X2, X2, X1

        LDR     X4, =vxMmuEarlyInit
        ADD     X4, X4, X2
        LDR     x0, =STATIC_MMU_TABLE_BASE
        LDR     x1, =LOCAL_MEM_LOCAL_ADRS
        LDR     x2, =IMA_SIZE
        BLR     X4

        /* now call usrInit (startType) */

        MOV     X0, X19
        LDR     X1, =usrInit
        BR      X1

#else /* !_WRS_CONFIG_LP64 */

#ifdef INCLUDE_LOAD_VXWORKS_FROM_32BIT_BOOTAPP
        B       el1_mode
#else
       .long 0xd503201f        /* NOP */
#endif /* INCLUDE_LOAD_VXWORKS_FROM_32BIT_BOOTAPP */

        /*
         * This binary code lower its exception level to EL1,
         * switch mode from AArch64 to AArch32 and call 32bit
         * routine el1_mode.
         *
         * Refer to sysALib64to32.s for more information about
         * this code and how to build it.
         */

        .long 0xaa0003e5        /* mov     x5, x0                          */
        .long 0xd5384241        /* mrs     x1, currentel                   */
        .long 0xf100303f        /* cmp     x1, #0xc                        */
        .long 0x540000a0        /* b.eq    ffffffff80100020 <el3_mode>     */
        .long 0xf100203f        /* cmp     x1, #0x8                        */
        .long 0x540001c0        /* b.eq    ffffffff8010004c <el2_mode>     */
        .long 0xf100103f        /* cmp     x1, #0x4                        */
        .long 0x540005a0        /* b.eq    ffffffff801000d0 <el1_mode>     */
el3_mode:
        .long 0xd280b621        /* mov     x1, #0x5b1                      */
        .long 0xd51e1101        /* msr     scr_el3, x1                     */
        .long 0xd51e115f        /* msr     cptr_el3, xzr                   */
        .long 0xd2807921        /* mov     x1, #0x3c9                      */
        .long 0xd51e4001        /* msr     spsr_el3, x1                    */
        .long 0xd53ec001        /* mrs     x1, vbar_el3                    */
        .long 0xd51cc001        /* msr     vbar_el2, x1                    */
        .long 0x10000080        /* adr     x0, ffffffff8010004c <el2_mode> */
        .long 0xaa0003fe        /* mov     x30, x0                         */
        .long 0xd51e403e        /* msr     elr_el3, x30                    */
        .long 0xd69f03e0        /* eret                                    */
el2_mode:
        .long 0xd5380001        /* mrs     x1, midr_el1                    */
        .long 0xd51c0001        /* msr     vpidr_el2, x1                   */
        .long 0xd53800a1        /* mrs     x1, mpidr_el1                   */
        .long 0xd51c00a1        /* msr     vmpidr_el2, x1                  */
        .long 0xd53ce101        /* mrs     x1, cnthctl_el2                 */
        .long 0xb2400421        /* orr     x1, x1, #0x3                    */
        .long 0xd51ce101        /* msr     cnthctl_el2, x1                 */
        .long 0xd51ce07f        /* msr     cntvoff_el2, xzr                */
        .long 0xd2801a61        /* mov     x1, #0xd3                       */
        .long 0xd51c4001        /* msr     spsr_el2, x1                    */
        .long 0xd2867fe1        /* mov     x1, #0x33ff                     */
        .long 0xd51c1141        /* msr     cptr_el2, x1                    */
        .long 0xd2810601        /* mov     x1, #0x830                      */
        .long 0xf2a618a1        /* movk    x1, #0x30c5, lsl #16            */
        .long 0xd51c1001        /* msr     sctlr_el2, x1                   */
        .long 0xd5381000        /* mrs     x0, sctlr_el1                   */
        .long 0xd2820081        /* mov     x1, #0x1004                     */
        .long 0x8a010000        /* and     x0, x0, x1                      */
        .long 0xd2810001        /* mov     x1, #0x800                      */
        .long 0xf2a61a01        /* movk    x1, #0x30d0, lsl #16            */
        .long 0xaa000021        /* orr     x1, x1, x0                      */
        .long 0xd5181001        /* msr     sctlr_el1, x1                   */
        .long 0xd51c117f        /* msr     hstr_el2, xzr                   */
        .long 0xd2a00601        /* mov     x1, #0x300000                   */
        .long 0xd5181041        /* msr     cpacr_el1, x1                   */
        .long 0xd2800001        /* mov     x1, #0x0                        */
        .long 0xd51c1101        /* msr     hcr_el2, x1                     */
        .long 0xd53cc001        /* mrs     x1, vbar_el2                    */
        .long 0xd518c001        /* msr     vbar_el1, x1                    */
        .long 0x10000080        /* adr     x0, ffffffff801000d0 <el1_mode> */
        .long 0xaa0003fe        /* mov     x30, x0                         */
        .long 0xd51c403e        /* msr     elr_el2, x30                    */
        .long 0xd69f03e0        /* eret                                    */

el1_mode:

        /* set initial stack pointer so stack grows down from start of code */

        ADR     sp, sysInit
        MOV     fp, #0                  /* initialize frame pointer */

        LDR     r1, =sysInit
        ADR     r4, sysInit
        SUB     r4, r4, r1

#ifdef _WRS_CONFIG_SMP

        /* Enable all PPI and SGI interrupts in the distributor */

        MRC     p15, 0, r1, c0, c0, 5
        AND     r2, r1, #0xff
        LSR     r1, r1, #8
        AND     r3, r1, #0xff
        LSL     r3, r3, #1
        ADDS    r1, r2, r3
        BEQ     2f          /* if core0 */
        B       3f
#endif /* _WRS_CONFIG_SMP */

2:

#ifdef INCLUDE_STANDALONE_DTB
        LDR     r0, =dt_blob_start
#else /* !INCLUDE_STANDALONE_DTB */
        MOV     r0, r5
#endif /* INCLUDE_STANDALONE_DTB */

        LDR     r1, =gpDtbInit
        ADD     r1, r1, r4
        STR     r0, [r1]
3:
        LDR     r1, =usrInit
        ADD     r1, r1, r4
        MOV     r0, #2      /* legacy BOOT_COLD */
        BX      r1
#endif /* _WRS_CONFIG_LP64 */

FUNC_END (sysInit)

