/* sysLib.c - Intel Agilex5 SoC system-dependent library */

/*
 * Copyright (c) 2023 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */


/*
DESCRIPTION
This library provides board-specific routines for Intel Agilex5 SoC series
processor.

INCLUDE FILES:

SEE ALSO:
\tb VxWorks Programmer's Guide: Configuration
*/



/* includes */

#include <vxWorks.h>
#include <boardLib.h>
#include <itl_agilex5.h>
#include "prjParams.h"

#ifdef INCLUDE_SHOW_ROUTINES
#include <stdio.h>
#include <stdlib.h>
#include <ioLib.h>
#include <vmLibCommon.h>
#include <pmapLib.h>
#include <vxFdtLib.h>

/* defines */

#define SYSMGR_SILICONID1_OFFSET        0

/* locals */

LOCAL void socInfoShow (void);
LOCAL void boardInfo (void);
#endif

LOCAL STATUS socEndMacGet (char *, unsigned int, UINT8 *, unsigned int);

/* imports */

IMPORT BOOL itlAgilex5Probe(char * compat);
IMPORT void cpuArmVerShow (void);

/* globals */

BOARD_FUNC_TBL itlAgilex5FuncTbl =
    {
    .earlyInit  = itlAgilex5EarlyInit,
    .init       = itlAgilex5Init,
    .reset      = itlAgilex5Reset,
    .model      = itlAgilex5Model,
    .usDelay    = itlAgilex5UsDelay,
    .nsDelay    = NULL,
    .nsRes      = NULL,
#ifdef _WRS_CONFIG_SMP
    .cpuEn      = itlAgilex5CpuEnable,
    .cpuAvail   = itlAgilex5CpuAvail,
    .cpuDis     = NULL,
#endif /*_WRS_CONFIG_SMP*/
#ifdef INCLUDE_SHOW_ROUTINES
    .infoShow   = boardInfo,
#else
    .infoShow   = NULL,
#endif
    /* .endMacGet  = */ socEndMacGet
};

/* locals */

LOCAL BOARD_DESC itlAgilex5BoardDesc =
    {
    /* .uVer     = */ BOARD_DESC_VER_3_0,
    /* .pCompat  = */ "intel,socfpga-agilex",
#if defined(INCLUDE_DEBUG_KPRINTF) || defined(INCLUDE_DEBUG_KPUTS)
    /* .uFlag    = */ BOARD_DESC_FLAG (BOARD_DESC_FLAG_DBG, 0),
#else
    /* .uFlag    = */ BOARD_DESC_FLAG (0, 0),
#endif
    /* .probe    = */ itlAgilex5Probe,
    /* .pFuncTbl = */ &itlAgilex5FuncTbl
    };

BOARD_DEF (itlAgilex5BoardDesc)

/*******************************************************************************
*
* socEndMacGet - get the MAC address
*
* This function gets the MAC address for a network controller which is
* specified by the network interface name.
*
* INPUT:
*     ifName      : A pointer pointing to a string which is the network
*                   interface name of the network controller
*     ifUnit      : The network interface index
*     ifMacAddr   : A pointer pointing to the memory that is used to store
*                   the MAC address
*     ifMacAddrLen: The length of the MAC address
*
* RETURNS: OK, or ERROR if it fails to get a MAC address
*/

LOCAL STATUS socEndMacGet
    (
    char *       ifName,
    unsigned int ifUnit,
    UINT8 *      ifMacAddr,
    unsigned int ifMacAddrLen
    )
    {
    STATUS  retVal = ERROR;
    /* TODO */
    return (retVal);
    }

#ifdef INCLUDE_SHOW_ROUTINES

/*******************************************************************************
*
* socInfoShow - print SOC information
*
* This routine prints SOC information.
*
* RETURNS: N/A
* ERRNO: N/A
*/

LOCAL void socInfoShow (void)
    {
    void * virtAddr;
    UINT32 phyaddr;
    int    offset;
    UINT32 *prop;
    UINT32 rev;

    offset = vxFdtPathOffset("/alt_soc/sysmgr");
    if (offset <= 0)
        {
        (void) printf("ERROR: no sysmgr unit present in this board\n");
        return;
        }
    prop = (UINT32*)vxFdtPropGet (offset, "reg", NULL);
    phyaddr = vxFdt32ToCpu (*prop);

    virtAddr = pmapGlobalMap ((PHYS_ADDR) phyaddr,(size_t) 0x1000,
        MMU_ATTR_SUP_RW | MMU_ATTR_CACHE_OFF | MMU_ATTR_CACHE_GUARDED);

    if (virtAddr == PMAP_FAILED)
        {
        (void) printf ("pmapGlobalMap returned ERROR.\n");
        return;
        }
    rev = (*(UINT32*)((char*)virtAddr + SYSMGR_SILICONID1_OFFSET))&0xFFFF;

    (void) printf("CPU:Agilex Rev %d, ", rev);
    return;
    }

/*******************************************************************************
*
* boardInfo - print board information
*
* This routine prints board information.
*
* RETURNS: N/A
* ERRNO: N/A
*
*/

LOCAL void boardInfo (void)
    {
    char * devName = "/eeprom/0";
    UINT32 length = 40;
    char   readbuf[40] = {0};
    char * pBuf = readbuf;
    int    fd = 0;

    /* open the EEPROM */

    if ((fd = open(devName, O_RDONLY, 0)) == ERROR)
        {
        (void) printf("eepromIoRead: open the EEPROM failed.\n");
        return;
        }

    /* read the data */

    if (read(fd, pBuf, length) == ERROR)
        {
        (void) printf("eepromIoRead: read the EEPROM failed.\n");
        goto Exit;
        }

    /* an empty EEPROM */

    if ((pBuf[8] >= 'A' && pBuf[8] <='Z')
         || (pBuf[8] >= 'a' && pBuf[8] <= 'z'))
        {
        pBuf[39] = 0; /* limit to 32 bytes of string board name */
        (void) printf("%s ", (char*)(&pBuf[8]));
        (void) printf("Rev %d, ", *((UINT32*)pBuf)); /* EEPROM revision */
        }
    else
        {
        (void) printf("Altera ");
        (void) printf("Agilex SoC GSRD ");
        (void) printf("(unknown revision), ");
        }

    socInfoShow ();

    cpuArmVerShow ();

Exit:
    (void) close (fd);
    return;
    }
#endif /* INCLUDE_SHOW_ROUTINES */
