/* vmGlobalMap.c - virtual memory global mapping library */

/*
 * Copyright (c) 2005, 2007, 2009-2010, 2014-2016, 2018-2023
 * Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION
This library provides the minimal virtual memory functionality for initializing
global MMU mappings for the kernel. These mappings are created at system startup
based on the system memory descriptor table - sysPhysmemDesc[] - of the BSP.
This functionality is enabled with the INCLUDE_MMU_GLOBAL_MAP component.

The mappings can be modified at runtime only if the vmBaseLib API is included
in the system with the INCLUDE_MMU_BASIC component.

The physical memory descriptor contains information used to initialize
the MMU attribute information in the translation table. The following state
bits may be or'ed together:

Supervisor access:
\ts
MMU_ATTR_PROT_SUP_READ | read access in supervisor mode
MMU_ATTR_PROT_SUP_WRITE | write access in supervisor mode
MMU_ATTR_PROT_SUP_EXE | executable access in supervisor mode
\te
User access:
\ts
MMU_ATTR_PROT_USR_READ | read access in user mode
MMU_ATTR_PROT_USR_WRITE | write access in user mode
MMU_ATTR_PROT_USR_EXE | executable access in user mode
\te
Validity attribute:
\ts
MMU_ATTR_VALID | page is valid.
MMU_ATTR_VALID_NOT | page is not valid

\te
Cache attributes:
\ts
MMU_ATTR_CACHE_OFF | cache turned off
MMU_ATTR_CACHE_COPYBACK | cache in copy-back mode
MMU_ATTR_CACHE_WRITETHRU | cache set in writethrough mode
MMU_ATTR_CACHE_GUARDED | page access set to guarded.
MMU_ATTR_CACHE_COHERENCY | page access set to cache coherent.
MMU_ATTR_CACHE_DEFAULT | default cache value, set with USER_D_CACHE_MODE
\te

Additionally, mask bits are or'ed together in the `initialStateMask' structure
element to describe which state bits are being specified in the `initialState'
structure element:

 MMU_ATTR_PROT_MSK
 MMU_ATTR_CACHE_MSK
 MMU_ATTR_VALID_MSK

If there is an error when mappings are initialized, the reason for the failure
is recorded in exception message area and the system is rebooted. The most
common reasons for failures are the invalid combination of state/stateMask,
or overlapping and conflicting entries in sysPhysmemDec[].

INCLUDE FILES:

SEE ALSO: vmBaseLib,
\tb VxWorks Platform Programmer's Guide:[Kernel Virtual Memory Management](https://docs.windriver.com/bundle/vxworks_platform_programmers_guide_@vx_release/page/ntq1563884822063.html)
*/

#include <vxWorks.h>
#include <vsbConfig.h>

#if !defined (_WRS_DISABLE_BASEVM)

#include <mmuLib.h>
#include <private/vmLibP.h>
#include <private/adrSpaceLibP.h>
#include <sysLib.h>
#include <stdio.h>
#include <cacheLib.h>
#include <private/cacheLibP.h>  /* _func_cacheFuncsSet */
#include <excMsgLib.h>
#ifdef _WRS_CONFIG_KASAN
#include <private/kasanP.h>
#endif

/* globals */

STATUS (*_func_vmInvPageMap)
        (
        PHYS_ADDR    physicalAddr,  /* physical address to map */
        size_t       len,       /* length in bytes to map */
        BOOL         noAlias    /* disallow aliasing */
        );
STATUS (*_func_vmInvUnMap)
        (
        PHYS_ADDR   physicalAddr,   /* physical address to unmap */
        size_t  len             /* length to unmap */
        );
void   (*_func_vmBaseFuncsSet) (void);

MMU_ARCH_ATTR   mmuInvalidState;
MMU_ARCH_ATTR * mmuProtStateTransTbl;
MMU_ARCH_ATTR * mmuCacheStateTransTbl;
MMU_ARCH_ATTR * mmuValidStateTransTbl;
MMU_ARCH_ATTR * mmuMaskTransTbl;

VM_CONTEXT  sysVmContext;           /* kernel vm context */

/* locals */

LOCAL size_t    vmPageSize;

/* forward declarations */

/******************************************************************************
*
* vmGlobalMapInit - initialize global mapping
*
* This routine creates and installs a virtual memory context with mappings
* defined for each contiguous memory segment defined in <pMemDescArray>.
* In the standard VxWorks configuration, an instance of PHYS_MEM_DESC (called
* `sysPhysMemDesc') is defined in the BSP; the variable is passed to
* vmGlobalMapInit() by the system configuration mechanism.
*
* This routine is called only once during system initialization. It
* should never be called by application code.
*
* If <enable> is TRUE, the MMU is enabled upon return.
*
* ERRNO:
* \is
* \i S_mmuLib_INVALID_ATTR
* <cacheDefault> introduces arch specific restrictions for invalid SPL, cache
* or protection combinations.
* \ie
*
* RETURNS: A pointer to a newly created virtual memory context, or NULL if
* memory cannot be mapped.
*
* SEE ALSO: vmBaseLibInit()
*
* \NOMANUAL
*/

VM_CONTEXT_ID vmGlobalMapInit
    (
    PHYS_MEM_DESC * pMemDescArray,    /* pointer to array of mem descs */
    int         numDescArrayElements, /* no. of elements in pMemDescArray */
    BOOL        enable,               /* enable virtual memory */
    UINT        cacheDefault          /* default data cache mode */
    )
    {
    int         ix;
    PHYS_MEM_DESC * thisDesc;
    MMU_ARCH_ATTR   archDepStateMask;
    MMU_ARCH_ATTR   archDepState;
    UINT	    archIndepStateMask;
    UINT	    archIndepState;
    UINT	    mmuDefaultCacheIndex;

    vmPageSize = MMU_PAGE_SIZE_GET();              /* req: VX7-7402 */

    /*
     * Initialize the cache default entry.
     * The index to the cache is what is stored.
     */
                                                    /* req: VX7-7403 */
    if (cacheDefault & CACHE_WRITETHROUGH)
        {
        mmuDefaultCacheIndex = STATE_TO_CACHE_INDEX (MMU_ATTR_CACHE_WRITETHRU);
        }
    else if (cacheDefault & CACHE_COPYBACK)
        {
        mmuDefaultCacheIndex = STATE_TO_CACHE_INDEX (MMU_ATTR_CACHE_COPYBACK);
        }
    else
	{
	mmuDefaultCacheIndex = STATE_TO_CACHE_INDEX (MMU_ATTR_CACHE_OFF);
	}

    mmuCacheStateTransTbl [STATE_TO_CACHE_INDEX(MMU_ATTR_CACHE_DEFAULT)] =
               mmuDefaultCacheIndex;

    /* create translation table for default virtual memory context */
                                                    /* req: VX7-7404 */
    sysVmContext.mmuTransTbl = MMU_TRANS_TBL_CREATE ();

    if (sysVmContext.mmuTransTbl == NULL)
        {
        return (NULL);                            /* req: VX7-7405 */
        }

    /* Establish Direct Mapped Region */

    if (mmuLibFuncs.mmuDirectPageMap != NULL)       /* req: VX7-7406 */
        {
        mmuLibFuncs.mmuDirectPageMap (sysVmContext.mmuTransTbl,
                                      lowestRAMAddress,
                                      (size_t)(highestRAMAddress -
                                               lowestRAMAddress));
        }

    /* set up kernel mapping of physical to virtual memory */
                                                    /* req: VX7-7407 */
    for (ix = 0; ix < numDescArrayElements; ix++)
        {
        thisDesc = &pMemDescArray[ix];

        /* if physical address is NONE, ignore entry */

        if (thisDesc->physicalAddr == (PHYS_ADDR) NONE)
            {                                     /* req: VX7-7408 */
            continue;
            }

        if ((NOT_PAGE_ALIGNED (thisDesc->virtualAddr)) ||
            (NOT_PAGE_ALIGNED (thisDesc->physicalAddr)) ||
            (NOT_PAGE_ALIGNED (thisDesc->len)))
            {                                     /* req: VX7-7409 */
            errno = S_vmLib_NOT_PAGE_ALIGNED;
            return (NULL);
            }

        archIndepStateMask = (DEFAULT_VM_MAP_ATTR_MASK &
                              ~thisDesc->initialStateMask)
                             | (thisDesc->initialStateMask);
        archIndepState     = (DEFAULT_VM_MAP_ATTR & ~thisDesc->initialStateMask)
                             | (thisDesc->initialStateMask
                               & thisDesc->initialState);

        if (mmuDataNoExec)
            {
            archIndepState &= ~MMU_ATTR_PROT_SUP_EXE;
            }
                                                  /* req: VX7-7410 */
        if (vmAttrToArchConvert (archIndepStateMask,
                                 archIndepState,
                                 &archDepStateMask,
                                 &archDepState) == ERROR)
            {                                   /* req: VX7-7411 */
            (void) excMsgPrintf ("vmGlobalMapInit: invalid "
                              "attributes - sysPhysMemDesc[%d]\n", ix);
            return (NULL);
            }

        /* map pages */
                                                  /* req: VX7-7412 */
        if (MMU_GLOBAL_PAGE_MAP (thisDesc->virtualAddr, thisDesc->physicalAddr,
                                 thisDesc->len / vmPageSize, VM_PG_OPTIMIZE,
                                 archDepState) == ERROR)
            {
            return (NULL);                      /* req: VX7-7413 */
            }

#ifdef _WRS_CONFIG_KASAN
        /* re-establish the shadow mapping for kernel system memory */
        kasanEarlyShadow (thisDesc->virtualAddr, thisDesc->len, archIndepState);
#endif

        /* increment the physical page mapping in the inverse page table */
                                                  /* req: VX7-7414 */
        if ((_func_vmInvPageMap != NULL) &&
            (_func_vmInvPageMap (thisDesc->physicalAddr, thisDesc->len,
                 FALSE) == ERROR))
            {
            return (NULL);                      /* req: VX7-7415 */
            }
        }

    if (mmuDataNoExec)
        {
        extern char     wrs_kernel_text_start [];
        extern char     etext[];
        VIRT_ADDR       startAddr;
        VIRT_ADDR       endAddr;

        if (vmAttrToArchConvert (DEFAULT_VM_MAP_ATTR_MASK, DEFAULT_VM_MAP_ATTR,
                                &archDepStateMask, &archDepState) == ERROR)
            {
            return (NULL);
            }

        startAddr = (VIRT_ADDR) ROUND_DOWN ((ULONG) wrs_kernel_text_start, vmPageSize);
        endAddr   = (VIRT_ADDR) ROUND_UP ((ULONG) etext, vmPageSize);

        if (MMU_STATE_SET (sysVmContext.mmuTransTbl, startAddr,
                           (endAddr - startAddr) /  vmPageSize, archDepStateMask,
                           archDepState, FALSE) == ERROR)
            {
            return (NULL);
            }
        }

    /* Call the function to set a task's MMU context, if installed. */

    if  (_func_taskMmuCurrentSet != NULL)
        {                                         /* req: VX7-7416 */
        _func_taskMmuCurrentSet (sysVmContext.mmuTransTbl);
        }

    /* Set the current context to the newly created one. */

    MMU_CURRENT_SET (sysVmContext.mmuTransTbl);     /* req: VX7-7417 */

    /* update cache function pointers */

    cacheMmuAvailable = TRUE;                       /* req: VX7-7418 */

    if (_func_cacheFuncsSet != NULL)
    _func_cacheFuncsSet ();                         /* req: VX7-7419 */

    if (_func_vmBaseFuncsSet != NULL)
    _func_vmBaseFuncsSet();                        /* req: VX7-7420 */

    /* enable MMU if needed */

    if (enable)
        {                                          /* req: VX7-7421 */
        if (MMU_ENABLE (TRUE) == ERROR)
            {                                    /* req: VX7-7423 */
            return (NULL);
            }
        }
                                                   /* req: VX7-7421 */
#ifdef _WRS_CONFIG_KASAN
    /*
     * mark the whole kernel system memory range accessible.
     * The step must be done here, where the new MMU table has taken effect and
     * MMU has been enabled.
     */
    for (ix = 0; ix < numDescArrayElements; ix++)
        {
        thisDesc = &pMemDescArray[ix];
        if (thisDesc->physicalAddr == (PHYS_ADDR) NONE)
            {                                     /* req: VX7-7408 */
            continue;
            }
        kasanUnpoison ((void *) (thisDesc->virtualAddr), thisDesc->len);
        }
#endif
    return (&sysVmContext);                      /* req: VX7-7422 */
    }

/****************************************************************************
*
* vmAttrToArchConvert - convert MMU state to architecture dependent
*
* This routine convert the architecture-independent MMU state value to its
* corresponding architecture-dependent value. It sets <*pDepStateMask> and
* <*pDepState> to the architecture-dependent MMU mask and state, respectively.
*
* RETURNS: OK on success, or ERROR if <indepState> is not a valid combination
* of MMU states, or <indepStateMask> is not a valid combination of MMU state
* masks.
*
* ERRNO:
* \is
* \i S_vmLib_BAD_STATE_PARAM
* <indepState> is not a valid combination of MMU states.
* \i S_vmLib_BAD_MASK_PARAM
* <indepStateMask> is not a valid combination of MMU state masks.
* \ie
*
* \NOMANUAL
*/

STATUS vmAttrToArchConvert
    (
    UINT32 indepStateMask,          /* Arch independent MMU state Mask */
    UINT32 indepState,              /* Arch independent MMU state */
    MMU_ARCH_ATTR * pDepStateMask,  /* to return arch dependent mask */
    MMU_ARCH_ATTR * pDepState       /* to return arch dependent state */
    )
    {

    MMU_ARCH_ATTR archDepStateMask  = 0;
    MMU_ARCH_ATTR archDepState      = 0;
    MMU_ARCH_ATTR archDepProtState  = 0;
    MMU_ARCH_ATTR archDepCacheState = 0;
    MMU_ARCH_ATTR archDepValidState = 0;
    BOOL      validMask     = FALSE;
    MMU_ARCH_ATTR defaultIndex      = 0;
    MMU_ARCH_ATTR index             = 0;

    if ((indepState & (~MMU_ATTR_ALL_MSK)) != 0)
        {
        errno = S_vmLib_BAD_STATE_PARAM;
        return (ERROR);
        }
                                                /* req: VX7-7425 */
    if ((indepStateMask & MMU_ATTR_PROT_MSK) == MMU_ATTR_PROT_MSK)
        {
        validMask = TRUE;

        archDepProtState =
            mmuProtStateTransTbl[STATE_TO_PROT_INDEX(indepState)];

        /* if returned an invalid state, or SUP READ is not set return error */

        if ((archDepProtState == mmuInvalidState) ||
            (~indepState & MMU_ATTR_PROT_SUP_READ))
            {                                 /* req: VX7-7426 */
            errno = S_vmLib_BAD_STATE_PARAM;
            return (ERROR);
            }
        else
            {
            archDepState = archDepProtState;    /* req: VX7-7425 */
            }
        }

    /*
     * if necessary get architecture dependent cache state and
     * check its validity
     */
                                                /* req: VX7-7427 */
    if ((indepStateMask & MMU_ATTR_CACHE_MSK) == MMU_ATTR_CACHE_MSK)
        {
        validMask = TRUE;

        /*
         * MMU_ATTR_CACHE_DEFAULT must not be set in conjunction with any
         * other cache settings
         */

        if ((indepState & MMU_ATTR_CACHE_DEFAULT) &&
            (indepState & (MMU_ATTR_CACHE_OFF | MMU_ATTR_CACHE_COPYBACK)))
            {                                 /* req: VX7-7428 */
            errno = S_vmLib_BAD_STATE_PARAM;
            return (ERROR);
            }

        defaultIndex = 0;
                                              /* req: VX7-7429 */
        if (indepState & MMU_ATTR_CACHE_DEFAULT)
            {
            defaultIndex =
                mmuCacheStateTransTbl [STATE_TO_CACHE_INDEX(MMU_ATTR_CACHE_DEFAULT)];
            }

        if (defaultIndex == mmuInvalidState)
            {                               /* req: VX7-7430 */
            errno = S_vmLib_BAD_STATE_PARAM;
            return (ERROR);
            }

        if ((indepState & MMU_ATTR_CACHE_MSK) & (UINT)~(MMU_ATTR_CACHE_DEFAULT))
            {                                 /* req: VX7-7431 */
    	    index = defaultIndex
    	    	    + STATE_TO_CACHE_INDEX(((indepState & MMU_ATTR_CACHE_MSK) &
    	    	    (UINT)~(MMU_ATTR_CACHE_DEFAULT)));
            }
        else
            {				      /* req: VX7-7429 */
    	    index = defaultIndex;
            }
                                                /* req: VX7-7427 */
        archDepCacheState = mmuCacheStateTransTbl [index];

        if (archDepCacheState == mmuInvalidState)
            {                                 /* req: VX7-7432 */
            errno = S_vmLib_BAD_STATE_PARAM;
            return (ERROR);
            }
        else
            {                                 /* req: VX7-7431 */
            archDepState |= archDepCacheState;
            }
        }

    /*
     * if necessary get architecture dependent valid state and
     * check its validity
     */
                                                /* req: VX7-7433 */
    if ((indepStateMask & MMU_ATTR_VALID_MSK) == MMU_ATTR_VALID_MSK)
        {
        validMask = TRUE;

        archDepValidState =
                mmuValidStateTransTbl [STATE_TO_VALID_INDEX(indepState)];

        /* if returned an invalid state, return error */

        if (archDepValidState == mmuInvalidState)
            {                                 /* req: VX7-7434 */
            errno = S_vmLib_BAD_STATE_PARAM;
            return (ERROR);
            }
        else
            {                                   /* req: VX7-7433 */
            archDepState |= archDepValidState;
            }
        }

    /* get the architecture dependent state mask */

    if (validMask == TRUE)                      /* req: VX7-7435 */
        {
        archDepStateMask = mmuMaskTransTbl [STATEMASK_TO_INDEX(indepStateMask)];
        }

    /*
     * If this architecture uses special purpose attributes then use
     * the architecture specific translation mechanism.
     */
                                                /* req: VX7-7437 */
    if ((indepStateMask & MMU_ATTR_SPL_MSK) == MMU_ATTR_SPL_MSK)
        {
        validMask = TRUE;

#ifndef _WRS_CONFIG_ITL_SOC_AGILEX5
        if (MMU_SPL_TRANSLATE (indepState, indepStateMask,
                               &archDepState, &archDepStateMask) == ERROR)
            {
            errno = S_vmLib_BAD_STATE_PARAM;
            return (ERROR);                   /* req: VX7-7438 */
            }
#endif
        }

    if (validMask == FALSE)
        {
        /* No valid arch dependent mask was passed. So return ERROR. */

        errno = S_vmLib_BAD_MASK_PARAM;      /* req: VX7-7436 */
        return (ERROR);
        }

    *pDepStateMask = archDepStateMask;          /* req: VX7-7439 */
    *pDepState     = archDepState;

    return (OK);
    }
#endif  /* !_WRS_DISABLE_BASEVM */
