/* vxbFdtSynopsysDwGpio.c - VxBus driver for Synopsys DesignWare GPIO */

/*
 * Copyright (c) 2014, 2016-2018, 2021-2022 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */
/*
DESCRIPTION
This module implements a driver for the DesignWare GPIO IP core which
presents on Renesas RZ/N1 SoC and other platforms.
The driver has been verified on the platforms below:

Renesas RZ/N Series

The number of actual pins and banks may vary from the different SoC 
implementations. For Example, there are 170 pins, 3 banks and 2 ports per bank
in Renesas RZ/N1 implmentation. And in that SoC, all pins in Port A in all banks
share 8 interrupts, but each interrupt can be occupied by 1 pin exclusively
at one time. This driver detects which interrupt is used by the specific GPIO in
in connecting stage(calling vxbGpioIntConnect()), so before connecting interrupt
to a GPIO pin that's managed by this driver, the interrupt multiplexing must have
been set, and keep unchanged until vxbGpioIntDisConnect() is called for that 
pin. Usually, the interrupt multiplexing is manged by pinmux subsystem, if 
necessary, please refer to the corresponding pinmux driver for detail.

This driver supports various hardware implementations by just changing the DTS
data. Examples are given in the end of the DESCRIPTION.

To add the driver to the vxWorks image, add the following component to the
kernel configuration.

\cs
vxprj component add DRV_FDT_SYNOPSYS_DW_GPIO
\ce

Each GPIO device should be bound to a device tree node which requires
below properties:

\cs
compatible:     Specify the programming model for the device.
                It should be set to "snps,designware-gpio" and is used
                by vxbus GEN2 for device driver selection.

reg:            Specify the addresses of the device's resources within
                the address space defined by its parent bus. The second address
                entry is optional. If it exists, it represents the IO
                multiplexing module to figure out which interrupt the GPIO pin
                connects to.

interrupts:     Specify interrupt vector of the interrupts that are generated
                by this device. There may be more than 1 entry in it where the
                several interrupts are shared by GPIO pins, and 1 interrupt
                can be occupied by 1 pin only at a time. In such situation, the
                driver is recommended to be configured to support all GPIO pins
                in a single device instance, so the interrupt resources can be
                freely used by the GPIO pins.

interrupt-parent: This property is available to define an interrupt
                parent. if it is missing from a device, it's interrupt parent
                is assumed to be its device tree parent.

debounce-time:  Specify debounce time for the GPIO of this device if the 
                controller supports. All gpio pins have the same debounce-time
                if it supports. Because of all GPIO pins share the same debounce
                time,so function vxbGpioSetDebounce () only support this value 
                or 0. The debounce time is usually 1 clock period of the GPIO
                module. It's optional. If it's missed, which means the hardware
                doesn't support debounce. If this property exists, but no value
                is given, a default value of 20000 is used. Otherwise, the 
                specified value is used. The unit is micro second.

port-width:     Specify the number of pins implemented in the particular GPIO 
                controller. When the driver is configured to support several 
                ports and several banks in one single vxbus device instance, 
                the port-width should be the number of pin in Port A. If it's
                missed, the default value is 32.

total-num:      Specify the total pin number that the driver supports. For 
                example, in Renesas RZ/N1, the total pin number is 170. It can
                be greater than or equal to the port-width. If it's missed,
                the default values equals to port-width.

bank-num:       Specify the number of GPIO banks in 1 device instance. It can be
                missed if only 1 bank supported in a device instance.

bank-shift:     Specify the address range size of each bank. The size is the
                power of this value based on 2. It can be missed if only 1 bank
                supported in a device instance.

porta-num:      Specify the number of pins that in Port A which support
                interrupt. If it's missed, the port-width is used.

map-table:      Specify the mapping relationship between the chip pin numbering
                and the GPIO pin numbering. The schematics usually uses the chip
                pin number to represet pin, but driver needs to know which bank,
                which port contains the pin so can index to register bit. The
                map table is used for this purpose. The index of this table is
                the chip pin number, and the table values prepresent the GPIO
                pin number. Each table entry occupies 1 byte. If it's missed,
                the pin value passed to this driver is used as GPIO numbering
                directly. For exmaple, 33 represents Pin 1 in Bank 1 with 32-bit
                port-width.
                The numbers in this table are hexadecimal, but the 0x prefix or
                H postfix is missed.

\ce

Below are some examples:

\cs
        gpio0: gpio@FF708000
            {
            compatible = "snps,designware-gpio";
            gpio-controller;
            #gpio-cells = <4>;
            #interrupt-cells = <1>;
            reg = <0xFF708000 0x1000>;
            interrupts = <196>;
            interrupt-parent = <&intc>;
            port-width = <29>

            /@
             @ All GPIO pins share the same debounce time, make sure your
             @ debounce-time field here is correct. debounce-time must be the
             @ same as other gpio banks.
             @/

            debounce-time = <20000>;
            };
\ce

Note: The corresponding device instance of the dts node above supports 1 bank,
      and there are 29 pins in this bank, all the pins share the same interrupt.

\cs
        gpio: gpio@5000b000
            {
            compatible = "snps,designware-gpio";
            #gpio-cells = <4>;
            #interrupt-cells = <1>;
            pinmux-0 = <&gpio0_mux>;
            reg = <0x5000b000 0x3000>,
                  <0x51000000 0x1000>;
            interrupts = <135 0 0x104>,
                         <136 0 0x104>,
                         <137 0 0x104>,
                         <138 0 0x104>,
                         <139 0 0x104>,
                         <140 0 0x104>,
                         <141 0 0x104>,
                         <142 0 0x104>;
            interrupt-parent = <&gic>;
            clocks = <&bgpio1_pclk>,
                     <&bgpio2_pclk>,
                     <&bgpio3_pclk>;
            clock-names = "bgpio1_pclk", "bgpio2_pclk", "bgpio3_pclk";
            port-width = <32>;
            total-num  = <170>;
            bank-num   = <3>;
            bank-shift = <12>;
            porta-num  = <96>;
            map-table  = [
                00 /@   0: BGPIO1A[0]  @/ 60 /@   1: BGPIO1B[0]  @/ 61 /@   2: BGPIO1B[1]  @/
                01 /@   3: BGPIO1A[1]  @/ ....(The rest elements are skipped.
                The number of the elements must equals to total-num.)
              ];
\ce

Note: The corresponding device instance of the dts node above supports 3 banks,
      and there are 170 pins in these 3 banks, with 96 pins in Port A that 
      supports interrupts, 32 pins in each port except the Port B in the 3rd 
      bank, while there are only 10 pins (170 - (32*5)) in it. There are 8 
      interrupts shared by 96 pins in Port A of all banks.

INCLUDE FILES: vxBus.h vxbLib.h vxbGpioLib.h

SEE ALSO: vxBus
*/

/* includes */

#include <vxWorks.h>
#include <string.h>
#include <stdlib.h>
#include <semLib.h>
#include <stdio.h>
#include <ioLib.h>
#include <cacheLib.h>
#include <hwif/vxBus.h>
#include <hwif/vxbus/vxbLib.h>
#include <hwif/buslib/vxbFdtLib.h>
#include <subsys/timer/vxbTimerLib.h>
#include <subsys/gpio/vxbGpioLib.h>
#include <vxAtomicLib.h>
#include "vxbFdtSynopsysDwGpio.h"

/* defines */

#undef GPIO_DBG_ON
#ifdef  GPIO_DBG_ON

#include <private/kwriteLibP.h>     /* _func_kprintf */

#define GPIO_DBG_IRQ         0x00000001
#define GPIO_DBG_RW          0x00000002
#define GPIO_DBG_ERR         0x00000004
#define GPIO_DBG_ALL         0xffffffff
#define GPIO_DBG_OFF         0x00000000
LOCAL UINT32 gpioDbgMask =   GPIO_DBG_ALL;

#undef LOCAL
#define LOCAL

#define GPIO_DBG(mask, ...)            \
    if ((gpioDbgMask & mask) || (mask == GPIO_DBG_ALL))     \
        if (_func_kprintf != NULL)                           \
            (* _func_kprintf)(__VA_ARGS__)
#else
#define GPIO_DBG(mask, ...)
#endif  /* GPIO_DBG_ON */

#ifdef ARMBE8
#    define SWAP32 vxbSwap32
#    define SWAP16 vxbSwap16
#else
#    define SWAP32
#    define SWAP16
#endif /* ARMBE8 */

#undef CSR_READ_4
#define CSR_READ_4(pDrvCtrl, addr)                              \
    SWAP32(vxbRead32 (GPIO_HANDLE(pDrvCtrl),                    \
        (UINT32 *)((char *)GPIO_BAR(pDrvCtrl) + addr)))

#undef CSR_WRITE_4
#define CSR_WRITE_4(pDrvCtrl, addr, data)                       \
    vxbWrite32 (GPIO_HANDLE(pDrvCtrl),                          \
        (UINT32 *)((char *)GPIO_BAR(pDrvCtrl) + addr), SWAP32(data))

#undef CSR_SETBIT_4
#define CSR_SETBIT_4(pDrvCtrl, offset, val)                     \
    CSR_WRITE_4(pDrvCtrl, offset, CSR_READ_4(pDrvCtrl, offset) | (val))

#undef CSR_CLRBIT_4
#define CSR_CLRBIT_4(pDrvCtrl, offset, val)                     \
    CSR_WRITE_4(pDrvCtrl, offset, CSR_READ_4(pDrvCtrl, offset) & ~(val))

#undef CSR_SETCLRBIT_4
#define CSR_SETCLRBIT_4(pDrvCtrl, offset, msk, val)             \
    CSR_WRITE_4(pDrvCtrl, offset, (CSR_READ_4(pDrvCtrl, offset) & ~(msk)) | val)

#undef INT_CFG_READ_4
#define INT_CFG_READ_4(pDrvCtrl, addr)                          \
    vxbRead32 (INT_HANDLE(pDrvCtrl),                            \
        (UINT32 *)((char *)INT_BAR(pDrvCtrl) + addr))

#define GPIO_PIN_TO_BIT(gpio)           (1 << (gpio))

/* forward declarations */

LOCAL STATUS dwGpioProbe (VXB_DEV_ID pDev);
LOCAL STATUS dwGpioAttach (VXB_DEV_ID pDev);
LOCAL void   dwGpioCtrlInit (DW_GPIO_DRV_CTRL * pDrvCtrl);
LOCAL STATUS dwGpioPinReset (VXB_GPIOCTRL * pCtrl, UINT32 id);
LOCAL STATUS dwGpioPinAlloc (VXB_GPIOCTRL * pCtrl, UINT32 id);
LOCAL STATUS dwGpioPinFree (VXB_GPIOCTRL * pCtrl, UINT32 id);
LOCAL UINT32 dwGpioPinGetDir (VXB_GPIOCTRL * pCtrl, UINT32 id);
LOCAL STATUS dwGpioPinSetDir (VXB_GPIOCTRL * pCtrl, UINT32 id, UINT32 dir);
LOCAL UINT32 dwGpioPinGetValue (VXB_GPIOCTRL * pCtrl, UINT32 id);
LOCAL STATUS dwGpioPinSetValue (VXB_GPIOCTRL * pCtrl, UINT32 id, UINT32 value);
LOCAL STATUS dwGpioPinSetDebounce (VXB_GPIOCTRL * pCtrl, UINT32 id, UINT32 us);
LOCAL STATUS dwGpioIntConfig (VXB_GPIOCTRL * pCtrl, UINT32 id,
                              INTR_TRIGER trig, INTR_POLARITY pol);
LOCAL void   dwGpioISR (DW_GPIO_DRV_CTRL * pDrvCtrl);
LOCAL void   dwGpioExclusiveISR (DW_GPIO_INT_NODE * pNode);
LOCAL STATUS dwGpioIntConnect (VXB_GPIOCTRL * pCtrl, UINT32 id,
                               VOIDFUNCPTR pIsr, void * pArg);
LOCAL STATUS dwGpioIntDisConnect (VXB_GPIOCTRL * pCtrl, UINT32 id,
                                  VOIDFUNCPTR pIsr, void * pArg);
LOCAL STATUS dwGpioIntEnable (VXB_GPIOCTRL * pCtrl, UINT32 id,
                              VOIDFUNCPTR pIsr, void * pArg);
LOCAL STATUS dwGpioIntDisable (VXB_GPIOCTRL * pCtrl, UINT32 id,
                               VOIDFUNCPTR pIsr, void * pArg);
LOCAL __inline__ UINT32 dwGpioPinConv (UINT32 * bank, UINT32 * pin,
                                       DW_GPIO_DRV_CTRL * pDrvCtrl);

#ifdef _WRS_CONFIG_DEBUG_FLAG
LOCAL void   dwGpioShow (VXB_GPIOCTRL * pCtrl, UINT32 verbose);
#endif /* _WRS_CONFIG_DEBUG_FLAG */

LOCAL VXB_DRV_METHOD dwGpioMethodList[] =
    {
    /* DEVICE API */

    { VXB_DEVMETHOD_CALL(vxbDevProbe),  (FUNCPTR)dwGpioProbe  },
    { VXB_DEVMETHOD_CALL(vxbDevAttach), (FUNCPTR)dwGpioAttach },
    VXB_DEVMETHOD_END
    };

LOCAL VXB_FDT_DEV_MATCH_ENTRY dwGpioMatch[] =
    {
        {
        "snps,designware-gpio",                    /* compatible */
        NULL
        },
        {}                              /* empty terminated list */
    };

/* globals */

VXB_DRV vxbFdtSynopsysDwGpioDrv =
    {
    { NULL } ,
    DW_GPIO_DRIVER_NAME,                    /* Name */
    "Synopsys DesignWare GPIO Controller",  /* Description */
    VXB_BUSID_FDT,                          /* Class */
    0,                                      /* Flags */
    0,                                      /* Reference count */
    dwGpioMethodList                        /* Method table */
    };

VXB_DRV_DEF(vxbFdtSynopsysDwGpioDrv)

/*******************************************************************************
*
* dwGpioProbe - probe for device presence at specific address
*
* Check for DesignWare GPIO (or compatible) device at the specified base
* address. We assume one is present at that address, but
* we need to verify.
*
* RETURNS: OK if probe passes and assumed a valid (or compatible) device.
* ERROR otherwise.
*
*/

LOCAL STATUS dwGpioProbe
    (
    VXB_DEV_ID pDev
    )
    {
    return vxbFdtDevMatch (pDev, dwGpioMatch, NULL);
    }

/*******************************************************************************
*
* dwGpioAttach - attach DesignWare GPIO device
*
* This is the DesignWare GPIO initialization routine.
*
* RETURNS: OK, or ERROR if initialization failed.
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioAttach
    (
    VXB_DEV_ID  pDev
    )
    {
    VXB_FDT_DEV *       pFdtDev = vxbFdtDevGet (pDev);
    VXB_GPIOCTRL *      pCtrl   = NULL;
    DW_GPIO_DRV_CTRL *  pDrvCtrl= NULL;
    VXB_RESOURCE_ADR *  pResAdr = NULL;
    VXB_RESOURCE *      pRes    = NULL;
    VXB_RESOURCE *      pIntCfgRes = NULL;
    UINT32              intCellSize = 0;
    UINT32 *            prop;
    int                 len;
    int                 intcOffset;
    UINT32              i;

    if (pFdtDev == NULL)
        {
        GPIO_DBG (GPIO_DBG_ERR, "dwGpioAttach: pFdtDev is NULL\n");
        return ERROR;
        }

    if ((pDrvCtrl = vxbMemAlloc (sizeof (DW_GPIO_DRV_CTRL))) == NULL)
        {
        GPIO_DBG (GPIO_DBG_ERR, "dwGpioAttach: pDrvCtrl alloc failed\n");
        return ERROR;
        }

    pCtrl = &(pDrvCtrl->gpioCtrl);
    pCtrl->pDev = pDev;
    pDrvCtrl->gpioDev = pDev;

    prop = (UINT32 *)vxFdtPropGet(pFdtDev->offset, "interrupt-parent", &len);
    if ((prop != NULL) && (len > 0))
        {
        intcOffset = vxFdtNodeOffsetByPhandle(vxFdt32ToCpu(*prop));
        prop =  (UINT32 *)vxFdtPropGet(intcOffset, "#interrupt-cells", &len);
        if ((prop != NULL) && (len > 0))
            {
            intCellSize = vxFdt32ToCpu(*prop);
            }
        }

    GPIO_DBG (GPIO_DBG_RW, "dwGpioAttach: #interrupt-cells=%d\n", intCellSize);

    if (intCellSize != 0)
        {
        prop =  (UINT32 *)vxFdtPropGet(pFdtDev->offset, "interrupts", &len);
        if ((prop != NULL) && (len > 0))
            {
            pDrvCtrl->intNum = (UINT32)len / (intCellSize * sizeof (UINT32));
            }
        }

    GPIO_DBG (GPIO_DBG_RW, "dwGpioAttach: inuNum: %d\n", pDrvCtrl->intNum);

    if (pDrvCtrl->intNum != 0)
        {
        pDrvCtrl->intRes = (VXB_RESOURCE **)vxbMemAlloc (
                                                 pDrvCtrl->intNum * 
                                                 sizeof (VXB_RESOURCE *));
        if (pDrvCtrl->intRes == NULL)
            {
            GPIO_DBG (GPIO_DBG_RW, 
                      "dwGpioAttach: Cannot allocate mem for IRQ resources\n");
            goto error;
            }
        }

    /* get memory resources */

    pRes = vxbResourceAlloc (pDev, VXB_RES_MEMORY, (UINT16)0);

    if (pRes == NULL)
       {
       goto error;
       }

    pResAdr = (VXB_RESOURCE_ADR *)pRes->pRes;

    if (pResAdr == NULL)
       {
       goto error;
       }

    pDrvCtrl->handle    = pResAdr->pHandle;
#ifdef _WRS_CONFIG_LP64
    pDrvCtrl->bar       = (UINT64)pResAdr->virtual;
#else
    pDrvCtrl->bar       = (UINT32)pResAdr->virtual;
#endif

    pIntCfgRes = vxbResourceAlloc (pDev, VXB_RES_MEMORY, (UINT16)1);
    if (pIntCfgRes != NULL)
       {
       pResAdr = (VXB_RESOURCE_ADR *)pIntCfgRes->pRes;
       pDrvCtrl->intCfgHandle = pResAdr->pHandle;
       pDrvCtrl->intCfgBar    = (void *)pResAdr->virtual;
       }

    /* get interrupt resources */

    for (i = 0; i < pDrvCtrl->intNum; i++)
        {
        pDrvCtrl->intRes[i] = vxbResourceAlloc (pDev, VXB_RES_IRQ, (UINT16)i);
        if (pDrvCtrl->intRes[i] == NULL)
            {
            goto error;
            }
        }

    prop = (void *)vxFdtPropGet (pFdtDev->offset, "debounce-time", &len);
    if (prop == NULL)
        {
        pDrvCtrl->debounceTime = GPIO_INVALID_DEBOUNCE_TIME;
        GPIO_DBG (GPIO_DBG_RW, 
                  "dwGpioAttach: debounce time is not supported.\n");
        }
    else if ((size_t)len == sizeof (UINT32))
        {
        pDrvCtrl->debounceTime = vxFdt32ToCpu (*((UINT32*) prop));
        GPIO_DBG (GPIO_DBG_RW, 
                  "dwGpioAttach: The specified debounce time is %u.\n", 
                  pDrvCtrl->debounceTime);
        }
    else
        {
        pDrvCtrl->debounceTime = GPIO_DEFAULT_DEBOUNCE_TIME;
        GPIO_DBG (GPIO_DBG_RW, 
                  "dwGpioAttach: The default debounce time is %u.\n", 
                  pDrvCtrl->debounceTime);
        }

    prop = (UINT32 *)vxFdtPropGet (pFdtDev->offset, "port-width", &len);
    if (prop != NULL)
        {
        pDrvCtrl->portWidth = vxFdt32ToCpu (*prop);
        if (pDrvCtrl->portWidth > DW_MAX_GPIO_BANK_WIDTH)
            pDrvCtrl->portWidth = DW_MAX_GPIO_BANK_WIDTH;
        }
    else
        {
        pDrvCtrl->portWidth = DW_MAX_GPIO_BANK_WIDTH;
        }

    GPIO_DBG (GPIO_DBG_RW,
              "dwGpioAttach: portWidth: %d\n", pDrvCtrl->portWidth);

    prop = (UINT32 *)vxFdtPropGet (pFdtDev->offset, "total-num", &len);
    if (prop != NULL)
        {
        pDrvCtrl->totalPinNum = vxFdt32ToCpu (*prop);
        }
    else
        {
        pDrvCtrl->totalPinNum = pDrvCtrl->portWidth;
        }

    GPIO_DBG (GPIO_DBG_RW,
              "dwGpioAttach: total pins: %d\n", pDrvCtrl->totalPinNum);

    prop = (UINT32 *)vxFdtPropGet (pFdtDev->offset, "bank-num", &len);
    if ((prop != NULL) && (len > 0))
        {
        pDrvCtrl->bankNum = vxFdt32ToCpu (*prop);
        }
    else
        {
        pDrvCtrl->bankNum = 1U;
        }

    GPIO_DBG (GPIO_DBG_RW,
              "dwGpioAttach: total banks: %u\n", pDrvCtrl->bankNum);

    if (pDrvCtrl->bankNum > 1U)
        {
        prop = (UINT32 *)vxFdtPropGet (pFdtDev->offset, "bank-shift", &len);
        if ((prop != NULL) && (len > 0))
            {
            pDrvCtrl->bankShift = vxFdt32ToCpu (*prop);
            GPIO_DBG (GPIO_DBG_RW,
                      "dwGpioAttach: The bank shift bits are: %u\n",
                      pDrvCtrl->bankShift);
            }
        else
            {
            GPIO_DBG (GPIO_DBG_ERR,
                      "dwGpioAttach: Valid bank-shift value is mandatory for "
                      "multiple banks\n");
            goto error;
            }
        }

    prop = (UINT32 *)vxFdtPropGet (pFdtDev->offset, "porta-num", &len);
    if ((prop != NULL) && (len > 0))
        {
        pDrvCtrl->portANum = vxFdt32ToCpu (*prop);
        }
    else
        {
        pDrvCtrl->portANum = pDrvCtrl->portWidth;
        }

    GPIO_DBG (GPIO_DBG_RW,
              "dwGpioAttach: total pins in Port A: %u\n", pDrvCtrl->portANum);

    prop = (UINT32 *)vxFdtPropGet (pFdtDev->offset, "map-table", &len);
    if ((prop != NULL) && ((UINT32)len == pDrvCtrl->totalPinNum))
        {
        pDrvCtrl->pMapTable = (UINT8 *)prop;
        }

    GPIO_DBG (GPIO_DBG_RW,
              "dwGpioAttach: Map Table: %p\n", pDrvCtrl->pMapTable);

    pCtrl->length = pDrvCtrl->totalPinNum;

    pCtrl->pValidBmp = vxbMemAlloc (ROUND_UP ((pCtrl->length + 7) / 8,
                                              sizeof (UINT32)));
    if (pCtrl->pValidBmp == NULL)
        {
        GPIO_DBG (GPIO_DBG_ERR, "dwGpioAttach: pValidBmp alloc failed.\n");
        goto error;
        }

    /* parse the GPIO cells */

    prop = (void *)vxFdtPropGet (pFdtDev->offset, "#gpio-cells", &len);
    if ((prop == NULL) || (len != 4))
        {
        GPIO_DBG (GPIO_DBG_ERR,
                  "dwGpioAttach: gpio-cells parse error, len=%d.\n", len);
        goto error;
        }

    pCtrl->gpioCells            = vxFdt32ToCpu (((UINT32*)prop)[0]);
    pCtrl->gpioAlloc            = dwGpioPinAlloc;
    pCtrl->gpioFree             = dwGpioPinFree;
    pCtrl->gpioGetDir           = dwGpioPinGetDir;
    pCtrl->gpioSetDir           = dwGpioPinSetDir;
    pCtrl->gpioGetValue         = dwGpioPinGetValue;
    pCtrl->gpioSetValue         = dwGpioPinSetValue;
    pCtrl->gpioSetDebounce      = dwGpioPinSetDebounce;
    pCtrl->gpioIntConnect       = dwGpioIntConnect;
    pCtrl->gpioIntDisConnect    = dwGpioIntDisConnect;
    pCtrl->gpioIntEnable        = dwGpioIntEnable;
    pCtrl->gpioIntDisable       = dwGpioIntDisable;
    pCtrl->gpioIntConfig        = dwGpioIntConfig;
#ifdef _WRS_CONFIG_DEBUG_FLAG
    pCtrl->gpioShow             = dwGpioShow;
#endif /* _WRS_CONFIG_DEBUG_FLAG */

    /* save pDrvCtrl in VXB_DEVICE structure */

    vxbDevSoftcSet (pDev, (void *)pDrvCtrl);

    dwGpioCtrlInit (pDrvCtrl);

    /* connect interrupts */

    if (pDrvCtrl->intNum == 1) /* all gpio pins share 1 interrupt */
        {
        (void)vxbIntConnect (pDev, pDrvCtrl->intRes[0], dwGpioISR, pDrvCtrl);
        (void)vxbIntEnable (pDev, pDrvCtrl->intRes[0]);
        }
    else if (pDrvCtrl->intNum > 1) /* 1 pin occupies 1 interrupt exclusively */
        {
        pDrvCtrl->pIntNode[0] = (DW_GPIO_INT_NODE *)vxbMemAlloc (
                                    (sizeof (DW_GPIO_INT_NODE) + 
                                     sizeof (DW_GPIO_INT_NODE_EXT)) * 
                                    pDrvCtrl->intNum);
        if (pDrvCtrl->pIntNode[0] == NULL)
            {
            for (i = 0; i < pDrvCtrl->intNum; i++)
                {
                vxbResourceFree (pDev, pDrvCtrl->intRes[i]);
                pDrvCtrl->intRes[i] = NULL;
                }
            vxbMemFree (pDrvCtrl->intRes);
            pDrvCtrl->intRes = NULL;
            pDrvCtrl->intNum = 0;
            }
        else
            {
            for (i = 0; i < pDrvCtrl->intNum; i ++)
                {
                if (i != 0)
                    {
                    pDrvCtrl->pIntNode[i] = (DW_GPIO_INT_NODE *)
                        ((UINT8 *)pDrvCtrl->pIntNode[0] + 
                         ((sizeof (DW_GPIO_INT_NODE) + 
                           sizeof (DW_GPIO_INT_NODE_EXT)) * (size_t)i));
                    }
                pDrvCtrl->pIntNode[i]->intExt->pDrvCtrl = pDrvCtrl;

                (void)vxbIntConnect (pDev, pDrvCtrl->intRes[i], 
                                     dwGpioExclusiveISR, pDrvCtrl->pIntNode[i]);
                (void)vxbIntEnable (pDev, pDrvCtrl->intRes[i]);
                }
            }
        }

    if (vxbGpioAddCtlr (pCtrl) == OK)
        {
        return OK;
        }

    /* follow through to release resources */

    /* disable & disconnect interrupts */

    for (i = 0; i < pDrvCtrl->intNum; i ++)
        {
        (void) vxbIntDisable (pDev, pDrvCtrl->intRes[i]);
        (void) vxbIntDisconnect (pDev, pDrvCtrl->intRes[i]);
        if (pDrvCtrl->intRes[i] != NULL)
            {
            vxbResourceFree (pDev, pDrvCtrl->intRes[i]);
            }
        }

    if (pDrvCtrl->pIntNode[0] != NULL)
        {
        vxbMemFree (pDrvCtrl->intRes);
        pDrvCtrl->intRes = NULL;
        vxbMemFree (pDrvCtrl->pIntNode[0]);
        }

error:

    /* The pDrvCtrl must have been allocated successfully here */

    if (pCtrl->pValidBmp != NULL)
        {
        vxbMemFree (pCtrl->pValidBmp);
        }

    if (pRes != NULL)
        {
        (void) vxbResourceFree (pDev, pRes);
        }

    if (pIntCfgRes != NULL)
        {
        (void) vxbResourceFree (pDev, pIntCfgRes);
        }

    if (pDrvCtrl->intRes != NULL)
        {
        for (i = 0; i < pDrvCtrl->intNum; i++)
            {
            if (pDrvCtrl->intRes[i] != NULL)
                (void) vxbResourceFree (pDev, pDrvCtrl->intRes[i]);
            }
        vxbMemFree (pDrvCtrl->intRes);
        }
    
    vxbMemFree (pDrvCtrl);

    return ERROR;
    }

/*******************************************************************************
*
* dwGpioCtrlInit - initialize DesignWare GPIO device
*
* This is the DesignWare GPIO Hardware initialization routine.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

LOCAL void dwGpioCtrlInit
    (
    DW_GPIO_DRV_CTRL * pDrvCtrl
    )
    {
    UINT32 bank;

    for (bank = 0; bank < pDrvCtrl->bankNum; bank ++)
        {
        /* clear interrupt registers */

        CSR_WRITE_4 (pDrvCtrl, INTEN_OFFSET (bank, pDrvCtrl->bankShift), 0);
        CSR_WRITE_4 (pDrvCtrl, INTMASK_OFFSET (bank, pDrvCtrl->bankShift), 
                     (1U << pDrvCtrl->portWidth) - 1U);
        CSR_WRITE_4 (pDrvCtrl, INTTYPELEVEL_OFFSET (bank, pDrvCtrl->bankShift),
                     0);
        CSR_WRITE_4 (pDrvCtrl, INTPOLARITY_OFFSET (bank, pDrvCtrl->bankShift), 
                     0);
        CSR_WRITE_4 (pDrvCtrl, EOI_OFFSET (bank, pDrvCtrl->bankShift), 
                     CSR_READ_4 (pDrvCtrl, 
                                 RAWINTSTATUS_OFFSET (bank, 
                                                      pDrvCtrl->bankShift)));
        if (pDrvCtrl->debounceTime != GPIO_INVALID_DEBOUNCE_TIME)
            {
            CSR_WRITE_4 (pDrvCtrl, DEBOUNCE_OFFSET (bank, pDrvCtrl->bankShift),
                         0);
            }
        }
    }

/*******************************************************************************
*
* dwGpioPinReset - reset a GPIO pin
*
* This routine resets a GPIO pin by setting direction to input and
* disabling interrupt.
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioPinReset
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id      /* pin to reset */
    )
    {
    
    UINT32             bank;
    UINT32             port;
    UINT32             pin;
    DW_GPIO_DRV_CTRL * pDrvCtrl = (DW_GPIO_DRV_CTRL *)pCtrl;

    pin = id - pCtrl->firstPin;

    GPIO_DBG (GPIO_DBG_RW, "dwGpioPinReset: id %u pin %u\n", id, pin);

    /* set to input */

    if (dwGpioPinSetDir (pCtrl, id, GPIO_DIR_INPUT) == ERROR)
        return ERROR;

    port = dwGpioPinConv (&bank, &pin, pDrvCtrl);

    /* clear debounce register */

    if (pDrvCtrl->debounceTime != GPIO_INVALID_DEBOUNCE_TIME)
        {
        CSR_CLRBIT_4 (pDrvCtrl, DEBOUNCE_OFFSET (bank, pDrvCtrl->bankShift), 
                      GPIO_PIN_TO_BIT (pin));
        }

    if (port != DW_GPIO_PORTA)
        return OK;

    /* disable interrupt */

    CSR_CLRBIT_4 (pDrvCtrl, INTEN_OFFSET (bank, pDrvCtrl->bankShift),
                  GPIO_PIN_TO_BIT (pin));

    CSR_SETBIT_4 (pDrvCtrl, INTMASK_OFFSET (bank, pDrvCtrl->bankShift),
                  GPIO_PIN_TO_BIT (pin));

    CSR_WRITE_4 (pDrvCtrl, EOI_OFFSET (bank, pDrvCtrl->bankShift),
                 GPIO_PIN_TO_BIT (pin));

    return OK;
    }

/*******************************************************************************
*
* dwGpioPinAlloc - allocate a GPIO pin
*
* This routine makes controller specific pin allocation.
* Since GPIO lib will do generic allocating, here only set to default state.
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioPinAlloc
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id      /* pin to allocate */
    )
    {
    GPIO_DBG (GPIO_DBG_RW, "dwGpioPinAlloc: id %u.\n", id);

    return dwGpioPinReset (pCtrl, id);
    }

/*******************************************************************************
*
* dwGpioPinFree - free a GPIO pin
*
* This routine makes controller specific pin release.
* Since GPIO lib will do generic free, only set to pin reset state.
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioPinFree
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id      /* pin to free */
    )
    {
    GPIO_DBG (GPIO_DBG_RW, "dwGpioPinFree: id %u.\n", id);

    return dwGpioPinReset (pCtrl, id);
    }

/*******************************************************************************
*
* dwGpioPinGetDir - get GPIO pin direction
*
* This routine gets the specified pin's direction.
*
* RETURNS: GPIO pin direction
*
* ERRNO: N/A
*/

LOCAL UINT32 dwGpioPinGetDir
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id
    )
    {
    UINT32             val;
    UINT32             bank;
    UINT32             port;
    UINT32             pin;
    DW_GPIO_DRV_CTRL * pDrvCtrl = (DW_GPIO_DRV_CTRL *)pCtrl;

    pin = id - pCtrl->firstPin;

    port = dwGpioPinConv (&bank, &pin, pDrvCtrl);

    GPIO_DBG (GPIO_DBG_RW, "dwGpioPinGetDir: id %u bank %u port %u pin %u\n", 
              id, bank, port, pin);

    val = CSR_READ_4 (pDrvCtrl, 
                      SWPORT_DDR_OFFSET(bank, pDrvCtrl->bankShift, port));
    if ((val & GPIO_PIN_TO_BIT (pin)) != 0U)
        return GPIO_DIR_OUTPUT;
    else
        return GPIO_DIR_INPUT;
    }

/*******************************************************************************
*
* dwGpioPinSetDir - set GPIO pin direction
*
* This routine sets the specified pin's direction.
*
* RETURNS: always OK
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioPinSetDir
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id,
    UINT32          dir
    )
    {
    UINT32             bank;
    UINT32             port;
    UINT32             pin;
    DW_GPIO_DRV_CTRL * pDrvCtrl = (DW_GPIO_DRV_CTRL *)pCtrl;

    pin = id - pCtrl->firstPin;

    port = dwGpioPinConv (&bank, &pin, pDrvCtrl);

    GPIO_DBG (GPIO_DBG_RW, "dwGpioPinSetDir: id %u bank %u port %u pin %u\n", 
              id, bank, port, pin);

    if (dir == GPIO_DIR_OUTPUT)
        {
        CSR_SETBIT_4 (pDrvCtrl, 
                      SWPORT_DDR_OFFSET (bank, pDrvCtrl->bankShift, port),
                      GPIO_PIN_TO_BIT (pin));
        }
    else
        {
        CSR_CLRBIT_4 (pDrvCtrl, 
                      SWPORT_DDR_OFFSET (bank, pDrvCtrl->bankShift, port),
                      GPIO_PIN_TO_BIT (pin));
        }

    return OK;
    }

/*******************************************************************************
*
* dwGpioPinGetValue - get GPIO pin value
*
* This routine gets the specified pin's value.
*
* RETURNS: GPIO pin's value
*
* ERRNO: N/A
*/

LOCAL UINT32 dwGpioPinGetValue
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id
    )
    {
    UINT32             val;
    UINT32             bank;
    UINT32             port;
    UINT32             pin;
    DW_GPIO_DRV_CTRL * pDrvCtrl = (DW_GPIO_DRV_CTRL *)pCtrl;

    pin = id - pCtrl->firstPin;

    port = dwGpioPinConv (&bank, &pin, pDrvCtrl);

    GPIO_DBG (GPIO_DBG_RW, "dwGpioPinGetValue: id %u bank %u port %u pin %u\n", 
              id, bank, port, pin);

    if (dwGpioPinGetDir (pCtrl, id) == GPIO_DIR_INPUT)
        {
        val = CSR_READ_4 (pDrvCtrl, 
                          EXTPORT_IDR_OFFSET (bank, pDrvCtrl->bankShift, port));
        }
    else
        {
        val = CSR_READ_4 (pDrvCtrl, 
                          SWPORT_DR_OFFSET (bank, pDrvCtrl->bankShift, port));
        }

    val &= GPIO_PIN_TO_BIT (pin);
    return ((val != 0) ? GPIO_VALUE_HIGH: GPIO_VALUE_LOW);
    }

/*******************************************************************************
*
* dwGpioPinSetValue - set GPIO pin value
*
* This routine sets the specified pin's value regardless it's output status.
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioPinSetValue
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id,
    UINT32          value
    )
    {
    UINT32             bank;
    UINT32             port;
    UINT32             pin;
    DW_GPIO_DRV_CTRL * pDrvCtrl = (DW_GPIO_DRV_CTRL *)pCtrl;

    pin = id - pCtrl->firstPin;

    port = dwGpioPinConv (&bank, &pin, pDrvCtrl);

    GPIO_DBG (GPIO_DBG_RW, "dwGpioPinSetValue: id %u bank %u port %u pin %u\n",
              id, bank, port, pin);

    /* 
     * May need to checkout the pin is set as output, but this should be rare
     * case. Checking the direction would impact the performance, so it's not
     * done here.
     */

    if (value == GPIO_VALUE_LOW)
        {
        CSR_CLRBIT_4 (pDrvCtrl, 
                      SWPORT_DR_OFFSET (bank, pDrvCtrl->bankShift, port),
                      GPIO_PIN_TO_BIT(pin));
        }
    else
        {
        CSR_SETBIT_4 (pDrvCtrl, 
                      SWPORT_DR_OFFSET (bank, pDrvCtrl->bankShift, port),
                      GPIO_PIN_TO_BIT(pin));
        }

    return OK;
    }

/*******************************************************************************
*
* dwGpioPinSetDebounce - set GPIO pin debounce
*
* This routine sets the specified pin debounce time, because of all GPIO pins in 
* DesignWare GPIO module share the same debounce time, so this routine will,
*   - set pin debounce function when [us] equals to 'debounce-time' specified 
*     in dts
*   - clear pin debounce function when [us] equals to 0
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioPinSetDebounce
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id,
    UINT32          us
    )
    {
    UINT32             bank;
#ifdef  GPIO_DBG_ON
    UINT32             port;
#endif /* GPIO_DBG_ON */
    UINT32             pin;
    DW_GPIO_DRV_CTRL * pDrvCtrl = (DW_GPIO_DRV_CTRL *)pCtrl;

    if (pDrvCtrl->debounceTime == GPIO_INVALID_DEBOUNCE_TIME)
        {
        GPIO_DBG (GPIO_DBG_ERR, 
                  "dwGpioPinSetDebounce: The debounce is not supported\n");
        return ERROR;
        }

    pin = id - pCtrl->firstPin;

#ifdef  GPIO_DBG_ON
    port = dwGpioPinConv (&bank, &pin, pDrvCtrl);
#else /* !GPIO_DBG_ON */
    (void) dwGpioPinConv (&bank, &pin, pDrvCtrl);
#endif /* GPIO_DBG_ON */

    GPIO_DBG (GPIO_DBG_RW, 
              "dwGpioPinSetDebounce: id %u bank %u port %u pin %u\n",
              id, bank, port, pin);

    if (us == 0)
        {
        /* clear debounce register */

        CSR_CLRBIT_4 (pDrvCtrl, 
                      DEBOUNCE_OFFSET (bank, pDrvCtrl->bankShift),
                      GPIO_PIN_TO_BIT (pin));
        }
    else if (us == pDrvCtrl->debounceTime)
        {
        /* set debounce register */

        CSR_SETBIT_4 (pDrvCtrl, 
                      DEBOUNCE_OFFSET (bank, pDrvCtrl->bankShift),
                      GPIO_PIN_TO_BIT (pin));            
        }
    else 
        {
        return ERROR;
        }

    return OK;
    }

/*******************************************************************************
*
* dwGpioIntConfig - configure GPIO pin interrupt
*
* This routine configures GPIO pin interrupt trigger level and polarity.
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioIntConfig
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id,
    INTR_TRIGER     trig,
    INTR_POLARITY   pol
    )
    {
    UINT32             bank;
    UINT32             port;
    UINT32             pin = id - pCtrl->firstPin;
    DW_GPIO_DRV_CTRL * pDrvCtrl = (DW_GPIO_DRV_CTRL *)(pCtrl);

    port = dwGpioPinConv (&bank, &pin, pDrvCtrl);

    GPIO_DBG (GPIO_DBG_RW, 
              "dwGpioIntConfig: id %u bank %u port %u pin %u trig %d pol %d\n",
              id, bank, port, pin, trig, pol);

    if (port != DW_GPIO_PORTA)
        {
        GPIO_DBG (GPIO_DBG_ERR, 
                  "dwGpioIntConfig: The port doesn't support interrupt\n",
                  id, bank, port, pin);
        return ERROR;
        }

    if ((trig == INTR_TRIGGER_CONFORM) && (pol == INTR_POLARITY_CONFORM))
        {
        /* do not change */

        return OK;
        }

    if (trig == INTR_TRIGGER_LEVEL)
        {
        CSR_CLRBIT_4 (pDrvCtrl, 
                      INTTYPELEVEL_OFFSET (bank, pDrvCtrl->bankShift), 
                      GPIO_PIN_TO_BIT(pin));
        }
    else if (trig == INTR_TRIGGER_EDGE)
        {
        CSR_SETBIT_4 (pDrvCtrl, 
                      INTTYPELEVEL_OFFSET (bank, pDrvCtrl->bankShift), 
                      GPIO_PIN_TO_BIT(pin));
        }
    else
        {
        GPIO_DBG (GPIO_DBG_ERR, 
                  "dwGpioIntConfig: The trigger mode is not supported\n",
                  id, bank, port, pin);
        return ERROR;
        }

    if (pol == INTR_POLARITY_HIGH)
        {
        CSR_SETBIT_4 (pDrvCtrl, 
                      INTPOLARITY_OFFSET (bank, pDrvCtrl->bankShift), 
                      GPIO_PIN_TO_BIT(pin));
        }
    else if (pol == INTR_POLARITY_LOW)
        {
        CSR_CLRBIT_4 (pDrvCtrl, 
                      INTPOLARITY_OFFSET (bank, pDrvCtrl->bankShift), 
                      GPIO_PIN_TO_BIT(pin));
        }
    else
        {
        GPIO_DBG (GPIO_DBG_ERR, 
                  "dwGpioIntConfig: The polarity is not supported\n",
                  id, bank, port, pin);
        return ERROR;
        }

    return OK;
    }

/*******************************************************************************
*
* dwGpioISR - GPIO interrupt handler
*
* This routine is GPIO interrupt handler.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

LOCAL void dwGpioISR
    (
    DW_GPIO_DRV_CTRL * pDrvCtrl
    )
    {
    DW_GPIO_INT_NODE *  pNode;
    UINT32              tmp;
    UINT32              mask = 0;
    UINT32              pin;
    UINT32              triggerMode;

    if (pDrvCtrl == NULL)
        {
        GPIO_DBG (GPIO_DBG_ERR, "dwGpioISR: The pDrvCtrl is NULL\n");
        return;
        }

    /* this ISR only works on single bank */

    tmp = CSR_READ_4 (pDrvCtrl, INTSTATUS_OFFSET (0, DW_GPIO_PORTA));
    if (tmp == 0)
        {
        GPIO_DBG (GPIO_DBG_ERR, "dwGpioISR: suspicious interrupt\n");
        return;
        }

    GPIO_DBG (GPIO_DBG_IRQ, "dwGpioISR: Entering...\n");

    /* save current mask status */

    mask = CSR_READ_4 (pDrvCtrl, INTMASK_OFFSET (0, DW_GPIO_PORTA));
    triggerMode = CSR_READ_4 (pDrvCtrl, INTTYPELEVEL_OFFSET (0, DW_GPIO_PORTA));

    /* mask all gpio pins in this bank */

    CSR_WRITE_4 (pDrvCtrl, INTMASK_OFFSET (0, DW_GPIO_PORTA), 
                 (1U << pDrvCtrl->portWidth) - 1U);

    for (pin = 0; pin < pDrvCtrl->portWidth; pin++)
        {
        if ((tmp & GPIO_PIN_TO_BIT (pin)) != 0)
            {
            pNode = pDrvCtrl->pIntNode[pin];

            if ((triggerMode & GPIO_PIN_TO_BIT (pin)) != 0)
                {
                /* edge trigger, clear first to allow subsequent interrupt */

                CSR_WRITE_4 (pDrvCtrl, EOI_OFFSET (0, DW_GPIO_PORTA),
                             GPIO_PIN_TO_BIT (pin));
                }

            if ((pNode != NULL) && (pNode->pIsr != NULL))
                {
                pNode->pIsr(pNode->pArg);
                }
            else
                {
                /* suspicious int */

                GPIO_DBG (GPIO_DBG_RW, "dwGpioISR: suspicious int %x. \n.", pin);
                }

            if ((triggerMode & GPIO_PIN_TO_BIT (pin)) == 0)
                {
                /* 
                 * Edge trigger, clear after ISR to avoid not clear interrupt 
                 * pending again.
                 */

                CSR_WRITE_4 (pDrvCtrl, EOI_OFFSET (0, DW_GPIO_PORTA), 
                GPIO_PIN_TO_BIT (pin));
                }
            }
        }

    CSR_WRITE_4 (pDrvCtrl, INTMASK_OFFSET (0, DW_GPIO_PORTA), mask);
    }

/******************************************************************************
*
* dwGpioExclusiveISR - GPIO interrupt handler for exclusive interrupt
*
* This function is defined for GPIO interrupt handler that 1 GPIO pin exclusively
* occupies 1 interrupt.
*
* RETURNS: N/A
*
* ERRNO
*/

LOCAL void dwGpioExclusiveISR
    (
    DW_GPIO_INT_NODE * pNode
    )
    {
    DW_GPIO_INT_NODE_EXT * pNodeExt;
    DW_GPIO_DRV_CTRL *     pDrvCtrl;
    UINT32                 bank;
    UINT32                 pin;

    if (pNode == NULL)
        return;

    pNodeExt = pNode->intExt;
    pDrvCtrl = pNodeExt->pDrvCtrl;
    pin = pNodeExt->gpioPin;

    GPIO_DBG (GPIO_DBG_RW, "%s: pNodeExt %p, pDrvCtrl %p, pin %d.\n", 
                          __FUNCTION__, pNodeExt, pDrvCtrl, pin);

    bank = pin / pDrvCtrl->portWidth;
    pin  = pin % pDrvCtrl->portWidth;

    if (pNodeExt->triggerMode == INTR_TRIGGER_EDGE)
        {
        CSR_WRITE_4 (pDrvCtrl, EOI_OFFSET (bank, pDrvCtrl->bankShift), 
                     GPIO_PIN_TO_BIT (pin));
        GPIO_DBG (GPIO_DBG_RW, "%s: edge interrupt %d EOI.\n", 
                          __FUNCTION__, pin);
        }

    if (pNode->pIsr != NULL)
        {
        pNode->pIsr (pNode->pArg);
        }

    if (pNodeExt->triggerMode == INTR_TRIGGER_LEVEL)
        {
        CSR_WRITE_4 (pDrvCtrl, EOI_OFFSET (bank, pDrvCtrl->bankShift), 
                     GPIO_PIN_TO_BIT (pin));
         GPIO_DBG (GPIO_DBG_RW, "%s: level interrupt %d EOI.\n", 
                          __FUNCTION__, pin);
        }

    return;
    }

/*******************************************************************************
*
* dwGpioIntConnect - connect GPIO interrupt
*
* This routine connects GPIO interrupt to user supplied ISR.
*
* RETURNS: OK, or ERROR if there is error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioIntConnect
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id,
    VOIDFUNCPTR     pIsr,           /* ISR */
    void *          pArg            /* parameter */
    )
    {
    DW_GPIO_DRV_CTRL *     pDrvCtrl;
    DW_GPIO_INT_NODE *     pNode;
    DW_GPIO_INT_NODE_EXT * pNodeExt;
    UINT32                 bank;
    UINT32                 readPin;
    UINT32                 globalPin;
    UINT32                 pin;
    UINT32                 i;

    pDrvCtrl = (DW_GPIO_DRV_CTRL *)pCtrl;

    if (pDrvCtrl->intNum == 0)
        {
        GPIO_DBG (GPIO_DBG_ERR, "%s: pDrvCtrl intNum is 0.\n", __FUNCTION__);
        return ERROR; /* does not support interrupt */
        }

    pin = id - pCtrl->firstPin;

    if (pDrvCtrl->pMapTable != NULL)
        {
        globalPin = (UINT32)pDrvCtrl->pMapTable[pin];
        }
    else
        {
        globalPin = pin;
        }

    if (globalPin >= pDrvCtrl->portANum)
        {
        GPIO_DBG (GPIO_DBG_ERR, "%s: pDrvCtrl GPIO pin(%d) is not in Port A.\n",
                  __FUNCTION__, pin);
        return ERROR; /* this gpio pin doesn't support interrupt */
        }

    (void) dwGpioPinConv (&bank, &pin, pDrvCtrl);

    if (pDrvCtrl->intNum == 1)
        {
        /* 
         * All GPIO pins share the same interrupt at the same time, the ISR needs
         * to find which pins cause interrupt.
         */

        GPIO_DBG (GPIO_DBG_RW, "%s: 1 interrupt per bank.\n", 
                  __FUNCTION__);

        if ((pNode = pDrvCtrl->pIntNode[pin]) == NULL)
            {
            pNode = vxbMemAlloc (sizeof (DW_GPIO_INT_NODE));
            if (pNode == NULL)
                return ERROR;
            
            pDrvCtrl->pIntNode[pin] = pNode;
            }
        else
            {
            /* do not connect twice */

            GPIO_DBG (GPIO_DBG_ERR,
                "dwGpioIntConnect: %d already connected.\n", id);
            return ERROR;
            }
        }
    else
        {
        /*
         * all GPIO pins share several interrupts, but 1 interrupt can only be  
         * connected by 1 pin at a time. So the pin is determined when 
         * connected. 
         */

        GPIO_DBG (GPIO_DBG_RW, "%s: %d interrupts per controller.\n", 
                  __FUNCTION__, pDrvCtrl->intNum);

        if (pDrvCtrl->intCfgBar == NULL)
            {
            GPIO_DBG (GPIO_DBG_ERR, "%s: intCfgBar is NULL.\n", 
                      __FUNCTION__);
            return ERROR;
            }

        for (i = 0; i < pDrvCtrl->intNum; i++)
            {
            readPin = INT_CFG_READ_4 (pDrvCtrl, DW_GPIO_INTCFG (i));
            if (readPin == globalPin)
                {
                GPIO_DBG (GPIO_DBG_RW, "%s: the %d int match.\n", 
                          __FUNCTION__, i);
                break;
                }
            }

        if (i >= pDrvCtrl->intNum)
            {
            GPIO_DBG (GPIO_DBG_ERR, "%s: No int match.\n", 
                      __FUNCTION__);
            return ERROR;
            }

        pNode = pDrvCtrl->pIntNode[i];
        pNodeExt = pNode->intExt;

        if ((CSR_READ_4 (pDrvCtrl, INTTYPELEVEL_OFFSET (bank, 
                                                         pDrvCtrl->bankShift)) &
             GPIO_PIN_TO_BIT (pin)) != 0)
            {
            pNodeExt->triggerMode = INTR_TRIGGER_EDGE;
            GPIO_DBG (GPIO_DBG_RW, "%s: edge trigger.\n", 
                      __FUNCTION__);
            }
        else
            {
            pNodeExt->triggerMode = INTR_TRIGGER_LEVEL;
            GPIO_DBG (GPIO_DBG_RW, "%s: level trigger.\n", 
                      __FUNCTION__);
            }

        pNodeExt->gpioPin = globalPin;
        }

    pNode->bEnabled = FALSE;
    pNode->pIsr = pIsr;
    pNode->pArg = pArg;

    return OK;
    }

/*******************************************************************************
*
* dwGpioIntDisConnect - disconnect GPIO interrupt
*
* This routine disconnects GPIO interrupt from user supplied ISR.
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioIntDisConnect
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id,
    VOIDFUNCPTR     pIsr,           /* ISR */
    void *          pArg            /* parameter */
    )
    {
    DW_GPIO_DRV_CTRL * pDrvCtrl;
    DW_GPIO_INT_NODE * pNode;
    UINT32             pin;
    UINT32             i;

    pDrvCtrl = (DW_GPIO_DRV_CTRL *) pCtrl;

    pin = id - pCtrl->firstPin;

    if (pDrvCtrl->pMapTable != NULL)
        {
        pin = (UINT32)pDrvCtrl->pMapTable[pin];
        }

    if (pDrvCtrl->intNum == 1)
        {
        if ((pNode = pDrvCtrl->pIntNode[pin]) == NULL)
            {
            /* not connected yet */

            GPIO_DBG (GPIO_DBG_ERR,
                "Interrupt for %d had not been connected yet.\n.",
                id);
            return ERROR;
            }

        pDrvCtrl->pIntNode[pin] = NULL;
        vxbMemFree(pNode);
        }
    else
        {
        DW_GPIO_INT_NODE_EXT *     pNodeExt;

        for (i = 0; i < pDrvCtrl->intNum; i++)
            {
            pNode = pDrvCtrl->pIntNode[i];
            pNodeExt = pNode->intExt;
            if (pin == pNodeExt->gpioPin)
                {
                break;
                }
            }

        if (i < pDrvCtrl->intNum)
            {
            pNode->pIsr = NULL;
            pNode->pArg = NULL;
            }
        else
            {
            return ERROR;
            }
        }

    return OK;
    }

/*******************************************************************************
*
* dwGpioIntEnable - enable GPIO interrupt
*
* This routine enables GPIO interrupt.
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioIntEnable
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id,
    VOIDFUNCPTR     pIsr,           /* ISR */
    void *          pArg            /* parameter */
    )
    {
    DW_GPIO_DRV_CTRL * pDrvCtrl;
    DW_GPIO_INT_NODE * pNode;
    UINT32             pin;
    UINT32             globalPin;
    UINT32             bank;
    UINT32             i;

    pDrvCtrl = (DW_GPIO_DRV_CTRL *) pCtrl;
    pin = id - pCtrl->firstPin;

    if (pDrvCtrl->pMapTable != NULL)
        globalPin = pDrvCtrl->pMapTable[pin];
    else
        globalPin = pin;

    if (globalPin >= pDrvCtrl->portANum)
        {
        GPIO_DBG (GPIO_DBG_ERR, "dwGpioIntEnable: pin %u not in Port A.\n", globalPin);
        return ERROR;
        }

    if (pDrvCtrl->intNum == 1)
        {
        if ((pNode = pDrvCtrl->pIntNode[pin]) == NULL)
            {
            /* not connected yet */

            GPIO_DBG (GPIO_DBG_ERR, "dwGpioIntEnable: %u not connected.\n", id);
            return ERROR;
            }
        }
    else
        {
        DW_GPIO_INT_NODE_EXT * pNodeExt;

        for (i = 0; i < pDrvCtrl->intNum; i++)
            {
            pNode = pDrvCtrl->pIntNode[i];
            pNodeExt = pNode->intExt;
            if (globalPin == pNodeExt->gpioPin)
                {
                break;
                }
            }
        if (i >= pDrvCtrl->intNum)
            {
            return ERROR; /* didn't find connected interrupt */
            }
        }

    if (pNode->bEnabled)
        {
        return OK;
        }

    (void) dwGpioPinConv (&bank, &pin, pDrvCtrl);

    /* set interrupt enable register */

    CSR_SETBIT_4(pDrvCtrl, INTEN_OFFSET (bank, pDrvCtrl->bankShift), 
                 GPIO_PIN_TO_BIT(pin));
    CSR_CLRBIT_4(pDrvCtrl, INTMASK_OFFSET (bank, pDrvCtrl->bankShift), 
                 GPIO_PIN_TO_BIT(pin));

    pNode->bEnabled = TRUE;

    return OK;
    }

/*******************************************************************************
*
* dwGpioIntDisable - disable GPIO interrupt
*
* This routine disables GPIO interrupt.
*
* RETURNS: OK, or ERROR if there is an error
*
* ERRNO: N/A
*/

LOCAL STATUS dwGpioIntDisable
    (
    VXB_GPIOCTRL *  pCtrl,
    UINT32          id,
    VOIDFUNCPTR     pIsr,           /* ISR */
    void *          pArg            /* parameter */
    )
    {
    DW_GPIO_DRV_CTRL * pDrvCtrl;
    DW_GPIO_INT_NODE * pNode;
    UINT32             pin = (UINT32)(id - pCtrl->firstPin);
    UINT32             globalPin;
    UINT32             bank;
    UINT32             i;

    pDrvCtrl = (DW_GPIO_DRV_CTRL *) pCtrl;
    pin = id - pCtrl->firstPin;

    if (pDrvCtrl->pMapTable != NULL)
        globalPin = (UINT32)pDrvCtrl->pMapTable[pin];
    else
        globalPin = pin;
    
    if (globalPin >= pDrvCtrl->portANum)
        {
        GPIO_DBG (GPIO_DBG_ERR, "dwGpioIntDisable: pin %d not in Port A.\n",
                  globalPin);
        return ERROR;
        }

    if (pDrvCtrl->intNum == 1)
        {
        if ((pNode = pDrvCtrl->pIntNode[pin]) == NULL)
            {
            /* not connected yet */

            GPIO_DBG (GPIO_DBG_ERR, "dwGpioIntDisable: %d not connected.\n",
                      id);
            return ERROR;
            }
        }
    else
        {
        DW_GPIO_INT_NODE_EXT * pNodeExt;

        for (i = 0; i < pDrvCtrl->intNum; i++)
            {
            pNode = pDrvCtrl->pIntNode[i];
            pNodeExt = pNode->intExt;
            if (globalPin == pNodeExt->gpioPin)
                break;
            }

        if (i >= pDrvCtrl->intNum)
            {
            return ERROR;
            }
        }

    if (!pNode->bEnabled)
        {
        /* already disabled */

        return OK;
        }

    pNode->bEnabled = FALSE;

    (void) dwGpioPinConv (&bank, &pin, pDrvCtrl);

    CSR_SETBIT_4(pDrvCtrl, INTMASK_OFFSET (bank, pDrvCtrl->bankShift), 
                 GPIO_PIN_TO_BIT(pin));

    /* clear interrupt enable register */

    CSR_CLRBIT_4(pDrvCtrl, INTEN_OFFSET (bank, pDrvCtrl->bankShift), 
                 GPIO_PIN_TO_BIT(pin));

    return OK;
    }

/*******************************************************************************
*
* dwGpioPinConv - convert the pin into bank, port and pin within the port
*
* This routine convert the pin into bank, port and pin within the port.
*
* RETURNS: Gpio's port
*
* ERRNO: N/A
*/

LOCAL __inline__ UINT32 dwGpioPinConv (UINT32 * bank, UINT32 * pin,
                                       DW_GPIO_DRV_CTRL * pDrvCtrl)
    {
    if (pDrvCtrl->pMapTable != NULL)
        *pin = pDrvCtrl->pMapTable[*pin];

    if (*pin < pDrvCtrl->portANum)
        {
        *bank = *pin / pDrvCtrl->portWidth;
        *pin  = *pin % pDrvCtrl->portWidth;
        return (UINT32)DW_GPIO_PORTA;
        }
    else
        {
        *pin  -= pDrvCtrl->portANum;
        *bank = *pin / pDrvCtrl->portWidth;
        *pin  = *pin % pDrvCtrl->portWidth;
        return (UINT32)DW_GPIO_PORTB;
        }
    }

#ifdef _WRS_CONFIG_DEBUG_FLAG
/*******************************************************************************
*
* dwGpioShow - show GPIO information
*
* This routine shows GPIO information.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

LOCAL void dwGpioShow
    (
    VXB_GPIOCTRL * pCtrl,
    UINT32         verbose
    )
    {
    DW_GPIO_DRV_CTRL * pDrvCtrl;
    UINT32             bank;
    UINT32             port;

    pDrvCtrl = (DW_GPIO_DRV_CTRL *) pCtrl;
    if (pDrvCtrl == NULL)
        return;

    if (verbose == 0)
        return;

    (void)printf("DesignWare compatible GPIO controller information:\n");

    for (bank = 0; bank < pDrvCtrl->bankNum; bank++)
        {
        (void)printf("\n================Bank %d information============\n", bank);
        (void)printf("IDCODE_OFFSET:      0x%x\n",
                     CSR_READ_4 (pDrvCtrl, 
                                 IDCODE_OFFSET (bank, pDrvCtrl->bankShift)));
        (void)printf("COMPVERSION_OFFSET: 0x%x\n",
                     CSR_READ_4 (pDrvCtrl, 
                                 COMPVERSION_OFFSET (bank, pDrvCtrl->bankShift)));
        (void)printf("CONFIGREG1_OFFSET:  0x%x\n",
                     CSR_READ_4 (pDrvCtrl, 
                                 CONFIGREG1_OFFSET (bank, pDrvCtrl->bankShift)));
        (void)printf("CONFIGREG2_OFFSET:  0x%x\n",
                     CSR_READ_4 (pDrvCtrl, 
                                 CONFIGREG2_OFFSET(bank, pDrvCtrl->bankShift)));

        for (port = DW_GPIO_PORTA; port <= DW_GPIO_PORTB; port += DW_GPIO_PORTB)
            {
            (void)printf("\n==============Port %c information==========\n", 
                         (port == DW_GPIO_PORTA)?'A':'B');
            (void)printf("SWPORT_DR_OFFSET:    0x%x\n",
                         CSR_READ_4 (pDrvCtrl, SWPORT_DR_OFFSET (bank, 
                                               pDrvCtrl->bankShift, port)));
            (void)printf("SWPORT_DDR_OFFSET:   0x%x\n",
                         CSR_READ_4 (pDrvCtrl, SWPORT_DDR_OFFSET (bank, 
                                               pDrvCtrl->bankShift, port)));
            (void)printf("EXTPORT_OFFSET:      0x%x\n",
                         CSR_READ_4 (pDrvCtrl, EXTPORT_IDR_OFFSET (bank, 
                                               pDrvCtrl->bankShift, port)));

            if (port != DW_GPIO_PORTA)
                {
                continue;
                }

            (void)printf("INTEN_OFFSET:        0x%x\n",
                         CSR_READ_4 (pDrvCtrl, INTEN_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));
            (void)printf("INTMASK_OFFSET:      0x%x\n",
                         CSR_READ_4 (pDrvCtrl, INTMASK_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));
            (void)printf("INTTYPELEVEL_OFFSET: 0x%x\n",
                         CSR_READ_4 (pDrvCtrl, INTTYPELEVEL_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));
            (void)printf("INTPOLARITY_OFFSET:  0x%x\n",
                         CSR_READ_4 (pDrvCtrl, INTPOLARITY_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));
            (void)printf("INTSTATUS_OFFSET:    0x%x\n",
                         CSR_READ_4 (pDrvCtrl, INTSTATUS_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));
            (void)printf("RAWINTSTATUS_OFFSET: 0x%x\n",
                         CSR_READ_4 (pDrvCtrl, RAWINTSTATUS_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));

            (void)printf("DEBOUNCE_OFFSET:     0x%x\n",
                         CSR_READ_4 (pDrvCtrl, DEBOUNCE_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));
            (void)printf("EOI_OFFSET:          0x%x\n",
                         CSR_READ_4 (pDrvCtrl, EOI_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));
            
            (void)printf("LSSYNC_OFFSET:       0x%x\n",
                         CSR_READ_4 (pDrvCtrl, LSSYNC_OFFSET (bank, 
                                               pDrvCtrl->bankShift)));
            }
        }
    }

#endif /* _WRS_CONFIG_DEBUG_FLAG */
