/* vxbNs16550Sio.c - National Semiconductor 16550 Serial Driver */

/*
 * Copyright (c) 2013-2021 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */
/*
DESCRIPTION

This is the VxBus driver for the NS16550 (or compatible) UART.
When packaged as pair of devices, also called a DUART (dual
universal asynchronous receiver-transmitter)
In general, this device includes an asynchronous
receiver/transmitter, a baud rate generator, and
modem control capability.

Only asynchronous serial operation is supported by this driver.
Modem control is available though the ioctl() interface.
The default serial settings are 8 data bits, 1 stop bit, no parity,
and software flow control.

TARGET-SPECIFIC PARAMETERS

This driver uses either the vxbFdtNs16550Sio adaptor for devices described in
the Flattened Device Tree, the vxbPciNs16550Sio platform adaptor for devices on
the PCI bus and/or the vxbIaNs16550Sio adaptor in the Intel PSL to instantiate
devices found in ACPI tables. The adaptor provides the vxBus vxbDevProbe() and
vxbDevAttach() methods to initialize and configure the driver.

All adaptor libraries provide this library with an interrupt and clock
resource handle, and the following parameters: 
\is

\i <regDelta>
The amount of address space between successive 8-bit registers.
If not specified, 1 byte will be used as the register interval.

\i <fifoSize>
The depth of the FIFO on this device.
If not specified and the device supports FIFO, the FIFO depth is set 16 bytes.
If not specified and the device doesn't support FIFO, the FIFO depth is set 1
byte.

\i <baud>
The initial baud rate.

\i <xtal>
The clock divisor for the baud rate.

\i <bar>
The base address in memory of the device. 

\ie

CONFIGURATION
To use the NS16550 SIO driver, configure VxWorks with the
DRV_SIO_FDT_NS16550, DRV_SIO_PCI_NS16550 or DRV_SIO_IA_NS16550 component.
The required components for this driver are
INCLUDE_VXBUS
INCLUDE_ISR_DEFER

INCLUDE FILES: vxbNs16550Sio.h

SEE ALSO: vxbFdtNs16550Sio, vxbPciNs16550Sio, tyLib
*/

/* includes */

#include <vxWorks.h>
#include <vsbConfig.h>
#include <intLib.h>
#include <errnoLib.h>
#include <errno.h>
#include <sioLib.h>
#include <ioLib.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ttyLib.h>

#ifdef    _WRS_CONFIG_SMP
#include <private/spinLockLibP.h>
#endif    /* _WRS_CONFIG_SMP */

#include <hwif/vxBus.h>
#include <subsys/int/vxbIntLib.h>
#include <hwif/drv/sio/vxbSioUtil.h>
#include <hwif/drv/sio/vxbNs16550Sio.h>

/*
 * When power management is included in the kernel, there is a potential of missing
 * an interrupt when returning from power states. The serial service function needs
 * to be called twice for each character typed on the console. This is for the
 * character RX'ed and the TX empty when the character is echo'd to the console.
 * This fix takes care of the problem by calling the service loop at least twice, in
 * case we miss one of the two interrupts.
 */

#define NS16550_INT_COUNT 2

#undef NS16550_DEBUG_ON
#ifdef NS16550_DEBUG_ON

/*
 * NOTE: when the NS16550_DEBUG_ON is on, the serial console have the risk of
 * deadlock. The logic is "the driver --> logMsg() --> driver". There is a msgQSend
 * in function logMsg, and the value of parameter timeout is WAIT_FOREVER. So when
 * the logMsg queue is full, the deadlock happens. To avoid it, we can define the
 * MAX_LOG_MSGS to a bigger value.
 */

/*
 * NOTE: printf() and logMsg() debugging cannot be used before the
 * call to ns16550vxbInstConnect().  Any use before that may cause the
 * system to crash before finishing the boot process.  To debug
 * the probe function, init function, and init2 function, either
 * use a hardware debugger, move the specified functionality to
 * ns16550vxbInstConnect(), or manually call the function after boot.
 */
LOCAL int ns16550vxbDebugLevel = 1000;

#ifndef NS16550_DBG_MSG
#define NS16550_DBG_MSG(level,fmt,a,b,c,d,e,f) \
        if ( ns16550vxbDebugLevel >= level ) \
            logMsg(fmt,a,b,c,d,e,f)
#endif /* NS16550_DBG_MSG */

#else /* NS16550_DEBUG_ON */

#define NS16550_DBG_MSG(level,fmt,a,b,c,d,e,f)

#endif /* NS16550_DEBUG_ON */

/* forward declarations of method functions */

void ns16550vxbSioChanGet(VXB_DEV_ID pDev, void * pArg);

IMPORT int vxbSioNextGet(void);
IMPORT void vxbUsDelay(int    counter);

/* local defines       */

#ifdef    _WRS_CONFIG_SMP
LOCAL BOOL ns16550SpinlockFuncReady = FALSE;

#define VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan)                \
    if (ns16550SpinlockFuncReady)                    \
    SPIN_LOCK_ISR_TAKE(&pChan->spinlockIsr)
#define VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan)                \
    if (ns16550SpinlockFuncReady)                    \
    SPIN_LOCK_ISR_GIVE(&pChan->spinlockIsr)
#define VXB_NS16550_SPIN_LOCK_ISR_HELD(pChan)                \
    (ns16550SpinlockFuncReady ? spinLockIsrHeld(&pChan->spinlockIsr) : FALSE)
#define VXB_NS16550_ISR_SET(pChan)                      \
    SPIN_LOCK_ISR_TAKE(&pChan->spinlockIsr)
#define VXB_NS16550_ISR_CLEAR(pChan)                    \
    SPIN_LOCK_ISR_GIVE(&pChan->spinlockIsr)
#define VXB_NS16550_SPIN_LOCK_READY                    \
    ns16550SpinlockFuncReady = TRUE
#else
#define VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan)                \
    SPIN_LOCK_ISR_TAKE(&pChan->spinlockIsr)
#define VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan)                \
    SPIN_LOCK_ISR_GIVE(&pChan->spinlockIsr)
#define VXB_NS16550_SPIN_LOCK_ISR_HELD(pChan)    FALSE
#define VXB_NS16550_ISR_SET(pChan)
#define VXB_NS16550_ISR_CLEAR(pChan)
#define VXB_NS16550_SPIN_LOCK_READY
#endif    /* _WRS_CONFIG_SMP */

#define REG_READ(reg, pchan)                      \
    if (pchan->flag & STRICT_32BIT_ACCESS)        \
        (void) vxbRead32 (pchan->handle,          \
        (UINT32 *) ((UINT8 *)pchan->bar +         \
        (reg * pchan->regDelta)));                \
    else                                          \
        (void) vxbRead8 (pchan->handle,           \
        (UINT8 *)pchan->bar +                     \
        (reg * pchan->regDelta));

#define REG_GET(reg, pchan, val)                  \
    if (pchan->flag & STRICT_32BIT_ACCESS)        \
        val = (UINT8) vxbRead32 (pchan->handle,   \
        (UINT32 *) ((UINT8 *)pchan->bar +         \
        (reg * pchan->regDelta)));                \
    else                                          \
        val = vxbRead8 (pchan->handle,            \
        (UINT8 *)pchan->bar +                     \
        (reg * pchan->regDelta));

#define REG_SET(reg, pchan, val)                  \
    if (pchan->flag & STRICT_32BIT_ACCESS)        \
        vxbWrite32 (pchan->handle,                \
            (UINT32 *) ((UINT8 *)pchan->bar +     \
            (reg * pchan->regDelta)), (UINT32)(val));       \
    else                                          \
        vxbWrite8 (pchan->handle,                 \
            (UINT8 *)pchan->bar +                 \
            (reg * pchan->regDelta), (UINT8)(val));

/* static forward declarations */

#ifndef    _VXBUS_BASIC_SIOPOLL
LOCAL   int     ns16550vxbCallbackInstall
    (
    SIO_CHAN *,
    int, STATUS (*)(),
    void *
    );
LOCAL   int     ns16550vxbTxStartup (NS16550VXB_CHAN *);
LOCAL   void    ns16550vxbIntRd2 (void * pData);
LOCAL   void    ns16550vxbIntWr2 (void * pData);
STATUS  ns16550vxbInt (VXB_DEV_ID pDev);
#endif    /* _VXBUS_BASIC_SIOPOLL */

LOCAL   STATUS  ns16550vxbDevInit (NS16550VXB_CHAN * pChan);
LOCAL   STATUS  ns16550vxbDummyCallback ();
LOCAL   STATUS  ns16550vxbInitChannel (NS16550VXB_CHAN *);
LOCAL   STATUS  ns16550vxbOpen (NS16550VXB_CHAN * pChan );
LOCAL   STATUS  ns16550vxbHup (NS16550VXB_CHAN * pChan );
LOCAL   STATUS  ns16550vxbModeSet (NS16550VXB_CHAN *, UINT);
LOCAL   STATUS  ns16550vxbOptsSet (NS16550VXB_CHAN *, UINT);
LOCAL   STATUS  ns16550vxbBaudSet (NS16550VXB_CHAN *, UINT);
LOCAL   STATUS  ns16550vxbIoctl (NS16550VXB_CHAN *, int, void *);
LOCAL   int     ns16550vxbPollInput (NS16550VXB_CHAN *, char *);
LOCAL   int     ns16550vxbPollOutput (NS16550VXB_CHAN *, char);

/* driver functions */

static SIO_DRV_FUNCS ns16550vxbSioDrvFuncs =
    {
    (int (*)())ns16550vxbIoctl,
#ifdef    _VXBUS_BASIC_SIOPOLL
    (int (*)())NULL,
    (int (*)())NULL,
#else    /* _VXBUS_BASIC_SIOPOLL */
    (int (*)())ns16550vxbTxStartup,
    (int (*)())ns16550vxbCallbackInstall,
#endif    /* _VXBUS_BASIC_SIOPOLL */
    (int (*)(SIO_CHAN *,char *))ns16550vxbPollInput,
    (int (*)(SIO_CHAN *,char))ns16550vxbPollOutput
    };


/******************************************************************************
*
* ns16550DrainFifo - drain the fifo
*
* This function drain the fifo.
*
* RETURNS: OK or ERROR when initialize failed
*
* ERRNO
*/

LOCAL void ns16550DrainFifo
    (
    NS16550VXB_CHAN *  pChan
    )
    {
    UINT8 fcrVal;

    if (pChan->drainFifo)
        {
        REG_GET (FCR, pChan, fcrVal);

        /* clear the FIFO */

        REG_SET (FCR, pChan, (fcrVal | FIFO_ENABLE));
        REG_SET (FCR, pChan, (fcrVal | RxCLEAR | TxCLEAR | FIFO_ENABLE));
        (void) vxbRead8 (pChan->handle, (UINT8 *) pChan->bar + RBR);

        /* retore the value of FCR */

        REG_SET(FCR, pChan, fcrVal);
        }
    }

/******************************************************************************
*
* ns16550Create - create ns16550-compatible channel device
*
* This is the ns16550vxb initialization function. It is called from the
* the vxbDevAttach method of the bus adapter library.
*
* RETURNS: OK or ERROR when initialize failed
*
* \NOMANUAL
*/

STATUS ns16550Create
    (
    NS16550VXB_CHAN *  pChan
    )
    {
#ifdef _WRS_CONFIG_CORE__IO_V2
    int err;
#endif

    /* get the channel number */

    pChan->channelNo = vxbSioNextGet();

    /* fetch local copies of IER, LCR, and MCR */

    REG_GET(IER, pChan, pChan->ier);

    REG_GET(LCR, pChan, pChan->lcr);
    REG_GET(MCR, pChan, pChan->mcr);

    /* add ier initial bits in case the bits are cleared at reset */

    pChan->ier |= pChan->ierInit;

#ifndef    _VXBUS_BASIC_SIOPOLL
    pChan->isrDefRd.func = ns16550vxbIntRd2;
    pChan->isrDefRd.pData = pChan;

    pChan->isrDefWr.func = ns16550vxbIntWr2;
    pChan->isrDefWr.pData = pChan;
#endif    /* _VXBUS_BASIC_SIOPOLL */

    /* Initialize spinlock. */

    SPIN_LOCK_ISR_INIT (&pChan->spinlockIsr, 0);

    if (ns16550vxbDevInit(pChan) != OK)
        {
        return ERROR;
        }

    VXB_NS16550_SPIN_LOCK_READY;

#ifdef _WRS_CONFIG_CORE__IO_V2
    err = usrTtyDevCreate (TTY_DEV_ONBOARD, &pChan->ttyIndex,
                           pChan->tyName, (SIO_CHAN *)pChan,
                           VXB_SIO_RX_BUF_SIZE, VXB_SIO_TX_BUF_SIZE,
                           pChan->channelNo);
    if (err != OK)
        {
        return ERROR;
        }
#else
    pChan->ttyIndex = pChan->channelNo;
    (void) snprintf_s (pChan->tyName, TTY_MAX_NAME_LENGTH, "%s%d", "/tyCo/",
                     pChan->channelNo);
    if (ttyDevCreate (pChan->tyName, (SIO_CHAN *)pChan,
                      VXB_SIO_TX_BUF_SIZE, VXB_SIO_TX_BUF_SIZE) != OK)
        {
        return ERROR;
        }
#endif /* _WRS_CONFIG_CORE__IO_V2 */

    return OK;
    }

/******************************************************************************
*
* ns16550Shutdown - shutdown ns16550vxb device
*
* This method disables the ns16550 device
*
* RETURNS: OK or ERROR when shutdown failed
*
* \NOMANUAL
*/

STATUS ns16550Shutdown
    (
    struct vxbDev * pDev
    )
    {
    NS16550VXB_CHAN * pChan;

    pChan = (NS16550VXB_CHAN *)vxbDevSoftcGet(pDev);

    while ( pChan != NULL )
        {
        (void) ns16550vxbDevInit(pChan);

        (void) vxbIntDisable(pDev,  pChan->intRes);

        pChan = pChan->pNext;
        }

    return OK;
    }

/******************************************************************************
*
* ns16550Detach - detach ns16550vxb device
*
* This method detaches the ns16550 driver from the device
*
* RETURNS: OK or ERROR when detach failed
*
* \NOMANUAL
*/

STATUS ns16550Detach
    (
    struct vxbDev * pDev
    )
    {
    NS16550VXB_CHAN * pChan, * pTmpChan;

    pChan = (NS16550VXB_CHAN *)vxbDevSoftcGet(pDev);

    while ( pChan != NULL )
        {
        if (pChan->adrRes != NULL)
            (void) vxbResourceFree(pDev, pChan->adrRes);

        if (pChan->intRes != NULL)
            {
            (void) vxbIntDisconnect(pDev, pChan->intRes);
            (void) vxbResourceFree(pDev, pChan->intRes);
            }

        usrTtyDevLegacyLinkRemove (pChan->channelNo);

        (void) ttyDevDelete(pChan->tyName);

        (void) ttyDevIndexFree(pChan->ttyIndex, TTY_DEV_ONBOARD);

        pTmpChan = pChan;

        pChan = pChan->pNext;

        vxbMemFree((char *)pTmpChan);
        }

    return OK;
    }

/******************************************************************************
*
* ns16550vxbDummyCallback - dummy callback function.
*
* This function is a dummy callback function.
*
* RETURNS: ERROR, always
*
* ERRNO
*/

LOCAL STATUS ns16550vxbDummyCallback (void)
    {
    return(ERROR);
    }

/******************************************************************************
*
* ns16550vxbDevInit - initialize an NS16550 channel
*
* This function initializes some SIO_CHAN function pointers and then resets
* the chip in a quiescent state.  Before this function is called, the BSP
* must already have initialized all the device addresses, etc. in the
* NS16550VXB_CHAN structure.
*
* RETURNS: OK, or ERROR if initialize failed
*
* ERRNO
*/

LOCAL STATUS ns16550vxbDevInit
    (
    NS16550VXB_CHAN * pChan    /* pointer to channel */
    )
    {
    /* set the non BSP-specific constants */

    pChan->getTxChar    = ns16550vxbDummyCallback;
    pChan->putRcvChar   = ns16550vxbDummyCallback;
    pChan->channelMode  = 0;    /* undefined */
    pChan->options      = (CLOCAL | CREAD | CS8);
#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
    pChan->mcr          = MCR_DTR|MCR_RTS;
#else
    pChan->mcr          = MCR_OUT2;
#endif
    pChan->pDrvFuncs    = &ns16550vxbSioDrvFuncs;

    return ns16550vxbInitChannel (pChan);
    }

/*******************************************************************************
*
* ns16550vxbInitChannel - initialize UART
*
* Initialize the number of data bits, parity and set the selected
* baud rate. Set the modem control signals if the option is selected.
*
* RETURNS:  OK, or ERROR if initialize failed
*
* ERRNO
*/

LOCAL STATUS ns16550vxbInitChannel
    (
    NS16550VXB_CHAN * pChan    /* pointer to channel */
    )
    {
    /* clear the FIFO */

    ns16550DrainFifo (pChan);

    /* set the requested baud rate */

    if (ns16550vxbBaudSet(pChan, (UINT)pChan->baudRate) != OK)
        {
        return ERROR;
        }

    /* set the options */

    return ns16550vxbOptsSet(pChan, pChan->options);
    }


/*******************************************************************************
*
* ns16550vxbOptsSet - set the serial options
*
* Set the channel operating mode to that specified.  All sioLib options
* are supported: CLOCAL, HUPCL, CREAD, CSIZE, PARENB, and PARODD.
* When the HUPCL option is enabled, a connection is closed on the last
* close() call and opened on each open() call.
*
* NOTE:
*
* This function disables the transmitter.  The calling function
* may have to re-enable it.
*
* RETURNS: OK, or ERROR
*
* ERRNO
*/

LOCAL STATUS ns16550vxbOptsSet
    (
    NS16550VXB_CHAN * pChan,   /* pointer to channel */
    UINT options            /* new hardware options */
    )
    {
    UINT8       value;

    if (pChan == NULL || options & 0xffffff00)
    return ERROR;

    VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);

    /* clear RTS and DTR bits */

#ifndef _WRS_CONFIG_ITL_SOC_AGILEX5
    pChan->mcr = (UINT8) (pChan->mcr & ~(MCR_RTS | MCR_DTR));
#endif

    switch (options & CSIZE)
        {
        case CS5:
            pChan->lcr = CHAR_LEN_5; break;
        case CS6:
            pChan->lcr = CHAR_LEN_6; break;
        case CS7:
            pChan->lcr = CHAR_LEN_7; break;
        case CS8:
        default:
            pChan->lcr = CHAR_LEN_8; break;
        }

    if (options & STOPB)
        pChan->lcr |= LCR_STB;
    else
        pChan->lcr |= ONE_STOP;

    switch (options & (PARENB | PARODD))
        {
        case PARENB|PARODD:
            pChan->lcr |= LCR_PEN;
            pChan->ier |= IER_ELSI;   /* Enable receiver line status interrupt.*/
            break;
        case PARENB:
            pChan->lcr |= (LCR_PEN | LCR_EPS);
            pChan->ier |= IER_ELSI;   /* Enable receiver line status interrupt.*/
            break;
        default:
        case 0:
            pChan->lcr |= PARITY_NONE;
            pChan->ier &= (UINT8) (~IER_ELSI);
            break;
        }

    REG_SET(IER, pChan, pChan->ierInit);

    if (!(options & CLOCAL))
        {
        /* !clocal enables hardware flow control(DTR/DSR) */

        pChan->mcr = (UINT8) (pChan->mcr | (MCR_DTR | MCR_RTS));
        pChan->ier = (UINT8) (pChan->ier & (~TxFIFO_BIT));
        pChan->ier |= IER_EMSI;    /* enable modem status interrupt */
        }
    else
        {
        /* disable modem status interrupt */

        pChan->ier = (UINT8) (pChan->ier & ~IER_EMSI);
        }

    REG_SET(LCR, pChan, pChan->lcr);
    REG_SET(MCR, pChan, pChan->mcr);

    /* now reset the channel mode registers */
    value = (RxCLEAR | TxCLEAR | FIFO_ENABLE | FCR_RXTRIG_L | FCR_RXTRIG_H);
    REG_SET(FCR, pChan, value);

    /* check FIFO support if fifoSize is unknown  */

    if (pChan->fifoSize == NS16550_FIFO_DEPTH_UNKNOWN)
        {
        /* test if FIFO is supported on the device */

        REG_GET(IIR, pChan, value);

        if ((value & (FCR_RXTRIG_L|FCR_RXTRIG_H)) ==
                            (FCR_RXTRIG_L|FCR_RXTRIG_H))
            {
            /*
             * FIFO is supported.
             * The max TX FIFO supported on this driver is 16 bytes, unless
             * BSP configures the depth through "fifoLen" parameter explicitly.
             */

            pChan->fifoSize = NS16550_FIFO_DEPTH_DEFAULT;
            }
        else
            {
            /* No FIFO support */

            pChan->fifoSize = NS16550_FIFO_DEPTH_NO;
            }
        }

    /* enable FIFO in DMA mode 1 */

    value = (FCR_DMA | RxCLEAR | TxCLEAR | FIFO_ENABLE);
    REG_SET(FCR, pChan, value);

    if (options & CREAD)
    pChan->ier |= RxFIFO_BIT;

    if (pChan->channelMode == SIO_MODE_INT)
        {
        REG_SET(IER, pChan, pChan->ier);
        }

    pChan->options = options;

    VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);

    return OK;
    }


/*******************************************************************************
*
* ns16550vxbHup - hang up the modem control lines
*
* Resets the RTS and DTR signals and clears both the receiver and
* transmitter sections.
*
* RETURNS: OK
*
* ERRNO
*/

LOCAL STATUS ns16550vxbHup
    (
    NS16550VXB_CHAN * pChan    /* pointer to channel */
    )
    {
    UINT8 value = 0;

    VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);

    pChan->mcr = (UINT8) (pChan->mcr & ~(MCR_RTS | MCR_DTR));
    REG_SET(MCR, pChan, pChan->mcr);
    value = (RxCLEAR | TxCLEAR);
    REG_SET(FCR, pChan, value);

    VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);

    return(OK);
    }


/*******************************************************************************
*
* ns16550vxbOpen - Set the modem control lines
*
* Set the modem control lines(RTS, DTR) TRUE if not already set.
* It also clears the receiver, transmitter and enables the FIFO.
*
* RETURNS: OK
*
* ERRNO
*/

LOCAL STATUS ns16550vxbOpen
    (
    NS16550VXB_CHAN * pChan    /* pointer to channel */
    )
    {
    UINT8   mask;
    UINT8   value = 0;

    REG_GET(MCR, pChan, mask);
    mask = (UINT8) (mask & (MCR_RTS | MCR_DTR));

    if (mask != (MCR_RTS | MCR_DTR))
        {
        /* RTS and DTR not set yet */

        VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);

        /* set RTS and DTR TRUE */

        pChan->mcr |= (MCR_DTR | MCR_RTS);
        REG_SET(MCR, pChan, pChan->mcr);

        /* clear Tx and receive and enable FIFO */
        value = (FCR_DMA | RxCLEAR | TxCLEAR | FIFO_ENABLE);
        REG_SET(FCR, pChan, value);

        VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);
        }

    return(OK);
    }


/******************************************************************************
*
* ns16550vxbBaudSet - change baud rate for channel
*
* This function sets the baud rate for the UART. The interrupts are disabled
* during chip access.
*
* RETURNS: OK, or ERROR if check DW_USR_BUSY timeout or baud rate out of range
*
* ERRNO
*/

LOCAL STATUS  ns16550vxbBaudSet
    (
    NS16550VXB_CHAN * pChan,   /* pointer to channel */
    UINT       baud         /* requested baud rate */
    )
    {
    int   divisor;
    UINT8   value;

    VXB_ASSERT(pChan->pDev != NULL, ERROR)

    if (baud == 0)
        {
        return ERROR;
        }

    if(pChan->pDivisorFunc==NULL)
        divisor = (int) ((pChan->xtal / ( 8UL * baud ) + 1) / 2);
    else
        divisor = (*pChan->pDivisorFunc)(pChan->xtal,baud);

    /*
     * check whether the desired baud rate is out of range based
     * on the XTAL for this device
     */

    if ((divisor == 0) || ((divisor & (int) 0xFFFF0000) != 0))
        {
        return ERROR;
        }

    NS16550_DBG_MSG (500,"ns16550vxbBaudSet(0x%x, %d), divisor = %d / 0x%08x\n",
                    (_Vx_usr_arg_t)pChan,baud,divisor,divisor,5,6);

    /* disable interrupts during chip access */

    VXB_NS16550_SPIN_LOCK_ISR_TAKE (pChan);

    if ((pChan->flag & DW_DLL_CHECK_USR) != 0)
        {
        UINT32 timeout = NS16550_TIMEOUT;

        while (timeout--)
            {
            REG_GET (DW_USR, pChan, value);
            if ((value & DW_USR_BUSY) != DW_USR_BUSY)
                break;
            if (timeout == 0)
                {
                NS16550_DBG_MSG ( \
                        500,
                        "ns16550vxbBaudSet check DW_USR_BUSY timeout\n",
                        1, 2, 3, 4, 5, 6);
                VXB_NS16550_SPIN_LOCK_ISR_GIVE (pChan);
                return ERROR;
                }
            VXB_NS16550_SPIN_LOCK_ISR_GIVE (pChan);
            vxbUsDelay(1);
            VXB_NS16550_SPIN_LOCK_ISR_TAKE (pChan);
            }
        }

    /* Enable access to the divisor latches by setting DLAB in LCR. */
    value = (LCR_DLAB | pChan->lcr);
    REG_SET(LCR, pChan, value);

    /* Set divisor latches. */
    value = (UINT8)divisor;
    REG_SET(DLL, pChan, value);
    value = (UINT8) (divisor >> 8);
    REG_SET(DLM, pChan, value);

    /* Restore line control register */

    REG_SET(LCR, pChan, pChan->lcr);

    pChan->baudRate = (int)baud;

    VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);

    return(OK);
    }


/*******************************************************************************
*
* ns16550vxbModeSet - change channel mode setting
*
* This driver supports both polled and interrupt modes and is capable of
* switching between modes dynamically.
*
* If interrupt mode is desired this function enables the channels receiver and
* transmitter interrupts. If the modem control option is TRUE, the Tx interrupt
* is disabled if the CTS signal is FALSE. It is enabled otherwise.
*
* If polled mode is desired the device interrupts are disabled.
*
* RETURNS: OK, or ERROR
*/

LOCAL STATUS ns16550vxbModeSet
    (
    NS16550VXB_CHAN * pChan,   /* pointer to channel */
    UINT    newMode         /* mode requested */
    )
    {
    UINT8   mask;
    BOOL    spinLockOwn;

    NS16550_DBG_MSG(500,"ns16550vxbModeSet(0x%x, %d)\n",
                (_Vx_usr_arg_t)pChan,newMode,3,4,5,6);

    if ((newMode != SIO_MODE_POLL) && (newMode != SIO_MODE_INT))
        return(ERROR);

    /* test if spinlock is already owned by the currect CPU core */

    spinLockOwn = VXB_NS16550_SPIN_LOCK_ISR_HELD(pChan);

    if (newMode == SIO_MODE_INT)
        {
        /* set the pointer of the  queue for deferred work */

        if ((pChan->queueId == NULL) && (!(pChan->flag & DISABLE_DEFER)))
            pChan->queueId = isrDeferQueueGet (pChan->pDev, 0, 0, 0);

        if (!spinLockOwn)
            {
            VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);
            }

        /* Enable appropriate interrupts */

        if (pChan->options & CLOCAL)
            {
            UINT8 ierLocal = (pChan->ier | RxFIFO_BIT | TxFIFO_BIT);
            REG_SET(IER, pChan, ierLocal);
            }
        else
            {
            REG_GET(MSR, pChan, mask);
            mask = (UINT8) (mask & MSR_CTS);

            /* if the CTS is asserted enable Tx interrupt */

            if (mask & MSR_CTS)
        pChan->ier |= TxFIFO_BIT;    /* enable Tx interrupt */
            else
        {
        /* disable Tx interrupt */

        pChan->ier = (UINT8) (pChan->ier & (~TxFIFO_BIT));
        }

            REG_SET(IER, pChan, pChan->ier);
            }
        }
    else
        {
        if (!spinLockOwn)
            {
            VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);
            }

        /* disable all ns16550vxb interrupts */
        REG_SET(IER, pChan, pChan->ierInit);
        }

    pChan->channelMode = (UINT16) newMode;

    if (!spinLockOwn)
        {
        VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);
        }

    NS16550_DBG_MSG(500,"ns16550vxbModeSet(0x%x, %d): Done\n",
                (_Vx_usr_arg_t)pChan,newMode,3,4,5,6);

    return(OK);
    }


/*******************************************************************************
*
* ns16550vxbIoctl - special device control
*
* Includes commands to get/set baud rate, mode (INT,POLL), hardware options
* (parity, number of data bits), and modem control (RTS/CTS and DTR/DSR).
* The ioctl command SIO_HUP is sent by ttyDrv when the last close() function
* call is made. Likewise SIO_OPEN is sent when the first open() function call
* is made.
*
* RETURNS: OK on success
*
* ERRNO: EIO on device error, ENOSYS on unsupported request
*/

LOCAL STATUS ns16550vxbIoctl
    (
    NS16550VXB_CHAN *  pChan,  /* pointer to channel */
    int            request,    /* request code */
    void *        pArg        /* some argument */
    )
    {
    FAST STATUS  status;
    long arg = (long)pArg;

    NS16550_DBG_MSG(500,"ns16550vxbIoctl(%p, %x, %p)\n",
                (_Vx_usr_arg_t)pChan,request,pArg,4,5,6);

    status = OK;

    switch (request)
        {
        case SIO_NAME_GET:
            (void)strncpy_s ((char *)pArg, TTY_MAX_NAME_LENGTH,
                             pChan->tyName, TTY_MAX_NAME_LENGTH);
            break;

        case SIO_BAUD_SET:
            status = (ns16550vxbBaudSet (pChan, (UINT)arg) == OK) ? OK : EIO;
            break;

        case SIO_BAUD_GET:
            *(int *)pArg = pChan->baudRate;
            break;

        case SIO_MODE_SET:
            status = (ns16550vxbModeSet (pChan, (UINT)arg) == OK) ? OK : EIO;
            break;

        case SIO_MODE_GET:
            *(int *)pArg = pChan->channelMode;
            break;

        case SIO_AVAIL_MODES_GET:
            *(int *)pArg = SIO_MODE_INT | SIO_MODE_POLL;
            break;

        case SIO_HW_OPTS_SET:
            status = (ns16550vxbOptsSet (pChan, (UINT)arg) == OK) ? OK : EIO;
            break;

        case SIO_HW_OPTS_GET:
            *(int *)pArg = (int)(pChan->options);
            break;

        case SIO_HUP:
            /* check if hupcl option is enabled */

            if (pChan->options & HUPCL)
        status = ns16550vxbHup (pChan);
            break;

        case SIO_OPEN:
            /* check if hupcl option is enabled */

            if (pChan->options & HUPCL)
        status = ns16550vxbOpen (pChan);
            break;

#ifdef  _WRS_CONFIG_SMP
        case SIO_DEV_LOCK:
            VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);
            break;

        case SIO_DEV_UNLOCK:
            VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);
            break;
#endif  /* _WRS_CONFIG_SMP */

        default:
            status = ENOSYS;
        }

    NS16550_DBG_MSG(500,"ns16550vxbIoctl(%p, %d, %p): status: %u\n",
                (_Vx_usr_arg_t)pChan,request,pArg,status,5,6);

    return(status);
    }


#ifndef    _VXBUS_BASIC_SIOPOLL
/*******************************************************************************
*
* ns16550vxbIntWr2 - deferred handle a transmitter interrupt
*
* This function handles write interrupts from the UART. It reads a character
* and puts it in the transmit holding register of the device for transfer.
*
* If there are no more characters to transmit, transmission is disabled by
* clearing the transmit interrupt enable bit in the IER(int enable register).
*
* RETURNS: N/A
*
* ERRNO
*/

LOCAL void ns16550vxbIntWr2
    (
    void * pData        /* pointer to channel */
    )
    {
    NS16550VXB_CHAN * pChan = (NS16550VXB_CHAN *) pData;
    UINT8    status;
    char    outChar;

    VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);

    /*
     * Clear the txDefer.
     * Another deferred job can be posted before completing the deferred job
     * since the txDefer is cleared prior to start the getTxChar.
     * But the job queue is already cleared before entering into
     * ns16550vxbIntWr2() and the maximum required job queue number for
     * transmit is one.
     */

    pChan->txDefer = FALSE;

    while (TRUE)
        {
        REG_GET(LSR, pChan, status);

        /* is the transmitter ready to accept a character? */

        if ((status & LSR_THRE) != 0x00)
            {
            int i;

            VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);

            /* Transmit characters upto the FIFO depth */

            for (i = 0; i < pChan->fifoSize; i++)
                {
                if ((*pChan->getTxChar) (pChan->getTxArg, &outChar) != ERROR)
                    {
                    /*
                     * Re-take the spinlock to mutual exclude the TX Reg
                     * access between the kprintf() in SMP mode.
                     */

                    VXB_NS16550_ISR_SET(pChan);

                    /* write char to Transmit Holding Reg */

                    REG_SET(THR, pChan, outChar);

                    VXB_NS16550_ISR_CLEAR(pChan);
                    }
                else
                    return;
                }
            }
        else
            {
            pChan->ier |= (TxFIFO_BIT);    /* indicates to enable Tx Int */
            REG_SET(IER, pChan, pChan->ier);
            VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);
            break;
            }

        VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);
        }
    }


/*******************************************************************************
*
* ns16550vxbIntRd2 - deferred handle a receiver interrupt
*
* This function handles read interrupts from the UART.
*
* RETURNS: N/A
*
* ERRNO
*/

LOCAL void ns16550vxbIntRd2
    (
    void * pData        /* pointer to channel */
    )
    {
    NS16550VXB_CHAN * pChan = (NS16550VXB_CHAN *) pData;
    char   inchar;
    char * pChar = &inchar;

    while (TRUE)
        {
        VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);

        if ((ns16550vxbPollInput (pChan, pChar)) == OK)
            {
            VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);

            (*pChan->putRcvChar) (pChan->putRcvArg, inchar);
            }
        else
            break;
        }

    /*
     * Another deferred job can be posted before completing the deferred job
     * if further Rx interrupt is raised before completing ns16550vxbIntRd2().
     * But the job queue is already cleared before entering into
     * this function and the maximum required job queue number for reception
     * is one.
     */

    pChan->ier |= RxFIFO_BIT;        /* indicates to enable Rx Int */
    REG_SET(IER, pChan, pChan->ier);

    VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);
    }


/********************************************************************************
*
* ns16550vxbInt - interrupt level processing
*
* This function handles four sources of interrupts from the UART. They are
* prioritized in the following order by the Interrupt Identification Register:
* Receiver Line Status, Received Data Ready, Transmit Holding Register Empty
* and Modem Status.
*
* When a modem status interrupt occurs, the transmit interrupt is enabled if
* the CTS signal is TRUE.
* 
* The vxbDevAttach method of the bus adapter library connects this function
* to the IRQ.
*
* RETURNS: OK
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS ns16550vxbInt
    (
    VXB_DEV_ID pDev
    )
    {
    int count = NS16550_INT_COUNT;
    FAST NS16550VXB_CHAN * pChan = (NS16550VXB_CHAN *)(vxbDevSoftcGet(pDev));
    FAST volatile char     intStatus;
    UINT8           iirValue, lsrValue;

    NS16550_DBG_MSG (800,"ns16550vxbInt(%p): %s\n",
                    (_Vx_usr_arg_t)pChan,pChan->tyName,3,4,5,6);

    while (( pChan != NULL ) && count > 0)
        {
        VXB_NS16550_ISR_SET(pChan);

        /* read the Interrupt Status Register (Int. Ident.) */

        REG_GET(IIR, pChan, iirValue);
        intStatus = (char)iirValue;
        intStatus = intStatus & 0x0f;

        /*
         * This UART chip always produces level active interrupts,
         * and the IIR only indicates the highest priority interrupt.
         * In the case that receive and transmit interrupts happened at
         * the same time, we must clear both interrupt pending to keep the
         * edge-triggered interrupt (output from interrupt controller) from
         * locking up. One way of doing it is to disable all interrupts
         * at the beginning of the ISR and reenable them at the end.
         */

        REG_SET(IER, pChan, pChan->ierInit);

        NS16550_DBG_MSG (800,"ns16550vxbInt(%p): intStatus: %x\n",
                        (_Vx_usr_arg_t)pChan,intStatus,3,4,5,6);

        switch (intStatus)
            {
            case IIR_RLS:
                /* overrun,parity error and break interrupt */

                REG_GET(LSR, pChan, lsrValue); /* read LSR to reset interrupt */
                if(lsrValue & LSR_PE)
                    {
                    REG_READ(RBR, pChan);
                    }
                intStatus = (char)lsrValue;

                NS16550_DBG_MSG (800,"ns16550vxbInt(%p): RLS: %x\n",
                                (_Vx_usr_arg_t)pChan,intStatus,3,4,5,6);

                break;

            case IIR_RDA:           /* received data available */
            case IIR_TIMEOUT:
                {
                /*
                 * receiver FIFO interrupt. In some cases, IIR_RDA is
                 * not indicated in IIR register when there is more
                 * than one character in FIFO.
                 */

                /* indicate to disable Rx Int */

                pChan->ier = (UINT8) (pChan->ier & ~(RxFIFO_BIT));
                REG_SET(IER, pChan, pChan->ier);

                /*
                 * Another spinlock is taken in isrDeferJobAdd().
                 * So release spinlock before entering into isrDeferJobAdd().
                 */

                VXB_NS16550_ISR_CLEAR(pChan);

                /* defer the job */

#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
                    ns16550vxbIntRd2(pChan);
#else
                if (pChan->queueId == NULL)
                    ns16550vxbIntRd2(pChan);
                else
                    isrDeferJobAdd (pChan->queueId, &pChan->isrDefRd);

                 NS16550_DBG_MSG (800,"ns16550vxbInt(%p): RX/TIMEOUT\n",
                                 (_Vx_usr_arg_t)pChan,2,3,4,5,6);

#endif
                goto nextChan;
                }

            case IIR_THRE:  /* transmitter holding register ready */

                /*
                 * If Tx deferred job is still in queue, no need to add
                 * another one and save the number of the job queue to prevent
                 * from the queue overflow.
                 */

                NS16550_DBG_MSG (800,"ns16550vxbInt(%p): THRE\n",
                                (_Vx_usr_arg_t)pChan,2,3,4,5,6);

                /* indicate to disable Tx Int */

                pChan->ier = (UINT8) (pChan->ier & ~(TxFIFO_BIT));

                if (!pChan->txDefer)
                    {
                    /* defer the job */

                    pChan->txDefer = TRUE;

                    REG_SET(IER, pChan, pChan->ier);

                    /*
                     * Another spinlock is taken in isrDeferJobAdd ().
                     * So release spinlock before entering into
                     * isrDeferJobAdd().
                     */

                    VXB_NS16550_ISR_CLEAR(pChan);

                    if (pChan->queueId == NULL)
                    ns16550vxbIntWr2(pChan);
                    else
                    isrDeferJobAdd (pChan->queueId, &pChan->isrDefWr);

                    NS16550_DBG_MSG (800,"ns16550vxbInt(%p): TX\n",
                                    (_Vx_usr_arg_t)pChan,2,3,4,5,6);

                    goto nextChan;
                    }

                break;

            case IIR_MSTAT: /* modem status register */
                {
                UINT8    msr;

                REG_GET(MSR, pChan, msr);

#define SPR_74889_FIX
#ifdef  SPR_74889_FIX

                /*
                 * If the CTS line changed, enable or
                 * disable Tx interrupt accordingly.
                 */
                if (msr & MSR_DCTS)
                    {
                    if (msr & MSR_CTS)
                    /* CTS was turned on */
                    pChan->ier |= TxFIFO_BIT;
                    else
                    /* CTS was turned off */
                    pChan->ier = (UINT8) (pChan->ier & (~TxFIFO_BIT));
                    }
#else   /* SPR_74889_FIX */
                /* if CTS is asserted by modem, enable Tx interrupt */

                if ((msr & MSR_CTS) && (msr & MSR_DCTS))
                    pChan->ier |= TxFIFO_BIT;
                else
                    pChan->ier &= (~TxFIFO_BIT);
#endif  /* SPR_74889_FIX */

                NS16550_DBG_MSG (800,"ns16550vxbInt(%p): MSTAT\n",
                                (_Vx_usr_arg_t)pChan,2,3,4,5,6);

                }
                break;

            default:
                break;
            }

        REG_SET(IER, pChan, pChan->ier); /* enable interrupts accordingly */

        VXB_NS16550_ISR_CLEAR(pChan);

nextChan:

        if (--count == 0)
            {
            pChan = pChan->pNext;
            count = NS16550_INT_COUNT;
            }
        }

    NS16550_DBG_MSG (800,"ns16550vxbInt(%p): done\n",
                    (_Vx_usr_arg_t)pChan,22,3,4,5,6);

    return OK;
    }


/*******************************************************************************
*
* ns16550vxbTxStartup - transmitter startup function
*
* Call interrupt level character output function and enable interrupt if it is
* in interrupt mode with no hardware flow control.
* If the option for hardware flow control is enabled and CTS is set TRUE,
* then the Tx interrupt is enabled.
*
* RETURNS: OK
*
* ERRNO: ENOSYS if in polled mode
*/

LOCAL int ns16550vxbTxStartup
    (
    NS16550VXB_CHAN * pChan    /* pointer to channel */
    )
    {
    UINT8 mask;
    BOOL  spinLockOwn;
    char  outChar;

    NS16550_DBG_MSG(500,"ns16550vxbTxStartup(0x%x)\n",
            (_Vx_usr_arg_t)pChan,2,3,4,5,6);

    /* test if spinlock is already owned by the currect CPU core */

    spinLockOwn = VXB_NS16550_SPIN_LOCK_ISR_HELD(pChan);

    if (pChan->channelMode == SIO_MODE_INT)
        {
        if (!spinLockOwn)
            {
            VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);
            }

        if (pChan->options & CLOCAL)
            {
            /* No modem control */

            /*
             * If there is a job queue already posted for transmit, not
             * necessary to enable Tx interrupt.
             */

            if (!pChan->txDefer)
                {
                /* avoid live lock in getTxChar */

                if (!spinLockOwn)
                    {
                    VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);
                    }

                if ((*pChan->getTxChar) (pChan->getTxArg, &outChar) != ERROR)
                    {
                    /* write char to Transmit Holding Reg */

                    if (!spinLockOwn)
                        {
                        VXB_NS16550_SPIN_LOCK_ISR_TAKE(pChan);
                        }

                    REG_SET(THR, pChan, outChar);
                    }
                else
                    {
                    return (ENOSYS);
                    }

                pChan->ier |= TxFIFO_BIT;
                }
            }
        else
            {
            REG_GET(MSR, pChan, mask);
            mask = (UINT8) (mask & MSR_CTS);

            /* if the CTS is asserted enable Tx interrupt */

            if (mask & MSR_CTS)
        {
        /*
         * If there is a job queue already posted for transmit, not
         * necessary to enable Tx interrupt.
         */

        if (!pChan->txDefer)
            pChan->ier |= TxFIFO_BIT;    /* enable Tx interrupt */
        }
            else
        {
        /* disable Tx interrupt */

        pChan->ier = (UINT8) (pChan->ier & (~TxFIFO_BIT));
        }
            }

        REG_SET(IER, pChan, pChan->ier);

        if (!spinLockOwn)
            {
            VXB_NS16550_SPIN_LOCK_ISR_GIVE(pChan);
            }

        NS16550_DBG_MSG(500,"ns16550vxbTxStartup(0x%x): started\n",
                (_Vx_usr_arg_t)pChan,2,3,4,5,6);

        return(OK);
        }
    else
        {
        NS16550_DBG_MSG(500,"ns16550vxbTxStartup(0x%x): ENOSYS\n",
                (_Vx_usr_arg_t)pChan,2,3,4,5,6);

        return(ENOSYS);
        }
    }
#endif    /* _VXBUS_BASIC_SIOPOLL */


#ifdef    _NS1650_VXB_POLL_DEBUG
UINT32    ns16550vxbPollBusyCount = 0;
UINT32    ns16550vxbPollXmitCount = 0;
UINT32    ns16550vxbPollFlowCount = 0;
#endif    /* _NS1650_VXB_POLL_DEBUG */


/******************************************************************************
*
* ns16550vxbPollOutput - output a character in polled mode
*
* This function transmits a character in polled mode.
*
* RETURNS: OK if a character sent
*
* ERRNO: EAGAIN if the output buffer is full
*/

LOCAL int ns16550vxbPollOutput
    (
    NS16550VXB_CHAN *  pChan,  /* pointer to channel */
    char            outChar /* char to send */
    )
    {
    UINT8 pollStatus;
    UINT8 msr;

    NS16550_DBG_MSG(1500,"ns16550vxbPollOutput(0x%x, '%c')\n",
                (_Vx_usr_arg_t)pChan,outChar,3,4,5,6);

    REG_GET(LSR, pChan, pollStatus);
    REG_GET(MSR, pChan, msr);

    /* is the transmitter ready to accept a character? */

    while ((pollStatus & LSR_THRE) == 0x00)
        {
#ifdef    _NS1650_VXB_POLL_DEBUG
        ++ns16550vxbPollBusyCount;
#endif    /* _NS1650_VXB_POLL_DEBUG */
        NS16550_DBG_MSG(1500,"ns16550vxbPollOutput(): not ready\n",
                1,2,3,4,5,6);
        REG_GET(LSR, pChan, pollStatus);
#if 0
        return(EAGAIN);
#endif
        }

    /* modem flow control ? */
    if ((!(pChan->options & CLOCAL)) && !(msr & MSR_CTS))
        {
#ifdef    _NS1650_VXB_POLL_DEBUG
        ++ns16550vxbPollFlowCount;
#endif    /* _NS1650_VXB_POLL_DEBUG */
        NS16550_DBG_MSG(1500,"ns16550vxbPollOutput(): flow control\n",
                1,2,3,4,5,6);
        return(EAGAIN);
        }

#ifdef    _NS1650_VXB_POLL_DEBUG
    ++ns16550vxbPollXmitCount;
#endif    /* _NS1650_VXB_POLL_DEBUG */

    REG_SET(THR, pChan, outChar);       /* transmit character */

    NS16550_DBG_MSG(1500,"ns16550vxbPollOutput(0x%x, '%c'): sent\n",
                (_Vx_usr_arg_t)pChan,outChar,3,4,5,6);

    return(OK);
    }


/******************************************************************************
*
* ns16550vxbPollInput - poll for input
*
* This function polls the device for input.
*
* RETURNS: OK if a character arrived
*
* ERRNO: EAGAIN if the input buffer if empty
*/

LOCAL int ns16550vxbPollInput
    (
    NS16550VXB_CHAN *  pChan,  /* pointer to channel */
    char *          pChar   /* pointer to char */
    )
    {
    UINT8 pollStatus;
    UINT8 *inChar = (UINT8 *)pChar;

    NS16550_DBG_MSG(1500,"ns16550vxbPollInput(0x%x, 0x%08x)\n",
            (_Vx_usr_arg_t)pChan,(_Vx_usr_arg_t)pChar,3,4,5,6);

    REG_GET(LSR, pChan, pollStatus);

    if ((pollStatus & LSR_DR) == 0x00)
        {
        NS16550_DBG_MSG(1500,"ns16550vxbPollInput(): LSR_DR\n",
                1,2,3,4,5,6);
        return(EAGAIN);
        }

    if (pollStatus & LSR_PE)
        {
        REG_READ(RBR, pChan);

        NS16550_DBG_MSG(1500,"ns16550vxbPollInput(): LSR_PE\n",
            1,2,3,4,5,6);
        return(EAGAIN);
    }

    /* got a character */

    REG_GET(RBR, pChan, *inChar);

    NS16550_DBG_MSG(1500,"ns16550vxbPollInput(0x%x, 0x%08x): got '%c'\n",
            (_Vx_usr_arg_t)pChan,(_Vx_usr_arg_t)pChar,*inChar,4,5,6);

    return(OK);
    }


#ifndef    _VXBUS_BASIC_SIOPOLL
/******************************************************************************
*
* ns16550vxbCallbackInstall - install ISR callbacks
*
* This function installs the callback functions to get/put chars for the
* driver.
*
* RETURNS: OK on success
*
* ERRNO: ENOSYS on unsupported callback type
*/

LOCAL int ns16550vxbCallbackInstall
    (
    SIO_CHAN *  pSioChan,       /* pointer to device to control */
    int         callbackType,   /* callback type(Tx or receive) */
    STATUS      (*callback)(),  /* pointer to callback function */
    void *      callbackArg     /* callback function argument */
    )
    {
    NS16550VXB_CHAN * pChan = (NS16550VXB_CHAN *)pSioChan;

    switch (callbackType)
        {
        case SIO_CALLBACK_GET_TX_CHAR:
            pChan->getTxChar    = callback;
            pChan->getTxArg     = callbackArg;
            return(OK);
        case SIO_CALLBACK_PUT_RCV_CHAR:
            pChan->putRcvChar   = callback;
            pChan->putRcvArg    = callbackArg;
            return(OK);
        default:
            return(ENOSYS);
        }

    }
#endif    /* _VXBUS_BASIC_SIOPOLL */

