/* alt_soc_agilex.c - Intel Agilex Processor Support Library */

/*
 * Copyright (c) 2022 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION
This library provides board-specific routines for the Intel Agilex.

*/

#include <vxWorks.h>
#include <stdio.h>
#include <boardLib.h>
#include <sysLib.h>
#include <string.h>
#include <intLib.h>
#include <vxLib.h>
#include <cacheLib.h>
#include <dllLib.h>
#include <pmapLib.h>
#include <vmLib.h>
#include <vxFdtLib.h>
#include <arch/arm/mmuArmLib.h>
#include <private/vmLibP.h>
#include <private/vxCpuLibP.h>
#include <sys/arm/psciLib.h>
#include <private/kwriteLibP.h>
#include <subsys/timer/vxbTimerLib.h>
#include <hwif/vxBus.h>

#include <itl_agilex5.h>

IMPORT VOID armSysToMonitor(FUNCPTR, FUNCPTR, int);
IMPORT VOID sysInit(VOID);
IMPORT int  sysStartType;

#ifdef _WRS_CONFIG_SMP
IMPORT STATUS armMonitorSpinRelease (UINT32);
IMPORT STATUS armMpCoreInit(UINT32, WIND_CPU_STATE *, FUNCPTR);
IMPORT UINT32 vxCpuIdGetByIndex(UINT32);
LOCAL VOID itlAgilex5PspExit(UINT32);
#endif /* _WRS_CONFIG_SMP */

#undef DEBUG_MSG
#define DEBUG_MSG(...)                                  \
    do                                                  \
        {                                               \
        if (_func_kprintf != NULL)                      \
            {                                           \
            (* _func_kprintf)(__VA_ARGS__);             \
            }                                           \
        }                                               \
    while ((FALSE))

#define CSR_READ_32(reg)        (*(volatile UINT32 *)(reg))
#define CSR_WRITE_32(reg, data) (*((volatile UINT32 *)(reg)) = (data))
#define CSR_SET_32(reg, bits)   CSR_WRITE_32(reg, CSR_READ_32(reg) | (bits))
#define CSR_CLR_32(reg, bits)   CSR_WRITE_32(reg, CSR_READ_32(reg) & (~(bits)))

#define CLKMGR_PHYS_ADRS 0x10d10000
#define CLKMGR_ADRS_LEN  0x1000
#define RSTMGR_PHYS_ADRS 0x10d11000
#define RSTMGR_ADRS_LEN  0x100
#define SYSMGR_PHYS_ADRS 0x10d12000
#define SYSMGR_ADRS_LEN  0x500

#define RSTMGR_PER0MODRST           (0x24)
#define RSTMGR_PER0MODRST_TSN0      (1u << 0)
#define RSTMGR_PER0MODRST_TSN1      (1u << 1)
#define RSTMGR_PER0MODRST_TSN2      (1u << 2)
#define RSTMGR_PER0MODRST_USB0      (1u << 3)
#define RSTMGR_PER0MODRST_USB1      (1u << 4)
#define RSTMGR_PER0MODRST_SDMMC     (1u << 7)
#define RSTMGR_PER0MODRST_TSN0ECC   (1u << 8)
#define RSTMGR_PER0MODRST_TSN1ECC   (1u << 9)
#define RSTMGR_PER0MODRST_TSN2ECC   (1u << 10)
#define RSTMGR_PER0MODRST_USB0ECC   (1u << 11)
#define RSTMGR_PER0MODRST_USB1ECC   (1u << 12)
#define RSTMGR_PER0MODRST_SDMMCOCP  (1u << 15)
#define RSTMGR_PER0MODRST_DMA       (1u << 16)
#define RSTMGR_PER0MODRST_DMAOCP    (1u << 21)
#define RSTMGR_PER0MODRST_EMACPTP   (1u << 22)

#define RSTMGR_PER1MODRST           (0x28)
#define RSTMGR_PER1MODRST_GPIO0     (1u << 24)
#define RSTMGR_PER1MODRST_GPIO1     (1u << 25)
#define RSTMGR_PER1MODRST_UART0     (1u << 16)
#define RSTMGR_PER1MODRST_UART1     (1u << 17)
#define RSTMGR_PER1MODRST_I2C4      (1u << 12)
#define RSTMGR_PER1MODRST_I2C3      (1u << 11)
#define RSTMGR_PER1MODRST_I2C2      (1u << 10)
#define RSTMGR_PER1MODRST_I2C1      (1u << 9)
#define RSTMGR_PER1MODRST_I2C0      (1u << 8)
#define RSTMGR_PER1MODRST_SPTIMER0  (1u << 6)
#define RSTMGR_PER1MODRST_SPTIMER1  (1u << 7)
#define RSTMGR_PER1MODRST_L4TIMER1  (1u << 5)
#define RSTMGR_PER1MODRST_L4TIMER0  (1u << 4)

#define CLKMGR_EN                   (0x7C)
#define CLKMGR_ENS                  (0x80)
#define CLKMGR_EN_GPIODBEN          (1u << 4)

#define UART_REG_THR                (0x0u)
#define UART_REG_IER                (0x4u)
#define UART_REG_FCR                (0x8u)
#define UART_REG_LCR                (0xCu)
#define UART_REG_LSR                (0x14u)
#define UART_REG_LCR_BIT_DLAB       (1u << 7)
#define UART_REG_LCR_BITS_DEFAULT   (3u << 0)
#define UART_REG_LSR_BIT_TEMT       (1u << 6)
#define NS16550_DEFAULT_BAUD        (115200)

#define SYSMGR_TSN0                   (0x44)
#define SYSMGR_TSN0_PHY_INTF_RGMII    (0x01u)
#define SYSMGR_TSN0_PHY_INTF_RESET    (0x03u)

/* Locals */

LOCAL VIRT_ADDR uartBaseAdrs = (VIRT_ADDR) PMAP_FAILED;
LOCAL VIRT_ADDR rstMgrBaseAdrs = (VIRT_ADDR) PMAP_FAILED;
LOCAL VIRT_ADDR sysMgrBaseAdrs = (VIRT_ADDR) PMAP_FAILED;
LOCAL VIRT_ADDR clkMgrBaseAdrs = (VIRT_ADDR) PMAP_FAILED;
LOCAL UINT32 l4SpClkFreq = 100000 * 1000; /* 100000 kHz */
LOCAL UINT32 timerClkFreq;
LOCAL UINT32 timerClkPerUs;

/* Local function prototypes */

LOCAL VOID itlAgilex5UartInit(VOID);
LOCAL VOID itlAgilex5EarlyDebugInit(VOID);
LOCAL int itlAgilex5PhysMemInfoGet(PHYS_MEM_DESC*, int);
LOCAL VOID itlAgilex5DebugInit(VOID);
LOCAL STATUS itlAgilex5DevInit(VOID);
LOCAL STATUS itlAgilex5Kwrite(char *, size_t);
LOCAL VOID itlAgilex5Core0Exit(UINT32);

/* Inline functions */

_WRS_INLINE UINT32 __inline__GetCntFreq (void)
    {
    UINT64  tempVal;
    __asm__ __volatile__ ("MRS %0, CNTFRQ_EL0"
                          : "=r" (tempVal)
                          : );
    return (UINT32)tempVal;
    }

_WRS_INLINE UINT32 __inline__GetVirtTimerValue (void)
    {
    UINT64  tempVal;
    __asm__ __volatile__ ("MRS %0, CNTV_TVAL_EL0"
                          : "=r" (tempVal)
                          : );
    return (UINT32)tempVal;
    }

/* Public functions */

/*******************************************************************************
*
* itlAgilex5Probe - Probe the board
*
*/

BOOL itlAgilex5Probe
    (
    char * compat
    )
    {
    int    offset;
    char * fdtCompat = NULL;
    int len;

    offset = vxFdtPathOffset("/");
    if (offset < 0)
        {
        return FALSE;
        }

    fdtCompat = ((char *) vxFdtPropGet (offset, "compatible", &len));
    if (fdtCompat == NULL)
        {
        return FALSE;
        }

    if (strcmp(compat, fdtCompat) != 0)
        {
        return FALSE;
        }

    return TRUE;
    }


/*******************************************************************************
*
* itlAgilex5EarlyInit - Early initialization
*
*/

VOID itlAgilex5EarlyInit(VOID)
    {
    if (boardFlagCheck(BOARD_DESC_FLAG_DBG, FALSE))
        {
        itlAgilex5EarlyDebugInit();
        }

    sysPhysMemDescNumEnt = itlAgilex5PhysMemInfoGet(sysPhysMemDesc, sysPhysMemDescNumEnt);
    if (sysPhysMemDescNumEnt == 0)
        {
        DEBUG_MSG("ERROR: sysPhysMemDescNumEnt is 0\n");
        }

    return;
    }


/*******************************************************************************
*
* itlAgilex5Init - PSL initialization
*
*/

VOID itlAgilex5Init(VOID)
    {
    UINT32 val;
    UINT32 mask;
    STATUS rc;

    if (boardFlagCheck(BOARD_DESC_FLAG_DBG, FALSE))
        {
        itlAgilex5DebugInit();
        }

    timerClkFreq = __inline__GetCntFreq();
    timerClkPerUs = timerClkFreq / 1000000;

    rc = itlAgilex5DevInit();
    if (rc != OK)
        {
        DEBUG_MSG("itlAgilex5DevInit() failed\n");
        return;
        }

    /* Disable the 4 watchdog timers */

    CSR_SET_32(rstMgrBaseAdrs + RSTMGR_PER1MODRST, 0x0000000f);

    /* Deassert DMA reset */

    mask = RSTMGR_PER0MODRST_DMA | RSTMGR_PER0MODRST_DMAOCP;
    CSR_CLR_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST, mask);
    for (;;)
        {
        val = CSR_READ_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST);
        if ((val & mask) == 0)
            {
            break;
            }
        }

    /* Deassert GPIO1 reset */

    mask = RSTMGR_PER1MODRST_GPIO1;
    CSR_CLR_32(rstMgrBaseAdrs + RSTMGR_PER1MODRST, mask);
    for (;;)
        {
        val = CSR_READ_32(rstMgrBaseAdrs + RSTMGR_PER1MODRST);
        if ((val & mask) == 0)
            {
            break;
            }
        }

    /* Deassert SD/MMC reset */

    mask = RSTMGR_PER0MODRST_SDMMC | RSTMGR_PER0MODRST_SDMMCOCP;
    CSR_CLR_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST, mask);
    for (;;)
        {
        val = CSR_READ_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST);
        if ((val & mask) == 0)
            {
            break;
            }
        }

    /* Deassert DW timer resets */

    mask = RSTMGR_PER1MODRST_SPTIMER1 | RSTMGR_PER1MODRST_SPTIMER0 |
           RSTMGR_PER1MODRST_L4TIMER1 | RSTMGR_PER1MODRST_L4TIMER0;
    CSR_CLR_32(rstMgrBaseAdrs + RSTMGR_PER1MODRST, mask);
    for (;;)
        {
        val = CSR_READ_32(rstMgrBaseAdrs + RSTMGR_PER1MODRST);
        if ((val & mask) == 0)
            {
            break;
            }
        }

    /* Assert Ethernet resets */

    mask = RSTMGR_PER0MODRST_TSN0 | RSTMGR_PER0MODRST_TSN1 |
           RSTMGR_PER0MODRST_TSN2 | RSTMGR_PER0MODRST_EMACPTP |
           RSTMGR_PER0MODRST_TSN0ECC | RSTMGR_PER0MODRST_TSN1ECC |
           RSTMGR_PER0MODRST_TSN2ECC;

    CSR_SET_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST, mask);
    for (;;)
        {
        val = CSR_READ_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST);
        if ((val & mask) == mask)
            {
            break;
            }
        }

    /* Set Ethernet 0 (TSN0) to RGMII */

    CSR_CLR_32(sysMgrBaseAdrs + SYSMGR_TSN0, SYSMGR_TSN0_PHY_INTF_RESET);
    CSR_SET_32(sysMgrBaseAdrs + SYSMGR_TSN0, SYSMGR_TSN0_PHY_INTF_RGMII);
    val = CSR_READ_32(sysMgrBaseAdrs + SYSMGR_TSN0);

    /* Deassert Ethernet resets */

    mask = RSTMGR_PER0MODRST_TSN0 | RSTMGR_PER0MODRST_TSN1 |
           RSTMGR_PER0MODRST_TSN2 | RSTMGR_PER0MODRST_EMACPTP |
           RSTMGR_PER0MODRST_TSN0ECC | RSTMGR_PER0MODRST_TSN1ECC |
           RSTMGR_PER0MODRST_TSN2ECC;

    CSR_CLR_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST, mask);
    for (;;)
        {
        val = CSR_READ_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST);
        if ((val & mask) == 0)
            {
            break;
            }
        }

    /* Deassert UART and I2C resets */

    mask = RSTMGR_PER1MODRST_I2C4 | RSTMGR_PER1MODRST_I2C3 |
        RSTMGR_PER1MODRST_I2C2 | RSTMGR_PER1MODRST_I2C1 |
        RSTMGR_PER1MODRST_I2C0 | RSTMGR_PER1MODRST_UART0 |
        RSTMGR_PER1MODRST_UART1;

    CSR_CLR_32(rstMgrBaseAdrs + RSTMGR_PER1MODRST, mask);
    for (;;)
        {
        val = CSR_READ_32(rstMgrBaseAdrs + RSTMGR_PER1MODRST);
        if ((val & mask) == 0)
            {
            break;
            }
        }

    /* Deassert USB resets */

    mask = RSTMGR_PER0MODRST_USB0 | RSTMGR_PER0MODRST_USB1 |
        RSTMGR_PER0MODRST_USB0ECC | RSTMGR_PER0MODRST_USB1ECC;

    CSR_CLR_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST, mask);
    for (;;)
        {
        val = CSR_READ_32(rstMgrBaseAdrs + RSTMGR_PER0MODRST);
        if ((val & mask) == 0)
            {
            break;
            }
        }
    }


/*******************************************************************************
*
* itlAgilex5Model - return the model name of the CPU board
*
* This routine returns the model name of the CPU board.
*
* RETURNS: A pointer to the string.
*
* ERRNO: N/A
*\NOMANUAL
*/

char * itlAgilex5Model (void)
    {
    int offset;

    if ((offset = vxFdtPathOffset("/")) >= 0)
        return((char *) vxFdtPropGet (offset, "model", NULL));
    else
        return("Unknown board");
    }


/*******************************************************************************
*
* itlAgilex5Reset - transfer control to the ROM monitor
*
* This routine transfers control to the ROM monitor.  Normally, it is called
* only by reboot() -- which services ^X -- and by fatal system errors such as
* exceptions at interrupt level.
*
* RETURNS: Does not return.
* ERRNO: N/A
*
*\NOMANUAL
*
*/

VOID itlAgilex5Reset
    (
    int startType
    )
    {
    STATUS rc;
    if ((sysWarmBootFunc == NULL) ||
       ((sysWarmBootFunc != NULL) && (startType == BOOT_CLEAR)))
        {
        goto cold;
        }


#ifdef _WRS_CONFIG_SMP
    armSysToMonitor((FUNCPTR)itlAgilex5Core0Exit, (FUNCPTR)itlAgilex5PspExit,
                    startType);
#else
    armSysToMonitor((FUNCPTR)itlAgilex5Core0Exit, NULL, startType);
#endif

cold:
    rc = vxPsciSysRst();
    if (rc != OK)
        {
        DEBUG_MSG ("vxPsciSysRst() failed\n");
        }

    vxbMsDelay (2000);
    DEBUG_MSG ("itlAgilex5Reset() failed\n");
    }

/*******************************************************************************
*
* itlAgilex5UsDelay - delay at least the specified amount of time (in microseconds)
* using a busy loop delay.
*
*/

VOID itlAgilex5UsDelay
    (
    int delay        /* length of time in microsec to delay */
    )
    {
    UINT32 oldVal;
    UINT32 newVal;
    UINT32 totalDelta;
    UINT32 decElapsed = 0;
    UINT32 maxTickCount = 0xffffffff;

    if(delay <= 0)
        {
        return;
        }

    totalDelta = (UINT32) delay * timerClkPerUs;

    oldVal = __inline__GetVirtTimerValue ();

    while (decElapsed < totalDelta)
        {
        newVal = __inline__GetVirtTimerValue ();

        if(newVal == oldVal)
            continue;

        /* no rollover */

        if (oldVal > newVal)
            decElapsed += (oldVal - newVal);

        /* rollover */

        else
            decElapsed += ((maxTickCount - newVal) + oldVal);

        oldVal = newVal;
        }
    }


#ifdef _WRS_CONFIG_SMP
/*******************************************************************************
*
* itlAgilex5CpuAvail - return the number of CPUs available for SMP
*
* This routine returns the number of CPUs available for SMP.
*
* RETURNS: the number of CPUs available for SMP.
*
* ERRNO: N/A
*\NOMANUAL
*/

UINT32 itlAgilex5CpuAvail (void)
    {
    return 4;
    }


IMPORT UINT32 vxMpidrGet(void);

/*******************************************************************************
*
* itlAgilex5CpuEnable - enable a multi core CPU
*
* This routine brings a CPU out of reset
*
* RETURNS: OK or ERROR
*\NOMANUAL
*/

STATUS itlAgilex5CpuEnable
    (
    UINT32           cpuId,
    WIND_CPU_STATE * cpuState
    )
    {
    STATUS rc;
    INT32 ret;

    if (cpuId >= itlAgilex5CpuAvail())
        {
        return ERROR;
        }

    rc = armMpCoreInit(cpuId, cpuState, NULL);
    if (rc != OK)
        {
        return ERROR;
        }

   if (sysStartType == BOOT_NORMAL)
        {
        return armMonitorSpinRelease(cpuId);
        }
    else
        {
        PHYS_ADDR physAddr;
        ULONG targetCpu;

        rc = vmTranslate(NULL, (VIRT_ADDR)sysInit, &physAddr);
        if (rc != OK)
            {
            DEBUG_MSG("vmTranslate() failed\n");
            return ERROR;
            }

        targetCpu = vxCpuIdGetByIndex(cpuId);
        ret = vxPsciCpuOn(targetCpu, physAddr, 0);
        if (ret != PSCI_SUCCESS)
            {
            return ERROR;
            }
        }

    return OK;
    }


/*******************************************************************************
*
* itlAgilex5PspExit - Secondary core shutdown routine
*
*/

LOCAL VOID itlAgilex5PspExit
    (
    UINT32 cpuId
    )
    {
    (void)intCpuLock();

    vxCpuOffHooksInvoke(cpuId);
    }


#endif /* ifdef _WRS_CONFIG_SMP */


/*******************************************************************************
*
* itlAgilex5Core0Exit - Core 0 shutdown routine
*
*/

LOCAL VOID itlAgilex5Core0Exit
    (
    UINT32 apCore
    )
    {
    (void)intCpuLock();

    vxCpuOffHooksInvoke(apCore);
    }


/*******************************************************************************
*
* itlAgilex5DebugInit - initialize debug console using pmapGlobalMap()
*
*/

LOCAL VOID itlAgilex5DebugInit(VOID)
    {
    int          offset;
    PHYS_ADDR    phyAddr;
    size_t       len;
    STATUS       rc;

    offset = vxFdtNodeOffsetByCompatible (0, "altr,designware-uart");
    if (offset < 0)
        {
        return;
        }

    rc = vxFdtDefRegGet(offset, 0, &phyAddr, &len);
    if (rc != OK)
        {
        return;
        }

    uartBaseAdrs = (VIRT_ADDR) pmapGlobalMap(phyAddr, len, VXB_REG_MAP_MMU_ATTR);
    if (uartBaseAdrs == (VIRT_ADDR) PMAP_FAILED)
        {
        return;
        }

    itlAgilex5UartInit();
    _func_kwrite = itlAgilex5Kwrite;
    }


/*******************************************************************************
*
* itlAgilex5EarlyDebugInit - Initialize debug console.
*
*/

LOCAL VOID itlAgilex5EarlyDebugInit(VOID)
    {
    int          offset;
    PHYS_ADDR    phyAddr;
    size_t       len;
    STATUS       rc;

    offset = vxFdtNodeOffsetByCompatible(0, "altr,designware-uart");
    if (offset < 0)
        {
        return;
        }

    rc = vxFdtDefRegGet(offset, 0, &phyAddr, &len);
    if (rc != OK)
        {
        return;
        }

    /* Early mapping: virtual = physical */

    uartBaseAdrs = (VIRT_ADDR) phyAddr;

    itlAgilex5UartInit();
    _func_kwrite  = itlAgilex5Kwrite;
    }


/*******************************************************************************
*
* itlAgilex5UartInit - Initialize UART
*
*/

LOCAL VOID itlAgilex5UartInit(VOID)
    {
    UINT32       xtal;
    UINT32       divisor;
    UINT32       baud;

    xtal = l4SpClkFreq;
    baud = NS16550_DEFAULT_BAUD;
    divisor = (UINT32) ((xtal / ( 8UL * baud ) + 1) / 2);

    /* Disable interrupts */

    CSR_WRITE_32(uartBaseAdrs + UART_REG_IER , 0x0);
    CSR_WRITE_32(uartBaseAdrs + UART_REG_FCR , 0x0);
    CSR_WRITE_32(uartBaseAdrs + UART_REG_LCR , UART_REG_LCR_BIT_DLAB |
                                              UART_REG_LCR_BITS_DEFAULT);
    /* Set baud rate divisor */

    CSR_WRITE_32(uartBaseAdrs + UART_REG_THR , divisor);
    CSR_WRITE_32(uartBaseAdrs + UART_REG_LCR , UART_REG_LCR_BITS_DEFAULT);
    }


/*******************************************************************************
*
* itlAgilex5Kwrite - write characters into UART with poll mode.
*
* RETURNS: OK, or ERROR if failed to write.
*
*/

LOCAL STATUS itlAgilex5Kwrite
    (
    char * buffer,
    size_t len
    )
    {
    if ((buffer == NULL) || (len == 0))
        return OK;

    while (len-- > 0)
        {
        if (*buffer == '\n')
            {
            while((CSR_READ_32 (uartBaseAdrs + UART_REG_LSR) &
                   UART_REG_LSR_BIT_TEMT) == 0)
                ;
            CSR_WRITE_32 (uartBaseAdrs + UART_REG_THR, '\r');
            }

        while((CSR_READ_32 (uartBaseAdrs + UART_REG_LSR) &
               UART_REG_LSR_BIT_TEMT) == 0)
            ;
        CSR_WRITE_32 (uartBaseAdrs + UART_REG_THR, *buffer);

        buffer++;
        }

    return OK;
    }

/*******************************************************************************
*
* itlAgilex5PhysMemInfoGet - get physical memory information
*
* This routine obtains the physical memory information and write them into
* <pPhysMemDesc> table.
*
* NOTE:
*   1. The region number should NOT exceed <physMemDescNum>;
*   2. All the attributes must be correct value that required by ARCH;
*   3. The kernel must be in the first physical memory region;
*   3. This routine is used by usrMmuInit().
*
* Note that this routine can only be used in itlAgilex5EarlyInit().
*
* RETURNS: region number
*/

LOCAL int itlAgilex5PhysMemInfoGet
    (
    PHYS_MEM_DESC * pPhysMemDesc,
    int             physMemDescNum
    )
    {
    int    regNum;
    int    i = 0;
    UINT   uState   = MMU_ATTR_VALID | MMU_ATTR_SUP_RWX
                       | MMU_ATTR_SPL_3 | MMU_ATTR_CACHE_COPYBACK
                       | MMU_ATTR_CACHE_COHERENCY;

    UINT   uStateMask  = MMU_ATTR_VALID_MSK | MMU_ATTR_PROT_MSK
                          | MMU_ATTR_SPL_MSK | MMU_ATTR_CACHE_MSK;

    regNum = vxFdtPhysMemInfoGet (pPhysMemDesc, physMemDescNum, uState);
    if (regNum > physMemDescNum)
        {
        regNum = physMemDescNum;
        }

    for (i = 0; i < regNum; i++)
        {
        pPhysMemDesc[i].initialStateMask = uStateMask;
        }

    return (regNum);
    }


/*******************************************************************************
*
* itlAgilex5DevInit - Initialize devices
*
*/

LOCAL STATUS itlAgilex5DevInit(VOID)
    {
    clkMgrBaseAdrs = (VIRT_ADDR) pmapGlobalMap(CLKMGR_PHYS_ADRS, CLKMGR_ADRS_LEN, VXB_REG_MAP_MMU_ATTR);
    if (clkMgrBaseAdrs == (VIRT_ADDR) PMAP_FAILED)
        {
        return ERROR;
        }

    rstMgrBaseAdrs = (VIRT_ADDR) pmapGlobalMap(RSTMGR_PHYS_ADRS, RSTMGR_ADRS_LEN, VXB_REG_MAP_MMU_ATTR);
    if (rstMgrBaseAdrs == (VIRT_ADDR) PMAP_FAILED)
        {
        return ERROR;
        }

    sysMgrBaseAdrs = (VIRT_ADDR) pmapGlobalMap(SYSMGR_PHYS_ADRS, SYSMGR_ADRS_LEN, VXB_REG_MAP_MMU_ATTR);
    if (sysMgrBaseAdrs == (VIRT_ADDR) PMAP_FAILED)
        {
        return ERROR;
        }

    return OK;
    }
