/* vxbHpnfcCore.h -  core code header for HPNFC driver */

/*
 * Copyright (c) 2012, 2014-2015, 2019, 2021 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify, or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

#ifndef __INCvxbHpnfcCoreh
#define __INCvxbHpnfcCoreh


#include <vxWorks.h>
#include <stdlib.h>
#include <string.h>
#include <logLib.h>
#include <vxAtomicLib.h>
#include <hwif/vxBus.h>
#include <hwif/util/vxbDmaBufLib.h>
#include <nand/vxbNandFlash.h>
#include <private/semLibP.h>
#include <hpnfc/vxbHpnfc.h>

#ifdef  __cplusplus
 extern "C" {
#endif

#define HPNFC_MAX_CS                         8

#define HPNFC_MIN_TX_BYTES                   4  /* must be multiple of 4 */
#define HPNFC_MIN_RX_BYTES                   4  /* must be multiple of 4 */

#define BIT(x)                               ((uint32_t)0x1 << (x))
#define GENMASK(h, l)                        (((~0u) << (l)) & \
                                              (~0u >> (32 - 1 - (h))))
/* register map */

#define HPNFC_CMD_REG0                       0x0000
#define   HPNFC_CMD_CT(x)                      (((UINT32)(x) & 0x3) << 30)
#define   HPNFC_CMD_CT_CDMA                    HPNFC_CMD_CT(0)
#define   HPNFC_CMD_CT_PIO                     HPNFC_CMD_CT(1)
#define   HPNFC_CMD_CT_GEN                     HPNFC_CMD_CT(3)
#define   HPNFC_CMD_TRD_NUM(x)                 (((UINT32)(x) & 0x7) << 24)
#define   HPNFC_CMD_INT                        (((UINT32)(x) & 0x1) << 20)
#define HPNFC_CMD_REG1                       0x0004
#define HPNFC_CMD_REG2                       0x0008
#define   HPNFC_CMD_LAYOUT_CMD(x)              (((UINT64)(x) & 0xff) << 16)

#define   HPNFC_CMD_LAYOUT_ADDR0(x)            (((UINT64)(x) & 0xff) << 16)
#define   HPNFC_CMD_LAYOUT_NUM_BYTES(x)        ((((UINT64)(x) & 0x3) - 1) << 11)

#define   HPNFC_CMD_LAYOUT_LAST_SEC_SIZE(x)    (((UINT64)(x) & 0xffff) << 40)
#define   HPNFC_CMD_LAYOUT_SEC_CNT(x)          (((UINT64)(x) & 0xff) << 32)
#define   HPNFC_CMD_LAYOUT_DIR_RD              ((UINT64)0x0 << 11)
#define   HPNFC_CMD_LAYOUT_DIR_WR              ((UINT64)0x1 << 11)

#define   HPNFC_CMD_LAYOUT_BANK(x)             (((UINT64)(x) & 0x3) << 8)
#define   HPNFC_CMD_LAYOUT_JEDEC               ((UINT64)0x1 << 7)
#define   HPNFC_CMD_LAYOUT_TWB                 ((UINT64)0x1 << 6)
#define   HPNFC_CMD_LAYOUT_TYPE(x)             (((UINT64)(x) & 0x3f) << 0)
#define   HPNFC_CMD_LAYOUT_TYPE_CMD_SEQ        HPNFC_CMD_LAYOUT_TYPE(0)
#define   HPNFC_CMD_LAYOUT_TYPE_ADDR_SEQ       HPNFC_CMD_LAYOUT_TYPE(1)
#define   HPNFC_CMD_LAYOUT_TYPE_DATA_SEQ       HPNFC_CMD_LAYOUT_TYPE(2)
#define HPNFC_CMD_REG3                       0x000c
#define HPNFC_CMD_STS_PTR                    0x0010
#define HPNFC_CMD_STS                        0x0014
#define HPNFC_CMD_STS_EXT                    0x0018
#define HPNFC_CMD_REG4                       0x0020
#define HPNFC_CMD_REG5                       0x0024
#define HPNFC_CMD_REG6                       0x0028
#define HPNFC_INTR_STS                       0x0110
#define HPNFC_INTR_EN                        0x0114
#define   HPNFC_INTR_GLOBAL                    BIT(31)
#define   HPNFC_INTR_PROT_ERR                  BIT(25)
#define   HPNFC_INTR_DI_CTX_ERR                BIT(24)
#define   HPNFC_INTR_ZQCL_ERR                  BIT(23)
#define   HPNFC_INTR_SDMA_ERR                  BIT(22)
#define   HPNFC_INTR_SDMA_TRIG                 BIT(21)
#define   HPNFC_INTR_CMD_IGNORED               BIT(20)
#define   HPNFC_INTR_CMD_REG_PAR_ERR           BIT(19)
#define   HPNFC_INTR_DDMA_TERR                 BIT(18)
#define   HPNFC_INTR_CDMA_TERR                 BIT(17)
#define   HPNFC_INTR_CDMA_IDLE                 BIT(16)
#define   HPNFC_INTR_ALL                       0x003ff000
#define   HPNFC_INTR_NONE                      0
#define HPNFC_CTRL_STS                       0x0118
#define   HPNFC_CTRL_STS_SDMA_PAUSED           BIT(16)
#define   HPNFC_CTRL_STS_INIT_FAIL             BIT(10)
#define   HPNFC_CTRL_STS_INIT_COMP             BIT(9)
#define   HPNFC_CTRL_STS_CTRL_BUSY             BIT(8)
#define   HPNFC_CTRL_STS_ZQCL_BUSY             BIT(4)
#define   HPNFC_CTRL_STS_MC_BUSY               BIT(3)
#define   HPNFC_CTRL_STS_CMD_ENG_BUSY          BIT(2)
#define   HPNFC_CTRL_STS_MDMA_BUSY             BIT(1)
#define   HPNFC_CTRL_STS_SDMA_BUSY             BIT(0)
#define HPNFC_TRD_STS                        0x0120
#define HPNFC_TRD_ERR_INTR_STS               0x0128
#define HPNFC_TRD_ERR_INTR_EN                0x0130
#define HPNFC_TRD_COMP_INTR_STS              0x0138
#define HPNFC_DMA_TGT_ERR_L                  0x0140
#define HPNFC_DMA_TGT_ERR_H                  0x0144
#define HPNFC_BOOT_STS                       0x0148
#define HPNFC_TRD_TO_INTR_STS                0x014c
#define HPNFC_TRD_TO_INTR_EN                 0x0154
#define HPNFC_ZQ_CAL_STS                     0x015c

#define HPNFC_TSFR_CFG0                      0x0400
#define HPNFC_TSFR_CFG1                      0x0404
#define HPNFC_LNG_POLL                       0x0408
#define HPNFC_SHT_POLL                       0x040c
#define HPNFC_RDST_CTRL0                     0x0410
#define HPNFC_RDST_CTRL1                     0x0414
#define HPNFC_LUN_STS_CMD                    0x0418
#define HPNFC_LUN_INTRLV_CMD                 0x041c
#define HPNFC_LUN_ADDR_OFST                  0x0420
#define HPNFC_NF_DEV_LAYOUT                  0x0424
#define HPNFC_ECC_CFG0                       0x0428
#define HPNFC_ECC_CFG1                       0x042c
#define HPNFC_DEV_CTRL                       0x0430
#define HPNFC_MP_CFG                         0x0434
#define HPNFC_CACHE_CFG                      0x0438
#define HPNFC_DMA_STGS                       0x043c
#define HPNFC_SDMA_SIZE                      0x0440
#define HPNFC_SDMA_TRD_NUM                   0x0444
#define HPNFC_TIME_OUT                       0x0448
#define HPNFC_SDMA_ADDR0                     0x044c
#define HPNFC_SDMA_ADDR1                     0x0450
#define HPNFC_FIFO_TRG_LVL                   0x0454
#define HPNFC_ZQ_CAL_CFG                     0x0458
#define HPNFC_ZQ_CAL_EN                      0x045c
#define HPNFC_LNG_CAL_PERIOD                 0x0460
#define HPNFC_SHT_CAL_PERIOD                 0x0464
#define HPNFC_REMAP_CTRL                     0x0480
#define HPNFC_REMAP_MASK                     0x0484
#define HPNFC_REMAP_ASDMA                      0x0488
#define HPNFC_REMAP_LOG_ADR                  0x048c
#define HPNFC_REMAP_PHYS_ADR                 0x0490
#define HPNFC_CTRL_DATA_CTRL                 0x0494

#define HPNFC_DI_CTRL                        0x0700
#define HPNFC_DI_INJECT0                     0x0704
#define HPNFC_DI_INJECT1                     0x0708
#define HPNFC_DI_ERR_REG_ADR                 0x070c
#define HPNFC_DI_INJECT2                     0x0710

#define HPNFC_CTRL_VER                       0x0800
#define HPNFC_CTRL_FEAT                      0x0804
#define HPNFC_MFG_ID                         0x0808
#define HPNFC_NF_DEV_AREAS                   0x080c
#define HPNFC_DEV_PARMS0                     0x0810
#define HPNFC_DEV_PARMS1                     0x0814
#define HPNFC_DEV_FEAT                       0x0818
#define HPNFC_DEV_BLKS_PER_LUN               0x081c
#define HPNFC_DEV_REV                        0x0820
#define HPNFC_ONFI_TMG_MDS0                  0x0824
#define HPNFC_ONFI_TMG_MDS1                  0x0828
#define HPNFC_ONFI_INTRLV_OP_ATTR            0x082c
#define HPNFC_ONFI_SYNC_OPT0                 0x0830
#define HPNFC_ONFI_SYNC_OPT1                 0x0834
#define HPNFC_BCH_CFG0                       0x0838
#define HPNFC_BCH_CFG1                       0x083c
#define HPNFC_BCH_CFG2                       0x0840
#define HPNFC_BCH_CFG3                       0x0844
#define HPNFC_DEV_FEAT1                      0x0850

#define HPNFC_PROT_CTRL0                     0x0900
#define HPNFC_PROT_DN0                       0x0904
#define HPNFC_PROT_UP0                       0x0908
#define HPNFC_PROT_CTRL1                     0x0910
#define HPNFC_PROT_DN1                       0x0914
#define HPNFC_PROT_UP1                       0x0918

#define HPNFC_WP_STGS                        0x1000
#define HPNFC_RBN_STS                        0x1004
#define   HPNFC_RBN_STS_RDY_MASK(cs)           (1 << (0x7 & (cs)))
#define HPNFC_CMN_STGS                       0x1008
#define   HPNFC_CMN_STGS_DEV_16BIT             BIT(8)
#define HPNFC_SKIP_BYTES_CFG                 0x100c
#define HPNFC_SKIP_BYTES_OFST                0x1010
#define HPNFC_TGL_TMG0                       0x1014
#define HPNFC_TGL_TMG1                       0x1018
#define HPNFC_ASYNC_TGL_TMG                  0x101c
#define HPNFC_SYNC_TMG                       0x1020
#define HPNFC_TMG0                           0x1024
#define HPNFC_TMG1                           0x1028
#define HPNFC_TMG2                           0x102c
#define HPNFC_TMG2                           0x102c
#define HPNFC_DLL_PHY_UPD_CNT                0x1030
#define HPNFC_DLL_PHY_CTRL                   0x1034

#define HPNFC_PHY_DQ_TMG                     0x2000
#define HPNFC_PHY_DQS_TMG                    0x2004
#define HPNFC_PHY_GATE_LPBK_CTRL             0x2008
#define HPNFC_PHY_DLL_MSTR_CTRL              0x200c
#define HPNFC_PHY_DLL_SLV_CTRL               0x2010
#define HPNFC_PHY_IE_TMG                     0x2014
#define HPNFC_PHY_OBS0                       0x2018
#define HPNFC_PHY_DLL_OBS0                   0x201c
#define HPNFC_PHY_DLL_OBS1                   0x2020
#define HPNFC_PHY_DLL_OBS2                   0x2024
#define HPNFC_PHY_STATIC_TGL                 0x2028
#define HPNFC_PHY_WR_DESKEW                  0x202c
#define HPNFC_PHY_WR_RD_DESKEW_CMD           0x2030
#define HPNFC_PHY_WR_DESKEW_PD_CTRL0         0x2034
#define HPNFC_PHY_WR_DESKEW_PD_CTRL1         0x2038
#define HPNFC_PHY_RD_DESKEW                  0x203c
#define HPNFC_PHY_VER                        0x2070
#define HPNFC_PHY_FEAT                       0x2074

#define HPNFC_PHY_CTRL                       0x2080
#define HPNFC_PHY_TSEL                       0x2084
#define HPNFC_PHY_GPIO_CTRL0                 0x2088
#define HPNFC_PHY_GPIO_CTRL1                 0x208c
#define HPNFC_PHY_GPIO_STS0                  0x2090
#define HPNFC_PHY_GPIO_STS1                  0x2094

typedef enum hpnfc_cs_type
   {
   HPNFC_CS_TYPE_OTHER = 0,
   HPNFC_CS_TYPE_NAND_8BIT,  /* 8-bit NAND Flash */
   HPNFC_CS_TYPE_NAND_16BIT  /* 16-bit NAND Flash */
   } HPNFC_CS_TYPE;

/* controller information */

typedef struct hpnfc
    {
    VXB_DEV_ID     pDev;
    void *         pRegHandle;
    void *         pSdmaHandle;
    VIRT_ADDR      uRegBase;
    VIRT_ADDR      uSdmaBase;
    VXB_RESOURCE * pIntRes;
    UINT32         uFlag;
    HPNFC_CS_TYPE  uCsType[HPNFC_MAX_CS];
    UINT8          uNandCsCnt;
    DL_LIST        chipList;
    SEMAPHORE      syncSem;
    /* variables below accommodate one chip select */
    UINT32         cmd;
    UINT32         bufIndex;
    UINT32         bufBytes;
    UINT8          buf[max(HPNFC_MIN_TX_BYTES, HPNFC_MIN_RX_BYTES)];
    } HPNFC, *pHPNFC;

/* controller access defines */

#define HPNFC_REG_READ(pHpnfc, offset)                    \
    vxbRead32 ((pHpnfc)->pRegHandle,                      \
               (void *)((pHpnfc)->uRegBase + (ULONG) (offset)))

#define HPNFC_REG_WRITE(pHpnfc, offset, data)             \
    vxbWrite32 ((pHpnfc)->pRegHandle,                     \
                (void *)((pHpnfc)->uRegBase + (ULONG)(offset)), (data))

#define HPNFC_REG_READ_BYTE(pHpnfc, offset)               \
    vxbRead8 ((pHpnfc)->pRegHandle,                       \
              (void *)((pHpnfc)->uRegBase + (ULONG) (offset)))

#define HPNFC_REG_WRITE_BYTE(pHpnfc, offset, data)        \
    vxbWrite8 ((pHpnfc)->pRegHandle,                      \
               (void *)((pHpnfc)->uRegBase + (ULONG) (offset)), (data))

#define HPNFC_REG_READ_WORD(pHpnfc, offset)               \
    vxbRead16 ((pHpnfc)->pRegHandle,                      \
               (void *)((pHpnfc)->uRegBase + (ULONG) (offset)))

#define HPNFC_REG_WRITE_WORD(pHpnfc, offset, data)        \
    vxbWrite16 ((pHpnfc)->pRegHandle,                     \
                (void *)((pHpnfc)->uRegBase + (ULONG) (offset)), (data))

/* SDMA access defines */

#define HPNFC_SDMA_READ(pHpnfc, offset)                   \
    vxbRead32 ((pHpnfc)->pSdmaHandle,                     \
               (void *)((pHpnfc)->uSdmaBase + (ULONG) (offset)))

#define HPNFC_SDMA_WRITE(pHpnfc, offset, data)            \
    vxbWrite32 ((pHpnfc)->pSdmaHandle,                    \
                (void *)((pHpnfc)->uSdmaBase + (ULONG)(offset)), (data))

#define HPNFC_SDMA_READ_BYTE(pHpnfc, offset)              \
    vxbRead8 ((pHpnfc)->pSdmaHandle,                      \
              (void *)((pHpnfc)->uSdmaBase + (ULONG) (offset)))

#define HPNFC_SDMA_WRITE_BYTE(pHpnfc, offset, data)       \
    vxbWrite8 ((pHpnfc)->pSdmaHandle,                     \
               (void *)((pHpnfc)->uSdmaBase + (ULONG) (offset)), (data))

#define HPNFC_SDMA_READ_WORD(pHpnfc, offset)              \
    vxbRead16 ((pHpnfc)->pSdmaHandle,                     \
               (void *)((pHpnfc)->uSdmaBase + (ULONG) (offset)))

#define HPNFC_SDMA_WRITE_WORD(pHpnfc, offset, data)       \
    vxbWrite16 ((pHpnfc)->pSdmaHandle,                    \
                (void *)((pHpnfc)->uSdmaBase + (ULONG) (offset)), (data))

#define HPNFC_DBG
#ifdef HPNFC_DBG
#   define HPNFC_DBG_MSG(fmt, a, b, c, d, e, f)                      \
    do                                                               \
        {                                                            \
        if (_func_logMsg != NULL)                                    \
            _func_logMsg (HPNFC_DRV_NAME": "fmt, a, b, c, d, e, f);  \
        } while (0)
#else /* !HPNFC_DBG */
#   define HPNFC_DBG_MSG(fmt, ...)
#endif /* HPNFC_DBG */

/* Function Declaration */

int vxbHpnfcFlashInit (pHPNFC pHpnfc);
void vxbHpnfcFlashClean (pHPNFC pHpnfc);
STATUS vxbHpnfcShow (VXB_DEV_ID pDev, UINT32 verbose);

#ifdef  __cplusplus
}
#endif

#endif /* __INCvxbHpnfcCoreh */

