/* vxbHpnfc.c -  Cadence HPNFC NAND Flash controller driver */

/*
 * Copyright (c) 2012-2014, 2016, 2019, 2021-2024 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify, or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION
This module, along with vxbHpnfcCore.c, implements a driver for Cadence
HPNFC NAND Flash controller. This module combination operates the controller
in what Cadence calls "Generic Work Mode." Interrupts, DMA and
hardware-based ECC are not used.

To add support for the HPNFC contoller, enable the following block device
management kernel components:

\cs
  DRV_VXB_HPNFC
  INCLUDE_MTD
  INCLUDE_MTD_SHOW
  INCLUDE_FTL
  INCLUDE_FTL_FORMAT
  INCLUDE_FTL_SHOW
\ce

and with these parameter values:

\cs
  MTD_CFG_STR "$hpnfc0#0:0x0,0x3fc,/ftl0#0,0"
\ce

If using DosFS with the NAND flash, many of these file-level kernel
components are useful or necessary:

\cs
  INCLUDE_DOSFS
  INCLUDE_RAWFS
  INCLUDE_ERF
  INCLUDE_DEVICE_MANAGER
  INCLUDE_FS_EVENT_UTIL
  INCLUDE_FS_MONITOR
  INCLUDE_XBD
  INCLUDE_XBD_BLK_DEV
  INCLUDE_DOSFS_DIR_FIXED
  INCLUDE_DOSFS_DIR_VFAT
  INCLUDE_DOSFS_FAT
  INCLUDE_DOSFS_FMT
  INCLUDE_DOSFS_CHKDSK
  INCLUDE_DOSFS_MAIN
\ce

This driver is bound using a device tree, and the device tree node for
the HPNFC NAND contoller should specify these properties:

\is
\i <compatible>
This property helps marry the device to the driver.  It should be set
to "cadence,hpnfc".

\i <reg>
This property specifies the physical base address and length of the
NAND contoller registers, and NAND contoller SDMA space.

\i <flag>
This property specifies hardware features which are currently disabled
or unsupported. The flag value should be 0xf to use interrupts and 0x1f
to use polling.
\ie

An example HPNFC controller FDT node is shown below:

\cs
        nand: nand@10b80000
            {
            compatible = "cadence,hpnfc";
            reg = <0x10b80000 0x10000>,    /@ regs @/
                  <0x10840000 0x10000>;    /@ sdma @/
            flag = <0xf>;
            };
\ce

INCLUDE FILES: vxbHpnfc.h vxbHpnfcCore.h

INTERNAL
This module combination supports a single 16-bit NAND flash on CS0 (i.e.,
chip select 0). Sensing and adapting to other configurations at run-time
is left for the day that more configurations are available to test.
*/

#include <hpnfc/vxbHpnfcCore.h>

#include <stdio.h>
#include <memLib.h>

#include <hwif/vxBus.h>
#include <hwif/buslib/vxbFdtLib.h>
#include <subsys/pinmux/vxbPinMuxLib.h>
#include <subsys/int/vxbIntLib.h>

/* defines */

/* globals */

UINT32 vxbHpnfcDummy;

/* locals */

/* VxBus methods */

LOCAL STATUS vxbHpnfcProbe (struct vxbDev * pDev);
LOCAL STATUS vxbHpnfcAttach (struct vxbDev * pDev);

/* Function declarations */

LOCAL STATUS vxbHpnfcHwInit (pHPNFC pHpnfc);

/* vxBus Gen 2 parameters */

LOCAL VXB_DRV_METHOD vxbHpnfcMethodList[] =
    {
    { VXB_DEVMETHOD_CALL(vxbDevProbe), vxbHpnfcProbe },
    { VXB_DEVMETHOD_CALL(vxbDevAttach), vxbHpnfcAttach },
    VXB_DEVMETHOD_END
    };

VXB_DRV vxbFdtHpnfcDrv =
    {
    { NULL } ,
    HPNFC_DRV_NAME,                 /* Name */
    "Cadence HPNFC NAND driver",    /* Description */
    VXB_BUSID_FDT,                  /* Class */
    0,                              /* Flags */
    0,                              /* Reference count */
    vxbHpnfcMethodList              /* Method table */
    };

LOCAL const VXB_FDT_DEV_MATCH_ENTRY vxbHpnfc_match[] =
    {
        {
        "cadence,hpnfc", /* compatible */
        (void *)NULL
        },
        {}
    };

VXB_DRV_DEF(vxbFdtHpnfcDrv)

/******************************************************************************
*
* vxbHpnfcProbe - probe for device match
*
* This routine checks the FDT device for a match to this driver.
*
* RETURNS: OK, or ERROR if no match.
*/

STATUS vxbHpnfcProbe
    (
    struct vxbDev * pDev
    )
    {
    STATUS                    status;
    VXB_FDT_DEV_MATCH_ENTRY * pEntry;

    /* check if the pDev pointer is valid */

    VXB_ASSERT (pDev != NULL, ERROR);

    status = vxbFdtDevMatch (pDev, vxbHpnfc_match, &pEntry);

    return (status);
    }

/*******************************************************************************
*
* vxbHpnfcAttach - attach HPNFC controller devices
*
* This routine initilizes an HPNFC controller and its NAND chips.
*
* RETURNS: OK, or ERROR if initialization fails.
*/

LOCAL STATUS  vxbHpnfcAttach
    (
    VXB_DEV_ID   pDev
    )
    {
    pHPNFC             pHpnfc;
    VXB_RESOURCE_ADR * pRegResAdr = NULL;
    VXB_RESOURCE_ADR * pSdmaResAdr = NULL;
    VXB_RESOURCE *     pRegRes;
    VXB_RESOURCE *     pSdmaRes;
    VXB_RESOURCE *     pIntRes;
    VXB_FDT_DEV *      pFdtDev;
    int                len;
    void *             pValue;

    pFdtDev = (VXB_FDT_DEV *)(vxbDevIvarsGet(pDev));

    pHpnfc = vxbMemAlloc (sizeof (HPNFC));

    if (pHpnfc == NULL)
        {
        HPNFC_DBG_MSG ("failed to allocate HPNFC information\n",
                       1, 2, 3, 4, 5, 6);
        return (ERROR);
        }

    bzero ((char *) pHpnfc, sizeof (HPNFC));
    pHpnfc->pDev = pDev;
    vxbDevSoftcSet(pDev, pHpnfc);

    /* acquire resources */

    pRegRes = vxbResourceAlloc(pDev, VXB_RES_MEMORY, 0);

    if(pRegRes == NULL)
        {
        HPNFC_DBG_MSG ("failed to allocate register\n", 1, 2, 3, 4, 5, 6);
        goto vxbHpnfcAttachError0;
        }

    pRegResAdr = (VXB_RESOURCE_ADR *)pRegRes->pRes;

    pHpnfc->uRegBase = pRegResAdr->virtual;
    pHpnfc->pRegHandle = pRegResAdr->pHandle;

    pSdmaRes = vxbResourceAlloc(pDev, VXB_RES_MEMORY, 1);

    if(pSdmaRes == NULL)
        {
        HPNFC_DBG_MSG ("failed to allocate SDMA\n", 1, 2, 3, 4, 5, 6);
        goto vxbHpnfcAttachError1;
        }

    pSdmaResAdr = (VXB_RESOURCE_ADR *)pSdmaRes->pRes;

    pHpnfc->uSdmaBase = pSdmaResAdr->virtual;
    pHpnfc->pSdmaHandle = pSdmaResAdr->pHandle;

    pIntRes = vxbResourceAlloc(pDev, VXB_RES_IRQ, 0);

    if(pIntRes == NULL)
        {
        HPNFC_DBG_MSG ("failed to allocate interrupt\n", 1, 2, 3, 4, 5, 6);
        goto vxbHpnfcAttachError2;
        }

    pHpnfc->pIntRes = pIntRes;

    /* obtain fdt parameters */

    pValue = (void *)vxFdtPropGet( pFdtDev->offset, "flag", &len);

    if (pValue != NULL)
        {
        pHpnfc->uFlag = vxFdt32ToCpu(*(UINT32 *)pValue);
        }

    DLL_INIT (&pHpnfc->chipList);

    semBInit (&pHpnfc->syncSem, SEM_Q_FIFO, SEM_EMPTY);

    /* initialize controller */

    if (vxbHpnfcHwInit (pHpnfc) != OK)
        {
        HPNFC_DBG_MSG ("failed to initializes HPNFC controller\n",
                       1, 2, 3, 4, 5, 6);
        goto vxbHpnfcAttachError3;
        }

    /* initialize flash chips attached to HPNFC */

    if (vxbHpnfcFlashInit (pHpnfc) != OK)
        {
        HPNFC_DBG_MSG ("hpnfc FlashInit fail\n", 1, 2, 3, 4, 5, 6);
        goto vxbHpnfcAttachError4;
        }

    return OK;

vxbHpnfcAttachError4:
    (void) vxbIntDisable (pHpnfc->pDev, pHpnfc->pIntRes);
vxbHpnfcAttachError3:
    (void)semTerminate (&pHpnfc->syncSem);
    (void)vxbResourceFree (pDev, pIntRes);
vxbHpnfcAttachError2:
    (void)vxbResourceFree (pDev, pSdmaRes);
vxbHpnfcAttachError1:
    (void)vxbResourceFree (pDev, pRegRes);
vxbHpnfcAttachError0:
    (void)vxbMemFree (pHpnfc);

    return ERROR;
    }

/*******************************************************************************
*
* vxbHpnfcInt - HPNFC interrupt handler
*
* This routine handles the HPNFC interrupts.
*
* RETURNS: N/A
*/

LOCAL void vxbHpnfcInt
    (
    pHPNFC pHpnfc
    )
    {
    UINT32 val;
    
    val = HPNFC_REG_READ (pHpnfc, HPNFC_INTR_STS);

    if ((val & HPNFC_INTR_SDMA_TRIG) != 0)
        {
        semGive (&pHpnfc->syncSem);
        }

    HPNFC_REG_WRITE (pHpnfc, HPNFC_INTR_STS, HPNFC_INTR_ALL);

    /* force write with a read */

    vxbHpnfcDummy = HPNFC_REG_READ (pHpnfc, HPNFC_INTR_STS);
    }

/*******************************************************************************
*
* vxbHpnfcHwInit - initialize HPNFC controller
*
* This routine initializes the HPNFC controller.
*
* RETURNS: OK.
*/

LOCAL STATUS vxbHpnfcHwInit
    (
    pHPNFC pHpnfc
    )
    {
    int i;

    if ((pHpnfc->uFlag & HPNFC_FLAG_INTDIS) == 0)
        {
        /* connect and enable interrupt */

        HPNFC_REG_WRITE (pHpnfc, HPNFC_INTR_EN, HPNFC_INTR_NONE);
        HPNFC_REG_WRITE (pHpnfc, HPNFC_INTR_STS, HPNFC_INTR_ALL);

        (void) vxbIntConnect (pHpnfc->pDev, pHpnfc->pIntRes, vxbHpnfcInt,
                              (void *)pHpnfc);

        (void) vxbIntEnable (pHpnfc->pDev, pHpnfc->pIntRes);

        HPNFC_REG_WRITE (pHpnfc, HPNFC_INTR_EN,
                         HPNFC_INTR_GLOBAL | HPNFC_INTR_SDMA_TRIG);
        }

    /* set up the chip selects */

    for (i = 0; i < HPNFC_MAX_CS; i++)
        {
        if (i == 0)
            {
            /* In the Intel simics model, the flash is 16-bits, and the
             * HPNFC_CMN_STGS register is set (by U-boot?) to 16-bits.
             * Even so, the controller interface presents the NAND flash
             * as if it is 8-bits for every command observed generated by
             * the stack except for NAND_CMD_CODE_STARD.  We therefore,
             * tell the stack to use the byteRd/byteWr routines through
             * uCsType[] and componsate for the NAND_CMD_CODE_STARD case
             * in vxbHpnfcUnitRd().
             */

            pHpnfc->uCsType[i] = HPNFC_CS_TYPE_NAND_8BIT;
            }
        }

    return (OK);
    }

#ifdef _WRS_CONFIG_DEBUG_FLAG
/*******************************************************************************
*
* vxbHpnfcShow - show HPNFC information
*
* This routine prints the HPNFC information.
*
* RETURNS: OK always
*/

STATUS vxbHpnfcShow
    (
    VXB_DEV_ID      pDev,
    UINT32          verbose
    )
    {
    DL_NODE *           pNode;
    pNAND_FLASH_INFO    pNandInfo;
    UINT32              i;
    pHPNFC              pHpnfc = (pHPNFC) vxbDevSoftcGet(pDev);

    printf ("            Initialization Status              : ");
    if (pHpnfc == NULL)
        {
        printf ("Failed\n");
        return (OK);
        }
    else
        printf ("Succeeded 0x%x\n", pHpnfc);

    printf ("            Register Base                      : %p\n",
            pHpnfc->uRegBase);
    printf ("            SDMA Base                          : %p\n",
            pHpnfc->uSdmaBase);

    printf ("            Prefecth/Postwrite Engine Enabled  : %s\n",
            ((pHpnfc->uFlag & HPNFC_FLAG_PFPWDIS) == 0) ? "Yes" : "No");
    printf ("            DMA Function Enabled               : %s\n",
            ((pHpnfc->uFlag & HPNFC_FLAG_DMADIS) == 0) ? "Yes" : "No");
    printf ("            NAND Cache Function Enabled        : %s\n",
            ((pHpnfc->uFlag & HPNFC_FLAG_CACHEDIS) == 0) ? "Yes" : "No");
    printf ("            Zero-Copy Function Enabled         : %s\n",
            ((pHpnfc->uFlag & HPNFC_FLAG_ZCOPYDIS) == 0) ? "Yes" : "No");

    i = 0;
    pNode = DLL_FIRST (&pHpnfc->chipList);
    while (pNode != NULL)
        {
        pNandInfo = (pNAND_FLASH_INFO) pNode;
        printf ("            ----------------------------------------------\n");
        printf ("            CS %d: \n", i++);
        (void)vxbNandFlashShow (pNandInfo, (int)verbose);
        pNode = DLL_NEXT (pNode);
        }

    return (OK);
    }

#endif /* _WRS_CONFIG_DEBUG_FLAG */
