/* vxbHpnfcCore.c - Cadence HPNFC NAND Flash controller core driver */

/*
 * Copyright (c) 2012-2014, 2016, 2021, 2023-2024 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify, or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION
See the main module, vxbHpnfc.c, for a description.

INCLUDE FILES: vxbHpnfc.h vxbHpnfcCore.h
*/

#include <stdio.h>
#include <logLib.h>
#include <hpnfc/vxbHpnfcCore.h>

#include <subsys/flash/vxbFlashLib.h>
#include <subsys/timer/vxbTimerLib.h>
#include <hwif/buslib/vxbFdtLib.h>
#include <taskLib.h>

/* defines */

/* #define HPNFC_OPS_DUMP - remove this code before formal release */

/* globals */

UINT64 vxbHpnfcBufRdCanary = 0;  /* for debugging, should not happen */
UINT64 vxbHpnfcBufWrCanary = 0;  /* for debugging, should not happen */

/* locals */

/* NAND Flash functions */

LOCAL int    vxbHpnfcCmdIssue (pNAND_FLASH_INFO, UINT16, UINT32);
LOCAL UINT32 vxbHpnfcDataSeqIssue (pNAND_FLASH_INFO, UINT8 *, UINT32, BOOL);
LOCAL UINT8  vxbHpnfcByteRd (pNAND_FLASH_INFO);
LOCAL int    vxbHpnfcByteWr (pNAND_FLASH_INFO, UINT8);
LOCAL UINT16 vxbHpnfcWordRd (pNAND_FLASH_INFO);
LOCAL int    vxbHpnfcWordWr (pNAND_FLASH_INFO, UINT16);
LOCAL int    vxbHpnfcBufRd  (pNAND_FLASH_INFO, pNAND_BUF_INFO);
LOCAL int    vxbHpnfcBufWr  (pNAND_FLASH_INFO, pNAND_BUF_INFO);
LOCAL int    vxbHpnfcRdyChk (pNAND_FLASH_INFO);

#ifdef HPNFC_OPS_DUMP
IMPORT STATUS stampit (const char * fmt, ...);

#define HPNFC_MAX_DUMP_BYTES 16
LOCAL UINT32 rdCount = 0;
LOCAL UINT32 wrCount = 0;
LOCAL UINT8  rdBytes[HPNFC_MAX_DUMP_BYTES];
LOCAL UINT8  wrBytes[HPNFC_MAX_DUMP_BYTES];
LOCAL char   dumpBuf[3 * HPNFC_MAX_DUMP_BYTES + 1];

/*******************************************************************************
*
* vxbHpnfcDump - dump data bytes
*
* This routine converts the rdBytes[] and wrBytes[] arrays to readable
* output.
*
* RETURNS: OK.
*/

STATUS vxbHpnfcDump (void)
    {
    int    i;
    int    dumpCount;
    char * pBuf = dumpBuf;

    if (rdCount != 0)
        {
        dumpCount = min ((int)rdCount, HPNFC_MAX_DUMP_BYTES);

        for (i = 0; i < dumpCount; i++)
            {
            pBuf += sprintf(pBuf, " %02x", rdBytes[i]);
            }

        stampit ("read %d:%s%s", rdCount, dumpBuf,
                 rdCount > HPNFC_MAX_DUMP_BYTES ? "..." : "");
        rdCount = 0;
        }

    if (wrCount != 0)
        {
        dumpCount = min ((int)wrCount, HPNFC_MAX_DUMP_BYTES);

        for (i = 0; i < dumpCount; i++)
            {
            pBuf += sprintf(pBuf, " %02x", wrBytes[i]);
            }

        stampit ("wrote %d:%s%s", wrCount, dumpBuf,
                 wrCount > HPNFC_MAX_DUMP_BYTES ? "..." : "");
        wrCount = 0;
        }

    return (OK);
    }
#endif /* HPNFC_OPS_DUMP */

/*******************************************************************************
*
* vxbHpnfcFlashInit - initialize Flash chips attached to HPNFC
*
* This routine initializes the Flash chips attached to the HPNFC.
*
* RETURNS: OK, or ERROR if failed.
*/

int vxbHpnfcFlashInit
    (
    pHPNFC pHpnfc
    )
    {
    UINT32           uCs = 0;
    pNAND_FLASH_INFO pNandInfo;
    VXB_FLASHCTRL *  pCtrl = NULL;

    while (uCs < HPNFC_MAX_CS)
        {
        if (pHpnfc->uCsType[uCs] != HPNFC_CS_TYPE_OTHER)
            {
            pNandInfo = vxbNandInfoCreate ();

            if (pNandInfo == NULL)
                {
                HPNFC_DBG_MSG ("failed to create NAND Flash information for "
                              "CS %d\n", uCs, 2, 3, 4, 5, 6);
                goto vxbHpnfcFlashInitError;
                }

            pNandInfo->pCtrl    = (void *) pHpnfc;
            pNandInfo->uCs      = uCs;
            pNandInfo->cmdIssue = vxbHpnfcCmdIssue;
            pNandInfo->byteRd   = vxbHpnfcByteRd;
            pNandInfo->byteWr   = vxbHpnfcByteWr;
            pNandInfo->wordRd   = vxbHpnfcWordRd;
            pNandInfo->wordWr   = vxbHpnfcWordWr;
            pNandInfo->bufRd    = vxbHpnfcBufRd;
            pNandInfo->bufWr    = vxbHpnfcBufWr;

            if (pHpnfc->uCsType[uCs] == HPNFC_CS_TYPE_NAND_16BIT)
                pNandInfo->uCap |= NAND_INFO_CAP_16BIT;

            if ((pHpnfc->uFlag & HPNFC_FLAG_CACHEDIS) != 0)
                pNandInfo->uCap |= NAND_INFO_CAP_CACHEDIS;

            if ((pHpnfc->uFlag & HPNFC_FLAG_ZCOPYDIS) != 0)
                pNandInfo->uCap |= NAND_INFO_CAP_ZCOPYDIS;

            pNandInfo->rdyChk = vxbHpnfcRdyChk;

            /* initialize NAND Flash information */

            if (vxbNandInfoInit (pNandInfo) != OK)
                {
                HPNFC_DBG_MSG ("failed to initialize NAND Flash information for "
                              "CS %d\n", uCs, 2, 3, 4, 5, 6);
                goto vxbHpnfcFlashInitError;
                }

            /* link the chip list */

            DLL_ADD (&pHpnfc->chipList, &pNandInfo->node);

            /* add to chip list of flash subsystem */

            if ((pCtrl = vxbMemAlloc(sizeof(VXB_FLASHCTRL)))== NULL)
                {
                HPNFC_DBG_MSG ("failed to add to  flash subsystem for "
                              "CS %d\n", uCs, 2, 3, 4, 5, 6);
                goto vxbHpnfcFlashInitError;
                }

            pCtrl->pDev = pHpnfc->pDev;
            pCtrl->unitId = uCs;
            snprintf (pCtrl->devName, sizeof(HPNFC_DRV_NAME) + 1,
                      HPNFC_DRV_NAME);
            pCtrl->flashChip = &pNandInfo->flashChip;
#ifdef _WRS_CONFIG_DEBUG_FLAG
            pCtrl->flashShow = vxbHpnfcShow;
#endif /* _WRS_CONFIG_DEBUG_FLAG */

            (void) vxbFlashChipAdd (pCtrl);
            }

        uCs++;
        }

    return (OK);

vxbHpnfcFlashInitError:

    vxbHpnfcFlashClean (pHpnfc);
    return (ERROR);
    }

/*******************************************************************************
*
* vxbHpnfcFlashClean - clean Flash chips attached to HPNFC
*
* This routine cleans the Flash chips attached to the HPNFC.
*
* RETURNS: N/A
*/

void vxbHpnfcFlashClean
    (
    pHPNFC   pHpnfc
    )
    {
    pNAND_FLASH_INFO    pCurNandInfo;
    pNAND_FLASH_INFO    pNextNandInfo;

    for (pCurNandInfo = (pNAND_FLASH_INFO) DLL_FIRST (&pHpnfc->chipList);
        pCurNandInfo != NULL; pCurNandInfo = pNextNandInfo)
        {
        pNextNandInfo = (pNAND_FLASH_INFO) DLL_NEXT (&pCurNandInfo->node);
        DLL_REMOVE (&pHpnfc->chipList, &pCurNandInfo->node);
        vxbNandInfoDel (pCurNandInfo);
        }
    }

/*******************************************************************************
*
* vxbHpnfcCmdIssue - issue command to HPNFC
*
* This routine issues the issue command to HPNFC.
*
* RETURNS: OK always.
*/

LOCAL int vxbHpnfcCmdIssue
    (
    pNAND_FLASH_INFO pNandInfo,
    UINT16           uCtrl,
    UINT32           uData
    )
    {
    pHPNFC pHpnfc = (pHPNFC)pNandInfo->pCtrl;
    UINT64 cmd23 = 0;
    UINT32 cmd0 = 0;

#ifdef HPNFC_OPS_DUMP
    /* dump data bytes from last command */

    vxbHpnfcDump();
#endif /* HPNFC_OPS_DUMP */

    if ((uCtrl & NAND_CMD_CTRL_CLE) != 0)
        {
#ifdef HPNFC_OPS_DUMP
        stampit("cmd cle 0x%02x", uData & 0xFF);
#endif /* HPNFC_OPS_DUMP */

        pHpnfc->cmd = uData;

        cmd23 |= HPNFC_CMD_LAYOUT_CMD(uData);
        cmd23 |= HPNFC_CMD_LAYOUT_BANK(0);
        cmd23 |= HPNFC_CMD_LAYOUT_TWB;
        cmd23 |= HPNFC_CMD_LAYOUT_TYPE_CMD_SEQ;

        pHpnfc->bufIndex = 0;
        }
    else if ((uCtrl & NAND_CMD_CTRL_ALE) != 0)
        {
#ifdef HPNFC_OPS_DUMP
        stampit("cmd ale 0x%02x", uData & 0xFF);
#endif /* HPNFC_OPS_DUMP */

        cmd23 |= HPNFC_CMD_LAYOUT_ADDR0(uData);
        cmd23 |= HPNFC_CMD_LAYOUT_NUM_BYTES(1);
        cmd23 |= HPNFC_CMD_LAYOUT_TWB;
        cmd23 |= HPNFC_CMD_LAYOUT_TYPE_ADDR_SEQ;
        }
    else
        {
        return (OK);
        }

    cmd0 |= HPNFC_CMD_CT_GEN;
    cmd0 |= HPNFC_CMD_TRD_NUM(0);

    HPNFC_REG_WRITE (pHpnfc, HPNFC_CMD_REG2, (ULONG)cmd23);
    HPNFC_REG_WRITE (pHpnfc, HPNFC_CMD_REG3, (ULONG)(cmd23 >> 32));
    HPNFC_REG_WRITE (pHpnfc, HPNFC_CMD_REG0, cmd0);

    pHpnfc->bufBytes = 0;
    pHpnfc->bufIndex = 0;
    
    return (OK);
    }

/*******************************************************************************
*
* vxbHpnfcSdmaWait - wait until SDMA is free
*
* This routine waits until SDMA is free.
*
* RETURNS: OK, always.
*/

LOCAL STATUS vxbHpnfcSdmaWait
    (
    pNAND_FLASH_INFO pNandInfo
    )
    {
    pHPNFC pHpnfc = (pHPNFC)pNandInfo->pCtrl;
    UINT32 uVal;

    if ((pHpnfc->uFlag & HPNFC_FLAG_INTDIS) == 0)
        {
        semTake (&pHpnfc->syncSem, WAIT_FOREVER);
        }
    else
        {
        do
            {
            uVal = HPNFC_REG_READ (pHpnfc, HPNFC_CTRL_STS);
            }
        while ((uVal & HPNFC_CTRL_STS_SDMA_BUSY) != 0);
        }

    return (OK);
    }

/*******************************************************************************
*
* vxbHpnfcDataSeqIssue - issue data sequence to HPNFC
*
* This routine issues the issue data sequence to HPNFC.
*
* RETURNS: Number of bytes transferred.
*/

LOCAL UINT32 vxbHpnfcDataSeqIssue
    (
    pNAND_FLASH_INFO pNandInfo,
    UINT8 *          pBuf,
    UINT32           nbytes,
    BOOL             isWrite
    )
    {
    pHPNFC pHpnfc = (pHPNFC)pNandInfo->pCtrl;
    UINT32 copyBytes;
    UINT32 cmdBytes;
    UINT64 cmd23 = 0;
    UINT32 cmd0 = 0;

    copyBytes = nbytes;
    cmdBytes = (UINT32)ROUND_UP (copyBytes, 4);

    cmd23 |= HPNFC_CMD_LAYOUT_LAST_SEC_SIZE (cmdBytes);
    cmd23 |= HPNFC_CMD_LAYOUT_SEC_CNT(1);
    cmd23 |= isWrite ? HPNFC_CMD_LAYOUT_DIR_WR : HPNFC_CMD_LAYOUT_DIR_RD;
    /* cmd23 |= HPNFC_CMD_LAYOUT_TWB; */
    cmd23 |= HPNFC_CMD_LAYOUT_TYPE_DATA_SEQ;

    cmd0 |= HPNFC_CMD_CT_GEN;
    cmd0 |= HPNFC_CMD_TRD_NUM(0);

    if (isWrite)
        {
        bcopy (pBuf, (char *)pHpnfc->uSdmaBase, copyBytes);
        }

    HPNFC_REG_WRITE (pHpnfc, HPNFC_CMD_REG2, (ULONG)cmd23);
    HPNFC_REG_WRITE (pHpnfc, HPNFC_CMD_REG3, (ULONG)(cmd23 >> 32));
    HPNFC_REG_WRITE (pHpnfc, HPNFC_CMD_REG0, cmd0);

    vxbHpnfcSdmaWait (pNandInfo);

    if (!isWrite)
        {
        bcopy ((char *)pHpnfc->uSdmaBase, pBuf, copyBytes);
        }

    return (cmdBytes);
    }

/*******************************************************************************
*
* vxbHpnfcUnitRd - read one unit from HPNFC
*
* This routine reads one unit from HPNFC.
*
* RETURNS: value of the unit read
*/

LOCAL UINT16 vxbHpnfcUnitRd
    (
    pNAND_FLASH_INFO pNandInfo,
    int              unitSize     /* 1 or 2 */
    )
    {
    pHPNFC pHpnfc = (pHPNFC)pNandInfo->pCtrl;
    UINT16 val;

    if (pHpnfc->bufBytes == 0)
        {
        pHpnfc->bufIndex = 0;
        pHpnfc->bufBytes = vxbHpnfcDataSeqIssue (pNandInfo, pHpnfc->buf,
                                                 HPNFC_MIN_RX_BYTES, FALSE);
        }

    if (unitSize == 1)
        {
        val = (UINT16)pHpnfc->buf[pHpnfc->bufIndex];

        if ((pHpnfc->cmd == NAND_CMD_CODE_STARD) &&
            ((HPNFC_REG_READ (pHpnfc, HPNFC_CMN_STGS) & 
              HPNFC_CMN_STGS_DEV_16BIT) != 0))
            {
            /* add extra increment/decrement to skip "dead" byte in MSB lane */

            pHpnfc->bufIndex += (UINT32)unitSize;
            pHpnfc->bufBytes -= (UINT32)unitSize;
            }
        }
    else
        {
        val = *(UINT16 *)&pHpnfc->buf[pHpnfc->bufIndex];
        }
        
    pHpnfc->bufIndex += (UINT32)unitSize;
    pHpnfc->bufBytes -= (UINT32)unitSize;

#ifdef HPNFC_OPS_DUMP
    if (rdCount < HPNFC_MAX_DUMP_BYTES)
        {
        if (unitSize == 1)
            {
            rdBytes[rdCount] = (UINT8)val;
            }
        else
            {
            *(UINT16 *)&rdBytes[rdCount] = val;
            }
        }
    rdCount += (UINT32)unitSize;
#endif /* HPNFC_OPS_DUMP */

    return (val);
    }

/*******************************************************************************
*
* vxbHpnfcUnitWr - write one unit from HPNFC
*
* This routine writes one unit from HPNFC.
*
* RETURNS: OK.
*/

LOCAL int vxbHpnfcUnitWr
    (
    pNAND_FLASH_INFO pNandInfo,
    UINT16           val,
    int              unitSize     /* 1 or 2 */
    )
    {
    pHPNFC pHpnfc = (pHPNFC)pNandInfo->pCtrl;

    if (unitSize == 1)
        {
        pHpnfc->buf[pHpnfc->bufIndex] = (UINT8)val;
        }
    else
        {
        *(UINT16 *)&pHpnfc->buf[pHpnfc->bufIndex] = val;
        }
        
    pHpnfc->bufIndex += (UINT32)unitSize;
    pHpnfc->bufBytes += (UINT32)unitSize;

    if (pHpnfc->bufBytes >= HPNFC_MIN_TX_BYTES)
        {
        vxbHpnfcDataSeqIssue (pNandInfo, pHpnfc->buf,
                              HPNFC_MIN_TX_BYTES, TRUE);
 
        pHpnfc->bufIndex = 0;
        pHpnfc->bufBytes = 0;
        }

#ifdef HPNFC_OPS_DUMP
    if (wrCount < HPNFC_MAX_DUMP_BYTES)
        {
        if (unitSize == 1)
            {
            wrBytes[wrCount] = (UINT8)val;
            }
        else
            {
            *(UINT16 *)&wrBytes[wrCount] = val;
            }
        }
    wrCount += (UINT32)unitSize;
#endif /* HPNFC_OPS_DUMP */

    return (OK);
    }

/*******************************************************************************
*
* vxbHpnfcByteRd - read one byte from HPNFC
*
* This routine reads one byte from HPNFC.
*
* RETURNS: value of the byte
*/

LOCAL UINT8 vxbHpnfcByteRd
    (
    pNAND_FLASH_INFO pNandInfo
    )
    {
    return ((UINT8)vxbHpnfcUnitRd (pNandInfo, 1));
    }
    
/*******************************************************************************
*
* vxbHpnfcByteWr - write one byte to HPNFC
*
* This routine writes one byte to HPNFC.
*
* RETURNS: OK.
*/

LOCAL int vxbHpnfcByteWr
    (
    pNAND_FLASH_INFO    pNandInfo,
    UINT8               uVal
    )
    {
    return (vxbHpnfcUnitWr (pNandInfo, (UINT16)uVal, 1));
    }

/*******************************************************************************
*
* vxbHpnfcWordRd - read one word from HPNFC
*
* This routine reads one word from HPNFC.
*
* RETURNS: value of the word.
*/

LOCAL UINT16 vxbHpnfcWordRd
    (
    pNAND_FLASH_INFO pNandInfo
    )
    {
    return (vxbHpnfcUnitRd (pNandInfo, 2));
    }

/*******************************************************************************
*
* vxbHpnfcWordWr - write one word to HPNFC
*
* This routine writes one word to HPNFC.
*
* RETURNS: OK.
*/

LOCAL int vxbHpnfcWordWr
    (
    pNAND_FLASH_INFO    pNandInfo,
    UINT16              uVal
    )
    {
    return (vxbHpnfcUnitWr (pNandInfo, uVal, 2));
    }

/*******************************************************************************
*
* vxbHpnfcBufRd - read data from NAND
*
* This routine reads the data from NAND.
*
* RETURNS: OK or ERROR if not enough buffer or misaligned request.
*/

LOCAL int vxbHpnfcBufRd
    (
    pNAND_FLASH_INFO pNandInfo,
    pNAND_BUF_INFO   pBufInfo
    )
    {
    FLASH_PAGE_SIZE_T i = pBufInfo->uStartId;
    FLASH_PAGE_SIZE_T j = pBufInfo->uStartOffset;
    FLASH_PAGE_SIZE_T uRet = pBufInfo->uLen;
    UINT32            copyBytes = 0;
    UINT32            cmdBytes = 0;

    for (; i < pBufInfo->uDescCnt; i++)
        {
        if (cmdBytes != copyBytes)
            {
            HPNFC_DBG_MSG ("misaligned bufRd request\n", 0, 0, 0, 0, 0, 0);
            vxbHpnfcBufRdCanary++;
            return (ERROR);
            }

        copyBytes = min (uRet, pBufInfo->pBufDesc[i].uLen);

        cmdBytes = vxbHpnfcDataSeqIssue (pNandInfo,
                                         &pBufInfo->pBufDesc[i].pBuf[j],
                                         copyBytes,
                                         FALSE);
#ifdef HPNFC_OPS_DUMP
        if (rdCount < HPNFC_MAX_DUMP_BYTES)
            {
            bcopy (&pBufInfo->pBufDesc[i].pBuf[j], &rdBytes[rdCount],
                   min (copyBytes, HPNFC_MAX_DUMP_BYTES - rdCount));
            }
        rdCount += (UINT32)copyBytes;
#endif /* HPNFC_OPS_DUMP */

        j += copyBytes - 1;
        uRet -= copyBytes;
        if (uRet == 0)
            {
            if (j == (pBufInfo->pBufDesc[i].uLen - 1))
                {
                pBufInfo->uStartId      = (UINT8)(i + 1);
                pBufInfo->uStartOffset  = 0;
                }
            else
                {
                pBufInfo->uStartId      = (UINT8)i;
                pBufInfo->uStartOffset  = (FLASH_PAGE_SIZE_T)(j + 1);
                }

            return (OK);
            }

        j = 0;
        }

    HPNFC_DBG_MSG ("not enough rx buffer\n", 0, 0, 0, 0, 0, 0);
    return (ERROR);
    }

/*******************************************************************************
*
* vxbHpnfcBufWr - write data to NAND
*
* This routine write the data to NAND.
*
* RETURNS: OK or ERROR if not enough buffer or misaligned request.
*/

LOCAL int vxbHpnfcBufWr
    (
    pNAND_FLASH_INFO pNandInfo,
    pNAND_BUF_INFO   pBufInfo
    )
    {
    FLASH_PAGE_SIZE_T i = pBufInfo->uStartId;
    FLASH_PAGE_SIZE_T j = pBufInfo->uStartOffset;
    FLASH_PAGE_SIZE_T uRet = pBufInfo->uLen;
    UINT32            copyBytes;
    UINT32            cmdBytes;

    for (; i < pBufInfo->uDescCnt; i++)
        {
        copyBytes = min (uRet, pBufInfo->pBufDesc[i].uLen);

        cmdBytes = vxbHpnfcDataSeqIssue (pNandInfo,
                                         &pBufInfo->pBufDesc[i].pBuf[j],
                                         copyBytes,
                                         TRUE);
#ifdef HPNFC_OPS_DUMP
        if (wrCount < HPNFC_MAX_DUMP_BYTES)
            {
            bcopy (&pBufInfo->pBufDesc[i].pBuf[j], &wrBytes[wrCount],
                   min (copyBytes, HPNFC_MAX_DUMP_BYTES - wrCount));
            }
        wrCount += (UINT32)copyBytes;
#endif /* HPNFC_OPS_DUMP */

        if (cmdBytes != copyBytes)
            {
            HPNFC_DBG_MSG ("misaligned bufWr request\n", 0, 0, 0, 0, 0, 0);
            vxbHpnfcBufWrCanary++;
            return (ERROR);
            }

        j += copyBytes - 1;
        uRet -= copyBytes;
        if (uRet == 0)
            {
            if (j == (pBufInfo->pBufDesc[i].uLen - 1))
                {
                pBufInfo->uStartId      = (UINT8)(i + 1);
                pBufInfo->uStartOffset  = 0;
                }
            else
                {
                pBufInfo->uStartId      = (UINT8)i;
                pBufInfo->uStartOffset  = (FLASH_PAGE_SIZE_T)(j + 1);
                }

            return (OK);
            }

        j = 0;
        }

    HPNFC_DBG_MSG ("not enough tx buffer\n", 0, 0, 0, 0, 0, 0);
    return (ERROR);
    }

/*******************************************************************************
*
* vxbHpnfcRdyChk - check ready/busy status
*
* This routine check read/busy status.
*
* RETURNS: ERROR or OK if ready
*/

LOCAL int vxbHpnfcRdyChk
    (
    pNAND_FLASH_INFO pNandInfo
    )
    {
    pHPNFC pHpnfc = (pHPNFC)pNandInfo->pCtrl;
    UINT32 uCs = pNandInfo->uCs;
    UINT32 uVal;

    uVal = HPNFC_REG_READ (pHpnfc, HPNFC_RBN_STS);

    return ((HPNFC_RBN_STS_RDY_MASK(uCs) & uVal) == 0 ? ERROR : OK);
    }
