/* vxbNandFlashOnfiCmdSetLib.c - NAND Flash ONFI Command Set Library */

/*
 * Copyright (c) 2012, 2014, 2016-2017, 2019-2022 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify, or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION

This library provides the command-set-specific routines for ONFI NAND Flash
chips.

ONFI(Open NAND Flash Interface) is one open standards for NAND flash memory
chips and devices that communicate with them, and supported widely by NAND Flash
Vendors. It is compatible with the existing NAND Flash interface, but adds one
self-described parameter page interface, and may include some control logic,
such as ECC, bad block management, and other low-level NAND management tasks.

Note:
    1. The ONFI-specified commands other than ONFI Signature Read and Parameter
    Page Read, are unsupported;
    2. Extended Parameter Page is unsupported now because it only includs the
    extended ECC information used by Block Abstracted NAND that is unsupported.

INCLUDE FILES: nand/vxbNandFlashLib.h

SEE ALSO:
\tb "Open NAND Flash Interface Specification",
\tb "Open NAND Flash Interface Specification:Block Abstracted NAND",
\tb "Open NAND Flash Interface Specification:NAND Connector"
\tb "VxWorks BSP and Driver Guide"
*/

#include <vsbConfig.h>

#include <nand/vxbNandFlashLib.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <hwif/vxbus/vxbLib.h>
#include <hwif/util/vxbAccess.h>

/* defines */

#if (_BYTE_ORDER == _BIG_ENDIAN)
#define NAND_ONFI_LE16_TO_CPU(x)    vxbSwap16 (x)
#define NAND_ONFI_LE32_TO_CPU(x)    vxbSwap32 (x)
#else /* !_BIG_ENDIAN */
#define NAND_ONFI_LE16_TO_CPU(x)    ((UINT16) (x))
#define NAND_ONFI_LE32_TO_CPU(x)    ((UINT32) (x))
#endif /* _BIG_ENDIAN */

/* ONFI signature */

#define ONFI_SIG_OFFSET             0x20
#define ONFI_SIG_LEN                4

/* revision defines */

#define NAND_ONFI_REV3_1            0x0080
#define NAND_ONFI_REV3_0            0x0040
#define NAND_ONFI_REV2_3            0x0020
#define NAND_ONFI_REV2_2            0x0010
#define NAND_ONFI_REV2_1            0x0008
#define NAND_ONFI_REV2_0            0x0004
#define NAND_ONFI_REV1_0            0x0002

#define NAND_ONFI_REV_SUPPORTED(x)  (((x) & 0xFE) != 0)

/* features defines */

#define NAND_ONFI_CAP_EVPP          0x1000  /* external Vpp */
#define NAND_ONFI_CAP_VOLADDR       0x0800  /* Volume addressing */
#define NAND_ONFI_CAP_NVDDR2        0x0400  /* NV-DDR2 */
#define NAND_ONFI_CAP_EZNAND        0x0200  /* EZ NAND */
#define NAND_ONFI_CAP_EPPRCLR       0x0100  /* program page register clear */
                                            /* enhancement */
#define NAND_ONFI_CAP_EXTPP         0x0080  /* extended parameter page */
#define NAND_ONFI_CAP_MPLANE_RD     0x0040  /* multi-plane read */
#define NAND_ONFI_CAP_NVDDR         0x0020  /* NV-DDR */
#define NAND_ONFI_CAP_O2EP_MV       0x0010  /* odd to even page Copyback */
#define NAND_ONFI_CAP_MPLANE_PGERA  0x0008  /* multi-plane program and erase */
#define NAND_ONFI_CAP_NSEQ_PPG      0x0004  /* non-sequential page */
                                            /* programming */
#define NAND_ONFI_CAP_MLUN          0x0002  /* multiple LUN */
#define NAND_ONFI_CAP_16BIT         0x0001  /* 16-bit data bus width */

/* Optional commands defines */

#define NAND_ONFI_CMD_LUN_FEATURES  0x1000  /* LUN Get and LUN Set Features */
#define NAND_ONFI_CMD_ODT_CFG       0x0800  /* ODT Configure */
#define NAND_ONFI_CMD_VOLSEL        0x0400  /* Volume Select */
#define NAND_ONFI_CMD_LUN_RST       0x0200  /* Reset LUN */
#define NAND_ONFI_CMD_SDA_MV        0x0100  /* Small Data Move */
#define NAND_ONFI_CMD_ROWADDR       0x0080  /* Change Row Address */
#define NAND_ONFI_CMD_ERDCOLCHG     0x0040  /* Change Read Column Enhanced */
#define NAND_ONFI_CMD_UIDRD         0x0020  /* Read Unique ID */
#define NAND_ONFI_CMD_MV            0x0010  /* Copyback */
#define NAND_ONFI_CMD_STA_ERD       0x0008  /* Read Status Enhanced */
#define NAND_ONFI_CMD_FEATURES      0x0004  /* Get and Set Features */
#define NAND_ONFI_CMD_CACHE_RD      0x0002  /* Read Cache command */
#define NAND_ONFI_CMD_CACHE_PPROG   0x0001  /* Page Cache Program command */

/* ONFI-JEDEC JTG primary advanced commands defines (Optional) */

#define NAND_ONFI_CMDADV_MPLANE_BLKERA  0x08  /* Multi-plane Block Erase */
#define NAND_ONFI_CMDADV_MPLANE_MV      0x04  /* Multi-plane Copyback Program */
#define NAND_ONFI_CMDADV_MPLANE_PG      0x02  /* Multi-plane Page Program */
#define NAND_ONFI_CMDADV_RNDOUT         0x01  /* Random Data Out */

/* Extended parameter page length defines (Optional) */

#define NAND_ONFI_EXTPP_LEN(x)          ((x) << 4)  /* length in bytes */

/* Device manufacturer defines (12 ASCII characters) */

#define NAND_ONFI_MFRSTR_LEN            12
#define NAND_ONFI_DEVSTR_PAD            0x20    /* Pad for Device String */

/* Device model defines (20 ASCII characters) */

#define NAND_ONFI_MODSTR_LEN             20

/* Date code defines (Optional) */

#define NAND_ONFI_DATECODE_YEAR(x)      \
    (2000 + ((x) & 0xF) + ((((x) >> 4) & 0xF) * 10))
#define NAND_ONFI_DATECODE_WEEK(x)      \
    (((x) & 0xF) + ((((x) >> 4) & 0xF) * 10))

/*
 * partial page are valid only for ONFI 1.0, 2.0 and 2.1, but obsoleted for ONFI
 * 2.2 and later.
 */

#define NAND_ONFI_PART_PAGE_SUP(uRev)               \
    ((((uRev) & (NAND_ONFI_REV1_0 |                 \
    NAND_ONFI_REV2_0 | NAND_ONFI_REV2_1)) != 0) &&  \
    (((uRev) & (NAND_ONFI_REV2_2 |                  \
    NAND_ONFI_REV2_3 | NAND_ONFI_REV3_0 | NAND_ONFI_REV3_1)) == 0))

/* Number of address cycles defines (Optional) */

#define NAND_ONFI_ADDRCYC_COL(x)        (((x) >> 4) & 0xF)
#define NAND_ONFI_ADDRCYC_RAW(x)        ((x) & 0xF)

/* Number of bits per cell defines */

#define NAND_ONFI_BITS_PER_CELL_VALID(x)    (((x) != 0) && ((x) != 0xFF))

/* Partial programming attributes defines */

#define NAND_ONFI_PARTPAGE_LIMIT        0x01 /* partial page programming has */
                                             /* constraints */
#define NAND_ONFI_PARTPAGE_OOB_AFTER_DA 0x10 /* partial page layout is */
                                             /* partial page data followed by */
                                             /* partial page spare constraints */

/* Number of plane address bits defines */

#define NAND_ONFI_PLANE_ADDR_BITS(x)    ((x) & 0xF)

/* Multi-plane operation attributes defines (Optional) */

#define NAND_ONFI_PLANEOP_ATTR_XNORBA   0x20  /* lower bit XNOR block address */
                                              /* restriction */
#define NAND_ONFI_PLANEOP_ATTR_CRD      0x10  /* Cache Read */
#define NAND_ONFI_PLANEOP_ATTR_BLKERA   0x08  /* Address restrictions for */
                                              /* cache operations */
#define NAND_ONFI_PLANEOP_ATTR_CPG      0x04  /* Cache Program */
#define NAND_ONFI_PLANEOP_ATTR_NO_BALIM 0x02  /* no block address restrictions */
#define NAND_ONFI_PLANEOP_ATTR_OVERLAP  0x01  /* Overlapped multi-plane */

/* EZ NAND support defines (Optional) */

#define NAND_ONFI_EZNAND_MV_ADJ         0x04 /* Copyback Adjacency */
#define NAND_ONFI_EZNAND_MV_OTHER_DST   0x02 /* Copyback for other planes & */
                                             /* LUNs */
#define NAND_ONFI_EZNAND_ATMOIC_ENDIS   0x01 /* enable/disable of automatic */
                                             /* retries */

/* timing mode support defines */

#define NAND_ONFI_TIMING_MODE_SUP(x, mode)   \
    (((x) & (1 << (mode))) != 0)

/* Source synchronous parameters for ONFI 2.0, 2.1, 2.2 and 2.3 */

#define NAND_ONFI_SRCSYNC_PARAM_EXIST(uRev) \
    (((uRev) & (NAND_ONFI_REV2_0 | NAND_ONFI_REV2_1 | \
    NAND_ONFI_REV2_2 | NAND_ONFI_REV2_3)) != 0)

/* Source synchronous features defines (Optional) */

#define NAND_ONFI_SRCSYNC_CLKSTOP  0x04 /* support CLK stopped for data input */
#define NAND_ONFI_SRCSYNC_TYPCAP   0x02 /* typical capacitance values present */
#define NAND_ONFI_SRCSYNC_TCADUSED 0x01 /* tCAD value to use */

/* NV-DDR parameters for ONFI 3.0 and 3.1 */

#define NAND_ONFI_NVDDR_PARAM_EXIST(uRev)   \
    (((uRev) & (NAND_ONFI_REV3_0 | NAND_ONFI_REV3_1)) != 0)

/* NV-DDR / NV-DDR2 features defines (Optional) */

#define NAND_ONFI_NVDDR_VPPENSEQ   0x08 /* requires Vpp enablement sequence */
#define NAND_ONFI_NVDDR_CLKSTOP    0x04 /* support CLK stopped for data input */
#define NAND_ONFI_NVDDR_TYPCAP     0x02 /* typical capacitance values present */
#define NAND_ONFI_NVDDR_TCADUSED   0x01 /* tCAD value to use */

/* Driver strength support defines */

#define NAND_ONFI_DRVSTR_OVEDIRVE_EXIST(uRev) \
    (((uRev) & (NAND_ONFI_REV2_0 | NAND_ONFI_REV2_1 | NAND_ONFI_REV2_2)) != 0)
#define NAND_ONFI_DRVSTR_OHM_EXIST(uRev) \
    (((uRev) & (NAND_ONFI_REV2_3 | NAND_ONFI_REV3_0 | NAND_ONFI_REV3_1)) != 0)
#define NAND_ONFI_DRVSTR_EXIST(uRev) \
    (NAND_ONFI_DRVSTR_OVEDIRVE_EXIST (uRev) &&  \
    NAND_ONFI_DRVSTR_OHM_EXIST (uRev))

/* valid for ONFI 2.0, 2.1 and 2.2 */

#define NAND_ONFI_DRVSTR_OVEDIRVE2  0x04 /* overdrive 2 drive strength */
#define NAND_ONFI_DRVSTR_OVEDIRVE1  0x02 /* overdrive 1 drive strength */

/* valid for ONFI 2.3, 3.0 and 3.1 */

#define NAND_ONFI_DRVSTR_18OHM      0x04 /* 18 Ohm drive strength */
#define NAND_ONFI_DRVSTR_25OHM      0x02 /* 25 Ohm drive strength */

/* valid for ONFI 2.0 and later */

#define NAND_ONFI_DRVSTR_SET        0x01 /* driver strength settings */

/* NV-DDR2 features defines (Optional) */

#define NAND_ONFI_NVDDR_EXTVREFQ_200MTS 0x20 /* external VREFQ required for */
                                             /* >= 200 MT/s */
#define NAND_ONFI_NVDDR_DIFFSIG_DQS     0x10 /* differential signaling for */
                                             /* DQS */
#define NAND_ONFI_NVDDR_DIFFSIG_REn     0x08 /* differential signaling for */
                                             /* RE_n */
#define NAND_ONFI_NVDDR2_30OHM_ODT      0x04 /* ODT value of 30 Ohms */
#define NAND_ONFI_NVDDR2_MATRIXTERM_ODT 0x02 /* matrix termination ODT */
#define NAND_ONFI_NVDDR2_SELFTERM_ODT   0x01 /* self-termination ODT */

/* Vendor specific data length */

#define NAND_ONFI_VENDOR_SPEC_LEN       88

/* Integrity CRC initial value */

#define ONFI_CRC_CALC_LEN               254
#define ONFI_CRC_CALC_INIT_VAL          0x4F4E
#define ONFI_CRC_CALC_MSB               0x8000  /* most significant bit */
#define ONFI_CRC_CALC_POLY              0x8005  /* generator polynomial */

/* typedefs */

/* ONFI parameter page structure */

typedef struct nand_onfi_param_page
    {

    /* Revision information and features block (0 ~ 31) */

    UINT8   uSig[ONFI_SIG_LEN]; /* signature */
    UINT16  uRev;               /* revision */
    UINT16  uCap;               /* features */
    UINT16  uOptCmd;            /* Optional commands */
    UINT8   uOptCmdAdv;         /* ONFI-JEDEC JTG primary advanced command */
                                /* support (Optional) */
    UINT8   uRes0;              /* reserved */
    UINT16  uExtParamPageLen;   /* Extended parameter page length (Optional) */
    UINT8   uNumParamPage;      /* ToTal Number of parameter pages (Optional) */
    UINT8   uRes1[17];          /* reserved */

    /* manufacturer information block (32 ~ 79) */

    char    pMfrStr[NAND_ONFI_MFRSTR_LEN];  /* Device manufacturer  */
    char    pModStr[NAND_ONFI_MODSTR_LEN];  /* Device model */
    UINT8   uJedecMfrId;                    /* JEDEC manufacturer ID */
    UINT8   uDateCodeYear;                  /* Year of Date code (Optional) */
    UINT8   uDateCodeWeek;                  /* Week of Date code (Optional) */
    UINT8   uRes2[13];                      /* Obsolete */

    /* memory organization block (80 ~ 127) */

    UINT32  uPageSize;      /* Number of data bytes per page */
    UINT16  uOobSize;       /* Number of spare bytes per page */

    /*
     * the following 2 parameters are valid only for ONFI 1.0, 2.0 and 2.1, but
     * obsoleted for ONFI 2.2 and later.
     */

    UINT32  uPartPageSize;  /* Number of data bytes per partial page */
    UINT16  uPartOobSize;   /* Number of spare bytes per partial page */

    UINT32  uPagesPerBlk;   /* Number of pages per block */
    UINT32  uBlksPerLun;    /* Number of blocks per LUN (logic unit) */
    UINT8   uLunCnt;        /* Number of LUN (logic unit) */
    UINT8   uAddrCyc;       /* Number of address cycles */
    UINT8   uBitPerCell;    /* Number of bits per cell */
    UINT16  uBadBlkPerLun;  /* Bad blocks maximum per LUN */
    UINT8   uBlkLifeVal;    /* Block endurance - value */
    UINT8   uBlkLifeMulti;  /* Block endurance - multiplier to 10 */
    UINT8   uValidBlkCnt;   /* Guaranteed valid blocks at beginning of target */
    UINT16  uValidBlkLife;  /* Block endurance for guaranteed valid blocks */
    UINT8   uNumPgPerPage;  /* Number of programs per page */

    /*
     * the Partial programming attributes are valid only for ONFI 1.0, 2.0 and
     * 2.1, but obsoleted for ONFI 2.2 and later.
     */

    UINT8   uPartProgAttr;

    UINT8   uNumEccCorr;    /* Number of bit ECC correctability per 512 bytes */
    UINT8   uPlaneAddrBits; /* Number of plane address bits */
    UINT8   uPlaneOpAttr;   /* Multi-plane operation attributes (Optional) */
    UINT8   uEzNandSup;     /* EZ NAND support (Optional) */
    UINT8   uRes5[12];      /* reserved */

    /* electrical parameter block (128 ~ 163) */

    UINT8   uIoPinMaxCap;       /* I/O pin maximum capacitance (pF) */
    UINT16  uSdrTimingMode;     /* SDR timing mode support */
    UINT8   uRes6[2];           /* Obsolete */
    UINT16  uMaxProgTime;       /* tPROG Maximum page program time (us) */
    UINT16  uMaxEraTime;        /* tBERS Maximum block erase time (us) */
    UINT16  uMaxRdTime;         /* tR Maximum page read time (us) */
    UINT16  uMinCcsTime;        /* tCCS Minimum change column setup time (ns) */

    /* Source synchronous or NV-DDR parameters (Optional) */

    union
        {
        UINT8   uRes7[3];      /* reserved for ONFI 1.0 */

        /* Source synchronous parameters for ONFI 2.0, 2.1, 2.2 and 2.3 */

        struct
            {
            UINT16  uSrcSyncTimingMode; /* Source synchronous timing mode */
            UINT8   uSrcSyncFeats;      /* Source synchronous features */
            } _WRS_PACK_ALIGN (1) srcSync;

        /* NV-DDR parameters for ONFI 3.0 and 3.1 */

        struct
            {
            UINT8   uNvDdrTimingMode;   /* NV-DDR timing mode  */
            UINT8   uNvDdr2TimingMode;  /* NV-DDR2 timing mode */
            UINT8   uNvDdrFeats;        /* NV-DDR / NV-DDR2 features */
            } nvDdr;
        } srcSyncOrNvDdr;

    /*
     * the following 3 parameters are avaliable if typical capacitance values
     * present bit is set (Optional)
     */

    UINT16  uClkInputPinCapTyp; /* CLK input pin typical capacitance (0.1 pF) */
    UINT16  uIoPinCapTyp;       /* IO pin typical capacitance (0.1 pF) */
    UINT16  uInputPinCapTyp;    /* input pin typical capacitance (0.1 pF) */

    /*
     * the following 2 parameters are valid and mandatory for ONFI 2.0 and later
     */

    UINT8   uInputPinCapMax;    /* input pin typical maximum capacitance (pF) */
    UINT8   uDrvStrengthSup;    /* Driver strength support */

    UINT16  uMaxMultiPlaneRdTime;   /* tR Maximum multi-plane page read time */
                                    /* in us (Optional) */
    UINT16  uTAdlVal;               /* tADL Program page register clear */
                                    /* enhancement tADL value in ns */
                                    /* (Optional) */
    UINT16  uTypEzNandRdTime;       /* tR Typical page read time for EZ NAND */
                                    /* in us (Optional) */

    UINT8   uNvDdr2Feats;           /* NV-DDR2 features (Optional) */
    UINT8   uNvDdr2WarmCyc;         /* NV-DDR2 warmup cycles (Optional) */
    UINT8   uRes8[4];               /* reserved */

    /* vendor block ( 164 ~ 255)*/

    UINT16  uVendorSpecRev;         /* Vendor specific Revision number */
    UINT8   uVendorSpec[NAND_ONFI_VENDOR_SPEC_LEN]; /* Vendor specific */
    UINT16  uCrc;                   /* Integrity CRC */
    } _WRS_PACK_ALIGN (1) NAND_ONFI_PARAM_PAGE, *pNAND_ONFI_PARAM_PAGE;

/* locals */

LOCAL UINT16 vxbNandOnfiEccCalc (UINT8 * pBuf, UINT32  uLen);
LOCAL UINT32 vxbNandOnfiStrFormat (char * pStr, UINT32  uLen);
#ifdef _WRS_CONFIG_DEBUG_FLAG
LOCAL void vxbNandOnfiStrPrintf (char * pStr, UINT32  uLen);
LOCAL void vxbNandOnfiTimingModePrintf (UINT16  uTimingMode);
LOCAL UINT32 vxbNandOnfiBlkLifeCalc (UINT8 uVal, UINT8 uMulti);
#endif /* _WRS_CONFIG_DEBUG_FLAG */

/* imports */

IMPORT void vxbUsDelay
    (
    int delayTime
    );

/* command sequence defines */

/* identify operations */

LOCAL NAND_CMD_SEQ onfiIdRdSeq =
    {
    4, {
            {
            NAND_CMD_TYPE_CMD,
            NAND_CMD_CTRL_CE | NAND_CMD_CTRL_CLE | NAND_CMD_CTRL_CHG,
            NAND_CMD_CODE_IDRD
            },
            {
            NAND_CMD_TYPE_COL_ADDR_NOADJ,
            NAND_CMD_CTRL_CE | NAND_CMD_CTRL_ALE | NAND_CMD_CTRL_CHG,
            0
            },
            {NAND_CMD_TYPE_RDY_CHK, 0, 0},
            {NAND_CMD_TYPE_OTHERRD, 0, 0}
        }
    };
LOCAL NAND_CMD_SEQ onfiPpRdSeqEntry =
    {
    3, {
            {
            NAND_CMD_TYPE_CMD,
            NAND_CMD_CTRL_CE | NAND_CMD_CTRL_CLE | NAND_CMD_CTRL_CHG,
            NAND_CMD_CODE_PPRD
            },
            {
            NAND_CMD_TYPE_COL_ADDR_NOADJ,
            NAND_CMD_CTRL_CE | NAND_CMD_CTRL_ALE | NAND_CMD_CTRL_CHG,
            0
            },
            {NAND_CMD_TYPE_RDY_CHK, 0, 0}
        }
    };
LOCAL NAND_CMD_SEQ onfiPpRdSeq =
    {
    1, {{NAND_CMD_TYPE_OTHERRD, 0, 0}}
    };

LOCAL UINT8 uOnfiSigStr[ONFI_SIG_LEN] = {'O', 'N', 'F', 'I'};

/*******************************************************************************
*
* vxbNandOnfiIdentify - identify NAND Flash with ONFI
*
* This routine identifies NAND Flash chip with ONFI and updates NAND Flash
* information.
*
* RETURNS: OK, or ERROR if failed
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS vxbNandOnfiIdentify
    (
    pNAND_FLASH_DESC    pNandDesc
    )
    {
    pNAND_ONFI_PARAM_PAGE   pOnfiInfo;
    UINT8                   pOnfiId[ONFI_SIG_LEN];
    UINT32                  uMfrStrLen;
    UINT32                  uModStrLen;
    BOOL                    bFound;

    /* read NAND ONFI ID(should OK always) */

    (void) vxbNandCmdIssue (pNandDesc, &onfiIdRdSeq,
                            0, ONFI_SIG_OFFSET, pOnfiId, ONFI_SIG_LEN, 0);
    if (bcmp ((char *) uOnfiSigStr, (char *) pOnfiId, ONFI_SIG_LEN) != 0)
        {
        NAND_ERR_MSG ("ONFI unsupported\n", 1, 2, 3, 4, 5, 6);
        return (ERROR);
        }

    /* allocate ONFI information */

    pOnfiInfo = (pNAND_ONFI_PARAM_PAGE) malloc (sizeof (NAND_ONFI_PARAM_PAGE));
    if (pOnfiInfo == NULL)
        {
        NAND_ERR_MSG ("failed to allocate ONFI information\n",
                      1, 2, 3, 4, 5, 6);
        return (ERROR);
        }

    /* begin to read ONFI parameter page */

    (void) vxbNandCmdIssue (pNandDesc, &onfiPpRdSeqEntry, 0, 0, NULL, 0, 0);
    bFound = FALSE;
    vxbUsDelay(100);
    while (1)
        {
        (void) vxbNandCmdIssue (pNandDesc, &onfiPpRdSeq,
                                0, 0, (UINT8 *) pOnfiInfo,
                                sizeof (NAND_ONFI_PARAM_PAGE), 0);
        if (bcmp ((char *) uOnfiSigStr, (char *) pOnfiInfo->uSig,
            ONFI_SIG_LEN) != 0)
            break;

        /* verify CRC to find one valid parameter page */

        if (vxbNandOnfiEccCalc ((UINT8 *) pOnfiInfo, ONFI_CRC_CALC_LEN) ==
            NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uCrc))
            {
            bFound = TRUE;
            break;
            }
        }

    if (!bFound)
        {
        NAND_ERR_MSG ("no valid ONFI parameter page found\n",
                      1, 2, 3, 4, 5, 6);
        free (pOnfiInfo);
        return (ERROR);
        }
    pNandDesc->pCmdSetSpec = (void *) pOnfiInfo;

    /* check ONFI revision */

    pOnfiInfo->uRev = NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uRev);
    if (!NAND_ONFI_REV_SUPPORTED (pOnfiInfo->uRev))
        {
        NAND_ERR_MSG ("unsupported ONFI revision (0x%08x)\n",
                      pOnfiInfo->uRev, 2, 3, 4, 5, 6);
        free (pOnfiInfo);
        return (ERROR);
        }
    pNandDesc->nandId.uCap |= NAND_CAP_ONFI;
    if (pNandDesc->nandId.uMfrID == NAND_MFRID_MICRON)
        pNandDesc->nandId.uCap |= NAND_CAP_BBSCAN_FLPG;

    /* format manufacturer and device name string */

    pNandDesc->nandId.pName = (char *) malloc (FLASH_CHIP_NAME_LEN);
    if (pNandDesc->nandId.pName == NULL)
        {
        NAND_ERR_MSG ("failed to allocate memory for device name string\n",
                      1, 2, 3, 4, 5, 6);
        free (pOnfiInfo);
        return (ERROR);
        }

    uMfrStrLen =
        vxbNandOnfiStrFormat (pOnfiInfo->pMfrStr, NAND_ONFI_MFRSTR_LEN);
    bcopy (pOnfiInfo->pMfrStr, pNandDesc->nandId.pName, uMfrStrLen);
    pNandDesc->nandId.pName[uMfrStrLen] = ' ';

    uModStrLen =
        vxbNandOnfiStrFormat (pOnfiInfo->pModStr, NAND_ONFI_MODSTR_LEN);
    if (uModStrLen < (FLASH_CHIP_NAME_LEN - 1 - (uMfrStrLen + 1)))
        {
        bcopy (pOnfiInfo->pModStr, &pNandDesc->nandId.pName[uMfrStrLen + 1],
                min(NAND_ONFI_MODSTR_LEN, uModStrLen));
        pNandDesc->nandId.pName[uMfrStrLen + 1 + uModStrLen] = EOS;
        }
    else
        {
        bcopy (pOnfiInfo->pModStr, &pNandDesc->nandId.pName[uMfrStrLen + 1],
                min(NAND_ONFI_MODSTR_LEN,
                    FLASH_CHIP_NAME_LEN - 1 - (uMfrStrLen + 1)));
        pNandDesc->nandId.pName[FLASH_CHIP_NAME_LEN - 1] = EOS;
        }

    /* transfer original data to valid information */

    pOnfiInfo->uCap = NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uCap);
    if ((pOnfiInfo->uCap & NAND_ONFI_CAP_16BIT) != 0)
        pNandDesc->nandId.uCap |= NAND_CAP_X16;
    if ((pOnfiInfo->uCap & NAND_ONFI_CAP_NSEQ_PPG) != 0)
        pNandDesc->nandId.uCap |= NAND_CAP_NSEQ_PROG;

    pOnfiInfo->uOptCmd = NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uOptCmd);
    if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_CACHE_PPROG) != 0)
        pNandDesc->nandId.uCap |= NAND_CAP_CACHE_PROG;
    if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_CACHE_RD) != 0)
        pNandDesc->nandId.uCap |= NAND_CAP_CACHE_RD;
    if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_MV) != 0)
        {
        pNandDesc->nandId.uCap |= NAND_CAP_MV;
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_O2EP_MV) == 0)
            pNandDesc->nandId.uCap |= NAND_CAP_MV_O2EP_LIMIT;
        }

    if ((pOnfiInfo->uCap & NAND_ONFI_CAP_EXTPP) != 0)
        pOnfiInfo->uExtParamPageLen =
            NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uExtParamPageLen);

    pOnfiInfo->uPageSize =
        NAND_ONFI_LE32_TO_CPU (pOnfiInfo->uPageSize);

    pNandDesc->nandId.uPageSize = (FLASH_PAGE_SIZE_T) pOnfiInfo->uPageSize;

    if (pOnfiInfo->uPageSize > NAND_SMALL_PAGE_SIZE)
        pNandDesc->nandId.uCap |= NAND_CAP_LP;

    pNandDesc->nandId.uOobSize = pOnfiInfo->uOobSize =
#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
        /* The Intel simics Spansion S34ML01G1 incorrectly reports it has
         * 0x48 bytes of spare bytes per page, when it actually has 0x40
         * bytes per page. Fix that here.
         */

        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uOobSize) & 0xfff0;
#else /* _WRS_CONFIG_ITL_SOC_AGILEX5 */
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uOobSize);
#endif /* _WRS_CONFIG_ITL_SOC_AGILEX5 */

    if (NAND_ONFI_PART_PAGE_SUP (pOnfiInfo->uRev))
        {
        pOnfiInfo->uPartPageSize =
            NAND_ONFI_LE32_TO_CPU (pOnfiInfo->uPartPageSize);
        pOnfiInfo->uPartOobSize =
            NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uPartOobSize);
        }

    pOnfiInfo->uPagesPerBlk = NAND_ONFI_LE32_TO_CPU (pOnfiInfo->uPagesPerBlk);
    pNandDesc->nandId.uBlkSize =
        pOnfiInfo->uPagesPerBlk * pNandDesc->nandId.uPageSize;

    pOnfiInfo->uBlksPerLun      =
        NAND_ONFI_LE32_TO_CPU (pOnfiInfo->uBlksPerLun);
    pNandDesc->nandId.uChipSize =
        (FLASH_SIZE_T) pOnfiInfo->uLunCnt * pOnfiInfo->uBlksPerLun *
        pNandDesc->nandId.uBlkSize;

    if (NAND_ONFI_ADDRCYC_COL (pOnfiInfo->uAddrCyc) == 1)
        pNandDesc->nandId.uCap |= NAND_CAP_COL_ADDR;
    else if (NAND_ONFI_ADDRCYC_COL (pOnfiInfo->uAddrCyc) == 2)
        pNandDesc->nandId.uCap |= NAND_CAP_COL_ADDR_2CYC;
    else
        {
        NAND_ERR_MSG ("unsupported column address cycle %d, %d cycles are "
                      "supported at most\n",
                      NAND_ONFI_ADDRCYC_COL (pOnfiInfo->uAddrCyc),
                      NAND_COL_ADDR_MAX_CYC, 3, 4, 5, 6);
        free (pOnfiInfo);
        return (ERROR);
        }

    if (NAND_ONFI_ADDRCYC_RAW (pOnfiInfo->uAddrCyc) == 1)
        pNandDesc->nandId.uCap |= NAND_CAP_PAGE_ADDR;
    else if (NAND_ONFI_ADDRCYC_RAW (pOnfiInfo->uAddrCyc) == 2)
        pNandDesc->nandId.uCap |= NAND_CAP_PAGE_ADDR_2CYC;
    else if (NAND_ONFI_ADDRCYC_RAW (pOnfiInfo->uAddrCyc) == 3)
        pNandDesc->nandId.uCap |= NAND_CAP_PAGE_ADDR_3CYC;
    else
        {
        NAND_ERR_MSG ("unsupported page address cycle %d, %d cycles are "
                      "supported at most\n",
                      NAND_ONFI_ADDRCYC_RAW (pOnfiInfo->uAddrCyc),
                      NAND_PAGE_ADDR_MAX_CYC, 3, 4, 5, 6);
        free (pOnfiInfo);
        return (ERROR);
        }

    if (NAND_ONFI_BITS_PER_CELL_VALID (pOnfiInfo->uBitPerCell) &&
        (pOnfiInfo->uBitPerCell > 1))
        {
        NAND_WARN_MSG ("MLC NAND with %d bit(s) in one cell is used as SLC "
                       "now\n", pOnfiInfo->uBitPerCell, 2, 3, 4, 5, 6);
        pNandDesc->nandId.uCap |= NAND_CAP_MLC;
        }

    pOnfiInfo->uBadBlkPerLun    =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uBadBlkPerLun);

    pOnfiInfo->uValidBlkLife    =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uValidBlkLife);
    pOnfiInfo->uSdrTimingMode   =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uSdrTimingMode);
    pOnfiInfo->uMaxProgTime     =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uMaxProgTime);
    pOnfiInfo->uMaxEraTime     =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uMaxEraTime);
    pOnfiInfo->uMaxRdTime     =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uMaxRdTime);
    pOnfiInfo->uMinCcsTime     =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uMinCcsTime);

    if ((!NAND_ONFI_NVDDR_PARAM_EXIST (pOnfiInfo->uRev)) &&
        NAND_ONFI_SRCSYNC_PARAM_EXIST (pOnfiInfo->uRev))
        pOnfiInfo->srcSyncOrNvDdr.srcSync.uSrcSyncTimingMode    =
            NAND_ONFI_LE16_TO_CPU (pOnfiInfo->srcSyncOrNvDdr.srcSync.uSrcSyncTimingMode);

    if ((NAND_ONFI_NVDDR_PARAM_EXIST (pOnfiInfo->uRev) &&
        ((pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats &
        NAND_ONFI_NVDDR_TYPCAP) != 0)) ||
        ((!NAND_ONFI_NVDDR_PARAM_EXIST (pOnfiInfo->uRev)) &&
        NAND_ONFI_SRCSYNC_PARAM_EXIST (pOnfiInfo->uRev) &&
        ((pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats &
        NAND_ONFI_SRCSYNC_TYPCAP) != 0)))
        {
        pOnfiInfo->uClkInputPinCapTyp   =
            NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uClkInputPinCapTyp);
        pOnfiInfo->uIoPinCapTyp         =
            NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uIoPinCapTyp);
        pOnfiInfo->uInputPinCapTyp      =
            (UINT8)NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uInputPinCapTyp);
        }

    pOnfiInfo->uMaxMultiPlaneRdTime =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uMaxMultiPlaneRdTime);
    pOnfiInfo->uTAdlVal             =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uTAdlVal);
    pOnfiInfo->uTypEzNandRdTime     =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uTypEzNandRdTime);
    pOnfiInfo->uVendorSpecRev     =
        NAND_ONFI_LE16_TO_CPU (pOnfiInfo->uVendorSpecRev);

    return (OK);
    }

#ifdef _WRS_CONFIG_DEBUG_FLAG
/*******************************************************************************
*
* vxbNandOnfiShow - show NAND ONFI information
*
* This routine prints the NAND ONFI information.
*
* RETURNS: OK always
*
* \NOMANUAL
*/

void vxbNandOnfiShow
    (
    pNAND_FLASH_DESC    pNandDesc
    )
    {
    pNAND_ONFI_PARAM_PAGE   pOnfiInfo =
        (pNAND_ONFI_PARAM_PAGE) pNandDesc->pCmdSetSpec;

    printf ("            \t\tSupported ONFI revisions         :");
    if ((pOnfiInfo->uRev & NAND_ONFI_REV1_0) != 0)
        printf (" 1.0");
    if ((pOnfiInfo->uRev & NAND_ONFI_REV2_0) != 0)
        printf (" 2.0");
    if ((pOnfiInfo->uRev & NAND_ONFI_REV2_1) != 0)
        printf (" 2.1");
    if ((pOnfiInfo->uRev & NAND_ONFI_REV2_2) != 0)
        printf (" 2.2");
    if ((pOnfiInfo->uRev & NAND_ONFI_REV2_3) != 0)
        printf (" 2.3");
    if ((pOnfiInfo->uRev & NAND_ONFI_REV3_0) != 0)
        printf (" 3.0");
    if ((pOnfiInfo->uRev & NAND_ONFI_REV3_1) != 0)
        printf (" 3.1");
    printf ("\n");

    if (pOnfiInfo->uCap  != 0)
        {
        printf ("            \t\tSupported features               :\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_16BIT) != 0)
            printf ("            \t\t\t16-bit data bus width\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_MLUN) != 0)
            printf ("            \t\t\tmultiple LUN operations\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_NSEQ_PPG) != 0)
            printf ("            \t\t\tnon-sequential page programming\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_MPLANE_PGERA) != 0)
            printf ("            \t\t\tmulti-plane program and erase "
                    "operations\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_O2EP_MV) != 0)
            printf ("            \t\t\todd to even page Copyback\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_NVDDR) != 0)
            printf ("            \t\t\tNV-DDR\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_MPLANE_RD) != 0)
            printf ("            \t\t\tmulti-plane read operations\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_EXTPP) != 0)
            printf ("            \t\t\textended parameter page\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_EPPRCLR) != 0)
            printf ("            \t\t\tprogram page register clear "
                    "enhancement\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_EZNAND) != 0)
            printf ("            \t\t\tEZ NAND\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_NVDDR2) != 0)
            printf ("            \t\t\tNV-DDR2\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_VOLADDR) != 0)
            printf ("            \t\t\tVolume addressing\n");
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_EVPP) != 0)
            printf ("            \t\t\texternal Vpp\n");
        }

    if (pOnfiInfo->uOptCmd != 0)
        {
        printf ("            \t\tSupported Optional Commands      :\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_CACHE_PPROG) != 0)
            printf ("            \t\t\tPage Cache Program Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_CACHE_RD) != 0)
            printf ("            \t\t\tRead Cache command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_FEATURES) != 0)
            printf ("            \t\t\tGet and Set Features Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_STA_ERD) != 0)
            printf ("            \t\t\tRead Status Enhanced Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_MV) != 0)
            printf ("            \t\t\tCopyback Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_UIDRD) != 0)
            printf ("            \t\t\tRead Unique ID Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_ERDCOLCHG) != 0)
            printf ("            \t\t\tChange Read Column Enhanced Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_ROWADDR) != 0)
            printf ("            \t\t\tChange Row Address Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_SDA_MV) != 0)
            printf ("            \t\t\tSmall Data Move Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_LUN_RST) != 0)
            printf ("            \t\t\tReset LUN Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_VOLSEL) != 0)
            printf ("            \t\t\tVolume Select Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_ODT_CFG) != 0)
            printf ("            \t\t\tODT Configure Command\n");
        if ((pOnfiInfo->uOptCmd & NAND_ONFI_CMD_LUN_FEATURES) != 0)
            printf ("            \t\t\tLUN Get and LUN Set Features Command\n");
        }

    if (pOnfiInfo->uOptCmdAdv != 0)
        {
        printf ("            \t\tSupported ONFI-JEDEC JTG primary advanced "
                "Commands :\n");
        if ((pOnfiInfo->uOptCmdAdv & NAND_ONFI_CMDADV_RNDOUT) != 0)
            printf ("            \t\t\tRandom Data Out\n");
        if ((pOnfiInfo->uOptCmdAdv & NAND_ONFI_CMDADV_MPLANE_PG) != 0)
            printf ("            \t\t\tMulti-plane Page Program\n");
        if ((pOnfiInfo->uOptCmdAdv & NAND_ONFI_CMDADV_MPLANE_MV) != 0)
            printf ("            \t\t\tMulti-plane Copyback Program\n");
        if ((pOnfiInfo->uOptCmdAdv & NAND_ONFI_CMDADV_MPLANE_BLKERA) != 0)
            printf ("            \t\t\tMulti-plane Block Erase Program\n");
        }

    if ((pOnfiInfo->uCap & NAND_ONFI_CAP_EXTPP) != 0)
        {
        if (pOnfiInfo->uExtParamPageLen != 0)
            printf ("            \t\tExtended parameter page length(byte) : "
                    "0x%x\n",
                    NAND_ONFI_EXTPP_LEN (pOnfiInfo->uExtParamPageLen));
        if (pOnfiInfo->uNumParamPage != 0)
            printf ("            \t\tToTal Number of parameter pages      : "
                    "0x%x\n", pOnfiInfo->uNumParamPage);
        }

    printf ("            \t\tDevice manufacturer                  : ");
    vxbNandOnfiStrPrintf (pOnfiInfo->pMfrStr, NAND_ONFI_MFRSTR_LEN);
    printf ("\n");

    printf ("            \t\tDevice model                         : ");
    vxbNandOnfiStrPrintf (pOnfiInfo->pModStr, NAND_ONFI_MODSTR_LEN);
    printf ("\n");

    printf ("            \t\tJEDEC manufacturer ID                : 0x%x\n",
            pOnfiInfo->uJedecMfrId);

    if ((pOnfiInfo->uDateCodeYear !=0) || (pOnfiInfo->uDateCodeWeek !=0))
        {
        printf ("            \t\tDate code                            : "
                "%d week %d\n",
                NAND_ONFI_DATECODE_YEAR (pOnfiInfo->uDateCodeYear),
                NAND_ONFI_DATECODE_WEEK (pOnfiInfo->uDateCodeWeek));
        }

    printf ("            \t\tNumber of data bytes per page        : 0x%x\n",
            pOnfiInfo->uPageSize);
    printf ("            \t\tNumber of OOB bytes per page         : 0x%x\n",
            pOnfiInfo->uOobSize);

    if (NAND_ONFI_PART_PAGE_SUP (pOnfiInfo->uRev))
        {
        printf ("            \t\tNumber of data bytes per partial page: 0x%x\n",
            pOnfiInfo->uPartPageSize);
        printf ("            \t\tNumber of OOB bytes per partial page : 0x%x\n",
                pOnfiInfo->uPartOobSize);
        }

    printf ("            \t\tNumber of pages per block            : 0x%x\n",
            pOnfiInfo->uPagesPerBlk);
    printf ("            \t\tNumber of blocks per LUN             : 0x%x\n",
            pOnfiInfo->uBlksPerLun);
    printf ("            \t\tNumber of LUN                        : 0x%x\n",
            pOnfiInfo->uLunCnt);
    printf ("            \t\tNumber of column/row address cycles  : %d / %d\n",
            NAND_ONFI_ADDRCYC_COL (pOnfiInfo->uAddrCyc),
            NAND_ONFI_ADDRCYC_RAW (pOnfiInfo->uAddrCyc));
    printf ("            \t\tNumber of bits per cell              : %d\n",
            pOnfiInfo->uBitPerCell);
    printf ("            \t\tBad blocks maximum per LUN           : 0x%x\n",
            pOnfiInfo->uBadBlkPerLun);
    printf ("            \t\tBlock endurance                      : %d times\n",
            vxbNandOnfiBlkLifeCalc (pOnfiInfo->uBlkLifeVal,
                                    pOnfiInfo->uBlkLifeMulti));
    printf ("            \t\tGuaranteed valid blocks at beginning : %d\n",
            pOnfiInfo->uValidBlkCnt);
    if (pOnfiInfo->uValidBlkLife != 0)
        printf ("            \t\tGuaranteed valid blocks endurance    : %d "
                "program/erase cycles at least\n", pOnfiInfo->uValidBlkLife);

    printf ("            \t\tNumber of programs per page          : %d\n",
            pOnfiInfo->uNumPgPerPage);

    if ((NAND_ONFI_PART_PAGE_SUP (pOnfiInfo->uRev)) &&
        (pOnfiInfo->uPartProgAttr != 0))
        {
        printf ("            \t\tMulti-plane operation attributes     :\n");
        if ((pOnfiInfo->uPartProgAttr & NAND_ONFI_PARTPAGE_LIMIT) != 0)
            {
            printf ("            \t\t\tpartial page programming has constraints"
                    "\n");

            if ((pOnfiInfo->uPartProgAttr & NAND_ONFI_PARTPAGE_OOB_AFTER_DA) !=
                0)
                printf ("            \t\t\tpartial page layout is partial page "
                        "data followed by partial page spare\n");
            }
        else
            printf ("            \t\t\tpartial page programming has not any "
                    "constraints\n");
        }

    if ((pOnfiInfo->uNumEccCorr != 0) && (pOnfiInfo->uNumEccCorr != 0xFF))
        printf ("            \t\tNumber of bit ECC correctability     : %d "
                "bit(s) per 512 bytes\n", pOnfiInfo->uNumEccCorr);

    printf ("            \t\tNumber of plane address bits         : %d\n",
            NAND_ONFI_PLANE_ADDR_BITS (pOnfiInfo->uPlaneAddrBits));

    if ((pOnfiInfo->uPlaneOpAttr != 0) && ((pOnfiInfo->uCap &
        (NAND_ONFI_CAP_MPLANE_PGERA | NAND_ONFI_CAP_MPLANE_RD)) != 0))
        {
        printf ("            \t\tMulti-plane operation attributes     :\n");
        if ((pOnfiInfo->uPlaneOpAttr & NAND_ONFI_PLANEOP_ATTR_OVERLAP) != 0)
            printf ("            \t\t\tOverlapped multi-plane support\n");
        else
            printf ("            \t\t\tConcurrent multi-plane support\n");

        if ((pOnfiInfo->uPlaneOpAttr & NAND_ONFI_PLANEOP_ATTR_NO_BALIM) != 0)
            printf ("            \t\t\tno block address restrictions\n");
        if ((pOnfiInfo->uPlaneOpAttr & NAND_ONFI_PLANEOP_ATTR_CPG) != 0)
            printf ("            \t\t\tprogram cache supported\n");
        if ((pOnfiInfo->uPlaneOpAttr & NAND_ONFI_PLANEOP_ATTR_BLKERA) != 0)
            printf ("            \t\t\tAddress restrictions for cache "
                    "operations\n");
        if ((pOnfiInfo->uPlaneOpAttr & NAND_ONFI_PLANEOP_ATTR_CRD) != 0)
            printf ("            \t\t\tread cache supported\n");
        if ((pOnfiInfo->uPlaneOpAttr & NAND_ONFI_PLANEOP_ATTR_XNORBA) != 0)
            printf ("            \t\t\tlower bit XNOR block address "
                    "restriction\n");
        }

    if (pOnfiInfo->uEzNandSup != 0)
        {
        printf ("            \t\tEZ NAND support                      :\n");
        if ((pOnfiInfo->uEzNandSup & NAND_ONFI_EZNAND_ATMOIC_ENDIS) != 0)
            printf ("            \t\t\tsupports enable/disable of automatic "
                    "retries\n");
        if ((pOnfiInfo->uEzNandSup & NAND_ONFI_EZNAND_MV_OTHER_DST) != 0)
            printf ("            \t\t\tsupports Copyback for other planes & "
                    "LUNs\n");
        if ((pOnfiInfo->uEzNandSup & NAND_ONFI_EZNAND_MV_ADJ) != 0)
            printf ("            \t\t\tRequires Copyback Adjacency(paired)\n");
        }

    printf ("            \t\tI/O pin capacitance, maximum         : %d pF\n",
            pOnfiInfo->uIoPinMaxCap);
    printf ("            \t\tSDR timing mode support              :");
    vxbNandOnfiTimingModePrintf (pOnfiInfo->uSdrTimingMode);
    printf ("\n");

    printf ("            \t\ttPROG Maximum page program time      : %d us\n",
            pOnfiInfo->uMaxProgTime);
    printf ("            \t\ttBERS Maximum block erase time       : %d us\n",
            pOnfiInfo->uMaxEraTime);
    printf ("            \t\ttR Maximum page read time            : %d us\n",
            pOnfiInfo->uMaxRdTime);
    printf ("            \t\ttCCS Minimum change column setup time: %d ns\n",
            pOnfiInfo->uMinCcsTime);

    if (NAND_ONFI_NVDDR_PARAM_EXIST (pOnfiInfo->uRev))
        {
        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_NVDDR) != 0)
            {
            printf ("            \t\tNV-DDR timing mode support           :");
            vxbNandOnfiTimingModePrintf (pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrTimingMode);
            printf ("\n");
            }

        if ((pOnfiInfo->uCap & NAND_ONFI_CAP_NVDDR2) != 0)
            {
            printf ("            \t\tNV-DDR2 timing mode support          :");
            vxbNandOnfiTimingModePrintf (pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdr2TimingMode);
            printf ("\n");
            }

        if (pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats != 0)
            {
            printf ("            \t\tNV-DDR / NV-DDR2 features            :\n");
            if ((pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats &
                NAND_ONFI_NVDDR_TCADUSED) != 0)
                printf ("            \t\t\ttCAD value to use\n");
            if ((pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats &
                NAND_ONFI_NVDDR_TYPCAP) != 0)
                printf ("            \t\t\ttypical capacitance values "
                        "present\n");
            if ((pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats &
                NAND_ONFI_NVDDR_CLKSTOP) != 0)
                printf ("            \t\t\tdevice supports CLK stopped for "
                        "data input\n");
            if ((pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats &
                NAND_ONFI_NVDDR_VPPENSEQ) != 0)
                printf ("            \t\t\tdevice requires Vpp enablement "
                        "sequence\n");
            }
        }
    else if (NAND_ONFI_SRCSYNC_PARAM_EXIST (pOnfiInfo->uRev))
        {
        printf ("            \t\tSource synchronous timing mode support:");
        vxbNandOnfiTimingModePrintf (pOnfiInfo->srcSyncOrNvDdr.srcSync.uSrcSyncTimingMode);
        printf ("\n");

        if (pOnfiInfo->srcSyncOrNvDdr.srcSync.uSrcSyncFeats != 0)
            {
            printf ("            \t\tSource synchronous features            :"
                    "\n");
            if ((pOnfiInfo->srcSyncOrNvDdr.srcSync.uSrcSyncFeats &
                NAND_ONFI_SRCSYNC_TCADUSED) != 0)
                printf ("            \t\t\ttCAD value to use\n");
            if ((pOnfiInfo->srcSyncOrNvDdr.srcSync.uSrcSyncFeats &
                NAND_ONFI_SRCSYNC_TYPCAP) != 0)
                printf ("            \t\t\ttypical capacitance values present"
                        "\n");
            if ((pOnfiInfo->srcSyncOrNvDdr.srcSync.uSrcSyncFeats &
                NAND_ONFI_SRCSYNC_TCADUSED) != 0)
                printf ("            \t\t\tdevice supports CLK stopped for "
                        "data input\n");
            }
        }

    if ((NAND_ONFI_NVDDR_PARAM_EXIST (pOnfiInfo->uRev) &&
        ((pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats &
        NAND_ONFI_NVDDR_TYPCAP) != 0)) ||
        ((!NAND_ONFI_NVDDR_PARAM_EXIST (pOnfiInfo->uRev)) &&
        NAND_ONFI_SRCSYNC_PARAM_EXIST (pOnfiInfo->uRev) &&
        ((pOnfiInfo->srcSyncOrNvDdr.nvDdr.uNvDdrFeats &
        NAND_ONFI_SRCSYNC_TYPCAP) != 0)))
        {
        printf ("            \t\tCLK input pin capacitance, typical   : "
                "%d times 0.1pF\n", pOnfiInfo->uClkInputPinCapTyp);
        printf ("            \t\tI/O pin capacitance, typical         : "
                "%d times 0.1pF\n", pOnfiInfo->uIoPinCapTyp);
        printf ("            \t\tInput pin capacitance, typical       : "
                "%d times 0.1pF\n", pOnfiInfo->uInputPinCapTyp);
        }

    if (NAND_ONFI_DRVSTR_EXIST (pOnfiInfo->uRev))
        {
        printf ("            \t\tInput pin capacitance, maximum       : "
                "%d times 0.1pF\n", pOnfiInfo->uInputPinCapMax);

        if (pOnfiInfo->uDrvStrengthSup != 0)
            {
            printf ("            \t\tDriver strength support              :\n");
            if ((pOnfiInfo->uDrvStrengthSup & NAND_ONFI_DRVSTR_SET) != 0)
                printf ("            \t\t\tsupports driver strength settings\n");

            if (NAND_ONFI_DRVSTR_OHM_EXIST (pOnfiInfo->uRev))
                {
                if ((pOnfiInfo->uDrvStrengthSup & NAND_ONFI_DRVSTR_25OHM) != 0)
                    printf ("            \t\t\tsupports 25 Ohm drive strength"
                            "\n");
                if ((pOnfiInfo->uDrvStrengthSup & NAND_ONFI_DRVSTR_18OHM) != 0)
                    printf ("            \t\t\tsupports 18 Ohm drive strength"
                            "\n");
                }
            else
                {
                if ((pOnfiInfo->uDrvStrengthSup & NAND_ONFI_DRVSTR_OVEDIRVE1) !=
                    0)
                    printf ("            \t\t\tsupports overdrive 1 drive "
                            "strength\n");
                if ((pOnfiInfo->uDrvStrengthSup & NAND_ONFI_DRVSTR_OVEDIRVE2) !=
                    0)
                    printf ("            \t\t\tsupports overdrive 2 drive "
                            "strength\n");
                }
            }
        }

    if ((pOnfiInfo->uCap & NAND_ONFI_CAP_MPLANE_RD) != 0)
        printf ("            \t\ttR Maximum multi-plane page read time: "
                "%d us\n", pOnfiInfo->uMaxMultiPlaneRdTime);

    if ((pOnfiInfo->uCap & NAND_ONFI_CAP_EPPRCLR) != 0)
        printf ("            \t\ttADL Program page register clear enhancement "
                "tADL value: %d ns\n", pOnfiInfo->uTAdlVal);

    if ((pOnfiInfo->uCap & NAND_ONFI_CAP_EZNAND) != 0)
        printf ("            \t\ttR Typical page read time for EZ NAND: "
                "%d us\n", pOnfiInfo->uTypEzNandRdTime);

    if (((pOnfiInfo->uCap & NAND_ONFI_CAP_NVDDR2) != 0) &&
        (pOnfiInfo->uNvDdr2Feats != 0))
        {
        printf ("            \t\tDriver strength support              :\n");
        if ((pOnfiInfo->uNvDdr2Feats & NAND_ONFI_NVDDR2_SELFTERM_ODT) != 0)
            printf ("            \t\t\tsupports self-termination ODT\n");
        if ((pOnfiInfo->uNvDdr2Feats & NAND_ONFI_NVDDR2_MATRIXTERM_ODT) != 0)
            printf ("            \t\t\tsupports matrix termination ODT\n");
        if ((pOnfiInfo->uNvDdr2Feats & NAND_ONFI_NVDDR2_30OHM_ODT) != 0)
            printf ("            \t\t\tsupports ODT value of 30 Ohms\n");
        if ((pOnfiInfo->uNvDdr2Feats & NAND_ONFI_NVDDR_DIFFSIG_REn) != 0)
            printf ("            \t\t\tsupports differential signaling for RE_n"
                    "\n");
        if ((pOnfiInfo->uNvDdr2Feats & NAND_ONFI_NVDDR_DIFFSIG_DQS) != 0)
            printf ("            \t\t\tsupports differential signaling for DQS"
                    "\n");
        if ((pOnfiInfo->uNvDdr2Feats & NAND_ONFI_NVDDR_EXTVREFQ_200MTS) != 0)
            printf ("            \t\t\texternal VREFQ required for >= 200 MT/s"
                    "\n");
        }

    printf ("            \t\tVendor specific Revision number      : %d\n",
            pOnfiInfo->uVendorSpecRev);
    }
#endif /* _WRS_CONFIG_DEBUG_FLAG */

/*******************************************************************************
*
* vxbNandOnfiEccCalc - calculate NAND ONFI ECC
*
* This routine calculates the NAND ONFI ECC.
*
* RETURNS: CRC result
*
* ERRNO: N/A
*/

LOCAL UINT16 vxbNandOnfiEccCalc
    (
    UINT8 * pBuf,
    UINT32  uLen
    )
    {
    UINT32  uOnfiCrc = ONFI_CRC_CALC_INIT_VAL;
    UINT32  uBit;
    UINT32  i;
    UINT32  j;

    for (i = 0; i < uLen; i++)
        {
        for (j = 0x80; j != 0; j >>= 1)
            {
            uBit        = uOnfiCrc & ONFI_CRC_CALC_MSB;
            uOnfiCrc  <<= 1;

            if ((pBuf[i] & j) != 0)
                uBit ^= ONFI_CRC_CALC_MSB;

            if (uBit != 0)
                uOnfiCrc ^= ONFI_CRC_CALC_POLY;
            }

        uOnfiCrc &= 0xFFFF;
        }

    return (UINT16)(uOnfiCrc);
    }


/*******************************************************************************
*
* vxbNandOnfiStrFormat - fromat NAND ONFI string
*
* This routine fromats the NAND ONFI string.
*
* RETURNS: fromated string length
*
* ERRNO: N/A
*/

LOCAL UINT32 vxbNandOnfiStrFormat
    (
    char *  pStr,
    UINT32  uLen
    )
    {
    UINT32  i;
    UINT32  uStrLen;

    /* remove pad from tail */

    for (i = uLen; i > 0; i--)
        {
        if (pStr[i - 1] != NAND_ONFI_DEVSTR_PAD)
            break;
        else
            pStr[i -1] = 0;
        }

    if (i == 0)
        return (0);

    uStrLen = i;
    for (i = 0; i < uStrLen; i++)
        {
        if (!isascii (pStr[i]))
            pStr[i] = '?';
        }

    return (uStrLen);
    }

#ifdef _WRS_CONFIG_DEBUG_FLAG
/*******************************************************************************
*
* vxbNandOnfiStrPrintf - print NAND ONFI string
*
* This routine prints the NAND ONFI string.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

LOCAL void vxbNandOnfiStrPrintf
    (
    char *  pStr,
    UINT32  uLen
    )
    {
    UINT32  i;

    for (i = 0; i < uLen; i++)
        {
        if (pStr[i] == EOS)
            break;

        printf ("%c", pStr[i]);
        }
    }

/*******************************************************************************
*
* vxbNandOnfiTimingModePrintf - print NAND ONFI timing mode
*
* This routine prints the NAND timing mode.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

LOCAL void vxbNandOnfiTimingModePrintf
    (
    UINT16  uTimingMode
    )
    {
    UINT8  i;

    for (i = 0; i < 16; i++)
        {
        if (NAND_ONFI_TIMING_MODE_SUP (uTimingMode, i))
            printf (" %d", i);
        }
    }

/*******************************************************************************
*
* vxbNandOnfiBlkLifeCalc - calculate NAND block life
*
* This routine calculates the NAND block life.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

LOCAL UINT32 vxbNandOnfiBlkLifeCalc
    (
    UINT8   uVal,
    UINT8   uMulti
    )
    {
    UINT32  uLife = uVal;
    UINT8   i;

    for (i = 0; i < uMulti; i++)
        uLife *= 10;

    return (uLife);
    }
#endif /* _WRS_CONFIG_DEBUG_FLAG */

