/* vxbSdMmcLib.c - Generic SD/SDIO bus Library */

/*
 * Copyright (c) 2012-2021, 2023 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION

This is the generic SD/SDIO bus library to implement neccessary SD commands for
a SD memory card or SDIO card which is compatible with SD memory card
specification 2.0.

The library can support up to 32 SD/MMC cards in a system.
*/

/* includes */

#include <vxWorks.h>
#include <intLib.h>
#include <stdio.h>
#include <logLib.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <cacheLib.h>
#include <iv.h>

#include <hwif/vxBus.h>
#include <hwif/vxbus/vxbBusType.h>
#include <subsys/int/vxbIntLib.h>
#include <subsys/int/vxbDyncIntLib.h>
#include <subsys/timer/vxbTimerLib.h>

#include <vxbSdMmcLib.h>

/* defines */

#ifdef  SD_LIB_DBG_ON

#include <ioLib.h>

IMPORT	_Vx_freq_t sysClkRateGet (void);
IMPORT  _Vx_ticks_t tickGet (void);

#ifdef  LOCAL
#undef  LOCAL
#define LOCAL
#endif /* LOCAL */

#define SD_LIB_DBG_BUSINIT       0x00000001
#define SD_LIB_DBG_CMD           0x00000002
#define SD_LIB_DBG_ERR           0x00000004
#define SD_LIB_DBG_INT           0x00000008
#define SD_LIB_DBG_ALL           0xffffffff
#define SD_LIB_DBG_OFF           0x00000000

UINT32 sdLibDbgMask = SD_LIB_DBG_OFF;

IMPORT FUNCPTR _func_logMsg;

#define SD_LIB_DBG(mask, string, a, b, c, d, e, f)               \
    if ((sdLibDbgMask & mask) || (mask == SD_LIB_DBG_ALL))       \
        if (_func_logMsg != NULL) \
            (* _func_logMsg)(string, a, b, c, d, e, f)
#else
#   define SD_LIB_DBG(mask, string, a, b, c, d, e, f)
#endif  /* SD_LIB_DBG_ON */

LOCAL UINT8 tuningBlockPattern[SD_TUNING_BLOCK_PATTERN_LEN] = {
    0xFF, 0x0F, 0xFF, 0x00, 0xFF, 0xCC, 0xC3, 0xCC,
    0xC3, 0x3C, 0xCC, 0xFF, 0xFE, 0xFF, 0xFE, 0xEF,
    0xFF, 0xDF, 0xFF, 0xDD, 0xFF, 0xFB, 0xFF, 0xFB,
    0xBF, 0xFF, 0x7F, 0xFF, 0x77, 0xF7, 0xBD, 0xEF,
    0xFF, 0xF0, 0xFF, 0xF0, 0x0F, 0xFC, 0xCC, 0x3C,
    0xCC, 0x33, 0xCC, 0xCF, 0xFF, 0xEF, 0xFF, 0xEE,
    0xFF, 0xFD, 0xFF, 0xFD, 0xDF, 0xFF, 0xBF, 0xFF,
    0xBB, 0xFF, 0xF7, 0xFF, 0xF7, 0x7F, 0x7B, 0xDE
};

BUSTYPE_DEF(vxbSdMmcBus, "SdMmc bus type");

/*******************************************************************************
*
* sdDeviceCreate - notify the bus subsystem of a device on SD bus
*
* This routine send SDIO cmd5 to target and check target's type. And store the
* cmdIssue pointer to SD_HARDWARE struct. The SD_HARDWARE struct is stored to
* the pVxbIvars field of VxBus device structure.
* Finally, sdDeviceCreate is called to announce the device to the system.
*
* RETURNS: OK, or ERROR
*
* ERRNO
*
* \NOMANUAL
*/

STATUS sdDeviceCreate
    (
    VXB_DEV_HANDLE pInst,
    void * pArg
    )
    {
    VXB_DEV_ID      pDev = NULL;
    SD_HARDWARE   * pSdHardWare;
    SD_HOST_SPEC  * pSdHostSpec;
    STATUS rc;

    /* initialize generic bus info */

    rc = vxbDevCreate (&pDev);

    if (rc != OK)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdDeviceAnnounce() - vxbDevStructAlloc not success\n",
                    0, 0, 0, 0, 0, 0);
        return(ERROR);
        }

    /* initialize bus-specific info */

    pSdHardWare = (SD_HARDWARE *) vxbMemAlloc(sizeof (SD_HARDWARE));
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdDeviceAnnounce - hwMemAlloc not success for pSdHardWare\n",
                    0, 0, 0, 0, 0, 0);
        vxbMemFree(pDev);
        return(ERROR);
        }

    /* init pSdHardWare's pHostSpec and vxbSdCmdIssue */

    rc = VXB_SDSPECINFO_GET(pInst, &(pSdHardWare->pHostSpec), &(pSdHardWare->vxbSdCmdIssue));
    if (rc != OK)
        {
        SD_LIB_DBG(SD_LIB_DBG_ERR,
                   "sdDeviceAnnounce - VXB_SDSPECINFO_GET not success\n",
                   0, 0, 0, 0, 0 ,0);
        vxbMemFree(pDev);
        vxbMemFree(pSdHardWare);
        return(ERROR);
        }

    vxbDevClassSet(pDev, VXB_BUSID_SDMMC);

    vxbDevIvarsSet(pDev,pSdHardWare);

    pSdHostSpec = (SD_HOST_SPEC *)(pSdHardWare->pHostSpec);
    if (pSdHostSpec == NULL)
        {
        SD_LIB_DBG(SD_LIB_DBG_ERR, "sdDeviceAnnounce - pSdHostSpec null\n",
                0, 0, 0, 0, 0, 0);
        vxbMemFree(pDev);
        vxbMemFree(pSdHardWare);
        return(ERROR);
        }

    /*
     * Tell host controller how to find SD card pDev.
     * After this, when freeing SD card pDev, we need set it to NULL.
     */

    pSdHostSpec->childDev = pDev;

    /* get host controller unit number */
    pSdHostSpec->unitNumber = vxbDevUnitGet(pInst);

    pSdHardWare->pHostDev = pInst;

    if (pSdHostSpec->vxbSdGetHostCap)
        {
        (void)(pSdHostSpec->vxbSdGetHostCap)(pInst, &(pSdHostSpec->hostCap));
        }

    if (SDMMC_HOST_IS_SPI(pSdHostSpec))
        {
        return sdSpiDeviceInit(pDev,0);
        }
    else
        {
        return sdDeviceInit(pDev,0);
        }
    }
/*******************************************************************************
*
* sdSpiDeviceInit - init a SD card in SPI mode
*
* This routine starts card identification in SPI mode.
*
* RETURNS: OK, or ERROR
*
* ERRNO
*
* \NOMANUAL
*/

STATUS sdSpiDeviceInit
    (
    VXB_DEV_HANDLE pDev,
    UINT32 reInit
    )
    {
    VXB_DEV_HANDLE pInst;
    SD_HARDWARE   * pSdHardWare;
    SD_HOST_SPEC  * pSdHostSpec;
    STATUS rc;
    UINT32  capbility;
    UINT32 timeOut = 0;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        return ERROR;

    pSdHostSpec = (SD_HOST_SPEC *)(pSdHardWare->pHostSpec);
    if (pSdHostSpec == NULL)
        return ERROR;

    capbility = pSdHostSpec->capbility;
    pInst = pSdHardWare->pHostDev;
    rc = sdCmdGoIdleState (pDev);
    if (rc == ERROR)
        {
        SD_LIB_DBG(SD_LIB_DBG_ERR,
                   "sdDeviceInit - sdCmdGoIdleState not success\n\n",
                   0, 0, 0, 0, 0, 0);
        pSdHostSpec->childDev = NULL;
        vxbMemFree(pDev);
        vxbMemFree(pSdHardWare);
        return(ERROR);
        }

    /* Only SD card is support in SPI mode, mmc isn't support after MMC V4.3 */

     pSdHardWare->isMmc = FALSE;
    (void)sdCmdGoIdleState (pDev);
    (void)sdCmdSendIfCond (pDev, capbility);
    if (pSdHardWare->version == SD_VERSION_UNKNOWN)
        {
        SD_LIB_DBG(SD_LIB_DBG_ERR,
                   "sdDeviceAnnounce - sdCmdGoIdleState not success\n\n",
                   0, 0, 0, 0, 0, 0);
        pSdHostSpec->childDev = NULL;
        vxbMemFree(pDev);
        vxbMemFree(pSdHardWare);
        return(ERROR);
        }

    while (timeOut++ < SDMMC_COMMAND_TIMEOUT)
        {
        rc = sdACmdSendOpCond (pDev, capbility, pSdHardWare->version, reInit);
        if ((rc == OK)&& (pSdHardWare->ocrValue == 0x0))
            {
            break;
            }

        if (pSdHardWare->hasException)
            {
            SD_LIB_DBG(SD_LIB_DBG_ERR, "sdDeviceInit - the device may get exception\n\n",
                       0, 0, 0, 0, 0, 0);

            /* Since the device may get some exception, quickly increase
               timeout to reduce loop time. */

            timeOut = timeOut * 2;
            }
        vxbMsDelay(1);
        }
    if (timeOut >= SDMMC_COMMAND_TIMEOUT)
        {
        SD_LIB_DBG(SD_LIB_DBG_ERR,
                   "sdDeviceAnnounce - sdACmdSendOpCond not success\n",
                   0, 0, 0, 0, 0, 0);
        pSdHostSpec->childDev = NULL;
        vxbMemFree(pDev);
        vxbMemFree(pSdHardWare);
        return(ERROR);
        }
    if (sdCmdSpiReadOcr(pDev) == OK)
        {
        pSdHardWare->voltage = (pSdHardWare->ocrValue) & OCR_VDD_VOL_MASK;
        }
    vxbDevNameSet (pDev, SD_CARD_NAME, TRUE);
    pSdHardWare->isComboCard = FALSE;
    pSdHardWare->isSdio = FALSE;
    pSdHardWare->isSpi = TRUE;

    rc = vxbDevAdd (pInst, pDev);
    if (rc == ERROR)
        {
        SD_LIB_DBG(SD_LIB_DBG_ERR,
                   "sdDeviceAnnounce - vxbDevAdd not success\n",
                   0, 0, 0, 0, 0, 0);
        pSdHostSpec->childDev = NULL;
        vxbMemFree(pDev);
        vxbMemFree(pSdHardWare);
        return(ERROR);
        }

    pSdHardWare->recover = TRUE;
    return OK;
    }

/*******************************************************************************
*
* sdDeviceInit - init a SD card
*
* This routine starts card identification.
*
* RETURNS: OK, or ERROR
*
* ERRNO
*
* \NOMANUAL
*/

STATUS sdDeviceInit
    (
    VXB_DEV_HANDLE pDev,
    UINT32 reInit
    )
    {
    VXB_DEV_HANDLE pInst;
    SD_HARDWARE   * pSdHardWare;
    SD_HOST_SPEC  * pSdHostSpec;
    STATUS rc;
    unsigned i = 0;
    UINT32  capbility;
    UINT32 timeOut = 0;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        return ERROR;

    pSdHostSpec = (SD_HOST_SPEC *)(pSdHardWare->pHostSpec);
    if (pSdHostSpec == NULL)
        return ERROR;

    capbility = pSdHostSpec->capbility;
    pInst = pSdHardWare->pHostDev;

    rc = sdCmdGoIdleState (pDev);
    if (rc == ERROR)
        {
        SD_LIB_DBG(SD_LIB_DBG_ERR,
                   "sdDeviceInit - sdCmdGoIdleState not success\n\n",
                   0, 0, 0, 0, 0, 0);
        pSdHostSpec->childDev = NULL;
        vxbMemFree(pDev);
        vxbMemFree(pSdHardWare);
        return(ERROR);
        }

    /* CMD8 is required to support SDHC or SDXC */

    if (!(pSdHostSpec->specFlag & SDHC_SPEC_FLAG_NO_SD))
        {
        (void) sdCmdSendIfCond (pDev, 0);
        }

    rc = sdioCmdIoSendOpCond(pDev, capbility & OCR_VDD_VOL_MASK);

#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
    if (TRUE)
#else
    if ((pSdHardWare->cmdErr & SDMMC_CMD_ERR_TIMEOUT) != 0x0)
#endif
        {

        /* To conform to JEDEC specification */
        rc = sdMmcCmdSendOpCond(pDev, 0);
        if ((rc == ERROR) ||
            ((pSdHardWare->cmdErr & SDMMC_CMD_ERR_TIMEOUT) != 0x0))
            pSdHardWare->isMmc = FALSE;
        else
            pSdHardWare->isMmc = TRUE;

#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
        pSdHardWare->isMmc = FALSE;
#endif
        if (pSdHardWare->isMmc)
            {
            sdCmdGoIdleState(pDev);
            while (timeOut++ < SDMMC_COMMAND_TIMEOUT)
                {
                rc = sdMmcCmdSendOpCond(pDev,
                                        (pSdHardWare->ocrValue |
                                        OCR_CARD_CAP_STS));
                if ((rc == OK) &&
                    ((pSdHardWare->ocrValue & (UINT32)OCR_CARD_PWRUP_STS) != 0x0))
                    {
                    pSdHardWare->voltage = (pSdHardWare->ocrValue) & OCR_VDD_VOL_MASK;
                    break;
                    }
                vxbMsDelay(1);
                }
            vxbDevNameSet (pDev, MMC_CARD_NAME, TRUE);
            }
        else
            {
            (void)sdCmdGoIdleState (pDev);
            (void)sdCmdSendIfCond (pDev, capbility);
            if (pSdHardWare->version == SD_VERSION_UNKNOWN)
                {
                SD_LIB_DBG(SD_LIB_DBG_ERR,
                           "sdDeviceAnnounce - sdCmdGoIdleState not success\n\n",
                           0, 0, 0, 0, 0, 0);
                pSdHostSpec->childDev = NULL;
                vxbMemFree(pDev);
                vxbMemFree(pSdHardWare);
                return(ERROR);
                }
            while (timeOut++ < SDMMC_COMMAND_TIMEOUT)
                {
                rc = sdACmdSendOpCond (pDev, capbility, pSdHardWare->version, reInit);

                if ((rc == OK) && ((pSdHardWare->ocrValue & (UINT32)OCR_CARD_PWRUP_STS) != 0x0))
                    {
                    pSdHardWare->voltage = (pSdHardWare->ocrValue) & OCR_VDD_VOL_MASK;
                    break;
                    }

                if (pSdHardWare->hasException)
                    {
                    SD_LIB_DBG(SD_LIB_DBG_ERR, "sdDeviceInit - the device get exception\n\n",
                               0, 0, 0, 0, 0, 0);

                    /* Since the device may get some exception, quickly increase
                       timeout to reduce loop time. */

                    timeOut = timeOut * 2;
                    }

                vxbMsDelay(1);
                }
            if (timeOut >= SDMMC_COMMAND_TIMEOUT)
                {
                SD_LIB_DBG(SD_LIB_DBG_ERR,
                           "sdDeviceAnnounce - sdACmdSendOpCond not success\n",
                           0, 0, 0, 0, 0, 0);
                pSdHostSpec->childDev = NULL;
                vxbMemFree(pDev);
                vxbMemFree(pSdHardWare);
                return(ERROR);
                }

            vxbDevNameSet (pDev, SD_CARD_NAME, TRUE);
            }

        pSdHardWare->isComboCard = FALSE;
        pSdHardWare->isSdio = FALSE;

        rc = vxbDevAdd (pInst, pDev);
        if (rc == ERROR)
            {
            SD_LIB_DBG(SD_LIB_DBG_ERR,
                       "sdDeviceAnnounce - vxbDevAdd not success\n",
                       0, 0, 0, 0, 0, 0);
            pSdHostSpec->childDev = NULL;
            vxbMemFree(pDev);
            vxbMemFree(pSdHardWare);
            return(ERROR);
            }

        pSdHardWare->recover = TRUE;

        /* coverity[leaked_storage] */

        /*
         * pDev and pSdHardWare are used with vxBus system,
         * so we can't free it. Set coverity keyword
         */

        return OK;
        }

    if (rc == OK)
        {
        UINT8 funcNum = 0;

        funcNum = (pSdHardWare->ocrValue >> 28) & 0x07;

        if (funcNum > 0)
            {
            rc = sdioCmdIoSendOpCond(pDev, capbility & 0x00FFFFFF);
            if (rc == ERROR)
                {
                SD_LIB_DBG (SD_LIB_DBG_ERR,
                            "sdDeviceAnnounce - sdioCmdIoSendOpCond not success\n",
                            0, 0, 0, 0, 0, 0);
                vxbMemFree(pDev);
                vxbMemFree(pSdHardWare);
                return(ERROR);
                }
            do{
               rc = sdCmdSendRelativeAddr (pDev);
               if (rc == ERROR)
                   {
                   SD_LIB_DBG (SD_LIB_DBG_ERR,
                               "sdDeviceAnnounce - sdCmdSendRelativeAddr not success\n",
                               0, 0, 0, 0, 0, 0);
                   vxbMemFree(pDev);
                   vxbMemFree(pSdHardWare);
                   return(ERROR);
                   }
               } while (pSdHardWare->rcaValue == 0);

            rc = sdCmdSelectCard (pDev);
            if (rc == ERROR)
                {
                SD_LIB_DBG (SD_LIB_DBG_ERR,
                            "sdDeviceAnnounce - sdCmdSelectCard not success\n",
                            0, 0, 0, 0, 0, 0);
                vxbMemFree(pDev);
                vxbMemFree(pSdHardWare);
                return(ERROR);
                }

            for (i = 0; i < funcNum; i++)
                {
                if (pDev == NULL)
                    {
                    rc = vxbDevCreate (&pDev);
                    if (rc == ERROR)
                        {
                        SD_LIB_DBG (SD_LIB_DBG_ERR,
                                    "sdDeviceAnnounce - vxbDevCreate not success\n",
                                    0, 0, 0, 0, 0, 0);
                        vxbMemFree(pSdHardWare);
                        return(ERROR);
                        }
                    }
                if (pSdHardWare == NULL)
                    {
                    pSdHardWare = (SD_HARDWARE *) vxbMemAlloc(sizeof (SD_HARDWARE));

                    if (pSdHardWare == NULL)
                        {
                        SD_LIB_DBG (SD_LIB_DBG_ERR,
                                    "sdDeviceAnnounce - hwMemAlloc not success for pSdHardWare\n",
                                    0, 0, 0, 0, 0, 0);
                        vxbMemFree(pDev);
                        return(ERROR);
                        }

                    rc = VXB_SDSPECINFO_GET(pInst, &(pSdHardWare->pHostSpec), &(pSdHardWare->vxbSdCmdIssue));
                    if (rc != OK)
                        {
                        SD_LIB_DBG (SD_LIB_DBG_ERR,
                                    "sdDeviceAnnounce - VXB_SDSPECINFO_GET not success\n",
                                    0, 0, 0, 0, 0, 0);
                        vxbMemFree(pDev);
                        vxbMemFree(pSdHardWare);
                        return(ERROR);
                        }

                    vxbDevClassSet(pDev, VXB_BUSID_SDMMC);

                    vxbDevIvarsSet(pDev,pSdHardWare);

                    pSdHostSpec = (SD_HOST_SPEC *)(pSdHardWare->pHostSpec);
                    if (pSdHostSpec == NULL)
                        {
                        SD_LIB_DBG (SD_LIB_DBG_ERR,
                                    "sdDeviceAnnounce - pSdHostSpec null\n",
                                    0, 0, 0, 0, 0, 0);
                        vxbMemFree(pDev);
                        vxbMemFree(pSdHardWare);
                        return(ERROR);
                        }

                    capbility = pSdHostSpec->capbility;
                    }

                pSdHardWare->isSdio = TRUE;
                pSdHardWare->isComboCard = FALSE;
                pSdHardWare->funcNum = i + 1;

                /* Need not check return status at here */

                rc = vxbDevAdd (pInst, pDev);
                if (rc == ERROR)
                    {
                    SD_LIB_DBG (SD_LIB_DBG_ERR,
                                "sdDeviceAnnounce - vxbDevAdd not success\n",
                                0, 0, 0, 0, 0, 0);
                    vxbMemFree(pDev);
                    vxbMemFree(pSdHardWare);
                    return(ERROR);
                    }

                pDev = NULL;
                pSdHardWare = NULL;
                }
            }
        }

    return OK;
    }

/*******************************************************************************
*
* sdCmdGoIdleState - issue CMD0 to reset all cards to idle state
*
* This routine issues CMD0 to reset all cards to idle state.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdGoIdleState
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;
    STATUS rc;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdGoIdleState() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }
    if (SDMMC_HOST_IS_SPI(pSdHardWare->pHostSpec))
        {
        SD_CMD_FORM (cmd, SDMMC_CMD_GO_IDLE_STATE, 0, SDMMC_SPI_RSP_R1, FALSE);
        }
    else
        {
        SD_CMD_FORM (cmd, SDMMC_CMD_GO_IDLE_STATE, 0, SDMMC_CMD_RSP_NONE, FALSE);
        }

    SD_LIB_DBG (SD_LIB_DBG_CMD, "sdCmdGoIdleState command\n", 0, 0, 0, 0, 0, 0);

    rc = sdCmdIssue(pDev, &cmd);

    /* Need at last 1ms delay for waiting for CMD0 respond */

    vxbMsDelay(1);
    return (rc);
    }

/*******************************************************************************
*
* sdMmcCmdSendOpCond - issue CMD1 to get mmc card OCR value
*
* This routine issues CMD1 to get mmc card OCR value.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdMmcCmdSendOpCond
    (
    VXB_DEV_HANDLE pDev,
    UINT32 ocrValue
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdMmcCmdSendOpCond() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    SD_CMD_FORM (cmd, MMC_CMD_SEND_OP_COND,
                 SDMMC_HOST_IS_SPI(pSdHardWare->pHostSpec) ? 0 : ocrValue,
                 SDMMC_CMD_RSP_R3 | SDMMC_SPI_RSP_R1, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdMmcCmdSendOpCond() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    pSdHardWare->ocrValue = cmd.cmdRsp[0];
    SD_LIB_DBG (SD_LIB_DBG_CMD,
                "sdMmcCmdSendOpCond() - ocrValue = 0x%x\n",
                pSdHardWare->ocrValue, 0, 0, 0, 0, 0);

    return (OK);
    }

/*******************************************************************************
*
* sdCmdAllSendCid - issue CMD2 to ask all cards to send their CIDs
*
* This routine issues CMD2 to ask all cards to send their CIDs.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdAllSendCid
    (
    VXB_DEV_HANDLE pDev,
    void * pCid
    )
    {
    SD_CMD cmd;

    SD_CMD_FORM (cmd, SDMMC_CMD_ALL_SEND_CID, 0, SDMMC_CMD_RSP_R2, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdCmdAllSendCid() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    bcopy ((char *)cmd.cmdRsp, (char *)pCid, CID_SIZE);
    SD_LIB_DBG (SD_LIB_DBG_CMD, "sdCmdAllSendCid() - cmdRsp[0] = 0x%x\n",
                cmd.cmdRsp[0], 0, 0, 0, 0, 0);

    return OK;
    }

/*******************************************************************************
*
* sdCmdSendRelativeAddr - issue CMD3 to ask card to publish a new RCA
*
* This routine issues CMD3 to ask card to publish a new RCA.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSendRelativeAddr
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendRelativeAddr() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    SD_CMD_FORM (cmd, SD_CMD_SEND_RELATIVE_ADDR, 0, SDMMC_CMD_RSP_R6, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendRelativeAddr() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    pSdHardWare->rcaValue = (UINT16)((cmd.cmdRsp[0] >> 16) & 0xffff);

    SD_LIB_DBG (SD_LIB_DBG_CMD, "sdCmdSendRelativeAddr() - rcaValue = 0x%x\n",
                pSdHardWare->rcaValue, 0, 0, 0, 0, 0);
    return (OK);
    }

/*******************************************************************************
*
* sdMmcCmdSetRelativeAddr - issue CMD3 to ask card to publish a new RCA
*
* This routine issues CMD3 to ask card to publish a new RCA.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdMmcCmdSetRelativeAddr
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdMmcCmdSetRelativeAddr() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    if (pSdHardWare->rcaValue == 0x0)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdMmcCmdSetRelativeAddr() - pSdHardWare rcaValue is 0x0\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    SD_CMD_FORM (cmd, MMC_CMD_SET_RELATIVE_ADDR, (UINT32)(pSdHardWare->rcaValue << 16),
                 SDMMC_CMD_RSP_R1, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdMmcCmdSetRelativeAddr() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    return (OK);
    }

/*******************************************************************************
*
* sdCmdSetDsr - issue CMD4 to program the DSR of all cards
*
* This routine issues CMD4 to program the DSR of all cards.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSetDsr
    (
    VXB_DEV_HANDLE pDev,
    UINT32        dsr
    )
    {
    SD_CMD cmd;

    SD_CMD_FORM (cmd, SDMMC_CMD_SET_DSR, dsr << 16, SDMMC_CMD_RSP_NONE, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* sdioCmdIoSendOpCond - issue CMD5 to ask SDIO card to publish OCR value
*
* This routine issues CMD5 to ask SDIO card to publish OCR value.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdioCmdIoSendOpCond
    (
    VXB_DEV_HANDLE pDev,
    UINT32 ocrValue
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdioCmdIoSendOpCond() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    SD_CMD_FORM (cmd, SDIO_CMD_SEND_OPCOND, ocrValue, SDMMC_CMD_RSP_R4, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdioCmdIoSendOpCond() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    pSdHardWare->ocrValue = cmd.cmdRsp[0];
    SD_LIB_DBG (SD_LIB_DBG_CMD, "sdioCmdIoSendOpCond() - ocrValue = 0x%x\n",
                pSdHardWare->ocrValue, 0, 0, 0, 0, 0);

    return (OK);
    }

/*******************************************************************************
*
* mmcSwitch - issue CMD6 to write MMC card's ext_csd register.
*
* This routine issue CMD6 to read/write MMC card's ext_csd register. Index is
* byte offset address in ext_csd register. Data is wrote to ext_csd register's
* value. Set is operation attribute.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mmcSwitch
    (
    VXB_DEV_HANDLE pDev,
    UINT8  index,
    UINT8  data,
    UINT8  set
    )
    {
    SD_CMD cmd;
    UINT32 arg = 0;
    UINT32 rspValue = 0;
    UINT32 timeOut = 0;

    arg = (UINT32)(MMC_SWITCH_MODE_WRITE_BYTE << 24) |
           (UINT32)(index << 16) |
           (UINT32)(data << 8) |
           set;

    SD_CMD_FORM (cmd, MMC_CMD_SWITCH, arg, SDMMC_CMD_RSP_R1B | SDMMC_SPI_RSP_R1, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "mmcSwitch() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    do
        {

        /* Need not check return status */

        (void)sdCmdSendStatus (pDev, &rspValue);
        timeOut++;
        if (timeOut >= SDMMC_COMMAND_TIMEOUT)
            return (ERROR);
        }while(CARD_STS_CUR_GET(rspValue) == CARD_STS_PRG);

    return (OK);
    }

/*******************************************************************************
*
* sdACmdSwitchFunc - send application specific command 6
*
* This routine sends application specific command 6. The input buffer and
* buffer length are provided by the caller.
*
* RETURNS: OK, or ERROR if command issue failed
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdACmdSwitchFunc
    (
    VXB_DEV_HANDLE pDev,
    void * buffer,
    UINT32 len,
    int    mode,
    int    group,
    int    val
    )
    {
    SD_CMD cmd;
    UINT32 arg = 0;

    arg = (UINT32)mode;
    arg |= 0xffffff;
    arg &= (UINT32)~(0xf << (group * 4));
    arg |= (UINT32)(val << (group * 4));

    SD_CMD_FORM (cmd, SD_CMD_SWITCH_FUNC, arg, SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, TRUE);
    SD_DATA_FORM (cmd, buffer, 1, len, TRUE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdACmdSwitchFunc() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    return OK;
    }

/*******************************************************************************
*
* sdCmdSelectCard - issue CMD7 to select a card
*
* This routine issues CMD7 to select a card.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSelectCard
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;
    STATUS rc;
    UINT32 i;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdioCmdIoSendOpCond() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    SD_CMD_FORM (cmd, SDMMC_CMD_SELECT_CARD,
                 SDMMC_CMD_ARG_RCA (pSdHardWare->rcaValue),
                 SDMMC_CMD_RSP_R1B, FALSE);

    for (i = 0; i < SDMMC_CMD_RETRY; i++)
        {
        rc = sdCmdIssue(pDev, &cmd);
        if (rc == ERROR)
            continue;
        else
            break;
        }

    return rc;
    }

/*******************************************************************************
*
* sdCmdDeselectCard - issue CMD7 to deselect a card
*
* This routine issues CMD7 to deselect a card.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdDeselectCard
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;

    SD_CMD_FORM (cmd, SDMMC_CMD_SELECT_CARD, 0, SDMMC_CMD_RSP_R1B, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* sdMmcACmdSendScr - issue ACMD51 to ask card to send its SCR
*
* This routine issues ACMD51 to ask card to send its SCR.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdMmcSendTuning
    (
    VXB_DEV_HANDLE pDev
    )
    {
    STATUS rc;
    UINT8 *sendData;
    SD_CMD cmd;

    sendData = (UINT8 *)cacheDmaMalloc (sizeof(UINT8) * SD_TUNING_BLOCK_PATTERN_LEN);
    if (sendData == NULL)
        return ERROR;

    SD_CMD_FORM (cmd, SDMMC_CMD_SEND_TUNING, 0, SDMMC_CMD_RSP_R1, TRUE);
    SD_DATA_FORM (cmd, sendData, 1, sizeof(UINT8) * SD_TUNING_BLOCK_PATTERN_LEN, TRUE);

    rc = sdCmdIssue(pDev, &cmd);
    (void)cacheDmaFree ((void *)sendData);

    return (rc);
    }

/*******************************************************************************
*
* sdCmdSendIfCond - issue CMD8 to ask card to send interface condition
*
* This routine issues CMD8 to ask card to send interface condition.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSendIfCond
    (
    VXB_DEV_HANDLE pDev,
    UINT32 capbility
    )
    {
    SD_CMD cmd;
    UINT32 arg = 0;
    STATUS rc;
    SD_HARDWARE * pSdHardWare;
    UINT8 result;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendIfCond() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    arg = (SD_CMD8_ARG_VHS_27_36 << SD_CMD8_ARG_VHS_SHIFT) |
           SD_CMD8_ARG_CHK_PATTERN;

    SD_CMD_FORM (cmd, SD_CMD_SEND_IF_COND, arg, SDMMC_CMD_RSP_R7 | SDMMC_SPI_RSP_R7, FALSE);

    rc = sdCmdIssue(pDev, &cmd);

    if (cmd.cmdErr & SDMMC_CMD_ERR_TIMEOUT)
        {
        pSdHardWare->version = SD_VERSION_100;
        SD_LIB_DBG (SD_LIB_DBG_CMD,
                    "sdCmdSendIfCond() - sd target version is SD_VERSION_100\n",
                    0, 0, 0, 0, 0, 0);
        return (OK);
        }

    if (rc == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendIfCond() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return (ERROR);
        }

    if (SDMMC_HOST_IS_SPI(pSdHardWare->pHostSpec))
        result = cmd.cmdRsp[1] & 0xFF;
    else
        result = cmd.cmdRsp[0] & 0xFF;

    if (result != SD_CMD8_ARG_CHK_PATTERN)
        {
        SD_LIB_DBG (SD_LIB_DBG_CMD,
                    "sdCmdSendIfCond() - sd target version is SD_VERSION_UNKNOWN\n",
                    0, 0, 0, 0, 0, 0);
        pSdHardWare->version = SD_VERSION_UNKNOWN;
        }
    else
        {
        SD_LIB_DBG (SD_LIB_DBG_CMD,
                    "sdCmdSendIfCond() - sd target version is SD_VERSION_200\n",
                    0, 0, 0, 0, 0, 0);
        pSdHardWare->version = SD_VERSION_200;
        }
    return (OK);
    }

/*******************************************************************************
*
* mmcCmdReadExtCsd - issue CMD8 to read MMC card's ext_csd register
*
* This routine issue CMD8 to read MMC card's ext_csd register.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mmcCmdReadExtCsd
    (
    VXB_DEV_HANDLE  pDev,
    UINT8 * pExtCsd
    )
    {
    SD_CMD cmd;
    UINT8 * pExtData;
    STATUS rc;

    pExtData = (UINT8 *)cacheDmaMalloc (sizeof(UINT8) * 512);
    if (pExtData == NULL)
        return ERROR;
    bzero((char *)pExtData, sizeof(UINT8) * 512);

    SD_CMD_FORM (cmd, MMC_CMD_SEND_EXT_CSD, 0, SDMMC_CMD_RSP_R1, TRUE);
    SD_DATA_FORM (cmd, pExtData, 1, (sizeof(UINT8) * 512), TRUE);

    rc = sdCmdIssue(pDev, &cmd);
    if (rc == OK)
        {
        bcopy ((char *)pExtData, (char *)(pExtCsd), (sizeof(UINT8) * 512));
        }

    (void)cacheDmaFree((void *)pExtData);
    return (rc);
    }

/*******************************************************************************
*
* sdCmdSendCsd - issue CMD9 to ask card to send its CSD
*
* This routine issues CMD9 to ask card to send its CSD.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSendCsd
    (
    VXB_DEV_HANDLE pDev,
    SDMMC_CSD * pCsd
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;
    UINT8 * pRegData = NULL;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendCsd() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    if (pCsd == NULL)
        return ERROR;
    if (SDMMC_HOST_IS_SPI(pSdHardWare->pHostSpec))
        {

        /* Read CSD is a simple read-block transaction in SPI mode */

        pRegData = (UINT8 *) malloc (sizeof(UINT8) * 18);
        if (pRegData == NULL)
            return ERROR;
        SD_CMD_FORM (cmd, SDMMC_CMD_SEND_CSD, SDMMC_CMD_ARG_RCA (pSdHardWare->rcaValue),
                     SDMMC_SPI_RSP_R1, TRUE);
        SD_DATA_FORM (cmd, pRegData, 1, 16, TRUE);
        }
    else
        SD_CMD_FORM (cmd, SDMMC_CMD_SEND_CSD, SDMMC_CMD_ARG_RCA (pSdHardWare->rcaValue),
                     SDMMC_CMD_RSP_R2 , FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendCsd() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        if ( pRegData != NULL )
            {
            (void) free(pRegData);
            }
        return (ERROR);
        }
    if (SDMMC_HOST_IS_SPI(pSdHardWare->pHostSpec))
        {
        bcopy ((char *)cmd.cmdData.buffer, (char *)pCsd, 16);
        }
    else
        {
        bcopy ((char *)cmd.cmdRsp, (char *)pCsd, 16);
        }
    if ( pRegData != NULL )
        {
        (void) free(pRegData);
        }
    return OK;
    }

/*******************************************************************************
*
* sdCmdSendCid - issue CMD10 to ask card to send its CID
*
* This routine issues CMD10 to ask card to send its CID.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSendCid
    (
    VXB_DEV_HANDLE pDev,
    void * pCid
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;
    UINT8 * pRegData = NULL;

    if (pCid == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendCid() - pCid is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendCid() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }
    if (SDMMC_HOST_IS_SPI(pSdHardWare->pHostSpec))
        {

        /* Read Cid is a simple read-block transaction in SPI mode */

        pRegData = (UINT8 *) malloc (sizeof(UINT8) * 18);
        if (pRegData == NULL)
            return ERROR;
        SD_CMD_FORM (cmd, SDMMC_CMD_SEND_CID, SDMMC_CMD_ARG_RCA (pSdHardWare->rcaValue),
                     SDMMC_SPI_RSP_R1, TRUE);
        SD_DATA_FORM (cmd, pRegData, 1, CID_SIZE, TRUE);
        }
    else
        SD_CMD_FORM (cmd, SDMMC_CMD_SEND_CID, SDMMC_CMD_ARG_RCA (pSdHardWare->rcaValue),
                     SDMMC_CMD_RSP_R2, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendCid() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        if ( pRegData != NULL )
            {
            (void) free(pRegData);
            }
        return (ERROR);
        }
    if (SDMMC_HOST_IS_SPI(pSdHardWare->pHostSpec))
        {
        bcopy ((char *)cmd.cmdData.buffer, (char *)pCid, CID_SIZE);
        }
    else
        {
        bcopy ((char *)cmd.cmdRsp, (char *)pCid, CID_SIZE);
        }
    if ( pRegData != NULL )
        {
        (void) free(pRegData);
        }
    return OK;
    }

/*******************************************************************************
*
* sdCmdSwitchVoltage - issue CMD11 to set card's voltage to 1.8V
*
* This routine issue CMD11 to set card's voltage to 1.8V.
* This routine is used for UHS-I supported card and mandatory for UHS-I card
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSwitchVoltage
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;

    SD_CMD_FORM (cmd, SD_CMD_SWITCH_VOLTAGE, 0, SDMMC_CMD_RSP_R1, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* sdCmdStopTransmission - issue CMD12 to force card to stop transmission
*
* This routine issues CMD12 to force card to stop transmission.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdStopTransmission
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;

    SD_CMD_FORM (cmd, SDMMC_CMD_STOP_TRANSMISSION, 0, SDMMC_CMD_RSP_R1B | SDMMC_SPI_RSP_R1B, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* sdCmdSendStatus - issue CMD13 to ask card to send its status
*
* This routine issues CMD13 to ask card to send its status.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSendStatus
    (
    VXB_DEV_HANDLE pDev,
    UINT32 * stsData
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;

    /* pDev could be null when we use software SD card detection */

    if (pDev == NULL)
        return ERROR;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSendStatus() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    SD_CMD_FORM (cmd, SDMMC_CMD_SEND_STATUS,
                 SDMMC_CMD_ARG_RCA (pSdHardWare->rcaValue),
                 SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R2, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        return (ERROR);

    *stsData = cmd.cmdRsp[0];
    return (OK);
    }

/*******************************************************************************
*
* sdCmdGoInactiveState - issue CMD15 to put card into the inactive state
*
* This routine issues CMD15 to put card into the inactive state.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdGoInactiveState
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdCmdGoInactiveState() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    SD_CMD_FORM (cmd, SDMMC_CMD_GO_INACTIVE_STATE,
                 SDMMC_CMD_ARG_RCA (pSdHardWare->rcaValue),
                 SDMMC_CMD_RSP_NONE, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* sdCmdSetBlockLen - issue CMD16 to set block length for block commands
*
* This routine issues CMD16 to set block length for block commands.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSetBlockLen
    (
    VXB_DEV_HANDLE pDev,
    UINT32 blkLen
    )
    {
    SD_CMD cmd;

    /* block length is always 512 bytes */

    SD_CMD_FORM (cmd, SDMMC_CMD_SET_BLOCKLEN, blkLen,
                 SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* mmcCmdSetBlockNum - issue CMD23 to set block length for block commands
*
* This routine issues CMD23 to set block length for block commands.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mmcCmdSetBlockNum
    (
    VXB_DEV_HANDLE pDev,
    UINT32 blkNum
    )
    {
    SD_CMD cmd;

    SD_CMD_FORM (cmd, MMC_CMD_SET_BLOCK_COUNT, blkNum,
                 SDMMC_CMD_RSP_R1, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* sdSendTuning - issue CMD19 to send the tuning data
*
* This routine issues CMD19 to send the tuning data.
* This routine is used for UHS-I supported card and mandatory for UHS-I card
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdSendTuning
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;
    STATUS rc;
    UINT8 *pBuf;

    pBuf = (UINT8 *)cacheDmaMalloc (SD_TUNING_BLOCK_PATTERN_LEN);
    if (pBuf == NULL)
        return ERROR;

    SD_CMD_FORM (cmd, SDMMC_CMD_SEND_TUNING, 0, SDMMC_CMD_RSP_R1, TRUE);
    SD_DATA_FORM (cmd, pBuf, 1, SD_TUNING_BLOCK_PATTERN_LEN, TRUE);

    rc = sdCmdIssue(pDev, &cmd);
    if (rc == OK)
        {
        if (memcmp (tuningBlockPattern, pBuf, SD_TUNING_BLOCK_PATTERN_LEN) != 0)
            rc = ERROR;
        }

    (void)cacheDmaFree ((void *)pBuf);
    return (rc);
    }

/*******************************************************************************
*
* mmcSendTuning - issue CMD21 to send the tuning data
*
* This routine issues CMD21 to send the tuning data.
* This routine is used for HS200 mode.
*
* RETURNS: OK, or ERROR if bad parameters or operation fails
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mmcSendTuning
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD         cmd;
    STATUS         rc;

    SD_CMD_FORM (cmd, SDMMC_CMD_SEND_TUNING_HS200, 0, SDMMC_CMD_RSP_R1, FALSE);
    rc = sdCmdIssue(pDev, &cmd);

    if (rc != OK)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "mmcSendTuning() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    return (OK);
    }

/*******************************************************************************
*
* mmcCmdSeEraseGrpStart - issue CMD35 to set erase group start address
*
* This routine issues CMD35 to set erase group start address.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mmcCmdSeEraseGrpStart
    (
    VXB_DEV_HANDLE pDev,
    UINT32 startAddr
    )
    {
    SD_CMD cmd;
    UINT32 cmdIndex;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "mmcCmdSeEraseGrpStart() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    if (pSdHardWare->isMmc)
        cmdIndex = MMC_CMD_ERASE_GROUP_START;
    else
        cmdIndex = SD_CMD_ERASE_WR_BLK_START;

    SD_CMD_FORM (cmd, cmdIndex, startAddr,
                 SDMMC_CMD_RSP_R1, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* mmcCmdSeEraseGrpStart - issue CMD35 to set erase group start address
*
* This routine issues CMD35 to set erase group start address.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mmcCmdSeEraseGrpEnd
    (
    VXB_DEV_HANDLE pDev,
    UINT32 endAddr
    )
    {
    SD_CMD cmd;
    UINT32 cmdIndex;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "mmcCmdSeEraseGrpEnd() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    if (pSdHardWare->isMmc)
        cmdIndex = MMC_CMD_ERASE_GROUP_END;
    else
        cmdIndex = SD_CMD_ERASE_WR_BLK_END;

    SD_CMD_FORM (cmd, cmdIndex, endAddr,
                 SDMMC_CMD_RSP_R1, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* mmcCmdSeEraseGrpStart - issue CMD35 to set erase group start address
*
* This routine issues CMD35 to set erase group start address.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mmcCmdSetEraseGrp
    (
    VXB_DEV_HANDLE pDev,
    UINT eraseArg
    )
    {
    SD_CMD cmd;
    UINT32 rspValue = 0;
    UINT32 timeOut = 0;

    SD_CMD_FORM (cmd, SDMMC_CMD_ERASE, eraseArg, SDMMC_CMD_RSP_R1B, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "mmcCmdSeEraseGrp() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }
    do
        {

        /* Need not check return status */

        (void)sdCmdSendStatus (pDev, &rspValue);
        timeOut++;
        if (timeOut >= SDMMC_COMMAND_TIMEOUT)
            return (ERROR);
        }while((!(rspValue & CARD_STS_READY_FOR_DATA)) ||
               (CARD_STS_CUR_GET(rspValue) == CARD_STS_PRG));

    return (OK);
    }

/*******************************************************************************
*
* mmcSendHpi - send HPI command to device
*
* This routine sends HPI command to device.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mmcSendHpi
    (
    VXB_DEV_HANDLE pDev,
    UINT32 cmdIndex,
    UINT32 * pStatus
    )
    {
    SD_CMD sdCmd;
    SD_HARDWARE * pSdHardWare;
    UINT32 argValue;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "mmcSendHpi() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    argValue = (UINT32)((pSdHardWare->rcaValue << 16) | 0x01);

    if (cmdIndex == SDMMC_CMD_STOP_TRANSMISSION)
        {
        SD_CMD_FORM (sdCmd, cmdIndex, argValue, SDMMC_CMD_RSP_R1B, FALSE);
        }
    else if (cmdIndex == SDMMC_CMD_SEND_STATUS)
        {
        SD_CMD_FORM (sdCmd, cmdIndex, argValue, SDMMC_CMD_RSP_R1, FALSE);
        }

    if (sdCmdIssue(pDev, &sdCmd) == ERROR)
        return (ERROR);

    if (pStatus != NULL)
        *pStatus = sdCmd.cmdRsp[0];
    return (OK);
    }

/*******************************************************************************
*
* sdCmdReadSingleBlock - issue CMD17 to read a single block
*
* This routine issues CMD17 to read a single block.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdReadSingleBlock
    (
    VXB_DEV_HANDLE   pDev,
    void *          buf,
    UINT64          blkNo,
    BOOL            highCapa
    )
    {
    return sdCmdReadBlock(pDev, buf, blkNo, 1, highCapa);
    }

#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
/*******************************************************************************
*
* sdCmdReadBlock - issue CMD17 to read blocks
*
* This routine issues CMD17 to read blocks.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdReadBlock
    (
    VXB_DEV_HANDLE   pDev,
    void *          buf,
    UINT64          blkNo,
    UINT32          numBlks,
    BOOL            highCapa
    )
    {
    SD_CMD cmd;
    UINT32 idx;
    UINT64 arg;
    SD_HARDWARE * pSdHardWare;
    UINT32 i;
    STATUS rc;
    UINT8* pBuf = (UINT8*) buf;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdReadBlock() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    idx = SDMMC_CMD_READ_SINGLE_BLOCK;

    if (highCapa)
        arg = blkNo;
    else
        arg = blkNo * pSdHardWare->blockSize;

    SD_LIB_DBG (SD_LIB_DBG_CMD,
                "sdCmdReadBlock() - arg = 0x%x and idx = %d\n",
                (UINT32) arg, idx, 0, 0, 0, 0);

    for (i = 0, pBuf = buf; i < numBlks; i++, pBuf += pSdHardWare->blockSize)
        {
        SD_CMD_FORM (cmd, idx, (arg+i), SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, TRUE);
        SD_DATA_FORM (cmd, pBuf, 1, pSdHardWare->blockSize, TRUE);
        rc = sdCmdIssue(pDev, &cmd);
        if (rc != OK)
            {
            break;
            }
        }
    return rc;
    }
#else
#error _WRS_CONFIG_ITL_SOC_AGILEX5 not defined
#endif

/*******************************************************************************
*
* sdCmdWriteSingleBlock - issue CMD24 to write a single block
*
* This routine issues CMD24 to write a single block.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdWriteSingleBlock
    (
    VXB_DEV_HANDLE   pDev,
    void *          buf,
    UINT64          blkNo,
    BOOL            isHighCapa
    )
    {
    return sdCmdWriteBlock(pDev, buf, blkNo, 1, isHighCapa);
    }


/*******************************************************************************
*
* sdCmdWriteMultipleBlock - issue CMD25 to write multiple blocks
*
* This routine issues CMD25 to writ multiple blocks.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdWriteMultipleBlock
    (
    VXB_DEV_HANDLE  pDev,
    void *          buf,
    UINT64          blkNo,
    UINT32          numBlks,
    BOOL            isHighCapa
    )
    {
    return sdCmdWriteBlock(pDev, buf, blkNo, numBlks, isHighCapa);
    }

#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
/*******************************************************************************
*
* sdCmdWriteBlock - issue CMD24 or CMD25 to write blocks
*
* This routine issues CMD24 or CMD25 to write blocks.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdWriteBlock
    (
    VXB_DEV_HANDLE  pDev,
    void *          buf,
    UINT64          blkNo,
    UINT32          numBlks,
    BOOL            isHighCapa
    )
    {
    SD_CMD cmd;
    UINT32 idx;
    UINT64 arg;
    SD_HARDWARE * pSdHardWare;
    UINT32 i;
    STATUS rc;
    UINT8* pBuf = (UINT8*) buf;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdWriteBlock() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    idx = SDMMC_CMD_WRITE_SINGLE_BLOCK;

    if (isHighCapa)
        arg = blkNo;
    else
        arg = blkNo * pSdHardWare->blockSize;

    SD_LIB_DBG (SD_LIB_DBG_CMD,
                "sdCmdWriteBlock() - arg = 0x%x and idx = %d\n",
                (UINT32) arg, idx, 0, 0, 0, 0);

    for (i = 0, pBuf = buf; i < numBlks; i++, pBuf += pSdHardWare->blockSize)
        {
        SD_CMD_FORM (cmd, idx, (arg+i), SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, TRUE);
        SD_DATA_FORM (cmd, pBuf, 1, pSdHardWare->blockSize, FALSE);
        rc = sdCmdIssue(pDev, &cmd);
        if (rc != OK)
            {
            break;
            }
        }
    return rc;
    }
#else
#error _WRS_CONFIG_ITL_SOC_AGILEX5 not defined
#endif

/*******************************************************************************
*
* sdCmdAppCmd - issue CMD55 to indicate an application command followed
*
* This routine issues CMD55 to indicate an application command followed.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdAppCmd
    (
    VXB_DEV_HANDLE  pDev
    )
    {
    SD_CMD         cmd;
    SD_HARDWARE *  pSdHardWare;
    STATUS         ret;
    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdAppCmd() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    SD_CMD_FORM (cmd, SDMMC_CMD_APP_CMD, SDMMC_CMD_ARG_RCA (pSdHardWare->rcaValue),
                 SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, FALSE);

    ret = sdCmdIssue(pDev, &cmd);
    if (cmd.cmdErr & SDMMC_CMD_ERR_TIMEOUT)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdAppCmd() - cmd timeout error\n",
                    0, 0, 0, 0, 0, 0);
        pSdHardWare->hasException = TRUE;
        }
     else
        {
        pSdHardWare->hasException = FALSE;
        }
    return (ret);
    }

/*******************************************************************************
*
* sdCmdSpiReadOcr - issue CMD58 to read the OCR register of a card
*
* This routine issues CMD58 to read the OCR register of a card. CCS bit is
* assigned to OCR[30]
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSpiReadOcr
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_CMD cmd;
    STATUS rc;
    UINT32 arg;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdCmdSpiReadOcr() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }
    arg = SD_SPI_CMD58_ARG_READ_OCR;
    SD_CMD_FORM (cmd, SDMMC_CMD_SPI_READ_OCR, arg, SDMMC_SPI_RSP_R3, FALSE);

    SD_LIB_DBG (SD_LIB_DBG_CMD, "sdCmdSpiReadOcr command\n", 0, 0, 0, 0, 0, 0);

    rc = sdCmdIssue(pDev, &cmd);
    if (rc == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSpiReadOcr() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return (ERROR);
        }
    pSdHardWare->ocrValue = cmd.cmdRsp[1];
    return (rc);
    }

/*******************************************************************************
*
* sdCmdSpiSetCrc - issue CMD59 to turn the CRC option on or off.
*
* This routine issues CMD59 to turn the CRC option on or off. A "1" in the CRC
* option bit will turn the option on, a "0" will turn it off.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdSpiSetCrc
    (
    VXB_DEV_HANDLE pDev,
    BOOL           usrSpiCrc
    )
    {
    SD_CMD cmd;
    STATUS rc;
    UINT32 arg =0;

    /* [0:0] CRC option */
    if (usrSpiCrc)
        arg = SD_SPI_CMD59_ARG_SET_CRC;
    SD_CMD_FORM (cmd, SDMMC_CMD_SPI_SET_CRC, arg, SDMMC_SPI_RSP_R1, FALSE);
    SD_LIB_DBG (SD_LIB_DBG_CMD, "sdCmdSpiSetCrc command\n", 0, 0, 0, 0, 0, 0);

    rc = sdCmdIssue(pDev, &cmd);
    if (rc == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdSpiSetCrc() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return (ERROR);
        }
    return (rc);
    }

/*******************************************************************************
*
* sdACmdSetBusWidth - issue ACMD6 to set the bus width for data transfer
*
* This routine issues ACMD6 to set the bus width for data transfer.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdACmdSetBusWidth
    (
    VXB_DEV_HANDLE   pDev,
    UINT32          width
    )
    {
    SD_CMD cmd;
    UINT32 arg;

    if (sdCmdAppCmd (pDev) == ERROR)
        return ERROR;

    if (width == SDMMC_BUS_WIDTH_1BIT)
        arg = SD_ACMD6_ARG_1BIT_BUS;
    else
        arg = SD_ACMD6_ARG_4BIT_BUS;

    SD_CMD_FORM (cmd, SD_ACMD_SET_BUS_WIDTH, arg, SDMMC_CMD_RSP_R1, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* sdACmdSendOpCond - issue ACMD41 to ask card to send its OCR
*
* This routine issues ACMD41 to ask card to send its OCR.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdACmdSendOpCond
    (
    VXB_DEV_HANDLE  pDev,
    UINT32 capbility,
    UINT32 version,
    UINT32 reInit
    )
    {
    SD_CMD cmd;
    UINT32 arg;
    SD_HARDWARE * pSdHardWare;
    VXB_DEV_HANDLE pParent;
    SD_HOST_SPEC * pSdHostSpec;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdACmdSendOpCond() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }
    pParent = pSdHardWare->pHostDev;

    if (sdCmdAppCmd (pDev) == ERROR)
        return (ERROR);

    arg = capbility & OCR_VDD_VOL_MASK;
    if (version == SD_VERSION_200)
        {
        if (capbility & OCR_CARD_CAP_STS)
            arg |= OCR_CARD_CAP_STS;
        }

   /*
    * Set ACMD41 argument to 3.3V if both 3.0V & 3.3V are not reported
    * by host controller's capabilities register
    */

    if (!(arg & (OCR_VDD_VOL_29_30 | OCR_VDD_VOL_30_31 |
                 OCR_VDD_VOL_32_33 | OCR_VDD_VOL_33_34)))
        arg |= (OCR_VDD_VOL_32_33 | OCR_VDD_VOL_33_34);

    pSdHostSpec = (SD_HOST_SPEC *)(pSdHardWare->pHostSpec);
    if (pSdHostSpec == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdACmdSendOpCond - pSdHostSpec == NULL\n",
                    0, 0, 0, 0, 0, 0);
        }
    else if ((pSdHostSpec->vxbSdCheckUhsSupport) &&
             (pSdHostSpec->vxbSdCheckUhsSupport(pParent)))
        {

        /* if host support UHS mode, we start UHS initialization */
        if (reInit == 0)
            {
            arg |= (OCR_VDD_VOL_S18R);
            SD_LIB_DBG (SD_LIB_DBG_BUSINIT,
                    "sdACmdSendOpCond - start UHS initialization.\n",
                    0, 0, 0, 0, 0, 0);
            }
        else
            {
            SD_LIB_DBG (SD_LIB_DBG_BUSINIT,
                    "sdACmdSendOpCond - UHS init failed, try non-UHS speed.\n",
                    0, 0, 0, 0, 0, 0);
            }
        }
    else
        {
        SD_LIB_DBG (SD_LIB_DBG_BUSINIT,
                    "sdACmdSendOpCond - start UHS initialization failed\n"
                    "    vxbSdCheckUhsSupport undefined or return FALSE\n",
                    0, 0, 0, 0, 0, 0);
        }
    if ((pSdHostSpec != NULL)&&(SDMMC_HOST_IS_SPI(pSdHostSpec)))
        {

        /* ACMD41 only send HCS bit, reserved bit shall be '0' in SPI mode */

        arg = SD_SPI_ACMD41_ARG_SEND_HCS;
        }
    SD_CMD_FORM (cmd, SD_ACMD_SEND_OP_COND, arg, SDMMC_CMD_RSP_R3 | SDMMC_SPI_RSP_R1, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        return (ERROR);

    pSdHardWare->ocrValue = cmd.cmdRsp[0];

    SD_LIB_DBG (SD_LIB_DBG_BUSINIT,
                 "sdACmdSendOpCond() - ACMD41 OCR = 0x%x\n",
                 cmd.cmdRsp[0], 0, 0, 0, 0, 0);

    return (OK);
    }

/*******************************************************************************
*
* sdACmdClrCardDetect - issue ACMD42 to clear card detect
*
* This routine issues ACMD42 to clear card detect.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdACmdClrCardDetect
    (
    VXB_DEV_HANDLE   pDev,
    UINT32          connect
    )
    {
    SD_CMD cmd;
    UINT32 arg;

    if (sdCmdAppCmd (pDev) == ERROR)
        return ERROR;

    arg = connect;
    cmd.cmdIdx = 0;
    SD_CMD_FORM (cmd, SD_ACMD_SET_CLR_CARD_DETECT, arg, SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, FALSE);

    return (sdCmdIssue(pDev, &cmd));
    }

/*******************************************************************************
*
* sdACmdSendWriteBlkNum - issue ACMD22 to ask card to send the num of write block
*
* This routine issues ACMD22 to ask card to send the number of the written
* (without errors) write blocks.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdACmdSendWriteBlkNum
    (
    VXB_DEV_HANDLE  pDev,
    UINT32 * wrBlkNum
    )
    {
    SD_CMD cmd;
    UINT32 * numWr;
    STATUS rc;

    numWr = (UINT32 *)cacheDmaMalloc (sizeof(UINT32));
    if (numWr == NULL)
        return ERROR;

    if (sdCmdAppCmd (pDev) == ERROR)
        {
        (void)cacheDmaFree((void *)numWr);
        return ERROR;
        }

    SD_CMD_FORM (cmd, SD_ACMD_SEND_NUM_WR_BLOCKS, 0, SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, TRUE);
    SD_DATA_FORM (cmd, numWr, 1, sizeof(UINT32), TRUE);

    rc = sdCmdIssue(pDev, &cmd);
    if (rc == OK)
        {
        bcopy ((char *)numWr, (char *)(wrBlkNum), sizeof(SDMMC_SCR));
        }

    (void)cacheDmaFree((void *)numWr);
    return (rc);
    }

/*******************************************************************************
*
* sdACmdSendScr - issue ACMD51 to ask card to send its SCR
*
* This routine issues ACMD51 to ask card to send its SCR.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdACmdSendScr
    (
    VXB_DEV_HANDLE  pDev,
    SDMMC_SCR * scrData
    )
    {
    SD_CMD cmd;
    SDMMC_SCR * scr;
    STATUS rc = ERROR;
    UINT32 i;

    scr = (SDMMC_SCR *)cacheDmaMalloc (sizeof(SDMMC_SCR));
    if (scr == NULL)
        return ERROR;

    for (i = 0; i < SDMMC_CMD_RETRY; i++)
        {
        if (sdCmdAppCmd (pDev) == ERROR)
            {
            vxbMsDelay(1);
            continue;
            }
        SD_CMD_FORM (cmd, SD_ACMD_SEND_SCR, 0, SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, TRUE);
        SD_DATA_FORM (cmd, scr, 1, sizeof(SDMMC_SCR), TRUE);

        rc = sdCmdIssue(pDev, &cmd);
        if (rc == ERROR)
            {
            vxbMsDelay(1);
            continue;
            }
        else
            break;
        }

    if (rc != ERROR)
        {
        bcopy ((char *)scr, (char *)(scrData), sizeof(SDMMC_SCR));
        }

    (void)cacheDmaFree((void *)scr);
    return (rc);
    }

/*******************************************************************************
*
* sdACmdSendSsr - issue ACMD13 to ask card to send its SSR
*
* This routine issues ACMD13 to ask card to send its SSR.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdACmdSendSsr
    (
    VXB_DEV_HANDLE  pDev,
    UINT8 * ssrData
    )
    {
    SD_CMD cmd;
    UINT8 * ssr;
    STATUS rc;

    ssr = (UINT8 *)cacheDmaMalloc (sizeof(UINT8) * SD_SSR_SIZE);
    if (ssr == NULL)
        return ERROR;

    if (sdCmdAppCmd (pDev) == ERROR)
        {
        (void)cacheDmaFree((void *)ssr);
        return ERROR;
        }

    SD_CMD_FORM (cmd, SD_ACMD_SD_STATUS, 0, SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R2, TRUE);
    SD_DATA_FORM (cmd, ssr, 1, sizeof(UINT8) * SD_SSR_SIZE, TRUE);

    rc = sdCmdIssue(pDev, &cmd);
    if (rc == OK)
        {
        bcopy ((char *)ssr, (char *)(ssrData), sizeof(UINT8) * SD_SSR_SIZE);
        }

    (void)cacheDmaFree((void *)ssr);
    return (rc);
    }

/*******************************************************************************
*
* sdioCmdIoRwDirect - issue CMD52 to read/write sdio card's register.
*
* This routine issues CMD52 to read/write sdio card's register. CMD52 must be
* input register offset, function number, rwOps. rwOps parameter include two
* part, R/W flag and RAW flag. R/W flag is defined at bit31 and RAW flag is
* defined at bit27. These flags's define is same with CMD52 description.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdioCmdIoRwDirect
    (
    VXB_DEV_HANDLE pDev,
    UINT32 regOffset,
    UINT32 funcNum,
    UINT32 rwOps,
    UINT8 * data
    )
    {
    SD_CMD cmd;
    UINT32 arg = 0;
    UINT8  cmdSts = 0;

    arg |= rwOps;
    arg |= SDIO_DIRECT_IO_REGADRS(regOffset);
    arg |= SDIO_DIRECT_IO_FUNC(funcNum);
    arg |= *data;

    SD_CMD_FORM (cmd, SDIO_CMD_DIRECT_IO, arg, SDMMC_CMD_RSP_R5, FALSE);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdioCmdIoRwDirect() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }
    cmdSts = (UINT8)((cmd.cmdRsp[0] >> 8) & 0xFF);
    if ((cmdSts &
        (SDIO_DIRECTIO_OUTRANGE | SDIO_DIRECTIO_ERR | SDIO_DIRECTIO_FUNCERR))
        != 0x0)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdioCmdIoRwDirect() - cmdSts = 0X%x return error\n",
                    cmdSts, 0, 0, 0, 0, 0);
        return ERROR;
        }
    if ((rwOps & (UINT32)SDIO_DIRECT_IO_RW) == 0x0)
        {
        *data = (UINT8)(cmd.cmdRsp[0] & 0xFF);
        SD_LIB_DBG (SD_LIB_DBG_CMD, "sdioCmdIoRwDirect() - REG DATA = 0X%x\n",
                    cmd.cmdRsp[0] & 0xFF, 0, 0, 0, 0, 0);
        }
    return (OK);
    }

/*******************************************************************************
*
* sdioCmdIoRwExtend - issue CMD53 to read/write sdio card's data.
*
* This routine issues CMD53 to read/write sdio card's data. CMD53 must be
* input data offset, function number, rwOps. rwOps parameter include two
* part, R/W flag and RAW flag. R/W flag is defined at bit31 and RAW flag is
* defined at bit27. These flags's define is same with CMD53 description.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdioCmdIoRwExtend
    (
    VXB_DEV_HANDLE pDev,
    UINT32 regOffset,
    UINT32 funcNum,
    UINT32 ops,
    SDDATA * pSdData
    )
    {
    SD_CMD cmd;
    UINT32 arg = 0;
    UINT8  cmdSts = 0;
    BOOL   isRead;

    arg |= ops;
    arg |= SDIO_DIRECT_IO_REGADRS(regOffset);
    arg |= SDIO_DIRECT_IO_FUNC(funcNum);

    if ((ops & SDIO_DIRECT_IOEXT_BLK) == 0x0)
        arg |= (pSdData->blkSize == 512) ? 0 : pSdData->blkSize;    /* byte mode */
    else
        arg |= SDIO_DIRECT_IOEXT_BLK | pSdData->blkNum;     /* block mode */

    if ((ops & (UINT32)SDIO_DIRECT_IO_RW) == 0x0)
        isRead = TRUE;
    else
        isRead = FALSE;

    SD_CMD_FORM (cmd, SDIO_CMD_DIRECT_EXT, arg, SDMMC_CMD_RSP_R5, TRUE);
    SD_DATA_FORM (cmd, pSdData->buffer, pSdData->blkNum, pSdData->blkSize, isRead);

    if (sdCmdIssue(pDev, &cmd) == ERROR)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR,
                    "sdioCmdIoRwExtend() - cmdIssue return error\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }
    cmdSts = (UINT8)((cmd.cmdRsp[0] >> 8) & 0xFF);
    if ((cmdSts &
        (SDIO_DIRECTIO_OUTRANGE | SDIO_DIRECTIO_ERR | SDIO_DIRECTIO_FUNCERR))
        != 0x0)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdioCmdIoRwExtend() - cmdSts = 0X%x return error\n",
                    cmdSts, 0, 0, 0, 0, 0);
        return ERROR;
        }

    return (OK);
    }

/*******************************************************************************
*
* sdErrCmdGet - Get the error command.
*
* This routine gets the command with an error when issuing multiple commands.
*
* RETURNS: the pointer to the error command or the last command if no errors
* occer.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

SD_CMD * sdErrCmdGet(SD_CMD * pSdCmdList)
    {
    if (pSdCmdList == NULL)
        {
        return NULL;
        }

    while ((pSdCmdList->cmdErr == 0) && (pSdCmdList->dataErr == 0) &&
           (SDMMC_CMD_IS_NOT_LAST_CMD(pSdCmdList->cmdFlag)))
        {
        pSdCmdList ++;
        }

    return pSdCmdList;
    }

/*******************************************************************************
*
* sdCmdIssue - the feature to issue one new SD/SDIO command
*
* This routine issues one new command.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdCmdIssue
    (
    VXB_DEV_HANDLE pDev,
    SD_CMD * pSdCmd
    )
    {
    SD_HARDWARE * pSdHardWare;
    STATUS rc;
    VXB_DEV_HANDLE pParent;
    SD_HOST_SPEC  * pSdHostSpec;
    SD_CMD * pSdErrCmd;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdIssue() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }
    if (pSdCmd == NULL)
        return ERROR;

    pParent = pSdHardWare->pHostDev;

    pSdHardWare->dataErr = 0x0;
    pSdHardWare->cmdErr = 0x0;

    rc = pSdHardWare->vxbSdCmdIssue(pParent, pSdCmd);

    pSdErrCmd = sdErrCmdGet(pSdCmd);

    pSdHardWare->cmdErr = pSdErrCmd->cmdErr;
    pSdHardWare->dataErr = pSdErrCmd->dataErr;

    if ((pSdHardWare->recover) && (rc != OK))
        {
        pSdHostSpec = (SD_HOST_SPEC *)(pSdHardWare->pHostSpec);
        if (pSdHostSpec == NULL)
            return ERROR;

        /* issue abort command for error recovery */

        if ((pSdCmd->hasData) && (pSdHostSpec->vxbSdAbortIssue))
            {
            SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdIssue() - Issue abort command\n",
                        0, 0, 0, 0, 0, 0);
            rc = pSdHostSpec->vxbSdAbortIssue(pParent);
            }

        /* try the original command again */

        pSdCmd->cmdErr = 0;
        pSdCmd->dataErr = 0;
        rc |= pSdHardWare->vxbSdCmdIssue(pParent, pSdCmd);

        pSdErrCmd = sdErrCmdGet(pSdCmd);

        pSdHardWare->cmdErr = pSdErrCmd->cmdErr;
        pSdHardWare->dataErr = pSdErrCmd->dataErr;

        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdIssue() - recover result=%d, cmdErr =%d, dataErr=%d\n",
                    rc, pSdErrCmd->cmdErr, pSdErrCmd->dataErr, 0, 0, 0);
        }

    return (rc);
    }

/*******************************************************************************
*
* vxbSdioReadCis - get all of the CIS value from one SDIO target device
*
* This routine get all of the CIS value from one SDIO target device
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS vxbSdioReadCis
    (
    VXB_DEV_HANDLE pDev
    )
    {
    STATUS ret;
    SDIO_FUNC_TUPLE * pCurr;
    SDIO_FUNC_TUPLE * pNext;
    SD_HARDWARE * pSdHardWare;
    UINT8 cisData = 0;
    UINT32 i;
    UINT32 ptr = 0;
    UINT8 tplCode = 0;
    UINT8  tplLink = 0;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "vxbSdioReadCis() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    for (i = 0; i < 3; i++)
        {
        ret = sdioCmdIoRwDirect(pDev,
                                (UINT32)(SDIO_FBR_BASE(pSdHardWare->funcNum) +
                                SDIO_FBR_CIS + i),
                                0, 0, &cisData);
        if (ret == ERROR)
            return (ERROR);

        ptr |= (UINT32)(cisData << (i * 8));
        }

    do {
        ret = sdioCmdIoRwDirect(pDev, ptr++, 0, 0, &tplCode);
        if (ret == ERROR)
            break;

        /* size of 0xFF means we're done */

        if (tplCode == 0xFF)
            break;

        /* 0x00 means no link field or data */

        if (tplCode == 0x00)
            continue;

        ret = sdioCmdIoRwDirect(pDev, ptr++, 0, 0, &tplLink);
        if (ret == ERROR)
            break;

        /* size of 0xFF means we're done */

        if (tplLink == 0xFF)
            break;

        pNext = (SDIO_FUNC_TUPLE *)malloc(sizeof(SDIO_FUNC_TUPLE));
        if (pNext == NULL)
            return ERROR;

        pCurr = pNext;

        if (pSdHardWare->pSdioFuncTuple == NULL)
            {
            pSdHardWare->pSdioFuncTuple = pNext;
            pCurr = pNext;
            }
        else if (pCurr != NULL)
            pCurr->next = pNext;

        pSdHardWare->pSdioFuncTuple = pCurr;

        if (tplLink != 0x0)
            {
            pSdHardWare->pSdioFuncTuple->pData = (UINT8 *)malloc(sizeof(UINT8) * tplLink);
            if (pSdHardWare->pSdioFuncTuple->pData == NULL)
                {
                free (pSdHardWare->pSdioFuncTuple);
                return ERROR;
                }
            }

        for (i = 0; i < tplLink; i++)
            {
            ret = sdioCmdIoRwDirect(pDev, ptr + i, 0, 0, (pCurr->pData + i));
            if (ret == ERROR)
                break;
            }
        if (ret == ERROR)
            {
            free(pCurr->pData);
            free(pCurr);
            break;
            }
        if (tplCode == SDIO_CISTPL_MANFID)
            {
            UINT16 * pData;
            pData = (UINT16 *)pCurr->pData;
            pSdHardWare->vxbSdioId.sdioDevId = *(pData);
            pData++;
            pSdHardWare->vxbSdioId.sdioVenderId = *(pData);
            }
        ptr += tplLink;
        } while (ret != ERROR);

    return ret;
    }

/*******************************************************************************
*
* vxbSdioSetFunc - set one SDIO function enable/disable
*
* This routine set one SDIO function enable/disable
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS vxbSdioSetFunc
    (
    VXB_DEV_HANDLE pDev,
    UINT8 func,
    BOOL  isEn
    )
    {
    STATUS ret;
    UINT8 ioRegData = 0;
    UINT32 timeOut = 0;
    SD_HARDWARE * pSdHardWare;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdioCmdIoSendOpCond() - pSdHardWare is NULL\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    ret = sdioCmdIoRwDirect(pDev, SDIO_IOEN_CCCR, 0, 0, &ioRegData);
    if (ret == ERROR)
        return (ERROR);
    ioRegData = (UINT8)(ioRegData | (1 << func));

    ret = sdioCmdIoRwDirect(pDev, SDIO_IOEN_CCCR, 0, (UINT32)SDIO_DIRECT_IO_RW, &ioRegData);
    if (ret == ERROR)
        return (ERROR);

    do
        {
        ret = sdioCmdIoRwDirect(pDev, SDIO_IOEN_CCCR, 0, 0, &ioRegData);
        if (ret == ERROR)
            return (ERROR);
        timeOut++;
        }while (((ioRegData & (1 << func)) == 0x0) && (timeOut < SDIO_OPREG_TIMEOUT));
    if (timeOut >= SDIO_OPREG_TIMEOUT)
        return (ERROR);
    return (OK);
    }

/*******************************************************************************
*
* sdPowerCycle - power cycle a SD card
*
* This routine performs a power cycle for a SD card.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*
*/

STATUS sdPowerCycle
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SD_HARDWARE   * pSdHardWare;
    SD_HOST_SPEC  * pSdHostSpec;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        return ERROR;

    pSdHostSpec = (SD_HOST_SPEC *)(pSdHardWare->pHostSpec);
    if (pSdHostSpec == NULL)
        return ERROR;

    if (pSdHostSpec->vxbSdPowerCycle)
        return pSdHostSpec->vxbSdPowerCycle(pSdHardWare->pHostDev);
    return ERROR;
    }

/*******************************************************************************
*
* sdmmcBlockNumGet - get the number of supported blocks for the request
*
* This routine gets the number of supported blocks for the request. The API is
* for controllers with HW issues for multiple I/O.
*
* RETURNS: Returns the number of supported blocks for the request.
*
* ERRNO: N/A
*
* \NOMANUAL
*
*/

UINT32 sdmmcBlockNumGet
    (
    VXB_DEV_HANDLE pDev,
    BOOL           isMmc,
    BOOL           isRead,
    UINT32         blkNum
    )
    {
    SD_HARDWARE   * pSdHardWare;
    SD_HOST_SPEC  * pSdHostSpec;

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO(pDev);
    if (pSdHardWare == NULL)
        return blkNum;

    pSdHostSpec = (SD_HOST_SPEC *)(pSdHardWare->pHostSpec);
    if (pSdHostSpec == NULL)
        return blkNum;

    if (pSdHostSpec->vxbMultiIoBlkNumGet)
        {
        return pSdHostSpec->vxbMultiIoBlkNumGet(pSdHardWare->pHostDev,
                                                isMmc, isRead, blkNum);
        }
    else
        {
        if(blkNum <= pSdHostSpec->maxBlocksPerRequest)
            {
            return blkNum;
            }
        else
            {
            return pSdHostSpec->maxBlocksPerRequest;
            }
        }
    }

/*******************************************************************************
*
* sdCmdGenCmd - issue CMD56 to get general infomation of SD card
*
* This routine issues CMD56 to get general information of the SD card. <pDev> is
* the handle of the SD card device. The implementation of CMD56 on different SD
* cards may be different. For CMD56, some SD cards only need one command, and
* some SD cards require multiple commands to be executed in sequence. Use the
* parameter <argCnt> to specify the number of transmission commands, and use the
* parameter <arg> to specify the parameters required for each transmission.
* According to SD specifications, BIT0 of argument is used to identify whether
* CMD56 is read or write operation this time:
*     "1" the host shall get a block of data from the card.
*     "0" the host shall send a block of data to the card.
* When CMD56 read operation, the general information of the returned SD card
* will be saved in pGenData. When CMD56 write operation, the sent data should be
* saved in pGenData. The length of pGenData is equal to block size. In case of
* SDHC, SDXC and SDUC Cards, block length is fixed to 512 bytes.
*
* RETURNS: OK or ERROR if any command fails.
*
* ERRNO: N/A
*/

STATUS sdCmdGenCmd
    (
    VXB_DEV_HANDLE pDev,      /* device handle */
    UINT32       * arg,       /* arguments list */
    UINT32         argCnt,    /* the number of arguments */
    UINT8        * pGenData   /* Read-write buffer */
    )

    {
    SD_CMD  * pCmd;
    STATUS    rc;
    UINT32    cmdIndex = 0;

    if ((pDev == NULL) || (arg == NULL) || (pGenData == NULL))
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdGenCmd() - Parameter error.\n",
                    0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    pCmd = (SD_CMD *)vxbMemAlloc(sizeof(SD_CMD) * argCnt);
    if (pCmd == NULL)
        {
        return ERROR;
        }

    for (cmdIndex = 0; cmdIndex < argCnt; cmdIndex++)
        {
        SD_CMD_FORM (pCmd[cmdIndex], SDMMC_CMD_GEN_CMD, arg[cmdIndex],
                     SDMMC_CMD_RSP_R1 | SDMMC_SPI_RSP_R1, TRUE);
        SD_DATA_FORM (pCmd[cmdIndex], pGenData, 1, SDMMC_BLOCK_SIZE, arg[cmdIndex] & 0x1);

        if (cmdIndex < (argCnt - 1))
            {
            SD_CMD_FLAG_FORM(pCmd[cmdIndex], SDMMC_CMD_FLAG_NEXT_CMD);
            }
        }

    rc = sdCmdIssue (pDev, pCmd);
    if (rc != OK)
        {
        SD_LIB_DBG (SD_LIB_DBG_ERR, "sdCmdGenCmd() - Parameter error.\n",
                    0, 0, 0, 0, 0, 0);
        }

    vxbMemFree(pCmd);
    return (rc);
    }

#ifdef  SD_LIB_DBG_ON

/*******************************************************************************
*
* sdMmcBlockReadTest - sdmmc block read test
*
* This routine tests sdmmc block read.
* Example1: sdMmcBlockReadTest "sdStorage",0,0,16,1
* Example2: sdMmcBlockReadTest "sdStorage",0,2048,65535,0
*
* RETURNS: OK or ERROR if any command fails.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdMmcBlockReadTest
    (
    char *          devicName,       /* device name such as "sdStorage" */
    UINT32          unitNum,         /* device unit number */
    unsigned int    beginBlockIndex, /* first block index to read*/
    unsigned int    totalBlockNum,   /* total block number to read*/
    BOOL            printBuf         /* print buffer content flag*/
    )
    {
    unsigned int        startTick;
    unsigned int        endTick;
    unsigned int        totalTick = 0;
    unsigned char *     buf;
    unsigned int        totalSize;
    unsigned int        i;

    VXB_DEV_HANDLE pDev = vxbDevAcquireByName(devicName, unitNum);
    vxbDevRelease(pDev);

    /* size should be 512*N */
    totalSize = totalBlockNum*512;

    buf=malloc((size_t)totalSize);
    if(buf==NULL)
        {
        printf("Not enough memory\n");
        return ERROR;
        }

    startTick = (unsigned int)tickGet();
    sdCmdReadBlock (pDev, (void *)buf, (UINT64)beginBlockIndex,
                                 totalBlockNum, TRUE);
    endTick = (unsigned int)tickGet();

    totalTick = endTick - startTick;
    if(totalTick>0)
        {
        printf("total %d ticks (%d.%03d seconds for %lld bytes), "
                "speed is %lld bytes/sec\n",
               totalTick,
               (totalTick/(UINT32)sysClkRateGet()),
               (totalTick * 1000 /(UINT32)sysClkRateGet() % 1000),
               ((long long)totalSize ),
               ((long long)totalSize ) / totalTick * sysClkRateGet());
        }

    if(printBuf)
        {
        i=0;
        while(i<totalSize)
            {
            if((i%512)==0)
                {
                printf("\nBlock #%d:\n",i/512);
                }
            printf("0x%03x: %02x %02x %02x %02x %02x %02x %02x %02x  "
                    "%02x %02x %02x %02x %02x %02x %02x %02x \n",
                    i,buf[i],buf[i+1],buf[i+2],buf[i+3],buf[i+4],buf[i+5],
                    buf[i+6],buf[i+7],buf[i+8],buf[i+9],buf[i+10],buf[i+11],
                    buf[i+12],buf[i+13],buf[i+14],buf[i+15]);
            i=i+16;

            }
        }

    free (buf);

    return OK;
    }

/*******************************************************************************
*
* sdMmcFileWriteTest - sdmmc file read test
*
* This routine tests sdmmc file write speed.
* Example: sdMmcFileWriteTest "/sd1", 100
*
* RETURNS: OK or ERROR if any command fails.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

int sdMmcFileWriteTest(char *  path, UINT32 Msize)
{
    char        fileName[128];
    int         fd,ret;
    char *      buffer;
    UINT32      startTick;
    UINT32      endTick;
    UINT32      totalTicks;
    UINT32      bufSize;
    int         clkRate;
    float       speed;

    bufSize = Msize*1024*1024;

    buffer = malloc ( bufSize );
    if(buffer==NULL)
        {
        printf("Not enough memory\n");
        return ERROR;
    }
    memset ( buffer, 0xab, bufSize );

    sprintf ( fileName, "%s/test%dM.txt", path, Msize );
    if((fd = open (fileName, O_RDWR | O_CREAT, 0777))==ERROR)
        {
        printf("create file error: %s\n", fileName);
        return ERROR;
        }
    close(fd);
    if ((fd = open (fileName, O_RDWR , 0777)) == ERROR)
        {
        printf("open error: %s\n", fileName);
        return ERROR;
        }
    startTick = tickGet( );

    ret = (int)write ( fd, buffer ,bufSize);
    if ( (UINT32)ret != bufSize)
        {
            printf("write error\n");
            return ERROR;
        }
    endTick = tickGet( );

    totalTicks = endTick - startTick;

    clkRate = (int)sysClkRateGet();
    speed=((float)Msize/(float)totalTicks)*(float)clkRate;
    printf("## write %dM bytes in %d ticks, sysClkRateGet()=%d \n",
            Msize,(endTick-startTick), sysClkRateGet());
    printf("## write speed=%f MB/s\n", speed);

    close(fd);
    free(buffer);

    return OK;

}

/*******************************************************************************
*
* sdMmcFileReadTest - sdmmc file read test
*
* This routine tests sdmmc file read speed.
* Example: sdMmcFileReadTest "/sd1", 100
*
* RETURNS: OK or ERROR if any command fails.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

int sdMmcFileReadTest(char *  path, UINT32 Msize)
{
    char        fileName[128];
    int         fd,ret;
    char *      buffer;
    UINT32      startTick;
    UINT32      endTick;
    UINT32      totalTicks;
    UINT32      bufSize;
    int         clkRate;
    float       speed;

    bufSize = Msize*1024*1024;

    buffer = malloc ( bufSize );
    if(buffer==NULL)
        {
        printf("Not enough memory\n");
        return ERROR;
    }

    memset ( buffer, 0xab, bufSize );

    sprintf ( fileName, "%s/test%dM.txt", path, Msize );

    if ((fd = open( fileName, O_RDONLY, 0777 )) == ERROR)
        {
        printf("open for read error: %s\n", fileName);
        return ERROR;
        }
    startTick = tickGet( );
    startTick = tickGet( );


    ret = (int)read ( fd, buffer ,bufSize);
    if ( (UINT32)ret != bufSize)
        {
            printf("read error\n");
            return ERROR;
        }

    endTick = tickGet( );

    totalTicks = endTick - startTick;

    clkRate = (int)sysClkRateGet();
    speed=((float)Msize/(float)totalTicks)*(float)clkRate;

    printf("## read %dM bytes in %d ticks, sysClkRateGet()=%d \n",
            Msize,(endTick-startTick), sysClkRateGet());
    printf("## read speed=%f MB/s\n", speed);

    close(fd);
    free(buffer);
    return OK;

}
#endif

