/* vxbMcpSdhcCtrl.c - Microchip pfsoc CDNS SDHC host controller driver  */

/*
 * Copyright (c) 2021 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION

This is the vxbus compliant Microchip Secure Digital Host Controller (SDHC)
driver.

The SDHC provides an interface between the Microchip host controller and
SD/MMC memory cards.

To add the driver to the vxWorks image, add the following component to the
kernel configuration.

\cs
vxprj component add DRV_MCP_SDHC_CTRL
\ce

This driver is bound to device tree, and the device tree node must specify
below parameters:

\is

\i <comptible>
This parameter specifies the name of the SDHC controller driver.
It should be one of:
  "cdns,sd4hc": 4-bit SDHCI on Microchip pfsoc device.

\i <reg>
This parameter specifies the register base address and length of this module.

\i <interrupt-parent>
This parameter specifies the offset of interrupt controller.

\i <interrupts>
This parameter specifies the interrupt number of this module.

\i <clocks>
This parameter specifies the clock for this module.

\ie

Optional Properties :

\is

\i <bus-width>
This parameter is used to select maximum bus-width of host controller supported
by eMMC/SD device.

\i <max-frequency>
This parameter is used to specify maximum operating clock frequency.

\i <disable-wp>
When set no physical WP line is present. This property should only be specified
when the controller has a dedicated write-protect detection logic. If a GPIO is
always used for the write-protect detection logic it is sufficient to not
specify wp-gpios property in the absence of a WP line.

\i <no-1-8-v>
When present, denotes that 1.8v card voltage is not supported on this system,
even if the controller claims it is.

\i <cap-sd-highspeed>
This parameter specifies that SD high-speed timing is supported.

\i <card-detect-delay>
This parameter specifies the delay time in milli-seconds before detecting card
after card insert event.

\i <sd-uhs-sdr12>
This parameter specifies that SD UHS SDR12 speed is supported.

\i <sd-uhs-sdr25>
This parameter specifies that SD UHS SDR25 speed is supported.

\i <sd-uhs-sdr50>
This parameter specifies that SD UHS SDR50 speed is supported.

\i <sd-uhs-sdr104>
This parameter specifies that SD UHS SDR104 speed is supported.

\i <cap-mmc-highspeed>
This parameter specifies that MMC high-speed timing is supported.

\i <mmc-ddr-3_3v>
This parameter specifies that eMMC high-speed DDR mode(3.3V I/O) is supported.

\i <non-removable>
This parameter specifies that non-removable slot (like eMMC); assume always present.

\i <no-sd>
This parameter specifies if or not send SD special commands (such as CMD8) during
initialization. By default, CMD8 is sent to device. But to some SDHC, if it is
connected with (e)MMC, CMD8 may cause not to work. In this case, need set it in
dts.

\i <no-sdio>
This parameter specifies if or not send SDIO special commands during initialization.

\i <voltage-ranges>
This parameter specifies the voltage range in case there are software
transparent level shifters on the outputs of the controller. Two cells are
required, first cell specifies minimum slot voltage (mV), second cell specifies
maximum slot voltage (mV). Several ranges could be specified.
(Not currently in use)

\ie

An example of device node is show below:
\cs
    emmc: sdhc@20008000
        {
        compatible = "cdns,sd4hc";
        reg = <0x0 0x20008000 0x0 0x1000>;
        interrupt-parent = <&L1>;
        interrupts = <88 89>;
        pinctrl-names = "default";
        clocks = <&clkcfg 6>;
        bus-width = <4>;
        cap-mmc-highspeed;
        mmc-ddr-3_3v;
        max-frequency = <200000000>;
        non-removable;
        no-sd;
        no-sdio;
        voltage-ranges = <3300 3300>;
        status = "okay";
        };

    sdcard: sd@20008000
        {
        compatible = "cdns,sd4hc";
        reg = <0x0 0x20008000 0x0 0x1000>;
        interrupt-parent = <&L1>;
        interrupts = <88>;
        pinctrl-names = "default";
        clocks = <&clkcfg 6>;
        bus-width = <4>;
        disable-wp;
        no-1-8-v;
        cap-mmc-highspeed;
        cap-sd-highspeed;
        card-detect-delay = <200>;
        sd-uhs-sdr12;
        sd-uhs-sdr25;
        sd-uhs-sdr50;
        sd-uhs-sdr104;
        max-frequency = <200000000>;
        status = "okay";
        };
\ce

*/

/* includes */

#include <vxWorks.h>
#include <stdio.h>
#include <string.h>
#include <semLib.h>
#include <sysLib.h>
#include <taskLib.h>
#include <cacheLib.h>
#include <intLib.h>
#include <errno.h>
#include <hwif/vxBus.h>
#include <hwif/buslib/vxbFdtLib.h>
#include <subsys/clk/vxbClkLib.h>
#include <subsys/timer/vxbTimerLib.h>
#include <subsys/int/vxbIntLib.h>
#include <vxbSdMmcLib.h>
#include <vxbSdhcCtrl.h>

/* defines */

#ifdef  MCP_SDHC_DBG_ON

#undef  LOCAL
#define LOCAL

#define MCP_SDHC_DBG_ERR            0x00000001
#define MCP_SDHC_DBG_INIT           0x00000002
#define MCP_SDHC_DBG_ALL            0xffffffff
#define MCP_SDHC_DBG_OFF            0x00000000

UINT32 mcpSdhcDbgMask = MCP_SDHC_DBG_OFF;

#define MCP_SDHC_DBG(mask, ...)                                         \
    do                                                                  \
        {                                                               \
        if ((mcpSdhcDbgMask & mask) || (mask == MCP_SDHC_DBG_ALL))      \
            {                                                           \
            if (_func_kprintf != NULL)                                  \
                {                                                       \
                (* _func_kprintf)(__VA_ARGS__);                         \
                }                                                       \
            }                                                           \
        }                                                               \
    while ((FALSE))
#else
#define MCP_SDHC_DBG(...)

#endif  /* MCP_SDHC_DBG_ON */

#define MCP_SDHC_TUNING_TIMES           5
#define MCP_SDHC_TUNING_MIN             0
#define MCP_SDHC_TUNING_MAX             40
#define MCP_SDHC_TUNING_STEP            1

/* MCP host SRS offset */

#define MCP_SRS_OFFSET                  0x200

/* MCP host specific version register bits */

#define MCP_SDHC_SVN_SHIFT              16
#define MCP_SDHC_SVN_MASK               (0xff << MCP_SDHC_SVN_SHIFT)

#define MCP_SDHC_SVN_10                 0      /* Specification Version 1.00 */
#define MCP_SDHC_SVN_20                 1      /* Specification Version 2.00 */
#define MCP_SDHC_SVN_30                 2      /* Specification Version 3.00 */
#define MCP_SDHC_SVN_40                 3      /* Specification Version 4.00 */

/* MCP SDHC general information register */

#define MCP_HRS_GIR                     0x0

/* MCP SDHC general information register bits */

#define MCP_HRS_GIR_SWR_SHIFT           0
#define MCP_HRS_GIR_SWR_MASK            (0x1 << MCP_HRS_GIR_SWR_SHIFT)

/* MCP SDHC debounce setting register */

#define MCP_HRS_DBSR                    0x4

/* MCP SDHC eMMC control register */

#define MCP_HRS_EMMC_CTRLR              0x18

/* MCP SDHC eMMC control registerr bits */

#define MCP_HRS_EMMC_CTRLR_EVT_SHIFT   8
#define MCP_HRS_EMMC_CTRLR_EVT_MASK    (0x3f << MCP_HRS_EMMC_CTRLR_EVT_SHIFT)

#define MCP_HRS_EMMC_CTRLR_ETR_SHIFT   15
#define MCP_HRS_EMMC_CTRLR_ETR_MASK    (0x1 << MCP_HRS_EMMC_CTRLR_ETR_SHIFT)

/* MCP SDHC eMMC Mode */

#define MCP_TIMING_SD                   0x0

#define MCP_TIMING_MMC_SDR              0x2
#define MCP_TIMING_MMC_DDR              0x3
#define MCP_TIMING_MMC_HS200            0x4
#define MCP_TIMING_MMC_HS400            0x5
#define MCP_TIMING_MMC_HS400ES          0x6
#define MCP_TIMING_MMC_LEGACY           0x7

#define HAVE_IOMUX                      (1U << 0)
#define DRV_8BIT                        (1U << 1)

#define BOARD_MICROCHIP_PFSOC           (1)

/* typedefs */

typedef struct mcpSdhcCfg
    {
    UINT32    uBoardType;
    UINT32    uSdhcFlag;
    UINT32    uCtrlFlag;
    } MCP_SDHC_CFG;

/* locals */

LOCAL STATUS mcpSdhcCtrlProbe (VXB_DEV_ID pDev);
LOCAL STATUS mcpSdhcCtrlAttach (VXB_DEV_ID pDev);
LOCAL STATUS mcpSdhcSpecInfoGet (VXB_DEV_ID pDev,
                                 void ** pHostSpec,
                                 VXB_SD_CMDISSUE_FUNC * pCmdIssue);
LOCAL void mcpSdhcCtrlMonitor (VXB_DEV_ID pDev);
LOCAL STATUS mcpSdhcTimingSetup (VXB_DEV_ID pDev, UINT32 uhsSpeed);
LOCAL STATUS mcpSdhcCtrlInit (VXB_DEV_ID pDev);
#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
LOCAL STATUS mcpSdhcPhyInit(VXB_DEV_ID pDev);
#else
LOCAL STATUS mcpSdhcTuningSet (VXB_DEV_ID pDev);
#endif


LOCAL VXB_DRV_METHOD vxbMcpSdhcCtrlMethods[] =
    {
    {VXB_DEVMETHOD_CALL(vxbDevProbe), mcpSdhcCtrlProbe},
    {VXB_DEVMETHOD_CALL(vxbDevAttach), mcpSdhcCtrlAttach},
    {VXB_DEVMETHOD_CALL(vxbSdspecinfoGet), mcpSdhcSpecInfoGet},
    VXB_DEVMETHOD_END
    };

LOCAL const MCP_SDHC_CFG  mcpSdhcCfg =
    {
    BOARD_MICROCHIP_PFSOC,
    SDHC_MISS_CARD_INS_INT_WHEN_INIT,
    DRV_8BIT
    };

/* MCP SDHC driver */

VXB_DRV vxbMcpSdhcCtrlDrv =
    {
    {NULL},
    "mchpSdhc",               /* Name */
    "Microchip SDHC driver",  /* Description */
    VXB_BUSID_FDT,            /* Class */
    0,                        /* Flags */
    0,                        /* Reference count */
    vxbMcpSdhcCtrlMethods     /* Method table */
    };

LOCAL const VXB_FDT_DEV_MATCH_ENTRY mcpSdhcMatch[] =
    {
    {
    "cdns,sd4hc", /* compatible */
    (void *)&mcpSdhcCfg
    },
    {} /* Empty terminated list */
    };

VXB_DRV_DEF(vxbMcpSdhcCtrlDrv);

IMPORT UCHAR erfLibInitialized;

/*******************************************************************************
*
* MCP_SDHC_READ_4 - read MCP SDHC controller MMCSD Subsystem registers
*
* This routine reads MCP SDHC controller MMCSD Subsystem registers.
*
* RETURNS: register contents
*
* ERRNO: N/A
*
* \NOMANUAL
*/

static inline UINT32 MCP_SDHC_READ_4
    (
    SDHC_DEV_CTRL    * pCtrl,
    UINT32             addr
    )
    {
    return vxbRead32 (pCtrl->regHandle, (UINT32 *)(pCtrl->bar + addr));
    }

/*******************************************************************************
*
* MCP_SDHC_WRITE_4 - write MCP SDHC controller MMCSD Subsystem registers.
*
* This routine writes MCP SDHC controller MMCSD Subsystem registers.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

static inline void MCP_SDHC_WRITE_4
    (
    SDHC_DEV_CTRL    * pCtrl,
    UINT32             addr,
    UINT32             value
    )
    {
    vxbWrite32 (pCtrl->regHandle, (UINT32 *)(pCtrl->bar + addr), value);
    }

/*******************************************************************************
*
* mcpSdhcSpecInfoGet - get host controller spec info
*
* This routine gets host controller spec info.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS mcpSdhcSpecInfoGet
    (
    VXB_DEV_ID             pDev,        /* Device information */
    void **                pHostSpec,   /* Host controller spec info */
    VXB_SD_CMDISSUE_FUNC * pCmdIssue    /* Command issue function */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_CTRL * pSdHostCtrl;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));
    if (pSdHostCtrl == NULL)
        {
        return ERROR;
        }

    *pHostSpec = (void *)(&(pSdHostCtrl->sdHostSpec));

    if (pSdHostCtrl->polling)
        {
        *pCmdIssue = sdhcCtrlCmdIssuePoll;
        }
    else
        {
        *pCmdIssue = sdhcCtrlCmdIssue;
        }

    return (OK);
    }

/*******************************************************************************
*
* mcpSdhcCtrlMonitor - MCP SDHC insert status monitor
*
* This function will check MCP SDHC insert/remove status. If target device
* insert status is TRUE, the function will run mcpSdhcDevAdd to add target
* device to vxbus system. If FLASE, will run mcpSdhcDevRemove to remove the
* target device from vxbus system.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL void mcpSdhcCtrlMonitor
    (
    VXB_DEV_ID pDev     /* Device information */
    )
    {
    BOOL rc;
    SD_HOST_CTRL  * pSdHostCtrl;
    SDHC_DEV_CTRL * pDrvCtrl;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return;
        }
    else
        {
        pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));
        if (pSdHostCtrl == NULL)
            {
            return;
            }
        }

    while (erfLibInitialized == FALSE)
        {
        (void)taskDelay (sysClkRateGet ());
        }

    while(1)
        {
        rc = pSdHostCtrl->sdHostOps.vxbSdCardInsertSts (pDev);

        if (rc && (pSdHostCtrl->attached == FALSE))
            {
            sdhcDevAdd (pDev);
            }
        if (!rc && (pSdHostCtrl->attached == TRUE))
            {
            sdhcDevRemove (pDev);
            }
        }
    }


#ifndef _WRS_CONFIG_ITL_SOC_AGILEX5
/*******************************************************************************
*
* mcpSdhcTuningValSet - setup tuning value
*
* This routine set the tuning value for the controller.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL STATUS mcpSdhcTuningValSet
    (
    SDHC_DEV_CTRL    * pCtrl,
    UINT32             val
    )
    {
    UINT32 reg;
    UINT32 timeOut = 0;

    /* delay a bit for card to be ready for next tuning due to errors */

    vxbMsDelay(1);

    /* set tune value */

    reg = MCP_SDHC_READ_4 (pCtrl, MCP_HRS_EMMC_CTRLR);
    reg &= (UINT32)(~MCP_HRS_EMMC_CTRLR_EVT_MASK);
    reg |= val;

    do
        {
        /* tune request */

        reg |= MCP_HRS_EMMC_CTRLR_ETR_MASK;
        MCP_SDHC_WRITE_4 (pCtrl, MCP_HRS_EMMC_CTRLR, reg);

        vxbUsDelay(1);

        reg = MCP_SDHC_READ_4 (pCtrl, MCP_HRS_EMMC_CTRLR);
        timeOut++;
        }
    while (((reg & MCP_HRS_EMMC_CTRLR_ETR_MASK) != 0) &&
           (timeOut < MCP_SDHC_TUNING_TIMES));

    if (timeOut >= MCP_SDHC_TUNING_TIMES)
        {
        return ERROR;
        }
    else
        {
        return OK;
        }
    }

/*******************************************************************************
*
* mcpSdhcTuningSet - setup the tuning
*
* This routine setup the tuning for the controller.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL STATUS mcpSdhcTuningSet
    (
    VXB_DEV_ID pDev
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_CTRL  * pSdHostCtrl;
    SD_HARDWARE   * pSdHardware;
    UINT32          minValue;
    UINT32          maxValue;
    UINT32          setValue;
    STATUS          ret = OK;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));

    pSdHardware = (SD_HARDWARE *)GET_HARDWAREINFO (pSdHostCtrl->sdHostSpec.childDev);
    if (pSdHardware == NULL)
        {
        return ERROR;
        }

    if (pDrvCtrl->tuningStep != 0)
        {
        /* find the mininum delay first which can pass tuning */

        minValue = pDrvCtrl->tuningMin;
        while (minValue < pDrvCtrl->tuningMax)
            {
            if (mcpSdhcTuningValSet(pDrvCtrl, minValue) != OK)
                {
                return ERROR;
                }
            if (sdhcTuningExecute(pDev) == OK)
                {
                break;
                }

            minValue += pDrvCtrl->tuningStep;
            }

        /* find the maxinum delay which can not pass tuning */

        maxValue = minValue + pDrvCtrl->tuningStep;
        while (maxValue < pDrvCtrl->tuningMax)
            {
            if (mcpSdhcTuningValSet(pDrvCtrl, maxValue) != OK)
                {
                return ERROR;
                }
            if (sdhcTuningExecute (pDev) != OK)
                {
                maxValue -= pDrvCtrl->tuningStep;
                break;
                }

            maxValue += pDrvCtrl->tuningStep;
            }

        setValue = (minValue + maxValue) / 2;
        }
    else
        {
        setValue = pDrvCtrl->tuningMin;
        }

    if (mcpSdhcTuningValSet(pDrvCtrl, setValue) != OK)
        {
        MCP_SDHC_DBG (MCP_SDHC_DBG_ERR, "set tuning value fail. set=%d\n",
                      setValue);
        return ERROR;
        }

    ret = sdhcTuningExecute (pDev);
    if (ret != OK)
        {
        MCP_SDHC_DBG (MCP_SDHC_DBG_ERR, "tuning fail. min=%d,max=%d,set=%d\n",
                      minValue, maxValue, setValue);
        }

    return ret;
    }
#endif // _WRS_CONFIG_ITL_SOC_AGILEX5

/*******************************************************************************
*
* mcpSdhcCtrlInit - MCP SDHC controller specific initialization
*
* This routine performs initialization of MCP SDHC controller.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL STATUS mcpSdhcCtrlInit
    (
    VXB_DEV_ID pDev
    )
    {
    SDHC_DEV_CTRL    * pDrvCtrl;
    SD_HOST_CTRL     * pHostCtrl;
    VXB_FDT_DEV      * pFdtDev;
    UINT32             regVal;
    UINT32             dpValue;
    STATUS             rc;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    pFdtDev = vxbFdtDevGet (pDev);
    if (pFdtDev == NULL)
        {
        return ERROR;
        }

    pHostCtrl = (SD_HOST_CTRL *) pDrvCtrl;

    /* reset core */

    MCP_SDHC_WRITE_4 (pDrvCtrl, MCP_HRS_GIR, MCP_HRS_GIR_SWR_MASK);
    do
        {
        regVal = MCP_SDHC_READ_4 (pDrvCtrl, MCP_HRS_GIR);
        }
    while ((regVal & MCP_HRS_GIR_SWR_MASK) != 0);

#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
    rc = mcpSdhcPhyInit(pDev);
    if (rc != OK)
        {
        return ERROR;
        }
    dpValue = 0x32;
#else
    dpValue = pDrvCtrl->sdHostCtrl.curClkFreq * 30 / 1000;
#endif
    /* set debounce period 30ms */

    MCP_SDHC_WRITE_4 (pDrvCtrl, MCP_HRS_DBSR, dpValue);

    /* tuning configuration initialize */

    pDrvCtrl->tuningMin  = MCP_SDHC_TUNING_MIN;
    pDrvCtrl->tuningMax  = MCP_SDHC_TUNING_MAX;
    pDrvCtrl->tuningStep = MCP_SDHC_TUNING_STEP;

    if (sdhcCtrlInstConnect (pHostCtrl) == ERROR)
        {
        return ERROR;
        }

    MCP_SDHC_DBG (MCP_SDHC_DBG_INIT, "mcpSdhcCtrlInit: bus-width %d Bit \n",
                  pHostCtrl->sdHostSpec.busWidth, 0, 0, 0, 0, 0);

    pHostCtrl->sdHostSpec.vxbSdBusWidthSetup
                = pHostCtrl->sdHostOps.vxbSdBusWidthSetup;
    pHostCtrl->sdHostSpec.vxbSdCardWpCheck
                = pHostCtrl->sdHostOps.vxbSdCardWpCheck;
    pHostCtrl->sdHostSpec.vxbSdClkFreqSetup
                = pHostCtrl->sdHostOps.vxbSdClkFreqSetup;
    pHostCtrl->sdHostSpec.vxbSdResumeSet
                = pHostCtrl->sdHostOps.vxbSdResumeSet;
    pHostCtrl->sdHostSpec.vxbSdVddSetup = pHostCtrl->sdHostOps.vxbSdVddSetup;

    pHostCtrl->sdHostSpec.capbility = pHostCtrl->capbility;
    pHostCtrl->sdHostSpec.directBio = pHostCtrl->directBio;
    pHostCtrl->sdHostSpec.vxbSdGetHostCap = sdhcGetCap;
    pHostCtrl->sdHostSpec.vxbSdSigVolSwitch = sdhcSigVolSwitch;
    pHostCtrl->sdHostSpec.vxbSdCheckUhsSupport = sdhcCheckUhsSupport;
    pHostCtrl->sdHostSpec.vxbSdTimingSetup = mcpSdhcTimingSetup;
#ifndef _WRS_CONFIG_ITL_SOC_AGILEX5
    pHostCtrl->sdHostSpec.vxbSdTuningSet = mcpSdhcTuningSet;
#endif

    pDrvCtrl->monTaskId = taskSpawn ("sdBusMonitor", 100, 0,
                                     8192, (FUNCPTR)mcpSdhcCtrlMonitor,
                                     (_Vx_usr_arg_t)pDev,
                                     0, 0, 0, 0, 0, 0, 0, 0, 0);

    if (pDrvCtrl->monTaskId == TASK_ID_ERROR)
        {
        MCP_SDHC_DBG (MCP_SDHC_DBG_ERR, "Create monitor task fault\n",
                      0, 0, 0, 0, 0, 0);
        return ERROR;
        }

    /* setup the interrupt mask */

    pDrvCtrl->intMask = (IRQ_DATA | IRQ_CMD);
    pDrvCtrl->intMask |= IRQ_AC12E;

    if (pDrvCtrl->sdHostCtrl.dmaMode == SDHC_DMA_MODE_PIO)
        {
        pDrvCtrl->intMask |= (IRQ_BRR | IRQ_BWR);
        }

    pDrvCtrl->intMask |= IRQ_CINS;
    pDrvCtrl->intMask |= IRQ_CRM;
    pDrvCtrl->intMask |= IRQ_CINT;

    CSR_WRITE_4 (pDev, SDHC_IRQSTATEN, pDrvCtrl->intMask);

    /* enable SDHC interrupts */

    if (pDrvCtrl->sdHostCtrl.polling == FALSE)
        {

        /* connect and enable interrupt */

        if (pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdIsr == NULL)
            {
            (void)taskDelete (pDrvCtrl->monTaskId);
            return ERROR;
            }

        rc = vxbIntConnect (pDev,
                            pDrvCtrl->intRes,
                            (VOIDFUNCPTR)pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdIsr,
                            pDev);
        if (rc == ERROR)
            {
            (void)taskDelete (pDrvCtrl->monTaskId);
            return ERROR;
            }

        rc = vxbIntEnable (pDev, pDrvCtrl->intRes);
        if (rc == ERROR)
            {
            (void)taskDelete (pDrvCtrl->monTaskId);
            return ERROR;
            }

        CSR_WRITE_4 (pDev, SDHC_IRQSIGEN, pDrvCtrl->intMask);
        }

    if (pDrvCtrl->flags & SDHC_MISS_CARD_INS_INT_WHEN_INIT)
        {
        regVal = CSR_READ_4 (pDev, SDHC_PRSSTAT);
        if (regVal & PRSSTAT_CINS)
            {
            (void)semGive (pDrvCtrl->sdHostCtrl.devChange);
            }
        }

    return OK;
    }

/*******************************************************************************
*
* mcpSdhcTimingSetup - setup timing on host side
*
* This routine setup timing on host side.
*
* RETURNS: OK if succeeded, ERROR if failed.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL STATUS mcpSdhcTimingSetup
    (
    VXB_DEV_ID        pDev,        /* Device information */
    UINT32            timing       /* Timing Mode */
    )
    {
    UINT32          reg;
    UINT32          uhs = 0;
    UINT32          mmcMode = 0;
    SDHC_DEV_CTRL * pDrvCtrl;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    switch (timing)
        {
        case SDMMC_TIMING_UHS_SDR12:
        case SDMMC_TIMING_MMC_LEGACY:
        case SDMMC_TIMING_MMC_HS:
            uhs = AUTOC12ERR_UHS_SDR12;
            if (timing == SDMMC_TIMING_MMC_LEGACY)
                {
                mmcMode = MCP_TIMING_MMC_LEGACY;
                }
            else
                {
                mmcMode = MCP_TIMING_MMC_SDR;
                }
            break;
        case SDMMC_TIMING_UHS_SDR25:
            uhs = AUTOC12ERR_UHS_SDR25;
            mmcMode = MCP_TIMING_MMC_SDR;
            break;
        case SDMMC_TIMING_UHS_SDR50:
            uhs = AUTOC12ERR_UHS_SDR50;
            mmcMode = MCP_TIMING_MMC_SDR;
            break;
        case SDMMC_TIMING_UHS_SDR104:
            uhs = AUTOC12ERR_UHS_SDR104_HS200;
            mmcMode = MCP_TIMING_MMC_SDR;
            break;
        case SDMMC_TIMING_UHS_DDR50:
        case SDMMC_TIMING_MMC_DDR52:
            uhs = AUTOC12ERR_UHS_DDR;
            mmcMode = MCP_TIMING_MMC_DDR;
            break;
        case SDMMC_TIMING_MMC_HS200:
            uhs = AUTOC12ERR_UHS_SDR104_HS200;
            mmcMode = MCP_TIMING_MMC_HS200;
            break;
        case SDMMC_TIMING_MMC_HS400:
            uhs = AUTOC12ERR_UHS_SDR104_HS200;
            mmcMode = MCP_TIMING_MMC_HS400;
            break;
        case SDMMC_TIMING_SD_HS:
            mmcMode = MCP_TIMING_SD;
            break;
        default:
            return ERROR;
        }

    reg = MCP_SDHC_READ_4 (pDrvCtrl, MCP_HRS_EMMC_CTRLR);
    reg &= (UINT32)(~(0x7));
    reg |= mmcMode;
    MCP_SDHC_WRITE_4 (pDrvCtrl, MCP_HRS_EMMC_CTRLR, reg);

    if (timing != SDMMC_TIMING_SD_HS)
        {
        reg = CSR_READ_4 (pDev, SDHC_AUTOC12ERR);
        reg &= ~(AUTOC12ERR_UHS_MASK);
        reg |= uhs;
        CSR_WRITE_4 (pDev, SDHC_AUTOC12ERR, reg);
        }

    return OK;
    }

/*******************************************************************************
*
* mcpSdhcCtrlProbe - probe MCP SDHC controller
*
* This function probe MCP SDHC controller.
*
* Parameter: pDev [in] - VXB_DEV_ID to be probed
*
* RETURNS: OK, or ERROR if the VXB_DEV is not match the driver
*
* \NOMANUAL
*/

LOCAL STATUS mcpSdhcCtrlProbe
    (
    VXB_DEV_ID  pDev    /* Device information */
    )
    {
    VXB_FDT_DEV_MATCH_ENTRY * pMatch;
    MCP_SDHC_CFG * pCfg;
    STATUS rc;

    rc = vxbFdtDevMatch (pDev, mcpSdhcMatch, &pMatch);
    if (rc == ERROR)
        {
        MCP_SDHC_DBG (MCP_SDHC_DBG_ERR, "Match device failed\n");
        return ERROR;
        }

    pCfg = (MCP_SDHC_CFG *)pMatch->data;

    if (sdhcCtrlProbe (pDev) == OK)
        {
        if (pCfg != NULL)
            {
            vxbDevDrvDataSet (pDev, (void *)pCfg);
            }
        return OK;
        }

    return ERROR;
    }

/*******************************************************************************
*
* mcpSdhcCtrlAttach - attach MCP SDHC controller
*
* This function attach MCP SDHC controller.
*
* Parameter: pDev [in] - VXB_DEV to be attached
*
* RETURNS: OK, or ERROR if the VXB_DEV is not match the driver
*
* \NOMANUAL
*/

LOCAL STATUS mcpSdhcCtrlAttach
    (
    VXB_DEV_ID  pDev    /* Device information */
    )
    {
    MCP_SDHC_CFG     * pCfg;
    SDHC_DEV_CTRL    * pDrvCtrl;
    STATUS             status;

    pDrvCtrl = (SDHC_DEV_CTRL *) vxbMemAlloc (sizeof (SDHC_DEV_CTRL));
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    pDrvCtrl->barOffset = MCP_SRS_OFFSET;
    vxbDevSoftcSet (pDev, pDrvCtrl);

    status = sdhcCtrlAttach (pDev);
    if (status != OK)
        {
        return status;
        }

    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdCardWpCheck = sdCardWpCheck;
    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdVddSetup = sdhcVddSetup;

    pCfg = (MCP_SDHC_CFG *)vxbDevDrvDataGet (pDev);
    pDrvCtrl->flags = pCfg->uSdhcFlag;

    /* set the endian mode */

    pDrvCtrl->bigEndian = FALSE;

    /* SDHC specific configurations */

    if (mcpSdhcCtrlInit (pDev) == ERROR)
        {
        goto FAILED;
        }

    return OK;

FAILED:
    vxbDevSoftcSet (pDev, NULL);

    if (pDrvCtrl != NULL)
        {
        vxbMemFree (pDrvCtrl);
        }

    return (ERROR);
    }


#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5

#define HRS00 0x00
#define HRS04 0x10
#define HRS05 0x14
#define HRS06 0x18
#define HRS07 0x1C
#define HRS09 0x24
#define HRS10 0x28
#define HRS16 0x40

#define GENMASK(h, l)       (((~0U) - (1U << (l)) + 1) & \
                             (~0U >> (32 - 1 - (h))))

#define BIT(x) (1U << x)

#define SDHCI_CDNS_HRS06        0x18        /* eMMC control */
#define   SDHCI_CDNS_HRS06_TUNE_UP      BIT(15)
#define   SDHCI_CDNS_HRS06_TUNE         GENMASK(13, 8)
#define   SDHCI_CDNS_HRS06_MODE         GENMASK(2, 0)
#define   SDHCI_CDNS_HRS06_MODE_SD      0x0
#define   SDHCI_CDNS_HRS06_MODE_MMC_SDR     0x2
#define   SDHCI_CDNS_HRS06_MODE_MMC_DDR     0x3
#define   SDHCI_CDNS_HRS06_MODE_MMC_HS200   0x4
#define   SDHCI_CDNS_HRS06_MODE_MMC_HS400   0x5
#define   SDHCI_CDNS_HRS06_MODE_MMC_HS400ES 0x6
/* PHY register values */
#define PHY_DQ_TIMING_REG                   0x2000
#define PHY_DQS_TIMING_REG                  0x2004
#define PHY_GATE_LPBK_CTRL_REG              0x2008
#define PHY_DLL_MASTER_CTRL_REG             0x200C
#define PHY_DLL_SLAVE_CTRL_REG              0x2010
#define PHY_CTRL_REG                        0x2080
#define USE_EXT_LPBK_DQS                    BIT(22)
#define USE_LPBK_DQS                        BIT(21)
#define USE_PHONY_DQS                       BIT(20)
#define USE_PHONY_DQS_CMD                   BIT(19)
#define SYNC_METHOD                         BIT(31)
#define SW_HALF_CYCLE_SHIFT                 BIT(28)
#define RD_DEL_SEL                          GENMASK(24, 19)
#define RD_DEL_SEL_DATA                     0x34
#define GATE_CFG_ALWAYS_ON                  BIT(6)
#define UNDERRUN_SUPPRESS                   BIT(18)
#define PARAM_DLL_BYPASS_MODE               BIT(23)
#define PARAM_PHASE_DETECT_SEL              GENMASK(22, 20)
#define PARAM_DLL_START_POINT               GENMASK(7, 0)
#define PARAM_PHASE_DETECT_SEL_DATA         0x2
#define PARAM_DLL_START_POINT_DATA          0x4
#define PARAM_DLL_START_POINT_DATA_SDR50    254

#define READ_DQS_CMD_DELAY      GENMASK(31, 24)
#define CLK_WRDQS_DELAY         GENMASK(23, 16)
#define CLK_WR_DELAY            GENMASK(15, 8)
#define READ_DQS_DELAY          GENMASK(7, 0)
#define READ_DQS_CMD_DELAY_DATA 0x0
#define CLK_WRDQS_DELAY_DATA    0x0
#define CLK_WR_DELAY_DATA       0x0
#define READ_DQS_DELAY_DATA     0x0

#define PHONY_DQS_TIMING        GENMASK(9, 4)
#define PHONY_DQS_TIMING_DATA   0x0

#define IO_MASK_ALWAYS_ON       BIT(31)
#define IO_MASK_END             GENMASK(29, 27)
#define IO_MASK_START           GENMASK(26, 24)
#define DATA_SELECT_OE_END      GENMASK(2, 0)
#define IO_MASK_END_DATA        0x5
/* DDR Mode */
#define IO_MASK_END_DATA_DDR    0x2
#define IO_MASK_START_DATA      0x0
#define DATA_SELECT_OE_END_DATA 0x1

LOCAL UINT32 mmcPhyRead
    (
    SDHC_DEV_CTRL* pDrvCtrl,
    UINT16 phyRegAddr
    )
    {
    UINT32 val;

    MCP_SDHC_WRITE_4(pDrvCtrl, HRS04, phyRegAddr);
    val = MCP_SDHC_READ_4(pDrvCtrl, HRS05);
    return val;
    }

LOCAL VOID mmcPhyWriteMask
    (
    SDHC_DEV_CTRL* pDrvCtrl,
    UINT16 phyRegAddr,
    UINT32 data,
    UINT32 mask
    )
    {
    UINT32 val;

    MCP_SDHC_WRITE_4(pDrvCtrl, HRS04, phyRegAddr);
    val = MCP_SDHC_READ_4(pDrvCtrl, HRS05);
    val &= ~mask;
    val |= data;
    MCP_SDHC_WRITE_4(pDrvCtrl, HRS05, val);
    }

LOCAL STATUS mcpSdhcPhyInit
    (
    VXB_DEV_ID pDev
    )
    {
    SDHC_DEV_CTRL    * pDrvCtrl;
    UINT32 val;
    UINT32 mask;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    /* Switch on DLL reset */

    val = MCP_SDHC_READ_4(pDrvCtrl, HRS09);
    val &= 0xfffffffe;
    MCP_SDHC_WRITE_4(pDrvCtrl, HRS09, val);

    /* Program PHY_DQS_TIMING_REG */

    val = (1U << 22) | (1U << 21) | (1U << 20) | (1U << 19);
    mask = val;
    mmcPhyWriteMask(pDrvCtrl, PHY_DQS_TIMING_REG, val, mask);

    /* Program PHY_GATE_LPBK_CTRL_REG */

    val = (1U << 31) | (0U << 28) | (52U << 19) | (1U << 18) | (1U << 6);
    mask = SYNC_METHOD | SW_HALF_CYCLE_SHIFT | RD_DEL_SEL | UNDERRUN_SUPPRESS |
            GATE_CFG_ALWAYS_ON;
    mmcPhyWriteMask(pDrvCtrl, PHY_GATE_LPBK_CTRL_REG, val, mask);

    /* Program PHY_DLL_MASTER_CTRL_REG */

    val = (1U << 23) | (2U << 20) | (4U << 0);
    mask = PARAM_DLL_BYPASS_MODE | PARAM_PHASE_DETECT_SEL |
            PARAM_DLL_START_POINT;
    mmcPhyWriteMask(pDrvCtrl, PHY_DLL_MASTER_CTRL_REG, val, mask);

    /* Program PHY_DLL_SLAVE_CTRL_REG */

    val = (0U << 24) | (0U << 16) | (0U << 8) | (0U << 0);
    mask = READ_DQS_CMD_DELAY | CLK_WRDQS_DELAY | CLK_WR_DELAY | READ_DQS_DELAY;
    mmcPhyWriteMask(pDrvCtrl, PHY_DLL_SLAVE_CTRL_REG, val, mask);

    /* Program PHY_CTRL_REG */

    MCP_SDHC_WRITE_4(pDrvCtrl, HRS04, 0x2080);
    val &= ~(0x3fU << 4);
    mask = PHONY_DQS_TIMING;
    mmcPhyWriteMask(pDrvCtrl, PHY_CTRL_REG, val, mask);

    /* Switch off DLL reset */

    val = MCP_SDHC_READ_4(pDrvCtrl, HRS09);
    val |= 0x00000001;
    MCP_SDHC_WRITE_4(pDrvCtrl, HRS09, val);

    /* Wait for PHY init. complete */

    for (;;)
        {
        val = MCP_SDHC_READ_4(pDrvCtrl, HRS09);
        if ((val & 0x00000002) == 0x00000002)
            {
            break;
            }
        }

    val = mmcPhyRead(pDrvCtrl, PHY_DQ_TIMING_REG) & 0x07FFFF8;
    val |= (0 << 31) | (0 << 27) | (0 << 24) | (1 << 0);
    mask = IO_MASK_ALWAYS_ON | IO_MASK_END | IO_MASK_START | DATA_SELECT_OE_END;
    mmcPhyWriteMask(pDrvCtrl,PHY_DQ_TIMING_REG, val, mask);

    val = MCP_SDHC_READ_4(pDrvCtrl, HRS09) & 0xFFFE7FF3;
    val |= (1 << 16) | (1 << 15) | (1 << 3) | (1 << 2);
    MCP_SDHC_WRITE_4(pDrvCtrl, HRS09, val);

    val = MCP_SDHC_READ_4(pDrvCtrl, HRS10) & 0xFFF0FFFF;
    val |= (0 << 16);
    MCP_SDHC_WRITE_4(pDrvCtrl, HRS10, val);

    val = (0 << 28) | (0 << 24) | (0 << 20) | (0 << 16) | (0 << 12) | (1 << 8) |
            (0 << 4) | (1 << 0);
    MCP_SDHC_WRITE_4(pDrvCtrl, HRS16, val);

    val = (9 << 16) | (0 << 0);
    MCP_SDHC_WRITE_4(pDrvCtrl, HRS07, val);

    return OK;
    }
#endif // _WRS_CONFIG_ITL_SOC_AGILEX5
