/* vxbSdhcCtrl.c - SDHC host controller common driver library */

/*
 * Copyright (c) 2012-2023 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
DESCRIPTION

This is the VxBus-compliant Secure Digital Host Controller (SDHC)
driver library which implements the functionality specified in
"SD Specifications Part A2 SD Host Controller Simplified Specification
Version 2.00".

The SDHC provides an interface between the host controller and the
SD/SDIO target device.

This file is not applicable to a particular VxBus driver. Rather, it
implements a number of functions for a standard SDHC controller.
However, many SDHC controllers follow the standard implementation.
Therefore, when a new SDHC controller driver is developed, only
those functions that differ from the standard need to be implemented.
Otherwise, the standard functions in this file are called as needed.

This driver is bound to a device tree, and the device tree node must specify
the following properties:

\is

\i <#address-cells>
This property specifies the number of register address parameters for a
device node if the parent of the device node does not specify the value
(for example, the device node for memory). The value of this property
must be 1.

\i <#size-cells>
This property specifies the number of register size parameters for a device
node if the parent of the device node does not specify the value (for example,
the device node for memory). The value of this property must be 1.

\i <compatible>
This property specifies the name of the SDHC driver.
For the LS1043ARDB, it is "fsl,sdhc".

\i <clocks>
This parameter specifies the clock that this module uses.

\i <reg>
This property specifies the register base address and length of this module.

\i <interrupt-parent>
This property is used to define an interrupt parent. If it is not defined for
a device, the interrupt parent is assumed to be the device tree parent for
the device.

\i <big-endian>
This property specifies the endianness of this module. Set the value to '0'
for little endian. For all other values, big endian is used. If this property
is not specified, little endian is used.

\i <little-endian>
This property is used for standard dts, which specifies the endianness
of this module. Specify this property if the controller is little endian.
If this property is not specified, big endian is used.

\i <voltage-ranges>
This property is used for standard dts, which specifies controller supporting
voltage ranges.

\i <sdhci,auto-cmd12>
This property is used for standard dts, which specifies controller supporting
atuo CMD12.

\i <status>
The initial status of the controller.

\i <dma-mode>
This property specifies the mode of data transfer, 0 means using SDMA,
and 1 means using PIO mode. If this property is not specified, SDMA is used.

\i <disable-wp>
When set no physical WP line is present. This property should only be specified
when the controller has a dedicated write-protect detection logic. If a GPIO is
always used for the write-protect detection logic it is sufficient to not
specify wp-gpios property in the absence of a WP line.

\i <cd-gpios>
This property is used for standard dts, which specifies GPIO used for SD card
detection. Currently only first two arguments are used.

\i <wp-gpios>
This property is used for standard dts, which specifies GPIO used for SD card
write proctection. Currently only first two arguments are used.

\ie

An example of a device node as follows:

Legacy DTS:
\cs
esdhc: esdhc@1560000
    {
    #address-cells = <1>;
    #size-cells = <1>;
    compatible = "fsl,sdhc";
    clocks = <&esdhc_en>;
    reg = <0 0x1560000 0 0x1000>;
    interrupt-parent = <&intc>;
    big-endian = <1>;
    cd-gpios = <&gpio5 22 0 0 0>;
    wp-gpios = <&gpio5 21 0 0 0>;
    status = "disabled";
    };
\ce

Standard DTS:
\cs
esdhc: esdhc@1560000
    {
    #address-cells = <1>;
    #size-cells = <1>;
    compatible = "fsl,sdhc";
    clocks = <&esdhc_en>;
    reg = <0 0x1560000 0 0x1000>;
    interrupt-parent = <&intc>;
    voltage-ranges = <1800 1800 3300 3300>;
    sdhci,auto-cmd12;
    little-endian;
    status = "disabled";
    };
\ce

SEE ALSO: vxBus, xbd
\tb SD Specifications: Part A1 Physical Layer Simplified Specification Version 2.00,
\tb SD Specifications: Part A2 SD Host Controller Simplified Specification Version 2.00

INCLUDE FILES:
*/

/* includes */

#include <vxWorks.h>
#include <stdio.h>
#include <semLib.h>
#include <sysLib.h>
#include <tickLib.h>
#include <taskLib.h>
#include <cacheLib.h>
#include <string.h>
#include <errnoLib.h>
#include <hwif/vxBus.h>
#include <subsys/clk/vxbClkLib.h>
#include <subsys/gpio/vxbGpioLib.h>
#include <vxbSdMmcLib.h>
#include <vxbSdhcCtrl.h>
#ifdef _WRS_CONFIG_FDT
#include <vxFdtLib.h>
#include <hwif/buslib/vxbFdtLib.h>
#endif /* _WRS_CONFIG_FDT */

#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
LOCAL atomic64_t intStatus;
#endif

/* defines */

#ifdef  SDHC_DBG_ON

#ifdef  LOCAL
#undef  LOCAL
#define LOCAL
#endif

#define SDHC_DBG_IRQ            0x00000001
#define SDHC_DBG_RW             0x00000002
#define SDHC_DBG_XBD            0x00000004
#define SDHC_DBG_ERR            0x00000008
#define SDHC_DBG_INIT           0x00000010
#define SDHC_DBG_ALL            0xffffffff
#define SDHC_DBG_OFF            0x00000000

UINT32 sdhcDbgMask = SDHC_DBG_OFF;

IMPORT FUNCPTR _func_logMsg;

#define SDHC_DBG(mask, string, c, d, e, f)                  \
    if ((sdhcDbgMask & mask) && (_func_logMsg != NULL))     \
        _func_logMsg("%s(%d) " string,                      \
                    (_Vx_usr_arg_t) __FUNCTION__, (_Vx_usr_arg_t) __LINE__,   \
                    (_Vx_usr_arg_t) c,            (_Vx_usr_arg_t) d,          \
                    (_Vx_usr_arg_t) e,            (_Vx_usr_arg_t) f)
#else
#define SDHC_DBG(mask, string, c, d, e, f)
#endif  /* SDHC_DBG_ON */

UINT  sdhcDmaBlockNum = 2048;

/* externs */

IMPORT UCHAR erfLibInitialized;

IMPORT void vxbUsDelay (int);
IMPORT void vxbMsDelay (int);

/* locals */

LOCAL STATUS sdhcCtrlAbortIssue (VXB_DEV_HANDLE pDev);
LOCAL void sdhcCtrlIsrDefer (void * pData);
#ifdef _WRS_CONFIG_FDT
LOCAL STATUS sdhcCtrlGetGpioConfig (VXB_DEV_HANDLE pDev);
LOCAL STATUS sdhcCtrlFdtParse (VXB_DEV_HANDLE);
#endif /* _WRS_CONFIG_FDT */

/*******************************************************************************
*
* sdhcInstInit - first-level initialization function for the SDHC device
*
* This function performs the first-level initialization for the SDHC device.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlProbe
    (
    VXB_DEV_HANDLE pDev     /* vxbus device ID */
    )
    {
    return OK;
    }

/*******************************************************************************
*
* sdhcCtrlAttach - attach initialization for the SDHC device
*
* This function performs the attach initialization for the SDHC device.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlAttach
    (
    VXB_DEV_HANDLE pDev     /* vxbus device ID */
    )
    {
#ifdef _WRS_CONFIG_FDT
    VXB_FDT_DEV * pFdtDev = NULL;
    const void * pValue;
    VXB_CLK * pClk;
    int len;
    SD_HOST_CTRL  * pHostCtrl;
#endif
    SDHC_DEV_CTRL * pDrvCtrl;
    VXB_RESOURCE_ADR * pResAdr = NULL;
    VXB_RESOURCE     * pRes = NULL;
    BOOL               bNew = FALSE;

    /* create controller driver context structure for core */

    pDrvCtrl = (SDHC_DEV_CTRL*) vxbDevSoftcGet(pDev);

    if (pDrvCtrl == NULL)
        {
        pDrvCtrl = (SDHC_DEV_CTRL *) vxbMemAlloc (sizeof (SDHC_DEV_CTRL));

        if (pDrvCtrl == NULL)
            return ERROR;

        bNew = TRUE;

        vxbDevSoftcSet(pDev, pDrvCtrl);
        }

#ifdef _WRS_CONFIG_FDT
    if (VXB_BUSID_MATCH(vxbDevClassGet (pDev), VXB_BUSID_FDT))
    {
    pHostCtrl = (SD_HOST_CTRL *)pDrvCtrl;

    /* Retrieve the clock-frequency */

    pFdtDev = vxbFdtDevGet(pDev);
    if (pFdtDev == NULL)
        goto errHandle;

    pValue = vxFdtPropGet(pFdtDev->offset, "clock-frequency", &len);
    if(pValue != NULL)
        {
        pDrvCtrl->sdHostCtrl.curClkFreq = vxFdt32ToCpu(*(UINT32 *)pValue);
        }
    else
        {
        pValue = vxFdtPropGet(pFdtDev->offset, "assigned-clock-rates", &len);
        if(pValue != NULL)
            {
            pDrvCtrl->sdHostCtrl.curClkFreq = vxFdt32ToCpu(*(UINT32 *)pValue);
            }
        else
            {
            pClk = vxbClkGet (pDev, NULL);
            if (pClk == NULL)
                {
                if (bNew)
                    {
                    vxbMemFree (pDrvCtrl);
                    vxbDevSoftcSet(pDev, NULL);
                    }
                return (EAGAIN);
                }
            (void) vxbClkEnable (pClk);
            pDrvCtrl->sdHostCtrl.curClkFreq = (UINT32)vxbClkRateGet(pClk);
            }
        }

    SDHC_DBG (SDHC_DBG_INIT, "get host clock freq %d Hz\n",
              pDrvCtrl->sdHostCtrl.curClkFreq, 0, 0, 0);

    /* retrieve user-set-clock-frequency information if provided */

    pValue = vxFdtPropGet(pFdtDev->offset, "user-set-clock-frequency", &len);
    if(pValue != NULL)
        {
        pDrvCtrl->userSetClkFreq = vxFdt32ToCpu(*(UINT32 *)pValue);

        if ((pHostCtrl->sdHostSpec.workFreq == 0) ||
            (pHostCtrl->sdHostSpec.workFreq > pDrvCtrl->userSetClkFreq))
            {
            pHostCtrl->sdHostSpec.workFreq = pDrvCtrl->userSetClkFreq;
            }
        }
    else
        {
        pDrvCtrl->userSetClkFreq = 0;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "directBio", &len);
    if(pValue != NULL)
        pDrvCtrl->sdHostCtrl.directBio = vxFdt32ToCpu(*(UINT32 *)pValue);
    else
        pDrvCtrl->sdHostCtrl.directBio = 0;

    pValue = vxFdtPropGet (pFdtDev->offset, "disable-protect-check", &len);
    if(pValue != NULL)
        {
        pDrvCtrl->noWPCheck = (BOOL)vxFdt32ToCpu(*(UINT32 *)pValue);
        }
    else
        {
        pDrvCtrl->noWPCheck = FALSE;
        }
    }
#endif

    pRes = vxbResourceAlloc(pDev, VXB_RES_MEMORY, 0);
    if(pRes == NULL)
        goto errHandle;

    pResAdr = (VXB_RESOURCE_ADR *)pRes->pRes;
    if (pResAdr == NULL)
        goto errHandle;

    pDrvCtrl->regHandle = ((VXB_RESOURCE_ADR *)pRes->pRes)->pHandle;
#if defined (_WRS_CONFIG_LP64)
    pDrvCtrl->bar = (UINT64)pResAdr->virtual;
#else
    pDrvCtrl->bar = (UINT32)pResAdr->virtual;
#endif

    pRes = vxbResourceAlloc(pDev, VXB_RES_IRQ, 0);
    if (pRes == NULL)
        goto errHandle;

    pDrvCtrl->intRes = pRes;

#ifdef _WRS_CONFIG_FDT
    if (VXB_BUSID_MATCH(vxbDevClassGet (pDev), VXB_BUSID_FDT))
        {
        pValue = vxFdtPropGet(pFdtDev->offset, "dma-mode", &len);
        if(pValue != NULL)
            pDrvCtrl->sdHostCtrl.dmaMode = vxFdt32ToCpu(*(UINT32 *)pValue);
        else
            pDrvCtrl->sdHostCtrl.dmaMode = SDHC_DMA_MODE_ADMA2;

        pValue = vxFdtPropGet(pFdtDev->offset, "userUhsMode", &len);
        if (pValue == NULL)
            {
            pDrvCtrl->userUhsMode = UHSMODE_AUTOSELECT;
            }
        else
            {
            pDrvCtrl->userUhsMode = vxFdt32ToCpu(*(UINT32 *)pValue);
            }
        }
    else
#endif /* _WRS_CONFIG_FDT */
        {
        pDrvCtrl->bigEndian = FALSE;
        pDrvCtrl->sdHostCtrl.dmaMode = SDHC_DMA_MODE_ADMA2;
        }

    pDrvCtrl->sdHostCtrl.pDev = pDev;
    pDrvCtrl->sdHostCtrl.attached = FALSE;
    pDrvCtrl->sdHostCtrl.isInsert = FALSE;

#ifdef _WRS_CONFIG_FDT

    /* parse the commn FDT properities */

    (void) sdhcCtrlFdtParse (pDev);

#endif /* _WRS_CONFIG_FDT */

    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdBusWidthSetup = sdhcCtrlBusWidthSetup;
    if (pDrvCtrl->noWPCheck == FALSE)
        pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdCardWpCheck = sdhcCtrlCardWpCheck;
    else
        pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdCardWpCheck = NULL;
    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdCardInsertSts = sdhcCtrlCardInsertSts;
    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdCmdPrepare = sdhcCtrlCmdPrepare;
#ifdef _WRS_CONFIG_FDT
    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdHostCtrlInit = sdhcCtrlInit;
#endif /* _WRS_CONFIG_FDT */
    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdClkFreqSetup = sdhcCtrlClkFreqSetup;
    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdIsr = (VXB_SD_ISR)sdhcCtrlIsr;
    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdVddSetup = NULL;
    pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdResumeSet = NULL;
    pDrvCtrl->sdHostCtrl.sdHostSpec.vxbSdAbortIssue = sdhcCtrlAbortIssue;

    pDrvCtrl->intSts = 0;
    pDrvCtrl->isrDefer.func = sdhcCtrlIsrDefer;
    pDrvCtrl->isrDefer.pData = (void *) pDev;

    /* set the pointer of the queue for deferred work */

    pDrvCtrl->queueId = isrDeferQueueGet (pDev, 0, 0, 0);
    if (pDrvCtrl->queueId == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "isrDeferQueueGet fail\n", 0, 0, 0, 0);
        goto errHandle;
        }

    pDrvCtrl->noR1bDataDone = FALSE;

    return OK;
errHandle:
    if (pRes != NULL)
        {
        (void)vxbResourceFree (pDev, pRes);
        }

    if (bNew)
        {
        vxbMemFree (pDrvCtrl);
        vxbDevSoftcSet(pDev, NULL);
        }

    return ERROR;
    }

/*******************************************************************************
*
* sdhcInstConnect - third-level initialization function for the SDHC device
*
* This function performs the third-level initialization for the SDHC device.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlInstConnect
    (
    SD_HOST_CTRL * pSdHostCtrl    /* SD host controller */
    )
    {
    STATUS rc;

    /*
     * The devChange semaphore is used by the interrupt service routine
     * to inform the card monitor task that a state change has occurred.
     */

    pSdHostCtrl->devChange = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);
    if (pSdHostCtrl->devChange == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "semBCreate devChange fail\n", 0, 0, 0, 0);
        return ERROR;
        }

    pSdHostCtrl->cmdDone = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);
    if (pSdHostCtrl->cmdDone == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "semBCreate cmdDone fail\n", 0, 0, 0, 0);
        goto err;
        }

    pSdHostCtrl->dataDone = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);
    if (pSdHostCtrl->dataDone == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "semBCreate dataDone fail\n", 0, 0, 0, 0);
        goto err;
        }

    pSdHostCtrl->pioReady = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);
    if (pSdHostCtrl->pioReady == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "semBCreate pioReady fail\n", 0, 0, 0, 0);
        goto err;
        }

    pSdHostCtrl->tuningDone = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);
    if (pSdHostCtrl->tuningDone == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "semBCreate tuningDone fail\n", 0, 0, 0, 0);
        goto err;
        }

    pSdHostCtrl->hostDevSem = semMCreate(SEM_Q_PRIORITY |
                                         SEM_DELETE_SAFE |
                                         SEM_INVERSION_SAFE);
    if (pSdHostCtrl->hostDevSem == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "semBCreate pioReady fail\n", 0, 0, 0, 0);
        goto err;
        }

    /* per-device init */

    if (pSdHostCtrl->sdHostOps.vxbSdHostCtrlInit != NULL)
        {
        rc = pSdHostCtrl->sdHostOps.vxbSdHostCtrlInit(pSdHostCtrl->pDev);
        if (rc == ERROR)
            {
            goto err;
            }
        }

    return OK;

err:
    if (pSdHostCtrl->dataDone != NULL)
        {
        if (semDelete(pSdHostCtrl->dataDone) != OK)
            {
            SDHC_DBG (SDHC_DBG_ERR, "semDelete dateDone fail\n", 0, 0, 0, 0);
            }
        }

    if (pSdHostCtrl->cmdDone != NULL)
        {
        if (semDelete(pSdHostCtrl->cmdDone) != OK)
            {
            SDHC_DBG (SDHC_DBG_ERR, "semDelete cmdDone fail\n", 0, 0, 0, 0);
            }
        }
    if (pSdHostCtrl->devChange != NULL)
        {
        if (semDelete(pSdHostCtrl->devChange) != OK)
            {
            SDHC_DBG (SDHC_DBG_ERR, "semDelete devChange fail\n", 0, 0, 0, 0);
            }
        }
    if (pSdHostCtrl->pioReady != NULL)
        {
        if (semDelete(pSdHostCtrl->pioReady) != OK)
            {
            SDHC_DBG (SDHC_DBG_ERR, "semDelete pioReady fail\n", 0, 0, 0, 0);
            }
        }
    if (pSdHostCtrl->tuningDone != NULL)
        {
        if (semDelete(pSdHostCtrl->tuningDone) != OK)
            {
            SDHC_DBG (SDHC_DBG_ERR, "semDelete tuningDone fail\n", 0, 0, 0, 0);
            }
        }
    if (pSdHostCtrl->hostDevSem != NULL)
        {
        if (semDelete(pSdHostCtrl->hostDevSem) != OK)
            {
            SDHC_DBG (SDHC_DBG_ERR, "semDelete hostDevSem fail\n", 0, 0, 0, 0);
            }
        }
    return ERROR;
    }

/*******************************************************************************
*
* sdhcCtrlInit - initialization per SDHC specification
*
* This function performs controller initialization per SDHC specification.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlInit
    (
    VXB_DEV_HANDLE pDev     /* vxbus device ID */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    UINT32 hostCap;
    UINT32 proctl;
    int flag;
    bus_size_t alignment;
    UINT32 breaktime;
#ifdef _WRS_CONFIG_FDT
    VXB_FDT_DEV * pFdtDev;
    const void * pValue;
    UINT32 * pVolValue;
    int len;
    BOOL volFromFdt = FALSE;
    UINT32 i = 0;
    UINT32 count = 0;
    UINT32 volValue = 0;
#endif /* _WRS_CONFIG_FDT */

    /* reset eSDHC controller */

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ERROR;

    /*
     * Below information about K70 SDHC errata is from Mask Set Errata for
     * Mask 4N96B, NXP Semiconductors.
     *
     * e3980: SDHC: Glitch is generated on card clock with software reset or
     * clock divider change
     * Description: A glitch may occur on the SDHC card clock when the software
     * sets the RSTA bit (software reset) in the system control register. It
     * can also be generated by setting the clock divider value. The glitch
     * produced can cause the external card to switch to an unknown state. The
     * occurrence is not deterministic.
     * Workaround: A simple workaround is to disable the SD card clock before
     * the software reset, and enable it when the module resumes the normal
     * operation. The Host and the SD card are in a master-slave relationship.
     * The Host provides clock and control transfer across the interface.
     * Therefore, any existing operation is discarded when the Host controller
     * is reset.
     * The recommended flow is as follows:
     *  1. Software disable bit[3], SDCLKEN, of the System Control Register
     *  2. Trigger software reset and/or set clock divider
     *  3. Check bit[3], SDSTB, of the Present State Register for stable clock
     *  4. Enable bit[3], SDCLKEN, of the System Control Register.
     * Using the above method, the eSDHC cannot send command or transfer data
     * when there is a glitch in the clock line, and the glitch does not cause
     * any issue.
     */

    if (pDrvCtrl->flags & SDHC_HOST_CTRL_KINETIS)
        {
        /* stop SD_CLK before software reset */

        CSR_CLRBIT_4 (pDev, SDHC_SYSCTL, FSL_SYSCTL_SDCLKEN);
        }

    CSR_WRITE_4 (pDev, SDHC_SYSCTL, SYSCTL_RSTA);

    breaktime = 0;
    while ((CSR_READ_4 (pDev, SDHC_SYSCTL) & SYSCTL_RSTA)
           && (breaktime < 5000))
        {
        vxbMsDelay (1);
        breaktime ++;
        if (breaktime == 5000)
            return ERROR;
        }

    if (pDrvCtrl->flags & SDHC_HOST_CTRL_KINETIS)
        {
        /* wait until the internal card clock is stable */

        breaktime = 0;
        while (((CSR_READ_4 (pDev, SDHC_PRSSTAT) & FSL_PRSSTAT_SDSTB) == 0)
               && (breaktime < 5000))
            {
            vxbMsDelay (1);
            breaktime ++;
            if (breaktime == 5000)
                return ERROR;
            }

        vxbUsDelay (20000);

        /* re-enable SD_CLK */

        CSR_SETBIT_4 (pDev, SDHC_SYSCTL, FSL_SYSCTL_SDCLKEN);
        vxbUsDelay (20000);

        /*
         * INITA - Initialization Active
         * When this bit is set, 80 SD-clocks are sent to the card. After
         * the 80 clocks are sent, this bit is self cleared. This bit is
         * very useful during the card power-up period when 74 SD-clocks
         * are needed and the clock auto gating feature is enabled.
         */

        CSR_SETBIT_4 (pDev, SDHC_SYSCTL, SYSCTL_INITA);

        breaktime = 0;
        while ((CSR_READ_4 (pDev, SDHC_SYSCTL) & SYSCTL_INITA)
               && (breaktime < 5000))
            {
            vxbMsDelay (1);
            breaktime ++;
            if (breaktime == 5000)
                return ERROR;
            }
        }

    vxbMsDelay(1);

    /* disable all interrupts */

    CSR_WRITE_4 (pDev, SDHC_IRQSIGEN, 0);
    CSR_WRITE_4 (pDev, SDHC_IRQSTATEN, 0);
    CSR_WRITE_4 (pDev, SDHC_IRQSTAT, IRQ_ALL);

    /* enable all clock domains */

    CSR_SETBIT_4 (pDev, SDHC_SYSCTL,
                  SYSCTL_SDCLK_EN | SYSCTL_INTCLK_EN);

    /* setup data transfer width to 1-bit */

    if (pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdBusWidthSetup != NULL)
        pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdBusWidthSetup(pDev,
                                           SDMMC_BUS_WIDTH_1BIT);

    /* get host controller version */

    pDrvCtrl->sdHostCtrl.version =
            (CSR_READ_4 (pDev, SDHC_HOSTVER) >> 16) & 0xffff;

    /*
     * Reg PROCTL (0x28) bits 4:3 DMA Select
     * DMA Select:
     *     00 : SDMA is selected
     *     01 : 32bit address ADMA1 is selected
     *     10 : 32bit address ADMA2 is selected
     *     11 : 64bit address ADMA2 is selected
     */

    proctl = CSR_READ_4 (pDev, SDHC_PROCTL);

    /* get host controller capabilities */

    hostCap = CSR_READ_4 (pDev, SDHC_HOSTCAPBLT);


    if(sdhcDmaBlockNum > SDHC_MAX_RW_SECTORS)
        {
        sdhcDmaBlockNum = SDHC_MAX_RW_SECTORS;
        }
    else if(sdhcDmaBlockNum == 0)
        {
        pDrvCtrl->sdHostCtrl.dmaMode = SDHC_DMA_MODE_PIO;
        }

    /*
     * DMA mode is firstly got from dts file. If it is not set, ADMA2 is
     * set as default. If SDHC does not support ADMA2, use ADMA1 instead.
     * Currently ADMA1 is not support, so just set to SDMA in this case.
     * If SDHC does not support SDMA, use PIO instead.
     */

    switch (pDrvCtrl->sdHostCtrl.dmaMode)
        {
        case SDHC_DMA_MODE_ADMA2:
            if (((pDrvCtrl->flags & SDHC_DMA_FSL_STYLE) == 0) &&
                (hostCap & HOSTCAPBLT_ADMA2))
                {
                proctl &= (UINT32) (~(0x3<<3));
                proctl |= (0x2 << 3);
                break;
                }
            else if ((pDrvCtrl->flags & SDHC_DMA_FSL_STYLE) &&
                 (hostCap & HOSTCAPBLT_ADMA2_B20))
                {
                proctl &= (UINT32) (~(0x3<<8));
                proctl |= (0x2 << 8);
                break;
                }
            else
                pDrvCtrl->sdHostCtrl.dmaMode = SDHC_DMA_MODE_ADMA1;

            /* fall through */

        case SDHC_DMA_MODE_ADMA1:
            pDrvCtrl->sdHostCtrl.dmaMode = SDHC_DMA_MODE_SDMA;

            /* fall through */

        case SDHC_DMA_MODE_SDMA:
            if (!(hostCap & HOSTCAPBLT_DMAS))
                {
                pDrvCtrl->sdHostCtrl.dmaMode = SDHC_DMA_MODE_PIO;
                }

            /* fall through */

        case SDHC_DMA_MODE_PIO:
            break;

        default:
            pDrvCtrl->sdHostCtrl.dmaMode = SDHC_DMA_MODE_PIO;
            break;
        }

    CSR_WRITE_4 (pDev, SDHC_PROCTL, proctl);
    SDHC_DBG (SDHC_DBG_INIT, "hostCap = 0x%x dmaMode = %d proctl=0x%x\n",
              hostCap, pDrvCtrl->sdHostCtrl.dmaMode, proctl, 0);

    /* determine voltage supported */

#ifdef _WRS_CONFIG_FDT

    pFdtDev = vxbFdtDevGet(pDev);
    if (pFdtDev != NULL)
        {
        pValue = vxFdtPropGet(pFdtDev->offset, "voltage-ranges", &len);
        if (pValue != NULL)
           {
           if ((len != 0) && (len % 8 == 0))
               {
                pVolValue = (UINT32 *)pValue;
                count = (UINT32)(len / 4);
                volFromFdt = TRUE;
               }
           }
       }

    if (volFromFdt == TRUE)
        {
        for (i = 0; i < count; i++)
            {
            volValue = vxFdt32ToCpu(*pVolValue);
            pVolValue++;
            if (volValue == 3300)
                {
                pDrvCtrl->sdHostCtrl.capbility |= OCR_VDD_VOL_32_33 | OCR_VDD_VOL_33_34;
                pDrvCtrl->sdHostCtrl.vdd = SDMMC_VOLTAGE_33V;
                }
            else if (volValue == 3000)
                {
                pDrvCtrl->sdHostCtrl.capbility |= OCR_VDD_VOL_29_30 | OCR_VDD_VOL_30_31;
                if (pDrvCtrl->sdHostCtrl.vdd != SDMMC_VOLTAGE_33V)
                    pDrvCtrl->sdHostCtrl.vdd = SDMMC_VOLTAGE_30V;
                }
            else
                if ((pDrvCtrl->sdHostCtrl.vdd != SDMMC_VOLTAGE_33V) &&
                    (pDrvCtrl->sdHostCtrl.vdd != SDMMC_VOLTAGE_30V))
                    pDrvCtrl->sdHostCtrl.vdd = SDMMC_VOLTAGE_18V;
            }
        }
    else
#endif /* _WRS_CONFIG_FDT */
        {
        if (hostCap & HOSTCAPBLT_VS30)
            pDrvCtrl->sdHostCtrl.capbility |= OCR_VDD_VOL_29_30 | OCR_VDD_VOL_30_31;
        if (hostCap & HOSTCAPBLT_VS33)
            pDrvCtrl->sdHostCtrl.capbility |= OCR_VDD_VOL_32_33 | OCR_VDD_VOL_33_34;
        if (hostCap & HOSTCAPBLT_VS18)
            pDrvCtrl->sdHostCtrl.capbility |= OCR_VDD_VOL_S18A;

        /* determine the highest SD bus voltage suppored by the controller */

        if (hostCap & HOSTCAPBLT_VS33)
            pDrvCtrl->sdHostCtrl.vdd = SDMMC_VOLTAGE_33V;
        else if (hostCap & HOSTCAPBLT_VS30)
            pDrvCtrl->sdHostCtrl.vdd = SDMMC_VOLTAGE_30V;
        else
            pDrvCtrl->sdHostCtrl.vdd = SDMMC_VOLTAGE_18V;
        }

    /* high capacity SD card is supported */

    pDrvCtrl->sdHostCtrl.capbility |= OCR_CARD_CAP_STS;

    /* determine high speed SD card supported */

    if (hostCap & HOSTCAPBLT_HSS)
        pDrvCtrl->highSpeed = TRUE;

    pDrvCtrl->sdHostCtrl.sdHostDmaParentTag = vxbDmaBufTagParentGet (pDev, 0);

    if ((!(pDrvCtrl->flags & SDHC_HW_SNOOP)) && (!(pDrvCtrl->dmaCoherent)))
        {
        flag = VXB_DMABUF_ALLOCNOW;
        alignment = (bus_size_t)_CACHE_ALIGN_SIZE;
        }
    else
        {
        flag = VXB_DMABUF_MAPCREATE_NOBOUNCE_BUF | VXB_DMABUF_CACHESNOOP_BUS;
        alignment = 1;
        }

    if (pDrvCtrl->flags & SDHC_DMABUF_FORCED_BOUNCE)
        {
        flag |= VXB_DMABUF_FORCED_BOUNCE;
        }

    if (pDrvCtrl->sdHostCtrl.dmaMode == SDHC_DMA_MODE_ADMA2)
        {
        pDrvCtrl->sdHostCtrl.dmaHdl = (void *) sdhcAdma2DescCreate(pDev);

        pDrvCtrl->sdHostCtrl.sdHostDmaTag = vxbDmaBufTagCreate
                   (pDev,
                    pDrvCtrl->sdHostCtrl.sdHostDmaParentTag, /* parent */
                    alignment,                               /* alignment */
                    0,                                       /* boundary */
                    VXB_SPACE_MAXADDR_32BIT,                 /* lowaddr */
                    VXB_SPACE_MAXADDR,                       /* highaddr */
                    NULL,                                    /* filter */
                    NULL,                                    /* filterarg */
                    sdhcDmaBlockNum * SDMMC_BLOCK_SIZE,      /* max size */
                    ADMA2_NB_DESC,                           /* nSegments */
                    ADMA2_SEG_LEN,                           /* max seg size */
                    flag,                                    /* flags */
                    NULL,                                    /* lockfunc */
                    NULL,                                    /* lockarg */
                    NULL);
        }
    else if (pDrvCtrl->sdHostCtrl.dmaMode == SDHC_DMA_MODE_SDMA)
        {
        pDrvCtrl->sdHostCtrl.sdHostDmaTag = vxbDmaBufTagCreate
                   (pDev,
                    pDrvCtrl->sdHostCtrl.sdHostDmaParentTag, /* parent */
                    alignment,                               /* alignment */
                    0,                                       /* boundary */
                    VXB_SPACE_MAXADDR_32BIT,                 /* lowaddr */
                    VXB_SPACE_MAXADDR,                       /* highaddr */
                    NULL,                                    /* filter */
                    NULL,                                    /* filterarg */
                    sdhcDmaBlockNum * SDMMC_BLOCK_SIZE,      /* max size */
                    1,                                       /* nSegments */
                    sdhcDmaBlockNum * SDMMC_BLOCK_SIZE,      /* max seg size */
                    flag,                                    /* flags */
                    NULL,                                    /* lockfunc */
                    NULL,                                    /* lockarg */
                    NULL);
        }

    if (pDrvCtrl->sdHostCtrl.dmaMode == SDHC_DMA_MODE_SDMA ||
        pDrvCtrl->sdHostCtrl.dmaMode == SDHC_DMA_MODE_ADMA2)
        {
        if (pDrvCtrl->sdHostCtrl.sdHostDmaTag == NULL)
            {
            SDHC_DBG (SDHC_DBG_ERR, "sdHostDmaTag create fault\n", 0, 0, 0, 0);
            return ERROR;
            }

        if (vxbDmaBufMapCreate (pDev, pDrvCtrl->sdHostCtrl.sdHostDmaTag, 0,
                                &(pDrvCtrl->sdHostCtrl.sdHostDmaMap)) == NULL)
            {
            return ERROR;
            }

        }

    if(pDrvCtrl->sdHostCtrl.dmaMode != SDHC_DMA_MODE_PIO)
        {
        pDrvCtrl->sdHostCtrl.sdHostSpec.maxBlocksPerRequest = sdhcDmaBlockNum;
        }
    else
        {
        pDrvCtrl->sdHostCtrl.sdHostSpec.maxBlocksPerRequest = SDHC_MAX_RW_SECTORS;
        }

    if (pDrvCtrl->flags & SDHC_FIFO_ENDIANESS_REVERSE)
       {

       /*
        * If endianess of internal buffer data port is differernt
        * from CPU's endiness, we swap the handle.
        */

       /* VXB_HANDLE_SWAP(x) adds a value to x.  We want that value to be as
        * indicated, not multiplied by the size of the pointer entity.  Use
        * a (char *) to guarantee that
        */

        pDrvCtrl->pioHandle = (void *)
                              VXB_HANDLE_SWAP ((UINT8 *)pDrvCtrl->regHandle);
        }

    return OK;
    }

/*******************************************************************************
*
* sdhcCtrlIsrDefer - handle interrupt in task context
*
* This routine handles interrupts of eSDHC.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL void sdhcCtrlIsrDefer
    (
    void * pData     /* vxbus device ID */
    )
    {
    VXB_DEV_ID pDev = (VXB_DEV_ID) pData;
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_CTRL * pSdHostCtrl;
    SD_CMD * pSdCmd;
    UINT32 intSts;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return;

    pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));
    pSdCmd = pSdHostCtrl->pCurCmd;

    while (1)
        {
#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
        intSts = (UINT32) vxAtomicGet(&intStatus);
        vxAtomicSet(&intStatus, 0);
#else
        intSts = CSR_READ_4 (pDev, SDHC_IRQSTAT);
#endif

        if ((intSts & CSR_READ_4 (pDev, SDHC_IRQSTATEN)) == 0)
            break;

        SDHC_DBG (SDHC_DBG_IRQ, "isrDefer get IRQSTAT = 0x%x\n",
                  intSts, 0, 0, 0);

        /* clear interrupts */

        CSR_WRITE_4 (pDev, SDHC_IRQSTAT, intSts);

        /* card insertion */

        if (intSts & IRQ_CINS)
            {
            SDHC_DBG (SDHC_DBG_IRQ, "card insert interrupt\n", 0, 0, 0, 0);

            /* turn off card insertion interrup */

            pDrvCtrl->intMask &= ~IRQ_CINS;
            pDrvCtrl->intMask |= IRQ_CRM;
            CSR_WRITE_4(pDev, SDHC_IRQSTATEN, pDrvCtrl->intMask);
            (void)semGive (pDrvCtrl->sdHostCtrl.devChange);
            }

        /* card removal */

        if (intSts & IRQ_CRM)
            {
            SDHC_DBG (SDHC_DBG_IRQ, "card removal interrupt\n", 0, 0, 0, 0);

            /* turn off card removal interrupt */

            pDrvCtrl->intMask &= ~IRQ_CRM;
            pDrvCtrl->intMask |= IRQ_CINS;
            CSR_WRITE_4(pDev, SDHC_IRQSTATEN, pDrvCtrl->intMask);
            (void)semGive (pDrvCtrl->sdHostCtrl.devChange);
            }

        /* command interrupts */

        if (intSts & IRQ_CMD)
            {
            if (intSts & IRQ_CMD_ERR)
                {
                if (intSts & IRQ_CIE)
                    {
                    SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd index error\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    pSdCmd->cmdErr |= SDMMC_CMD_ERR_PROTOCOL;
                    }

                if (intSts & IRQ_CEBE)
                    {
                    SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd end of bit error\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    pSdCmd->cmdErr |= SDMMC_CMD_ERR_PROTOCOL;
                    }

                if ((intSts & (IRQ_CCE | IRQ_CTOE)) ==
                    (IRQ_CCE | IRQ_CTOE))
                    {
                    SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd line conflict\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    pSdCmd->cmdErr |= SDMMC_CMD_ERR_PROTOCOL;
                    }
                if (intSts & IRQ_CCE)
                    {
                    SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd CRC error\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    pSdCmd->cmdErr |= SDMMC_CMD_ERR_PROTOCOL;
                    }
                if (intSts & IRQ_CTOE)
                    {
                    SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd timeout error\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    pSdCmd->cmdErr |= SDMMC_CMD_ERR_TIMEOUT;
                    }
                }

            (void)semGive (pDrvCtrl->sdHostCtrl.cmdDone);
            }

        /* DMA interrupt */

        if ((intSts & IRQ_DINT) != 0x0)
            {
            SDHC_DBG (SDHC_DBG_IRQ, "DMA int 0x%08x\n",
                      CSR_READ_4(pDev, SDHC_DSADDR), 0, 0, 0);

            CSR_WRITE_4 (pDev, SDHC_DSADDR, pDrvCtrl->sdmaAddr);

            /* setup next SDMA address */
            pDrvCtrl->sdmaAddr += pDrvCtrl->sdmaAddrBound;

            if (!(intSts & IRQ_TC))
                continue;
            }

        /* data interrupts */

        if (intSts & IRQ_DATA)
            {
            SDHC_DBG (SDHC_DBG_IRQ, "data interrupt\n", 0, 0, 0, 0);

            if (intSts & IRQ_DATA_ERR)
                {
                if (intSts & IRQ_DEBE)
                    {
                    SDHC_DBG (SDHC_DBG_ERR, "CMD%d data end of bit error\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    pSdCmd->dataErr |= SDMMC_DATA_ERR_PROTOCOL;
                    }

                if (intSts & IRQ_DCE)
                    {
                    SDHC_DBG (SDHC_DBG_ERR, "CMD%d data CRC error\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    pSdCmd->dataErr |= SDMMC_DATA_ERR_PROTOCOL;
                    }

                if (intSts & IRQ_DTOE)
                    {
                    SDHC_DBG (SDHC_DBG_ERR, "CMD%d data timeout error\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    pSdCmd->dataErr |= SDMMC_DATA_ERR_TIMEOUT;
                    }
                }
            (void)semGive (pDrvCtrl->sdHostCtrl.dataDone);
            }

        /* PIO mode */

        if (intSts & (IRQ_BRR |IRQ_BWR))
            {

           /*
            * BRR flag indicates valid data exists in the host side buffer.
            * BWR flag indicates space is available for write data.
            * These 2 bits will be set again after write 1 to clear BRR and
            * BWR becasue of interrupt source. It causes to enter ISR
            * repeatedly. So disable BRR and BWR interrupt here. Enable
            * BRR and BWR interrupt again before issue transfer command.
            */

            CSR_CLRBIT_4 (pDev, SDHC_IRQSIGEN, (IRQ_BRR | IRQ_BWR));
            pDrvCtrl->intMask &= ~ (IRQ_BRR | IRQ_BWR);

            if ((pDrvCtrl->tuningInProgress == 1) && (intSts & IRQ_BRR))
                {
                SDHC_DBG (SDHC_DBG_IRQ, "semGive tuningDone!\n", 0, 0, 0, 0);
                (void) semGive (pDrvCtrl->sdHostCtrl.tuningDone);
                return;
                }

            (void)semGive (pDrvCtrl->sdHostCtrl.pioReady);
            }

        /* other errors */

        if (pDrvCtrl->quirky & SDHC_QUIRKY_ACMD12)
            {
            if (intSts & IRQ_AC12E)
                {
                SDHC_DBG (SDHC_DBG_ERR, "Auto CMD12 error\n", 0, 0, 0, 0);
                pSdCmd->cmdErr |= SDMMC_CMD_ERR_ACMD12;
                }
            }
        } /* while (1) */

    /*
     * Now that we've consumed/drained all currently pending events,
     * unmask interrupts again so that we can catch anything new that
     * happens in the future.
     */

    CSR_WRITE_4(pDev, SDHC_IRQSIGEN, pDrvCtrl->intMask);

    return;
    }

/*******************************************************************************
*
* sdhcCtlrIsr - interrupt service routine
*
* This routine handles interrupts of eSDHC.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

void sdhcCtrlIsr
    (
    VXB_DEV_HANDLE pDev
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    UINT32 intSts = 0;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);

    /*
     * read interrupt status
     * We need to check that an event is actually pending. This is necessary in the
     * case where we're using legacy PCI INTx interrupts, because the SDHC
     * controller's interrupt pin might be shared with another device. If the
     * _other_ device raises an interrupt and the SDHC controller is idle,
     * then this function will still be called, but there won't be any work for us
     * to do, so we should just exit.
     * If we're using MSI interrupts instead of INTx, then this is redundant,
     * but it should have no bad side effects.
     */

    intSts = CSR_READ_4 (pDev, SDHC_IRQSTAT);
#ifdef _WRS_CONFIG_ITL_SOC_AGILEX5
    vxAtomicSet(&intStatus, intSts);
#endif
    CSR_WRITE_4 (pDev, SDHC_IRQSTAT, intSts);

    /* no event actually pending, just return */

    if (intSts == 0)
        return;

    /*
     * Now check to see if interrupts are masked off. If they are,
     * then we've already dispatched a task-level handler and are processing
     * events and we can just return. Otherwise, we need to proceed and
     * dispatch a task-level event below.
     * We may get here in the INTx case if another device sharing
     * our INTx interrupt pin signals an interrupt before the task is finished
     * handling everything.
     */

    if (CSR_READ_4 (pDev, SDHC_IRQSIGEN) == 0)
        return;

    /*
     * Mask off further interrupt events. DO NOT CLEAR IRQSIGSTAT.
     * We want any pending events in the IRQSTAT register to remain set
     * so that we can handle them later in task context - writing to
     * IRQSTATEN will clear all the IRQSTAT bits, which will break that.
     * This is mainly needed in the case where we're using legacy INTx
     * interrupts, because we need to de-assert the INTx pin so that we can
     * get out of this ISR and into a task so we can process pending events.
     */

    CSR_WRITE_4 (pDev, SDHC_IRQSIGEN, 0);

    /*
     * Queue up tIsrTask to handle the actual interrupt servicing work.
     * We don't necessarily need any special flags here to check if a
     * job has already been dispatched; we already know there's work
     * to do because interrupts were unmasked when we entered this
     * routine.
     */

    isrDeferJobAdd (pDrvCtrl->queueId, &pDrvCtrl->isrDefer);

    return;
    }

/*******************************************************************************
*
* sdhcCmdPrepare - prepare the command to be sent
*
* This function prepares the command to be sent.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlCmdPrepare
    (
    VXB_DEV_HANDLE pDev,     /* vxbus device ID */
    SD_CMD       * pSdCmd    /* pointer to command */
    )
    {
    SD_HOST_CTRL * pSdHostCtrl;
    SDHC_DEV_CTRL * pDrvCtrl;
    UINT32 xferTyp = 0;
    UINT32 i = 0;
    UINT32 j = 0;
    UINT32 mask = PRSSTAT_CIHB;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ERROR;
    else
        pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));

    /* check PRSSTAT[CDIHB] and PRSSTAT[CIHB] before writing to XFERTYP */

    if ((pSdCmd->hasData != 0) || (pSdCmd->rspType & SDMMC_CMD_RSP_BUSY))
        mask |= PRSSTAT_CDIHB;

    /* CMD0, CMD12, CMD13 doesn't need to wait for data */
    if (pSdCmd->cmdIdx == SDMMC_CMD_GO_IDLE_STATE ||
        pSdCmd->cmdIdx == SDMMC_CMD_STOP_TRANSMISSION ||
        pSdCmd->cmdIdx == SDMMC_CMD_SEND_STATUS)
        mask &= (UINT32)~PRSSTAT_CDIHB;

    while (CSR_READ_4(pDev, SDHC_PRSSTAT) & mask)
        {
        vxbMsDelay(1);
        i++;
        if (i > 150)
            {
            SDHC_DBG (SDHC_DBG_ERR, "CIHB/CDIHB busy\n", 3, 4, 5, 6);
            if ((CSR_READ_4(pDev, SDHC_PRSSTAT) & mask) == PRSSTAT_CIHB)
                {
                CSR_SETBIT_4 (pDev, SDHC_SYSCTL, SYSCTL_RSTC);
                while (CSR_READ_4 (pDev, SDHC_SYSCTL) & SYSCTL_RSTC)
                    {
                    vxbMsDelay(1);
                    j++;
                    if (j > 1000)
                        {
                        SDHC_DBG (SDHC_DBG_ERR, "CMD RESET fail\n", 3, 4, 5, 6);
                        return ERROR;
                        }
                    }

                SDHC_DBG (SDHC_DBG_ERR, "CMD RESET OK\n", 3, 4, 5, 6);
                }
            j= 0;
            if ((CSR_READ_4(pDev, SDHC_PRSSTAT) & mask) == PRSSTAT_CDIHB)
                {
                CSR_SETBIT_4 (pDev, SDHC_SYSCTL, SYSCTL_RSTD);
                while (CSR_READ_4 (pDev, SDHC_SYSCTL) & SYSCTL_RSTD)
                    {
                    vxbMsDelay(1);
                    j++;
                    if (j > 1000)
                        {
                        SDHC_DBG (SDHC_DBG_ERR, "DATA RESET fail\n", 3, 4, 5, 6);
                        return ERROR;
                        }
                    }
                SDHC_DBG (SDHC_DBG_ERR, "DATA RESET OK\n", 3, 4, 5, 6);
                }
            }
        }

    i = 0;

    /* CMD0, CMD12, CMD13 doesn't need to wait for DLA */
    if (pSdCmd->cmdIdx != SDMMC_CMD_GO_IDLE_STATE &&
        pSdCmd->cmdIdx != SDMMC_CMD_STOP_TRANSMISSION &&
        pSdCmd->cmdIdx != SDMMC_CMD_SEND_STATUS)
        mask = PRSSTAT_DLA;

    while (CSR_READ_4(pDev, SDHC_PRSSTAT) & mask)
        {
        vxbUsDelay(1);
        i++;
        if (i > 150000)
            {
            SDHC_DBG (SDHC_DBG_ERR, "DLA busy\n", 3, 4, 5, 6);
            return ERROR;
            }
        }

    /* setup command */

    xferTyp |= XFERTYP_CMDINX (pSdCmd->cmdIdx);

    /* setup command fields */

    if (pSdCmd->rspType & SDMMC_CMD_RSP_CMDIDX)
        xferTyp |= XFERTYP_CICEN;

    if (pSdCmd->rspType & SDMMC_CMD_RSP_CRC)
        xferTyp |= XFERTYP_CCCEN;

    if (pSdCmd->rspType & SDMMC_CMD_RSP_LEN136)
        xferTyp |= XFERTYP_RSPTYP_LEN136;
    else if (pSdCmd->rspType & SDMMC_CMD_RSP_BUSY)
        xferTyp |= XFERTYP_RSPTYP_LEN48B;
    else if (pSdCmd->rspType & SDMMC_CMD_RSP_LEN48)
        xferTyp |= XFERTYP_RSPTYP_LEN48;

    /* setup data fields */

    if (pSdCmd->hasData)
        {
        xferTyp |= XFERTYP_DATA_PRESENT;

        /* direction */

        if (pSdCmd->cmdData.isRead == TRUE)
            {
            xferTyp |= XFERTYP_DIR_READ;
            }

        if (pSdHostCtrl->dmaMode != SDHC_DMA_MODE_PIO)
            xferTyp |= XFERTYP_DMAEN;

        /* multipule block transfer */

        if (pSdCmd->cmdData.blkNum > 1)
            {
            xferTyp |= (XFERTYP_MULTI_BLK | XFERTYP_BCEN);

            /* automatically issue CMD12 */

            if ((pSdCmd->rspType & SDMMC_CMD_AUTOCMD12) != 0x0)
                xferTyp |= XFERTYP_AC12EN;
            }

        if (pSdHostCtrl->dmaMode == SDHC_DMA_MODE_SDMA)
            {
            pDrvCtrl->sdmaAddr = (UINT32) \
                    (pSdHostCtrl->sdHostDmaMap->fragList[0].frag);

            CSR_WRITE_4 (pDev, SDHC_DSADDR, pDrvCtrl->sdmaAddr);

            /* save SDMA next address and boundary */
            pDrvCtrl->sdmaAddrBound = (UINT32) \
                    (1 << (((CSR_READ_4 (pDev, SDHC_BLKATTR) >> \
                    BLKBUFFER_SIZE_SHIFT) & 0x7) + 12));
            pDrvCtrl->sdmaAddr += pDrvCtrl->sdmaAddrBound;
            pDrvCtrl->sdmaAddr = (UINT32) ROUND_DOWN (pDrvCtrl->sdmaAddr, \
                                                 pDrvCtrl->sdmaAddrBound);
            }
        else if (pSdHostCtrl->dmaMode == SDHC_DMA_MODE_ADMA2)
            {
            sdhcAdma2DescFill(pDev, pSdHostCtrl->dmaHdl, \
                              pSdHostCtrl->sdHostDmaMap);
            }
        else   /* PIO mode */
            {
            pDrvCtrl->pioDesc.bufOri = (char *) pSdCmd->cmdData.buffer;
            pDrvCtrl->pioDesc.bufCur = (char *) pSdCmd->cmdData.buffer;
            pDrvCtrl->pioDesc.blkSize =
                      pSdCmd->cmdData.blkSize * pSdCmd->cmdData.blkNum;
            CSR_SETBIT_4 (pDev, SDHC_IRQSIGEN, (IRQ_BRR | IRQ_BWR));
            CSR_SETBIT_4 (pDev, SDHC_IRQSTATEN, (IRQ_BRR | IRQ_BWR));
            pDrvCtrl->intMask |= (IRQ_BRR | IRQ_BWR);
            }

        CSR_WRITE_4 (pDev, SDHC_BLKATTR,
                     (pSdCmd->cmdData.blkNum << BLKATTR_CNT_SHIFT) |
                     pSdCmd->cmdData.blkSize |
                     (BLKBUFFER_SIZE_512K << BLKBUFFER_SIZE_SHIFT));

        /* setup transfer timeouts */

        CSR_CLRBIT_4 (pDev, SDHC_IRQSTATEN, IRQ_DTOE);
        CSR_CLRBIT_4 (pDev, SDHC_SYSCTL, SYSCTL_DTOCV_MASK);
        CSR_SETBIT_4 (pDev, SDHC_SYSCTL, \
                      SYSCTL_DTOCV_MAX << SYSCTL_DTOCV_SHIFT);
        CSR_SETBIT_4 (pDev, SDHC_IRQSTATEN, IRQ_DTOE);
        }

    /* issue the command */

    CSR_WRITE_4 (pDev, SDHC_CMDARG, (UINT32)pSdCmd->cmdArg);
    CSR_WRITE_4 (pDev, SDHC_XFERTYP, xferTyp);

    return OK;
    }

/*******************************************************************************
*
* sdhcCmdIssue - issue the command
*
* This function issues the command.
*
* RETURNS: OK or ERROR
*
* ERRNO: ENXIO, ENOLCK and EIO
*
* \NOMANUAL
*/

STATUS sdhcCtrlCmdIssue
    (
    VXB_DEV_HANDLE pDev,     /* vxbus device ID */
    SD_CMD       * pSdCmd    /* pointer to command */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_CTRL * pSdHostCtrl;
    SD_HARDWARE * pSdHardWare;
    UINT32 pollTimeOut = 0;
    STATUS rc = OK;
    BOOL hasNextCmd = FALSE;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        {
        (void)errnoSet (ENXIO);
        return ERROR;
        }

    pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));

    pSdHardWare = (SD_HARDWARE *)GET_HARDWAREINFO (pSdHostCtrl->sdHostSpec.childDev);
    if (pSdHardWare == NULL)
        {
        (void)errnoSet (ENXIO);
        return ERROR;
        }

    if (semTake (pSdHostCtrl->hostDevSem, WAIT_FOREVER) == ERROR)
        {
        (void)errnoSet (ENOLCK);
        return ERROR;
        }

    do
        {
        pSdHostCtrl->pCurCmd = pSdCmd;

        if (pSdCmd->cmdIdx != SDMMC_CMD_SEND_STATUS)
            pSdHostCtrl->cmdTick = tickGet();

        if ((pSdHostCtrl->dmaMode != SDHC_DMA_MODE_PIO) && (pSdCmd->hasData))
            {
            rc = vxbDmaBufMapLoad (pDev, pDrvCtrl->sdHostCtrl.sdHostDmaTag,
                                   pDrvCtrl->sdHostCtrl.sdHostDmaMap,
                                   pSdCmd->cmdData.buffer,
                                   pSdCmd->cmdData.blkNum * pSdCmd->cmdData.blkSize,
                                   0);
            if (rc == ERROR)
                {
                (void)semGive (pSdHostCtrl->hostDevSem);
                (void)errnoSet (EIO);
                return (rc);
                }

            if (pSdCmd->cmdData.isRead == TRUE)
                {
                (void)vxbDmaBufMapSync (pDev,
                            pSdHostCtrl->sdHostDmaTag,
                            pSdHostCtrl->sdHostDmaMap,
                            0,
                            pSdCmd->cmdData.blkNum * pSdCmd->cmdData.blkSize,
                            _VXB_DMABUFSYNC_DMA_PREREAD);
                }
            else
                {
                (void)vxbDmaBufMapSync (pDev,
                                pSdHostCtrl->sdHostDmaTag,
                                pSdHostCtrl->sdHostDmaMap,
                                0,
                                pSdCmd->cmdData.blkNum * pSdCmd->cmdData.blkSize,
                                _VXB_DMABUFSYNC_DMA_PREWRITE);
                }
            }

        if (pSdHostCtrl->sdHostOps.vxbSdCmdPrepare != NULL)
            {
            if (pSdHostCtrl->sdHostOps.vxbSdCmdPrepare (pDev, pSdCmd) == ERROR)
                {
                if ((pSdHostCtrl->dmaMode != SDHC_DMA_MODE_PIO) &&
                    (pSdCmd->hasData))
                    {
                    (void)vxbDmaBufMapUnload(pSdHostCtrl->sdHostDmaTag,
                                             pSdHostCtrl->sdHostDmaMap);
                    }
                (void)errnoSet (EIO);
                (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                return ERROR;
                }
            }

        if (pDrvCtrl->sdHostCtrl.polling == TRUE)
            {
            while (pollTimeOut < SDHC_PIO_TIMEOUT)
                {
                if ((CSR_READ_4 (pDev, SDHC_IRQSTAT) & IRQ_CC) != 0x0)
                    break;
                pollTimeOut++;
                }
            if (pollTimeOut >= SDHC_PIO_TIMEOUT)
                rc = ERROR;
            }
        else
            {
            if ((pDrvCtrl->tuningInProgress == 1) &&
               (pSdCmd->cmdIdx == SDMMC_CMD_SEND_TUNING_HS200 ||
                pSdCmd->cmdIdx == SDMMC_CMD_SEND_TUNING))
                {
                rc = semTake(pDrvCtrl->sdHostCtrl.tuningDone,
                            ((sysClkRateGet() * SDMMC_DATA_WAIT_MS) / 1000 + 1));

                SDHC_DBG (SDHC_DBG_RW, "CMD%d tuning done received\n",
                          pSdCmd->cmdIdx, 0, 0, 0);

                if (pSdCmd->cmdData.buffer == NULL)
                    {
                    while ((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_BREN))
                        {
                        (void) PIO_READ_4 (pDev, SDHC_DATPORT);
                        }
                    (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                    return OK;
                    }
                }
            else
                {
                rc = semTake (pDrvCtrl->sdHostCtrl.cmdDone,
                              ((sysClkRateGet() * SDMMC_CMD_WAIT_MS) / 1000 + 1));
                }
            }

        if (rc == ERROR)
            {
            SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd semTake timeout\n",
                      pSdCmd->cmdIdx, 0, 0, 0);
            pSdCmd->cmdErr |= SDMMC_CMD_ERR_TIMEOUT;
            if ((pSdHostCtrl->dmaMode != SDHC_DMA_MODE_PIO) && (pSdCmd->hasData))
                {
                (void) vxbDmaBufMapUnload(pSdHostCtrl->sdHostDmaTag,
                                          pSdHostCtrl->sdHostDmaMap);
                }
            (void)errnoSet (EIO);
            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
            return ERROR;
            }
        else
            {
            SDHC_DBG (SDHC_DBG_RW, "CMD%d cmdDone received\n",
                      pSdCmd->cmdIdx, 0, 0, 0);

            if (pSdCmd->cmdErr != 0)
                {
                if (((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_CIHB) != 0x0) ||
                    ((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_CDIHB) != 0x0) ||
                    (((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_DLA) != 0x0)))
                    {
                    CSR_SETBIT_4(pDev, SDHC_SYSCTL, SYSCTL_RSTC | SYSCTL_RSTD);
                    while ((CSR_READ_4(pDev, SDHC_SYSCTL) &
                        (SYSCTL_RSTC | SYSCTL_RSTD)) != 0);
                    }

                if ((pSdHostCtrl->dmaMode != SDHC_DMA_MODE_PIO) &&
                    (pSdCmd->hasData))
                    {
                    (void)vxbDmaBufMapUnload(pSdHostCtrl->sdHostDmaTag,
                                             pSdHostCtrl->sdHostDmaMap);
                    }
                (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                (void)errnoSet (EIO);
                return ERROR;
                }
            }

        if (pSdCmd->rspType != SDMMC_CMD_RSP_NONE)
            {
            if (pSdCmd->rspType & SDMMC_CMD_RSP_LEN136)
                {
                UINT32 cmdRsp[4];

                cmdRsp[0] = CSR_READ_4 (pDev, SDHC_CMDRSP0);
                cmdRsp[1] = CSR_READ_4 (pDev, SDHC_CMDRSP1);
                cmdRsp[2] = CSR_READ_4 (pDev, SDHC_CMDRSP2);
                cmdRsp[3] = CSR_READ_4 (pDev, SDHC_CMDRSP3);

                pSdCmd->cmdRsp[0] = be32toh ((cmdRsp[3] << 8) | (cmdRsp[2] >> 24));
                pSdCmd->cmdRsp[1] = be32toh ((cmdRsp[2] << 8) | (cmdRsp[1] >> 24));
                pSdCmd->cmdRsp[2] = be32toh ((cmdRsp[1] << 8) | (cmdRsp[0] >> 24));

                /* eliminate BE/LE confusion for Coverity */
                /* coverity[operator_confusion] */
                pSdCmd->cmdRsp[3] = be32toh (cmdRsp[0] << 8);
                }
            else
                {
                pSdCmd->cmdRsp[0] = CSR_READ_4 (pDev, SDHC_CMDRSP0);
                if (pSdHardWare->isMmc == TRUE)
                    {
                    if ((pSdCmd->rspType == SDMMC_CMD_RSP_R1) ||
                        (pSdCmd->rspType == SDMMC_CMD_RSP_R1B))
                        {
                        if (pSdCmd->cmdRsp[0] & CARD_ERROR_STATUS)
                            {
                            SDHC_DBG (SDHC_DBG_ERR,
                                 "Card is in error status. CMD%d status = 0x%x\n",
                                 pSdCmd->cmdIdx, pSdCmd->cmdRsp[0], 0, 0);
                            if ((pSdHostCtrl->dmaMode != SDHC_DMA_MODE_PIO) &&
                                (pSdCmd->hasData))
                                {
                                (void)vxbDmaBufMapUnload(pSdHostCtrl->sdHostDmaTag,
                                                         pSdHostCtrl->sdHostDmaMap);
                                }
                            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                            (void)errnoSet (EIO);
                            return ERROR;
                            }
                        }
                    }
                else
                    {
                    if ((pSdCmd->cmdIdx != SD_CMD_SEND_RELATIVE_ADDR) &&
                         (pSdCmd->cmdIdx != SD_CMD_SEND_IF_COND) &&
                        ((pSdCmd->rspType == SDMMC_CMD_RSP_R1) ||
                         (pSdCmd->rspType == SDMMC_CMD_RSP_R1B)))
                        {
                        if (pSdCmd->cmdRsp[0] & CARD_ERROR_STATUS)
                            {
                            SDHC_DBG (SDHC_DBG_ERR,
                                     "Card in error. CMD%d status = 0x%x\n",
                                     pSdCmd->cmdIdx, pSdCmd->cmdRsp[0], 0, 0);
                            if ((pSdHostCtrl->dmaMode != SDHC_DMA_MODE_PIO) &&
                                (pSdCmd->hasData))
                                {
                                (void)vxbDmaBufMapUnload(pSdHostCtrl->sdHostDmaTag,
                                                         pSdHostCtrl->sdHostDmaMap);
                                }
                            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                            (void)errnoSet (EIO);
                            return ERROR;
                            }
                        }
                    }
                }
            }

        if ((pSdHostCtrl->dmaMode == SDHC_DMA_MODE_PIO) && pSdCmd->hasData)
            {
            /* Data transfer in PIO mode */

            if (pSdCmd->cmdIdx != SDMMC_CMD_SEND_TUNING_HS200 &&
                pSdCmd->cmdIdx != SDMMC_CMD_SEND_TUNING)
                {
                rc = semTake (pSdHostCtrl->pioReady,
                              ((sysClkRateGet() * SDMMC_DATA_WAIT_MS) / 1000 + 1));
                }
            if (rc == ERROR)
                {
                SDHC_DBG (SDHC_DBG_ERR, "CMD%d PIO timeout\n",
                          pSdCmd->cmdIdx, 0, 0, 0);
                pSdCmd->cmdErr |= SDMMC_DATA_ERR_TIMEOUT;
                (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                (void)errnoSet (EIO);
                return ERROR;
                }
            else
                {

                if (pSdCmd->cmdData.isRead == TRUE)
                    {
                    SDHC_DBG (SDHC_DBG_RW, "pio read\n", 0, 0, 0, 0);
                    rc = sdhcCtrlPioRead (pDev);
                    if (rc == ERROR)
                        {
                        (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                        (void)errnoSet (EIO);
                        return ERROR;
                        }
                    }
                else
                    {
                    SDHC_DBG (SDHC_DBG_RW, "pio write\n", 0, 0, 0, 0);
                    rc = sdhcCtrlPioWrite (pDev);
                    if (rc == ERROR)
                        {
                        (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                        (void)errnoSet (EIO);
                        return ERROR;
                        }
                    }
                }
            }

        if ((pSdCmd->rspType & SDMMC_CMD_RSP_BUSY) && (pSdCmd->hasData == FALSE))
            {

            /* Need not check return status */
            /* NO data tran but has data int ack. such as sdCmdSelectCard */

            /*
             * Some boards such as fsl_imx6 SDMMC controller does not generate
             * data done interrupt for R1B type command. For this kind of
             * board, we use semTake() with NO_WAIT to save boot time.
             */

            if(pDrvCtrl->noR1bDataDone)
                {
                (void)semTake (pDrvCtrl->sdHostCtrl.dataDone, NO_WAIT);
                }
            else
                {
                (void)semTake (pDrvCtrl->sdHostCtrl.dataDone,
                                sysClkRateGet() / 4);
                }

            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
            return OK;
            }

        if ((pSdHostCtrl->dmaMode != SDHC_DMA_MODE_PIO) && pSdCmd->hasData)
            {
            /* Data transfer in DMA mode */

            rc = semTake (pDrvCtrl->sdHostCtrl.dataDone,
                          ((sysClkRateGet() * SDMMC_DATA_WAIT_MS) / 1000 + 1));
            if (rc == ERROR)
                {
                SDHC_DBG (SDHC_DBG_ERR, "CMD%d data semTake timeout\n",
                          pSdCmd->cmdIdx, 0, 0, 0);
                pSdCmd->dataErr |= SDMMC_DATA_ERR_TIMEOUT;
                (void)errnoSet (ENOLCK);
                }
            else
                {
                if (!pSdCmd->cmdData.isRead)
                    {
                    (void) vxbDmaBufMapSync (pDev,
                                pSdHostCtrl->sdHostDmaTag,
                                pSdHostCtrl->sdHostDmaMap,
                                0,
                                pSdCmd->cmdData.blkNum * pSdCmd->cmdData.blkSize,
                                _VXB_DMABUFSYNC_DMA_POSTWRITE);
                    }
                else
                    {
                    (void) vxbDmaBufMapSync (pDev,
                                pSdHostCtrl->sdHostDmaTag,
                                pSdHostCtrl->sdHostDmaMap,
                                0,
                                pSdCmd->cmdData.blkNum * pSdCmd->cmdData.blkSize,
                                _VXB_DMABUFSYNC_DMA_POSTREAD);
                    }

                (void) vxbDmaBufMapUnload(pSdHostCtrl->sdHostDmaTag,
                                          pSdHostCtrl->sdHostDmaMap);

                if (pSdCmd->dataErr != 0)
                    {
                    if (((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_CIHB) != 0x0) ||
                        ((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_CDIHB) != 0x0) ||
                        (((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_DLA) != 0x0)))
                        {
                        CSR_SETBIT_4(pDev, SDHC_SYSCTL, SYSCTL_RSTC | SYSCTL_RSTD);
                        while ((CSR_READ_4(pDev, SDHC_SYSCTL) &
                            (SYSCTL_RSTC | SYSCTL_RSTD)) != 0);
                        }
                    }
                }

            }

        if (pSdCmd->cmdErr & SDMMC_CMD_ERR_ACMD12)
            {
            SDHC_DBG (SDHC_DBG_ERR, "handle Auto CMD12 error\n", 0, 0, 0, 0);

            if (((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_CIHB) != 0x0) ||
                ((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_CDIHB) != 0x0) ||
                (((CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_DLA) != 0x0)))
                {
                CSR_SETBIT_4(pDev, SDHC_SYSCTL, SYSCTL_RSTC | SYSCTL_RSTD);
                while ((CSR_READ_4(pDev, SDHC_SYSCTL) &
                       (SYSCTL_RSTC | SYSCTL_RSTD)) != 0);
                }
           }

        if ((pSdCmd->dataErr != 0x0) || (pSdCmd->cmdErr != 0x0))
            {
            SDHC_DBG (SDHC_DBG_ERR, "CMD%d dataErr = 0x%x cmdErr = 0x%x\n",
                      pSdCmd->cmdIdx, pSdCmd->dataErr, pSdCmd->cmdErr, 0);
            (void)errnoSet (EIO);
            rc = ERROR;
            }

        hasNextCmd = (BOOL) SDMMC_CMD_IS_NOT_LAST_CMD(pSdCmd->cmdFlag);
        if (hasNextCmd)
            {
            pSdCmd ++;
            }
        }
    while (hasNextCmd);

    (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
    return (rc);
    }

/*******************************************************************************
*
* sdhcBusWidthSetup - setup the bus width
*
* This function setups the bus width.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlBusWidthSetup
    (
    VXB_DEV_HANDLE pDev,     /* vxbus device ID */
    UINT32         width     /* bus width */
    )
    {
    CSR_CLRBIT_4 (pDev, SDHC_PROCTL, PROCTL_DTW_4BIT | PROCTL_DTW_8BIT);

    switch (width)
        {
        case SDMMC_BUS_WIDTH_1BIT:
            CSR_SETBIT_4 (pDev, SDHC_PROCTL, PROCTL_DTW_1BIT);
            break;
        case SDMMC_BUS_WIDTH_4BIT:
            CSR_SETBIT_4 (pDev, SDHC_PROCTL, PROCTL_DTW_4BIT);
            break;
        case SDMMC_BUS_WIDTH_8BIT:
            CSR_SETBIT_4 (pDev, SDHC_PROCTL, PROCTL_DTW_8BIT);
            break;
        default:
            break;
        }
    return OK;
    }

/*******************************************************************************
*
* sdhcCtrlCardInsertSts - check if card is inserted or not
*
* This function checks if card is inserted or not.
*
* RETURNS: TRUE if card is inserted otherwise FALSE
*
* ERRNO: N/A
*
* \NOMANUAL
*/

BOOL sdhcCtrlCardInsertSts
    (
    VXB_DEV_HANDLE pDev     /* vxbus device ID */
    )
    {
    UINT32 prsStat;
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_CTRL * pSdHostCtrl;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        {
        return FALSE;
        }
    else
        pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));

    if (!pSdHostCtrl->polling)
        {
        if (pDrvCtrl->attrData.cdtype == SDHC_CD_SOFTWARE)
            {
            while ((tickGet() - pSdHostCtrl->cmdTick) < sysClkRateGet())
                {
                (void) taskDelay (1);
                }

            pSdHostCtrl->cmdTick = tickGet();
            if ((pSdHostCtrl->attached == TRUE) &&
                (sdCmdSendStatus(pSdHostCtrl->sdHostSpec.childDev, &prsStat)
                != OK))
                {
                pSdHostCtrl->isInsert = FALSE;
                }
            else
                {
                pSdHostCtrl->isInsert = TRUE;
                }
            }
        else
            {
            /* Need not check return status */

            (void)semTake(pSdHostCtrl->devChange, WAIT_FOREVER);
            if (pDrvCtrl->embedded == FALSE)
                {
                prsStat = CSR_READ_4 (pDev, SDHC_PRSSTAT);
                if (prsStat & PRSSTAT_CINS)
                    pSdHostCtrl->isInsert = TRUE;
                else
                    pSdHostCtrl->isInsert = FALSE;
                }
            else
                {
                pSdHostCtrl->isInsert = TRUE;
                }
            }
        }
    else
        {
        while (1)
            {
            (void)taskDelay (10 * sysClkRateGet());
            prsStat = CSR_READ_4 (pDev, SDHC_PRSSTAT);

            if (prsStat & PRSSTAT_CINS)
                {
                if (pSdHostCtrl->isInsert != TRUE)
                    {
                    pSdHostCtrl->isInsert = TRUE;
                    break;
                    }
                }
            else
                {
                if (pSdHostCtrl->isInsert != FALSE)
                    {
                    pSdHostCtrl->isInsert = FALSE;
                    break;
                    }
                }
            }
        }
    return (pSdHostCtrl->isInsert);
    }

/*******************************************************************************
*
* sdhcCardWpCheck - check if card is write protected
*
* This function checks if card is write protected.
*
* RETURNS: TRUE if card is write protected or FALSE
*
* ERRNO: N/A
*
* \NOMANUAL
*/

BOOL sdhcCtrlCardWpCheck
    (
    VXB_DEV_HANDLE pDev     /* vxbus device ID */
    )
    {
    BOOL wp;
    SDHC_DEV_CTRL * pDrvCtrl;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return FALSE;

    if (CSR_READ_4 (pDev, SDHC_PRSSTAT) & PRSSTAT_WPSPL)
        wp = FALSE;
    else
        wp = TRUE;

    return  (wp);
    }

/*******************************************************************************
*
* sdhcPioRead - read data from SD/MMC card using PIO
*
* This function reads data from SD/MMC card using PIO.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlPioRead
    (
    VXB_DEV_HANDLE   pDev     /* vxbus device ID */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    UINT32 blkSize;
    char * buffer;
    UINT32 data = 0;
    UINT32 i;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ERROR;

    blkSize = pDrvCtrl->pioDesc.blkSize;
    buffer = pDrvCtrl->pioDesc.bufCur;

    for (i = 0; i < blkSize; i += 4)
        {

        /*
         * BREN -Buffer Read Enable. This status bit is used for non-DMA read
         * transfers. This read only flag indicates that valid data exists in
         * the host side buffer. If this bit is set, valid data greater than
         * the watermark level exist in the buffer.
         */

        while (!(CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_BREN));

        data = PIO_READ_4 (pDev, SDHC_DATPORT);

        *(UINT32 *)buffer = data;
        buffer += 4;
        }

    pDrvCtrl->pioDesc.bufCur = buffer;

    SDHC_DBG (SDHC_DBG_RW, "bufOri = %p bufCur = %p\n",
              pDrvCtrl->pioDesc.bufOri, pDrvCtrl->pioDesc.bufCur, 0, 0);

    return OK;
    }

/*******************************************************************************
*
* sdhcPioWrite - write data to SD/MMC card using PIO
*
* This function writes data to SD/MMC card using PIO.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlPioWrite
    (
    VXB_DEV_HANDLE   pDev     /* vxbus device ID */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    UINT32 blkSize;
    char * buffer;
    UINT32 data;
    UINT32 i;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ERROR;

    blkSize = pDrvCtrl->pioDesc.blkSize;
    buffer = pDrvCtrl->pioDesc.bufCur;

    for (i = 0; i < blkSize; i += 4)
        {

        /*
         * BWEN - Buffer Write Enable. This status bit is used for non-DMA
         * write transfers. This read only flag indicates if space is available
         * for write data. If this bit is set, valid data greater than the
         * watermark level can be written to the buffer.
         */

        while (!(CSR_READ_4(pDev, SDHC_PRSSTAT) & PRSSTAT_BWEN));

        data = *(UINT32 *)buffer;
        buffer += 4;

        PIO_WRITE_4 (pDev, SDHC_DATPORT, data);

        }

    pDrvCtrl->pioDesc.bufCur = buffer;

    SDHC_DBG (SDHC_DBG_RW, "bufOri = %p bufCur = %p\n",
              pDrvCtrl->pioDesc.bufOri, pDrvCtrl->pioDesc.bufCur, 0, 0);

    return OK;
    }

/*******************************************************************************
*
* sdhcCmdIssuePoll - issue the command using polling mode
*
* This function issues the command using polling mode.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlCmdIssuePoll
    (
    VXB_DEV_HANDLE   pDev,     /* vxbus device ID */
    SD_CMD         * pSdCmd    /* pointer to command */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    volatile UINT32 intSts;
    UINT32 pollTimeOut = 0;
    STATUS rc = OK;
    BOOL hasNextCmd = FALSE;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ERROR;

    if (semTake (pDrvCtrl->sdHostCtrl.hostDevSem, WAIT_FOREVER) == ERROR)
        return ERROR;
    do
        {

        if (pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdCmdPrepare != NULL)
            {
            if (pDrvCtrl->sdHostCtrl.sdHostOps.vxbSdCmdPrepare (pDev, pSdCmd)
                == ERROR)
                {
                (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                return ERROR;
                }
            }

        /* wait for command to complete */

        if (pSdCmd->cmdIdx != SDMMC_CMD_STOP_TRANSMISSION)
            {
            while (pollTimeOut < POLL_TIMEOUT)
                {
                if ((CSR_READ_4 (pDev, SDHC_IRQSTAT) & IRQ_CC) != 0x0)
                    break;
                pollTimeOut++;
                }
            }

        if (pollTimeOut >= POLL_TIMEOUT)
            {
            SDHC_DBG (SDHC_DBG_ERR, "wait for command timeout\n", 0, 0, 0, 0);
            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
            return ERROR;
            }

        intSts = CSR_READ_4 (pDev, SDHC_IRQSTAT);

        if (intSts & IRQ_CMD_ERR)
            {
            if (intSts & IRQ_CIE)
                {
                SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd index error\n",
                          pSdCmd->cmdIdx, 0, 0, 0);
                pSdCmd->cmdErr |= SDMMC_CMD_ERR_PROTOCOL;
                }

            if (intSts & IRQ_CEBE)
                {
                SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd end of bit error\n",
                          pSdCmd->cmdIdx, 0, 0, 0);
                pSdCmd->cmdErr |= SDMMC_CMD_ERR_PROTOCOL;
                }

            if (intSts & IRQ_CCE)
                {
                SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd crc error\n",
                          pSdCmd->cmdIdx, 0, 0, 0);
                pSdCmd->cmdErr |= SDMMC_CMD_ERR_PROTOCOL;
                }

            if (intSts & IRQ_CTOE)
                {
                SDHC_DBG (SDHC_DBG_ERR, "CMD%d cmd timeout error\n",
                          pSdCmd->cmdIdx, 0, 0, 0);
                pSdCmd->cmdErr |= SDMMC_CMD_ERR_TIMEOUT;
                }

            CSR_WRITE_4 (pDev, SDHC_IRQSTAT, IRQ_ALL);
            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
            return ERROR;
            }

        if (pSdCmd->rspType != SDMMC_CMD_RSP_NONE)
            {
            if (pSdCmd->rspType & SDMMC_CMD_RSP_LEN136)
                {
                UINT32 cmdRsp[4];

                cmdRsp[0] = CSR_READ_4 (pDev, SDHC_CMDRSP0);
                cmdRsp[1] = CSR_READ_4 (pDev, SDHC_CMDRSP1);
                cmdRsp[2] = CSR_READ_4 (pDev, SDHC_CMDRSP2);
                cmdRsp[3] = CSR_READ_4 (pDev, SDHC_CMDRSP3);

                pSdCmd->cmdRsp[0] = be32toh ((cmdRsp[3] << 8) | (cmdRsp[2] >> 24));
                pSdCmd->cmdRsp[1] = be32toh ((cmdRsp[2] << 8) | (cmdRsp[1] >> 24));
                pSdCmd->cmdRsp[2] = be32toh ((cmdRsp[1] << 8) | (cmdRsp[0] >> 24));

            /* eliminate BE/LE confusion for Coverity */
            /* coverity[operator_confusion] */
                pSdCmd->cmdRsp[3] = be32toh (cmdRsp[0] << 8);
                }
            else
                {
                pSdCmd->cmdRsp[0] = CSR_READ_4 (pDev, SDHC_CMDRSP0);
                }
            }

        if (pSdCmd->hasData)
            {
            while (1)
                {
                intSts = CSR_READ_4 (pDev, SDHC_IRQSTAT);

                if (intSts & IRQ_TC)
                    {
                    SDHC_DBG (SDHC_DBG_IRQ, "CMD%d data complete\n",
                              pSdCmd->cmdIdx, 0, 0, 0);
                    break;
                    }

                if (intSts & IRQ_BRR)
                    {
                    SDHC_DBG (SDHC_DBG_IRQ, "CMD%d data buffer read ready\n",
                              pSdCmd->cmdIdx, 0, 0, 0);

                    if (CSR_READ_4 (pDev, SDHC_PRSSTAT) & PRSSTAT_BREN)
                        {
                        rc = sdhcCtrlPioRead (pDev);
                        if (rc == ERROR)
                            {
                            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                            return (ERROR);
                            }
                        }
                    }

                if (intSts & IRQ_BWR)
                    {
                    SDHC_DBG (SDHC_DBG_IRQ, "CMD%d data buffer write ready\n",
                              pSdCmd->cmdIdx, 0, 0, 0);

                    if (CSR_READ_4 (pDev, SDHC_PRSSTAT) & PRSSTAT_BWEN)
                        {
                        rc = sdhcCtrlPioWrite (pDev);
                        if (rc == ERROR)
                            {
                            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
                            return (ERROR);
                            }
                        }
                    }

                if (intSts & IRQ_DATA_ERR)
                    {
                    if (intSts & IRQ_DEBE)
                        {
                        SDHC_DBG (SDHC_DBG_ERR, "CMD%d data end of bit error\n",
                                  pSdCmd->cmdIdx, 0, 0, 0);
                        pSdCmd->dataErr |= SDMMC_DATA_ERR_PROTOCOL;
                        }

                    if (intSts & IRQ_DCE)
                        {
                        SDHC_DBG (SDHC_DBG_ERR, "CMD%d data crc error\n",
                                  pSdCmd->cmdIdx, 0, 0, 0);
                        pSdCmd->dataErr |= SDMMC_DATA_ERR_PROTOCOL;
                        }

                    if (intSts & IRQ_DTOE)
                        {
                        SDHC_DBG (SDHC_DBG_ERR, "CMD%d data timeout error\n",
                                  pSdCmd->cmdIdx, 0, 0, 0);
                        pSdCmd->dataErr |= SDMMC_DATA_ERR_TIMEOUT;
                        }

                    CSR_WRITE_4 (pDev, SDHC_IRQSTAT, IRQ_ALL);
                    (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);

                    return ERROR;
                    }
                };
            }
        CSR_WRITE_4 (pDev, SDHC_IRQSTAT, IRQ_ALL);

        hasNextCmd = (BOOL) SDMMC_CMD_IS_NOT_LAST_CMD(pSdCmd->cmdFlag);
        if (hasNextCmd)
            {
            pSdCmd ++;
            }
        }
    while(hasNextCmd);

    (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
    return OK;
    }

/*******************************************************************************
*
* sdhcCtrlClkSet - set the clock frequency
*
* This function sets the clock frequency.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlClkSet
    (
    VXB_DEV_HANDLE   pDev,     /* vxbus device ID */
    UINT32           clk       /* clock freqency */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_CTRL  * pSdHostCtrl;
    UINT32 val, dvs, sdClk;
    UINT32 clkSrc;
    UINT32 timeout = 1000;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ERROR;

    pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));

    clkSrc = pDrvCtrl->sdHostCtrl.curClkFreq;

    /* Calculate a divisor for SD clock frequency */

    if ((pSdHostCtrl->version & HOSTVER_SVN_MASK) <= HOSTVER_SVN_20)
        {
        dvs = 1;
        while (clk < (clkSrc / dvs))
            dvs *= 2;

        sdClk = dvs >> 1;
        sdClk <<= SYSCTL_DIV_SHIFT;
        SDHC_DBG (SDHC_DBG_INIT, "%d -> %d dvs=%d sdClk=0x%x\n",
                  clkSrc, clk, dvs, sdClk);
        val = CSR_READ_4(pDev, SDHC_SYSCTL);
        val &= (UINT32)~CLK_SETMASK;
        val &= (UINT32)~SYSCTL_SDCLK_EN;
        }
    else
        {

        /* 10-bit Divided Clock Mode is mandatory for host controller 3.0 */

        dvs = clkSrc / clk;
        if ((clkSrc % clk) != 0)
            dvs++;

        /*
         * Now dvs is our wanted divisor, the below calculation is
         * to change it to register format.
         */

        if (dvs == 1)
            dvs = 0;
        else
            dvs = (dvs / 2) + (dvs % 2);
        if (dvs > 1023)
            {
            SDHC_DBG (SDHC_DBG_ERR, "divider out of scope of 10 bits\n",
                      0, 0, 0, 0);
            return ERROR;
            }
        sdClk = ((dvs & 0xff) << 2) | ((dvs & 0x300) >> 8);
        sdClk <<= SYSCTL_10BITDIV_SHIFT;
        SDHC_DBG (SDHC_DBG_INIT, "%d -> %d dvs=%d sdClk=0x%x\n",
                  clkSrc, clk, dvs, sdClk);
        val = CSR_READ_4(pDev, SDHC_SYSCTL);
        val &= (UINT32)~CLK_SETMASK_10BIT;
        val &= (UINT32)~SYSCTL_SDCLK_EN;
        }

    /* Set SDCLK Frequency Select and Internal Clock Enable*/

    val |= sdClk;
    val |= SYSCTL_INTCLK_EN;
    CSR_WRITE_4(pDev, SDHC_SYSCTL, val);

   /*
    * Check Internal Clock Stable
    *
    * The spec doesn't say anything about this
    * internal clock stable timeout value,
    * so just a pick the value based on expierence.
    */

    while (timeout--)
        {
        val = CSR_READ_4 (pDev, SDHC_SYSCTL);
        if (val & SYSCTL_INTCLK_STABLE)
            break;
        vxbMsDelay(1);
        }

    if (timeout == 0)
        {
        return ERROR;
        }

    /* Set SD Clock On */

    val |= SYSCTL_SDCLK_EN;
    CSR_WRITE_4(pDev, SDHC_SYSCTL, val);

    return OK;
    }

/*******************************************************************************
*
* sdhcClkFreqSetup - setup the clock frequency
*
* This function sets up the clock frequency. If the BSP does not provide a clock
* setup function, this one is used.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlClkFreqSetup
    (
    VXB_DEV_HANDLE   pDev,     /* vxbus device ID */
    UINT32           clk       /* clock freqency */
    )
    {
    if (sdhcCtrlClkSet(pDev, clk) != OK)
        {
        return ERROR;
        }

    /* Set SDHC high speed enable */

    if (clk > SDMMC_CLK_FREQ_25MHZ)
        CSR_SETBIT_4(pDev, SDHC_PROCTL, PROCTL_HS_ENABLED);
    else
        CSR_CLRBIT_4(pDev, SDHC_PROCTL, PROCTL_HS_ENABLED);

    return OK;
    }

/*******************************************************************************
*
* sdhcCtrlAbortIssue - issue abort command to stop transaction
*
* This routine issues abort command to stop transaction.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL STATUS sdhcCtrlAbortIssue
    (
    VXB_DEV_HANDLE pDev     /* vxbus device ID */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_CTRL * pSdHostCtrl;
    UINT32 mask = PRSSTAT_CIHB;
    UINT32 i = 0;

    SDHC_DBG (SDHC_DBG_ERR, "start...\n", 0, 0, 0, 0);
    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }
    else
        {
        pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));
        }

    if (semTake (pSdHostCtrl->hostDevSem, WAIT_FOREVER) == ERROR)
        {
        return ERROR;
        }

    /* check PRSSTAT[CIHB] before issue abort command */

    while (CSR_READ_4(pDev, SDHC_PRSSTAT) & mask)
        {
        vxbUsDelay(1);
        i++;
        if (i > 10000)
            {
            (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
            return ERROR;
            }
        }

    /* issue abort command and software reset for CMD/DAT line */

    CSR_WRITE_4 (pDev, SDHC_XFERTYP, XFERTYP_CMDTYP_ABORT);
    CSR_SETBIT_4(pDev, SDHC_SYSCTL, SYSCTL_RSTC | SYSCTL_RSTD);
    while ((CSR_READ_4(pDev, SDHC_SYSCTL) & (SYSCTL_RSTC | SYSCTL_RSTD)) != 0);

    (void)semGive (pDrvCtrl->sdHostCtrl.hostDevSem);
    return OK;
}

/*******************************************************************************
*
* sdhcAdma2DescCreate - create data structures for ADMA2
*
* This routine creates data structures used for ADMA2
*
* RETURNS: SDHC_ADMA2_ID or NULL in case of failure
*
* ERRNO: N/A
*
* \NOMANUAL
*/

void * sdhcAdma2DescCreate
    (
    VXB_DEV_HANDLE pDev     /* vxbus device ID */
    )
    {
    SDHC_ADMA2_ID   pAdma2Hdl;
    size_t          bsize;

    pAdma2Hdl = (SDHC_ADMA2_ID)vxbMemAlloc(sizeof(SDHC_ADMA2_HDL));
    if (pAdma2Hdl == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "Hdl malloc(%d) fail\n",
                  sizeof(SDHC_ADMA2_HDL), 0, 0, 0);
        return NULL;
        }

    bsize = (ADMA2_NB_DESC * sizeof(SDHC_ADMA2_DESC));
    SDHC_DBG (SDHC_DBG_INIT, "nb_Desc=%d, size=%d\n",
              ADMA2_NB_DESC, bsize, 0, 0);

    pAdma2Hdl->descTag = vxbDmaBufTagCreate(pDev,
            NULL,                               /* parent */
            4,                                  /* alignment */
            0,                                  /* boundary */
            VXB_SPACE_MAXADDR_32BIT,            /* lowaddr */
            VXB_SPACE_MAXADDR,                  /* highaddr */
            NULL,                               /* filter */
            NULL,                               /* filterarg */
            bsize,                              /* max size */
            1,                                  /* nSegments */
            bsize,                              /* max seg size */
            VXB_DMABUF_NOCACHE,                 /* flags */
            NULL,                               /* lockfunc */
            NULL,                               /* lockarg */
            NULL);                              /* ppDmaTag */
    if (pAdma2Hdl->descTag == NULL)
        {
        SDHC_DBG (SDHC_DBG_ERR, "descTag create fail\n", 0, 0, 0, 0);
        vxbMemFree((void *) pAdma2Hdl);
        return NULL;
        }

    pAdma2Hdl->pDesc = (SDHC_ADMA2_DESC *) vxbDmaBufMemAlloc (pDev,
            pAdma2Hdl->descTag, NULL, VXB_DMABUF_MEMALLOC_CLEAR_BUF,
            &(pAdma2Hdl->descMap));
    if (pAdma2Hdl->pDesc == NULL)
        {
        SDHC_DBG(SDHC_DBG_ERR, "fail to alloc desc\n", 0, 0, 0, 0);
        vxbDmaBufTagDestroy(pAdma2Hdl->descTag);
        vxbMemFree((void *) pAdma2Hdl);
        return NULL;
        }

    if (vxbDmaBufMapLoad (pDev, pAdma2Hdl->descTag,
            pAdma2Hdl->descMap, pAdma2Hdl->pDesc, bsize, 0) != OK)
        {
        SDHC_DBG(SDHC_DBG_ERR, "fail to load map\n", 0, 0, 0, 0);
        vxbDmaBufMemFree (pAdma2Hdl->descTag, (void *)pAdma2Hdl->pDesc, pAdma2Hdl->descMap);
        vxbDmaBufTagDestroy(pAdma2Hdl->descTag);
        vxbMemFree((void *) pAdma2Hdl);
        return NULL;
        }

    pAdma2Hdl->descPhy= pAdma2Hdl->descMap->fragList[0].frag;
    SDHC_DBG (SDHC_DBG_INIT, "pAdma2Hdl=%p, pDesc=%p [0x%x]\n",
              pAdma2Hdl, pAdma2Hdl->pDesc, (UINT32) pAdma2Hdl->descPhy, 0);

    ADMA2_DESC_ATTR(pAdma2Hdl->pDesc, (ADMA2_ATTR_VALID | ADMA2_ATTR_END));
    ADMA2_DESC_LEN(pAdma2Hdl->pDesc, 0);
    ADMA2_DESC_ADDR(pAdma2Hdl->pDesc, 0);

    return (void *) pAdma2Hdl;
    }

/*******************************************************************************
*
* sdhcAdma2DescFill - fill descriptors before ADMA2 transfer
*
* This function fills descriptors before ADMA2 transfer
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

void sdhcAdma2DescFill
    (
    VXB_DEV_HANDLE  pDev,       /* vxbus device ID */
    void *          adma2Hdl ,  /* ADMA2 handle */
    VXB_DMA_MAP_ID  map         /* ADMA2 map */
    )
    {
    SDHC_ADMA2_DESC * pDesc;
    UINT32            regAerr, regSysa, regInt;
    int               nseg, idx;
    SDHC_ADMA2_ID     pAdma2Hdl = (SDHC_ADMA2_ID) adma2Hdl;

    if ((regAerr = CSR_READ_4 (pDev, SDHC_ADMA_ERRST)) != 0)
        {
        /* Clear Transfer Complete Status & ADMA Error Status - RW1C */

        regInt = CSR_READ_4 (pDev, SDHC_IRQSTAT);
        regInt &= (IRQ_TC | IRQ_ADMAE);
        if (regInt)
            CSR_WRITE_4 (pDev, SDHC_IRQSTAT, regInt);

        SDHC_DBG (SDHC_DBG_ERR,
                  "IRQSTAT=0x%x ADMA_ERR_STATUS=0x%x ADMA_SYS_ADDR=0x%x\n",
                  regInt, regAerr, CSR_READ_4 (pDev, SDHC_ADMA_SYSAD), 0);
        }

    nseg = map->nFrags;
    pDesc = pAdma2Hdl->pDesc;
    for (idx = 0; idx < nseg; idx++)
        {
        if (idx == nseg - 1)    /* last desc */
            ADMA2_DESC_ATTR(pDesc, (ADMA2_ATTR_TRAN | ADMA2_ATTR_VALID \
                                   | ADMA2_ATTR_END));
        else
            ADMA2_DESC_ATTR(pDesc, (ADMA2_ATTR_TRAN | ADMA2_ATTR_VALID));
        ADMA2_DESC_LEN(pDesc, map->fragList[idx].fragLen);
        ADMA2_DESC_ADDR(pDesc, map->fragList[idx].frag);
        SDHC_DBG (SDHC_DBG_RW, "<%d> addr=0x%x len=0x%x attr=0x%x\n",
                  idx, pDesc->addr, pDesc->len, pDesc->attr);
        pDesc++;
        }

    ADMA2_DESC_ATTR(pDesc, 0);
    ADMA2_DESC_LEN(pDesc, 0);
    ADMA2_DESC_ADDR(pDesc, 0);

    regSysa = (UINT32)pAdma2Hdl->descPhy;
    CSR_WRITE_4 (pDev, SDHC_ADMA_SYSAD, regSysa);
    SDHC_DBG (SDHC_DBG_RW, "Set ADMA_SYS_ADDR = 0x%x nseg = %d\n",
              regSysa, nseg, 0, 0);
    }

#ifdef _WRS_CONFIG_FDT

/*******************************************************************************
*
* sdhcCtrlHostCapParse - Parse host capability
*
* This routine parses host capability.
*
* RETURNS: OK.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL void sdhcCtrlHostCapParse
    (
    UINT64          capability,
    SD_HOST_CAP *   pHostCap    /* Host capability */
    )
    {
    UINT32  val = (UINT32)capability;

    /* D21 Hight Speed Support */

    if (val & HOSTCAPBLT_HSS)
        pHostCap->caps |= HOST_CAP_HS;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_HS;

    if (val & HOSTCAPBLT_VS18)
        pHostCap->caps |= HOST_CAP_VOLT_18V;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_VOLT_18V;

    if (val & HOSTCAPBLT_VS30)
        pHostCap->caps |= HOST_CAP_VOLT_30V;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_VOLT_30V;

    if (val & HOSTCAPBLT_VS33)
        pHostCap->caps |= HOST_CAP_VOLT_33V;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_VOLT_33V;

    pHostCap->caps |= HOST_CAP_4BIT;

    if (val & HOSTCAPBLT_8BIT)
        pHostCap->caps |= HOST_CAP_8BIT;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_8BIT;

    /* Parse D32-63 */

    val = (UINT32)(capability >> 32);

    /* D32 SDR50 Support */

    if (val & HOSTCAPBLT_SDR50)
        pHostCap->caps |= HOST_CAP_UHS_SDR50;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_UHS_SDR50;

    if (val & HOSTCAPBLT_SDR104)
        pHostCap->caps |= HOST_CAP_UHS_SDR104;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_UHS_SDR104;

    if (val & HOSTCAPBLT_DDR50)
        pHostCap->caps |= HOST_CAP_UHS_DDR50;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_UHS_DDR50;

    if (val & HOSTCAPBLT_TSDR50)
        pHostCap->caps |= HOST_CAP_SDR50_TUNE;
    else
        pHostCap->caps &= (UINT32)~ HOST_CAP_SDR50_TUNE;

    }

/*******************************************************************************
*
* sdhcCtrlGetHostCap - get host capability via FDT
*
* This routine gets host capability via FDT.
*
* RETURNS: OK.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlGetHostCap
    (
    VXB_DEV_HANDLE  pDev,       /* Device information */
    SD_HOST_CAP *   pHostCap    /* Host capability */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    VXB_FDT_DEV *   pFdtDev;
    const void *    pValue;
    UINT64          capability;
    UINT64          temp;
    int             len;
    UINT32 *        pCur;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    /* D0-31 */

    temp = CSR_READ_4 (pDev, SDHC_HOSTCAPBLT);

    capability = temp;

    /* D32-63 */

    temp = CSR_READ_4 (pDev, SDHC_HOSTCAPBLT2);

    capability |= (temp << 32);

    pFdtDev = vxbFdtDevGet(pDev);
    if (pFdtDev != NULL)
        {
        pValue = vxFdtPropGet(pFdtDev->offset, "sdhci-caps-mask", &len);
        if (pValue != NULL)
            {
            /* it must be 64 bit */

            if (len != sizeof(UINT64))
                {
                return ERROR;
                }

            pCur = (UINT32 *) pValue;
            temp = vxFdt32ToCpu (* pCur);
            temp = temp << 32;
            pCur++;
            temp |= vxFdt32ToCpu (* pCur);

            capability &= (UINT64)(~temp);
            }

        pValue = vxFdtPropGet(pFdtDev->offset, "sdhci-caps", &len);
        if (pValue != NULL)
            {
            /* it must be 64 bit */

            if (len != sizeof(UINT64))
                {
                return ERROR;
                }

            pCur = (UINT32 *) pValue;
            temp = vxFdt32ToCpu (* pCur);
            temp = temp << 32;
            pCur++;
            temp |= vxFdt32ToCpu (* pCur);

            capability |= temp;
            }
        }

    SDHC_DBG (SDHC_DBG_INIT,"capability = 0x%llx\n", capability, 0, 0, 0);
    sdhcCtrlHostCapParse(capability, pHostCap);

    if (pDrvCtrl->sdHostCtrl.curClkFreq >= MMC_CLK_FREQ_52MHZ &&
        pHostCap->caps & HOSTCAPBLT_DDR50)
        {
        pHostCap->caps |= (HOST_CAP_DDR_18V);
        pHostCap->caps |= (HOST_CAP_HS200_18V);
        }

    if ((pDrvCtrl->flags & SDHC_NOT_DO_TUNING_SDR50) ==
         SDHC_NOT_DO_TUNING_SDR50)
        {
        pHostCap->caps &= (UINT32)~ HOST_CAP_SDR50_TUNE;
        }

    return OK;
    }

/*******************************************************************************
*
* sdhcCtrlGetFdtHostCap - get host capability via FDT
*
* This routine gets host capability via FDT.
*
* RETURNS: OK.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlGetFdtHostCap
    (
    VXB_DEV_HANDLE  pDev,       /* Device information */
    SD_HOST_CAP *   pHostCap    /* Host capability */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    VXB_FDT_DEV   * pFdtDev;
    const void    * pValue;
    const void    * pValueTmp;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ERROR;

    pFdtDev = vxbFdtDevGet(pDev);
    if (pFdtDev == NULL)
        return OK;

    /*
     * bus-width: Number of data lines, can be <1>, <4>, or <8>.
     * The default will be <1> if the property is absent.
     */

    pValue = vxFdtPropGet(pFdtDev->offset, "bus-width", NULL);
    if (pValue != NULL)
        {
        if (vxFdt32ToCpu(*(UINT32 *)pValue) == 8)
            {
            pHostCap->caps |= HOST_CAP_8BIT;
            }
        else if (vxFdt32ToCpu(*(UINT32 *)pValue) == 4)
            {
            pHostCap->caps |= HOST_CAP_4BIT;
            }
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "sd-uhs-sdr50", NULL);
    if (pValue != NULL)
        {
        pHostCap->caps |= HOST_CAP_UHS_SDR50;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "sd-uhs-sdr104", NULL);
    if (pValue != NULL)
        {
        pHostCap->caps |= HOST_CAP_UHS_SDR104;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "sd-uhs-ddr50", NULL);
    if (pValue != NULL)
        {
        pHostCap->caps |= HOST_CAP_UHS_DDR50;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "mmc-hs200-1_8v", NULL);
    if (pValue != NULL)
        {
        pHostCap->caps |= HOST_CAP_HS200_18V;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "mmc-hs400-1_8v", NULL);
    if (pValue != NULL)
        {
        pHostCap->caps |= (HOST_CAP_HS200_18V | HOST_CAP_HS400_18V);
        }

    pValue = vxFdtPropGet (pFdtDev->offset, "cap-sd-highspeed", NULL);
    pValueTmp = vxFdtPropGet (pFdtDev->offset, "cap-mmc-highspeed", NULL);

    if ((pValue != NULL) || (pValueTmp != NULL))
        {
        pHostCap->caps |= HOST_CAP_HS;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "no-1-8-v", NULL);
    if (pValue != NULL)
        {
        pHostCap->caps &= ~(HOST_CAP_HS200_18V | HOST_CAP_HS400_18V);
        pHostCap->caps &= ~(HOST_CAP_UHS_DDR50 | HOST_CAP_UHS_SDR104 | HOST_CAP_UHS_SDR50);
        pHostCap->caps &= ~(HOST_CAP_DDR_18V);
        }

    return OK;
    }

/*******************************************************************************
*
* sdhcCtrlFdtParse - parse sdhc device-tree node
*
* This routine parses sdhc device-tree node.
*
* RETURNS: OK.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL STATUS sdhcCtrlFdtParse
    (
    VXB_DEV_HANDLE  pDev        /* Device information */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    VXB_FDT_DEV *   pFdtDev;
    const void *    pValue;
    UINT32 val;
    int len;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ERROR;

    pFdtDev = vxbFdtDevGet(pDev);
    if (pFdtDev == NULL)
        return ERROR;

    (void) sdhcCtrlGetGpioConfig (pDev);

    /*
     * bus-width: Number of data lines, can be <1>, <4>, or <8>.
     * If the property is absent,
     * the defaul will be <1> if new standard dts is used.
     * the defaul will be <4> if vxWorks legacy dts is used.
     */

    pValue = vxFdtPropGet(pFdtDev->offset, "bus-width", &len);
    if (pValue != NULL)
        {
        val = vxFdt32ToCpu(*(UINT32 *)pValue);
        }
    else
        {
        val = 4;
        }

    if (val != 8 && val != 4 && val != 1)
        {
        SDHC_DBG (SDHC_DBG_ERR ,"invalid bus-width %d\n", val, 0, 0, 0);
        return ERROR;
        }

    pDrvCtrl->sdHostCtrl.sdHostSpec.busWidth = val;

    pDrvCtrl->attrData.cdtype = SDHC_CD_NONE;

    pValue = vxFdtPropGet(pFdtDev->offset, "broken-cd", &len);
    if (pValue != NULL)
        {
        pDrvCtrl->attrData.cdtype = SDHC_CD_SOFTWARE;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "cd-gpios", &len);
    if (pValue != NULL)
        {
        pDrvCtrl->attrData.cdtype = SDHC_CD_GPIO;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "non-removable", &len);
    if (pValue != NULL)
        {
        pDrvCtrl->attrData.cdtype = SDHC_CD_PERMANENT;
        pDrvCtrl->embedded = TRUE;
        }
    else
        {
        pDrvCtrl->embedded = FALSE;
        }

    pValue = vxFdtPropGet (pFdtDev->offset, "card-detect-delay", &len);
    if (pValue != NULL)
        {
        pDrvCtrl->detectDelay = (UINT32)vxFdt32ToCpu (*(UINT32 *)pValue);
        }

    if (pDrvCtrl->attrData.cdtype == SDHC_CD_NONE)
        {
        pValue = vxFdtPropGet(pFdtDev->offset, "cd_type", &len);
        if (pValue != NULL)
            {
            pDrvCtrl->attrData.cdtype = (enum cdTypes) (vxFdt32ToCpu(*(UINT32 *)pValue));
            }
        else
            {
            pDrvCtrl->attrData.cdtype = SDHC_CD_CONTROLLER;
            }
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "wp_type", &len);
    if (pValue != NULL)
        {
        pDrvCtrl->attrData.wptype = (enum wpTypes) (vxFdt32ToCpu(*(UINT32 *)pValue));
        }
    else
        {
        pValue = vxFdtPropGet(pFdtDev->offset, "disable-wp", &len);
        if (pValue != NULL)
            {
            pDrvCtrl->attrData.wptype = SDHC_WP_NONE;
            }
        else
            {
            pValue = vxFdtPropGet(pFdtDev->offset, "wp-gpios", &len);
            if (pValue != NULL)
                {
                pDrvCtrl->attrData.wptype = SDHC_WP_GPIO;
                }
            else
                {
                pDrvCtrl->attrData.wptype = SDHC_WP_CONTROLLER;
                }
            }
        }

    pValue = vxFdtPropGet (pFdtDev->offset, "max-frequency", &len);
    if(pValue != NULL)
        {
        pDrvCtrl->maxDevClkFreq = vxFdt32ToCpu (*(UINT32 *)pValue);
        }
    else
        {
        pDrvCtrl->maxDevClkFreq = 0;
        }

    pValue = vxFdtPropGet (pFdtDev->offset, "no-sd", NULL);
    if (pValue != NULL)
        {
        pDrvCtrl->sdHostCtrl.sdHostSpec.specFlag |= SDHC_SPEC_FLAG_NO_SD;
        }
    else
        {
        pDrvCtrl->sdHostCtrl.sdHostSpec.specFlag &= (UINT32)~SDHC_SPEC_FLAG_NO_SD;
        }

    pValue = vxFdtPropGet (pFdtDev->offset, "no-sdio", NULL);
    if (pValue != NULL)
        {
        pDrvCtrl->sdHostCtrl.sdHostSpec.specFlag |= SDHC_SPEC_FLAG_NO_SDIO;
        }
    else
        {
        pDrvCtrl->sdHostCtrl.sdHostSpec.specFlag &= (UINT32)~SDHC_SPEC_FLAG_NO_SDIO;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "big-endian", &len);
    if(pValue != NULL)
        {
        if (len == 0)
            {
            pDrvCtrl->bigEndian = TRUE;
            }
        else
            {
            pDrvCtrl->bigEndian = (BOOL)vxFdt32ToCpu(*(UINT32 *)pValue);
            }
        }
    else
        {
        pDrvCtrl->bigEndian = FALSE;
        }

    pValue = vxFdtPropGet(pFdtDev->offset, "little-endian", &len);
    if(pValue != NULL)
        {
        pDrvCtrl->bigEndian = FALSE;
        }

    SDHC_DBG (SDHC_DBG_INIT, "cdType = %d, wpType = %d, bigEndian is %s\n",
              pDrvCtrl->attrData.cdtype, pDrvCtrl->attrData.wptype,
              ((pDrvCtrl->bigEndian) ? "True" : "False"), 0);

    pDrvCtrl->dmaCoherent = vxFdtIsDmaCoherent(pFdtDev->offset);
    return OK;
    }
#endif

/*******************************************************************************
*
* sdhcCtrlVddSetup - setup the SD bus voltage level and power it up
*
* This routine setups the SD bus voltage and powers it up. This
* routine will be called in the SDHC driver.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcCtrlVddSetup
    (
    VXB_DEV_HANDLE pDev,        /* Device information */
    UINT32         vdd          /* voltage level */
    )
    {
    UINT32 val = 0;

    SDHC_DBG (SDHC_DBG_ALL, "vdd = %d\n", vdd, 0, 0, 0);

    val = CSR_READ_4 (pDev, SDHC_PROCTL);

    val &= (UINT32)(~PROCTL_SDBUS_VOLT_MASK);

    /* setup power value and turn on sd bus power */

    if (vdd == SDMMC_VOLTAGE_18V)
        {
        SDHC_DBG (SDHC_DBG_ALL, "1.8V\n", 0, 0, 0, 0);
        val |= (PROCTL_SDBUS_VOLT_1P8 << PROCTL_VDD1_SEL_SHIFT);
        CSR_WRITE_4(pDev, SDHC_PROCTL, val);
        }
    else if (vdd == SDMMC_VOLTAGE_33V)
        {
        SDHC_DBG (SDHC_DBG_ALL, "3.3V\n", 0, 0, 0, 0);
        val |= (PROCTL_SDBUS_VOLT_3P3 << PROCTL_VDD1_SEL_SHIFT);
        CSR_WRITE_4(pDev, SDHC_PROCTL, val);
        }
    else if (vdd == SDMMC_VOLTAGE_30V)
        {
        SDHC_DBG (SDHC_DBG_ALL, "3.0V\n", 0, 0, 0, 0);
        val |= (PROCTL_SDBUS_VOLT_3P0 << PROCTL_VDD1_SEL_SHIFT);
        CSR_WRITE_4(pDev, SDHC_PROCTL, val);
        }
    else
        {
        SDHC_DBG (SDHC_DBG_ALL, "%dV error\n", val, 0, 0, 0);
        return ERROR;
        }

    val |= PROCTL_VDD1_PWR_EN;
    CSR_WRITE_4(pDev, SDHC_PROCTL, val);

    SDHC_DBG (SDHC_DBG_ALL, "SDHC_PROCTL=0x%08x\n", val, 0, 0, 0);
    return OK;
    }

/*******************************************************************************
*
* sdhcTuningStart - start tuning for SDHC controller
*
* This routine setup the tuning for the SDHC controller.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL void sdhcTuningStart
    (
    VXB_DEV_HANDLE pDev,            /* Device information */
    BOOL           isFirstTuning    /* First tuning or not */
    )
    {
    UINT32 reg;

    reg = CSR_READ_4(pDev, SDHC_AUTOC12ERR);
    reg |= (AUTOC12ERR_EXEC_TUNING);
    if (isFirstTuning)
        {
        reg &= (~AUTOC12ERR_TUNED_CLK);
        }

    CSR_WRITE_4 (pDev, SDHC_AUTOC12ERR, reg);

    CSR_WRITE_4 (pDev, SDHC_IRQSTATEN, IRQ_BRR);
    CSR_WRITE_4 (pDev, SDHC_IRQSIGEN, IRQ_BRR);
    }

/*******************************************************************************
*
* sdhcTuningComplete - invoked when SDHC tuning completion
*
* This routine should be invoked when tuning completion.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL void sdhcTuningComplete
    (
    VXB_DEV_HANDLE pDev     /* Device information */
    )
    {

    SDHC_DEV_CTRL * pDrvCtrl;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        return ;

    CSR_WRITE_4 (pDev, SDHC_IRQSTATEN, pDrvCtrl->intMask);
    CSR_WRITE_4 (pDev, SDHC_IRQSIGEN, pDrvCtrl->intMask);
    }

/*******************************************************************************
*
* sdhcTuningExecute - exectute tuning procedure for SDHC controller
*
* This routine exectute tuning procedure for the SDHC controller.
*
* RETURNS: OK or ERROR
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcTuningExecute
    (
    VXB_DEV_HANDLE pDev     /* Device information */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_CTRL  * pSdHostCtrl;
    SD_HARDWARE   * pSdHardware;
    SD_HOST_SPEC  * pHostSpec;
    UINT32          dmaMode;
    SD_CMD          cmd;
    UINT32          cmdOpt = SDMMC_CMD_SEND_TUNING;
    UINT32          length = 64;
    UINT32          times = 0;
    UINT32          val;

    SDHC_DBG (SDHC_DBG_ALL, "start...\n", 0, 0, 0, 0);

    if (pDev == NULL)
        {
        return ERROR;
        }

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));

    pSdHardware = (SD_HARDWARE *)GET_HARDWAREINFO (pSdHostCtrl->sdHostSpec.childDev);
    if (pSdHardware == NULL)
        {
        return ERROR;
        }

    pHostSpec = (SD_HOST_SPEC *)(pSdHardware->pHostSpec);
    if (pHostSpec == NULL)
        {
        return ERROR;
        }

    SDHC_DBG (SDHC_DBG_ALL, "pSdHardware->timing = %d\n",
              pSdHardware->timing, 0, 0, 0);

    if (pSdHardware->timing == SDMMC_TIMING_MMC_HS200)
        {
        cmdOpt = SDMMC_CMD_SEND_TUNING_HS200;
        if (pHostSpec->busWidth == SDMMC_BUS_WIDTH_8BIT)
            {
            length = 128;
            }
        }

    /* DMA is not enabled for TUNING */

    dmaMode = pDrvCtrl->sdHostCtrl.dmaMode;
    pDrvCtrl->sdHostCtrl.dmaMode = SDHC_DMA_MODE_PIO;

    pDrvCtrl->tuningInProgress = 1;

    /* 1. Set Execute Tuning = 1 */

    sdhcTuningStart(pDev, TRUE);

    SDHC_DBG (SDHC_DBG_ALL, "cmd = %d, length = %d",
              cmdOpt, length, 0, 0);

    SD_CMD_FORM (cmd, cmdOpt, 0, SDMMC_CMD_RSP_R1, TRUE);

    /* The standard tuning procedure doesn't need a buffer for reading. */

    SD_DATA_FORM (cmd, NULL, 1, length, TRUE);

    /*
     * 2. Issue SEND_TUNING_BLK Command (CMD19 for SD, CMD21 for eMMC).
     * 3. Wait for IRQSTAT[BRR], buffer read ready, to be set.
     * 4. Clear IRQSTAT[BRR].
     * 5. Check SYSCTL2[EXTN] to be cleared.
     * 6. Repeat steps 2-5, if EXTN is not cleared.
     */

    do
        {
        (void) sdCmdIssue(pSdHostCtrl->sdHostSpec.childDev, &cmd);

        times ++;

        val = CSR_READ_4(pDev, SDHC_AUTOC12ERR);

        SDHC_DBG (SDHC_DBG_ALL, "(%d) times, SDHC_AUTOC12ERR = 0x%08x\n",
                  times, val, 0, 0);

        if ((val & AUTOC12ERR_EXEC_TUNING) == 0)
            break;

        if (cmdOpt == SDMMC_CMD_SEND_TUNING)
            vxbMsDelay(1);
        }
    while(times < SDHC_TUNING_COUNT);

    if ((val & AUTOC12ERR_EXEC_TUNING) != 0)
        {
        val &= (~AUTOC12ERR_TUNED_CLK);
        val &= (~AUTOC12ERR_EXEC_TUNING);
        CSR_WRITE_4 (pDev, SDHC_AUTOC12ERR, val);
        SDHC_DBG (SDHC_DBG_ALL, "reach max (%d) times\n", times, 0, 0, 0);
        }

    pDrvCtrl->sdHostCtrl.dmaMode = dmaMode;
    pDrvCtrl->tuningInProgress = 0;

    sdhcTuningComplete(pDev);

    /*
     * 7. Check SYSCTL2[SMPCLKSEL], sampling clock select.
     * It's set value indicates tuning procedure success, and clears indicate
     * failure.
     */

    if ((val & AUTOC12ERR_TUNED_CLK) == 0)
        {
        SDHC_DBG (SDHC_DBG_ALL, "done fail\n", 0, 0, 0, 0);
        return ERROR;
        }

    SDHC_DBG (SDHC_DBG_ALL, "done successfully\n", 0, 0, 0, 0);
    return OK;
    }

#ifdef _WRS_CONFIG_FDT
/*******************************************************************************
*
* sdhcCtrlGetGpioConfig - get GPIO configuration
*
* This function gets GPIO configuration for card detection and write protection
* of SDHC controller.
* The "cd-gpios" has the highest priority. Then "cd_pin". If both don't exist in
* dts files, cdGpio/wpGpio in pDrvCtrl->attrData will be set to ERROR.
*
* RETURNS: ERROR if parameter check failure, OK otherwise.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL STATUS sdhcCtrlGetGpioConfig (VXB_DEV_HANDLE pDev)
    {
    VXB_FDT_DEV * pFdtDev;
    SDHC_DEV_CTRL * pDrvCtrl;
    const void * pValue;
    int len;
    int cdGpio;
    int wpGpio;

    pDrvCtrl = (SDHC_DEV_CTRL *)GET_DRVCTRL(pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    pFdtDev = vxbFdtDevGet(pDev);
    if (pFdtDev == NULL)
        {
        return ERROR;
        }

    /* must ensure GPIO driver is initializaed */

    cdGpio = vxbGpioGetByFdtIndex (pDev, "cd-gpios", 0);
    if (cdGpio >= 0)
        {
        pDrvCtrl->attrData.cdGpio = (UINT32)cdGpio;
        }
    else
        {
        pValue = vxFdtPropGet(pFdtDev->offset, "cd_pin", &len);
        if (pValue != NULL)
            {
            pDrvCtrl->attrData.cdGpio = vxFdt32ToCpu(*(UINT32 *)pValue);
            }
        else
            {
            pDrvCtrl->attrData.cdGpio = (UINT32) ERROR;
            }
        }

    wpGpio = vxbGpioGetByFdtIndex (pDev, "wp-gpios", 0);
    if (wpGpio >= 0)
        {
        pDrvCtrl->attrData.wpGpio = (UINT32)wpGpio;
        }
    else
        {
        pValue = vxFdtPropGet(pFdtDev->offset, "wp_pin", &len);
        if (pValue != NULL)
            {
            pDrvCtrl->attrData.wpGpio = vxFdt32ToCpu(*(UINT32 *)pValue);
            }
        else
            {
            pDrvCtrl->attrData.wpGpio = (UINT32) ERROR;
            }
        }

    SDHC_DBG (SDHC_DBG_INIT, "cdGpio=%d wpGpio=%d\n",
              pDrvCtrl->attrData.cdGpio, pDrvCtrl->attrData.wpGpio, 0, 0);
    return OK;
    }
#endif

/*******************************************************************************
*
* sdhcSdClockEnable - enable SD clock
*
* This routine enables SD clock.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL void sdhcSdClockEnable
    (
    VXB_DEV_ID pDev,    /* Device information */
    BOOL       enable   /* Enable or Disable */
    )
    {
    if (enable)
        {
        CSR_SETBIT_4 (pDev, SDHC_SYSCTL, SYSCTL_SDCLK_EN);
        }
    else
        {
        CSR_CLRBIT_4 (pDev, SDHC_SYSCTL, SYSCTL_SDCLK_EN);
        }
    }

/*******************************************************************************
*
* mcpSdhcSdBusPowerOff - turn off the SD Bus power
*
* This routine turns off the SD Bus power.
*
* RETURNS: None
*
* ERRNO: N/A
*
* \NOMANUAL
*/

LOCAL void sdhcSdBusPowerOff
    (
    VXB_DEV_ID	pDev    /* Device information */
    )
    {
    UINT32    value;

    /* TODO: set regulator vdd to 0 */

    value = CSR_READ_4 (pDev, SDHC_PROCTL);
    value &= (UINT32)(~PROCTL_POWERCTL_MASK);
    CSR_WRITE_4 (pDev, SDHC_PROCTL, value);


    SDHC_DBG (SDHC_DBG_INIT, "sdhcSdBusPowerOff done\n", 0, 0, 0, 0);
    }

/*******************************************************************************
*
* sdhcVddSetup - set signal voltage to desired level
*
* This routine sets SD signal voltage to desired level.
*
* RETURNS: OK if succeeded, ERROR if failed.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcVddSetup
    (
    VXB_DEV_ID    pDev,    /* Device information */
    UINT32        vdd      /* voltage level */
    )
    {
    STATUS        status = OK;
    UINT32        val;

    val = CSR_READ_4 (pDev, SDHC_AUTOC12ERR);

    if (vdd == SDMMC_VOLTAGE_33V)
        {
        val &= (UINT32)(~AUTOC12ERR_VDD_1P8);
        CSR_WRITE_4 (pDev, SDHC_AUTOC12ERR, val);

        if (sdhcCtrlVddSetup (pDev, vdd) == ERROR)
            {
            SDHC_DBG (SDHC_DBG_ALL, "sdhcVddSetup Vdd Setup error\n",
                      0, 0, 0, 0);
            return ERROR;
            }
        }
    else if (vdd == SDMMC_VOLTAGE_18V)
        {
        if (sdhcCtrlVddSetup (pDev, vdd) == ERROR)
            {
            SDHC_DBG (SDHC_DBG_ALL, "sdhcVddSetup Vdd Setup error\n",
                      0, 0, 0, 0);
            return ERROR;
            }

        val |= AUTOC12ERR_VDD_1P8;
        CSR_WRITE_4 (pDev, SDHC_AUTOC12ERR, val);
        }

    /* Wait 5 ms, we wait more */

    vxbMsDelay (10);

    return status;
    }

/*******************************************************************************
*
* sdhcSigVolSwitch - switch signal voltage to desired level
*
* This routine switches SD signal voltage to desired level.
*
* RETURNS: OK if succeeded, ERROR if failed.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcSigVolSwitch
    (
    VXB_DEV_ID    pDev,    /* Device information */
    UINT32        vdd      /* voltage level */
    )
    {
    SDHC_DEV_CTRL    * pDrvCtrl;
    VXB_DEV_ID         pCDev;
    STATUS             rc;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return ERROR;
        }

    pCDev = pDrvCtrl->sdHostCtrl.sdHostSpec.childDev;

    /* CMD11 */

    rc = sdCmdSwitchVoltage (pCDev);
    if (rc == ERROR)
        {
        SDHC_DBG (SDHC_DBG_ALL, "CMD11 Switch Voltage command error\n",
                  rc, 0, 0, 0);
        return (rc);
        }

    SDHC_DBG (SDHC_DBG_ALL, "sdhcSigVolSwitch set vdd %d \n",
              vdd, 0, 0, 0);

    vxbMsDelay (1);

    /* SD Clock Enable = 0 */

    sdhcSdClockEnable (pDev, FALSE);

    /* Clear SD Bus Power */

    sdhcSdBusPowerOff (pDev);

    (void) sdhcVddSetup (pDev,vdd);

    /* SD Clock Enable = 1 */

    sdhcSdClockEnable (pDev, TRUE);

    /* Wait 1 ms */

    vxbMsDelay (1);

    return OK;
    }

/*******************************************************************************
*
* sdhcDevAdd - SDHC insert device handle
*
* This function will handle the deivce insert condition.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

void sdhcDevAdd
    (
    VXB_DEV_ID pDev    /* Device information */
    )
    {
    SD_HOST_CTRL  * pSdHostCtrl;
    SDHC_DEV_CTRL * pDrvCtrl;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return;
        }

    pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));

    if (pSdHostCtrl->sdHostOps.vxbSdVddSetup != NULL)
        {
        pSdHostCtrl->sdHostOps.vxbSdVddSetup (pDev, pSdHostCtrl->vdd);
        }
    if (pSdHostCtrl->sdHostOps.vxbSdClkFreqSetup != NULL)
        {
        pSdHostCtrl->sdHostOps.vxbSdClkFreqSetup (pDev, SDMMC_CLK_FREQ_400KHZ);
        }
    if (pSdHostCtrl->sdHostOps.vxbSdBusWidthSetup != NULL)
        {
        pSdHostCtrl->sdHostOps.vxbSdBusWidthSetup (pDev, SDMMC_BUS_WIDTH_1BIT);
        }

    /* Need not check return status at here */

    (void)sdDeviceCreate (pDev, NULL);


    pSdHostCtrl->attached = TRUE;
    }

/*******************************************************************************
*
* sdhcDevRemove - SDHC remove handle
*
* This function will handle the deivce remvoe condition.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/

void sdhcDevRemove
    (
    VXB_DEV_ID pDev    /* Device information */
    )
    {
    VXB_DEV_ID      pDevList;
    SD_HOST_CTRL  * pSdHostCtrl;
    SDHC_DEV_CTRL * pDrvCtrl;
    int timeOut = 0;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return;
        }

    pSdHostCtrl = (SD_HOST_CTRL *)(&(pDrvCtrl->sdHostCtrl));

    if (((CSR_READ_4 (pDev, SDHC_PRSSTAT) & PRSSTAT_CDIHB) != 0x0) ||
        ((CSR_READ_4 (pDev, SDHC_PRSSTAT) & PRSSTAT_CIHB) != 0x0))
        {
        CSR_SETBIT_4 (pDev, SDHC_SYSCTL, SYSCTL_RSTC | SYSCTL_RSTD);
        while ((CSR_READ_4 (pDev, SDHC_SYSCTL) &
               (SYSCTL_RSTC | SYSCTL_RSTD)) != 0x0)
            {
            if (timeOut++ > SDMMC_COMMAND_TIMEOUT)
                {
                break;
                }
            (void)vxbMsDelay (1);
            }
        }
    pDevList = pSdHostCtrl->sdHostSpec.childDev;

    if (pDevList != NULL)
        {
        (void)vxbDevRemove (pDevList);
        pSdHostCtrl->sdHostSpec.childDev = NULL;
        }

    pSdHostCtrl->attached = FALSE;
    }

/*******************************************************************************
*
* sdCardWpCheck - Check card write protect status routine
*
* This routine checks card write protect status routine
*
* RETURNS: TRUE or FALSE
*
* ERRNO: N/A
*
* \NOMANUAL
*/

BOOL sdCardWpCheck
    (
    VXB_DEV_ID pDev
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return FALSE;
        }

    switch (pDrvCtrl->attrData.wptype)
        {
        case SDHC_WP_GPIO:
        /* GPIO mode is not supported */
        /* Fall through to SDHC_WP_CONTROLLER */
        case SDHC_WP_CONTROLLER:
            return (sdhcCtrlCardWpCheck(pDev));
        case SDHC_WP_NONE:
            return FALSE;
        default:
            return TRUE;
        }
    }

/*******************************************************************************
*
* sdhcCheckUhsSupport - check if host support UHS.
*
* This routine gives if host support UHS.
*
* RETURNS: TRUE if supported, FALSE if not supported.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

BOOL sdhcCheckUhsSupport
    (
    VXB_DEV_ID    pDev    /* Device information */
    )
    {
    SDHC_DEV_CTRL * pDrvCtrl;
    SD_HOST_SPEC  * pHostSpec = NULL;

    pDrvCtrl = (SDHC_DEV_CTRL *)vxbDevSoftcGet (pDev);
    if (pDrvCtrl == NULL)
        {
        return FALSE;
        }

    pHostSpec = &pDrvCtrl->sdHostCtrl.sdHostSpec;

    if ((pHostSpec->hostCap.caps & HOST_CAP_UHS_SDR50) ||
        (pHostSpec->hostCap.caps & HOST_CAP_UHS_DDR50) ||
        (pHostSpec->hostCap.caps & HOST_CAP_UHS_SDR104))
        {
        SDHC_DBG (SDHC_DBG_ALL, "sdhcCheckUhsSupport TRUE !\n",
                  0, 0, 0, 0);
        return TRUE;
        }

    SDHC_DBG (SDHC_DBG_ALL, "sdhcCheckUhsSupport FALSE !\n",
              0, 0, 0, 0);

    return FALSE;
    }

#ifdef _WRS_CONFIG_FDT
/*******************************************************************************
*
* sdhcGetCap - get capability
*
* This routine gets capability.
*
* RETURNS: OK.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcGetCap
    (
    VXB_DEV_ID         pDev,       /* Device information */
    SD_HOST_CAP      * pHostCap    /* Host capability */
    )
    {
    if (sdhcCtrlGetHostCap (pDev, pHostCap) != OK)
        {
        return ERROR;
        }

    if (sdhcCtrlGetFdtHostCap (pDev, pHostCap) != OK)
        {
        return ERROR;
        }

    SDHC_DBG (SDHC_DBG_ALL, "sdhcGetCap caps:0x%08x\n",
              pHostCap->caps, 0, 0, 0);

    return OK;
    }
#endif

/*******************************************************************************
*
* sdhcTimingSetup - setup timing on host side
*
* This routine setup timing on host side.
*
* RETURNS: OK if succeeded, ERROR if failed.
*
* ERRNO: N/A
*
* \NOMANUAL
*/

STATUS sdhcTimingSetup
    (
    VXB_DEV_HANDLE pDev,        /* Device information */
    UINT32         timing       /* Timing Mode */
    )
    {
    UINT32 reg;
    UINT32 uhs = 0;

    if (timing == SDMMC_TIMING_SD_HS)
        {
        return OK;
        }

    if (timing == SDMMC_TIMING_MMC_LEGACY ||
        timing == SDMMC_TIMING_MMC_HS     ||
        timing == SDMMC_TIMING_UHS_SDR12)
        uhs = AUTOC12ERR_UHS_SDR12;
    else if (timing == SDMMC_TIMING_MMC_DDR52)
        uhs = AUTOC12ERR_UHS_DDR;
    else if (timing == SDMMC_TIMING_UHS_SDR25)
        uhs = AUTOC12ERR_UHS_SDR25;
    else if (timing == SDMMC_TIMING_UHS_SDR50)
        uhs = AUTOC12ERR_UHS_SDR50;
    else if (timing == SDMMC_TIMING_MMC_HS200 ||
             timing == SDMMC_TIMING_UHS_SDR104)
        uhs = AUTOC12ERR_UHS_SDR104_HS200;
    else if (timing == SDMMC_TIMING_UHS_DDR50)
        uhs = (AUTOC12ERR_UHS_DDR);
    else
        return ERROR;

    reg = CSR_READ_4(pDev, SDHC_AUTOC12ERR);
    reg &= ~(AUTOC12ERR_UHS_MASK);
    reg |= uhs;
    CSR_WRITE_4 (pDev, SDHC_AUTOC12ERR, reg);

    return OK;
    }

#ifdef  SDHC_DBG_ON

/*******************************************************************************
*
* sdhcRegWrite - write sdhc register value
*
* This routine writes sdhc register value.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/
void sdhcRegWrite(VXB_DEV_HANDLE pDev, UINT32 offset, UINT32 value)
{
    UINT32 reg;

    CSR_WRITE_4 (pDev, offset, value);
    reg = CSR_READ_4(pDev, offset);
    printf("reg=0x%08x\n",reg);

}

/*******************************************************************************
*
* sdhcRegRead - read sdhc register value
*
* This routine reads sdhc register value.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/
void sdhcRegRead(VXB_DEV_HANDLE pDev, UINT32 offset)
{
    UINT32 reg;

    reg = CSR_READ_4(pDev, offset);
    printf("reg=0x%08x\n",reg);

}

/*******************************************************************************
*
* sdhcRegShow - show sdhc register value
*
* This routine shows sdhc register value.
* Example: sdhcRegShow "imxSdhc",1,0x00
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* \NOMANUAL
*/
void sdhcRegShow(char * hostCtrlName, UINT32 devUnit, UINT32 offset)
{
    VXB_DEV_HANDLE  pHostCtrlDev;
    UINT32          reg1;
    UINT32          reg2;
    UINT32          reg3;
    UINT32          reg4;
    SDHC_DEV_CTRL * pDrvCtrl;
    int             i;

    pHostCtrlDev= vxbDevAcquireByName(hostCtrlName, devUnit);
    vxbDevRelease(pHostCtrlDev);

    pDrvCtrl = (SDHC_DEV_CTRL*) vxbDevSoftcGet(pHostCtrlDev);
    printf("SDHC dev=0x%lx BAR=0x%lx\n", pHostCtrlDev, pDrvCtrl->bar);


    for(i=0;i<16; i++)
        {
        reg1 = CSR_READ_4(pHostCtrlDev, offset);
        reg2 = CSR_READ_4(pHostCtrlDev, offset+4);
        reg3 = CSR_READ_4(pHostCtrlDev, offset+8);
        reg4 = CSR_READ_4(pHostCtrlDev, offset+12);
        printf("%08x : %08x %08x %08x %08x\n",offset,reg1,reg2,reg3,reg4);

        offset=offset+16;
        }

}
#endif
