#include <vxWorks.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sllLib.h>
#include <hwif/vxBus.h>
#include <subsys/flash/vxbFlashLib.h>
#include <vxbFlashCommon.h>

LOCAL STATUS sfInit(VOID);
LOCAL VOID sfShutdown(VOID);
LOCAL STATUS sfRead(UINT32, UINT32);
LOCAL STATUS sfSectorIsErased(UINT32, BOOL*);
LOCAL STATUS sfEraseSector(UINT32);
LOCAL STATUS sfWrite(UINT32, UINT32);

LOCAL FLASH_CHIP* pFlashChip;
LOCAL UINT32 sectors;
LOCAL UINT32 sectorSize;
LOCAL UINT8* pRdBuf = NULL;
LOCAL UINT8* pWrBuf = NULL;
LOCAL BOOL verbose  = FALSE;
LOCAL BOOL init = FALSE;
LOCAL const UINT32 SPT_MAGIC = 0x57713427;

typedef struct
    {
    UINT32 magic;
    } SPT_LAYOUT;

STATUS qspiFlashReadTest(VOID)
    {
    STATUS rc;

    rc = sfInit();
    if (rc != OK)
        {
        return ERROR;
        }
    printf("Reading %d bytes from offset 0\n", sectorSize);
    rc = sfRead(0, sectorSize);
    if (rc != OK)
        {
        return ERROR;
        }

    printf("done\n");
    sfShutdown();
    return OK;
    }

STATUS qspiFlashSptShow(VOID)
    {
    STATUS rc;
    UINT32 offset;
    UINT32 readSize;
    SPT_LAYOUT* pSpt;

    rc = sfInit();
    if (rc != OK)
        {
        return ERROR;
        }

    readSize = sectorSize;
    for (offset = 0; offset < pFlashChip->uChipSize; offset += readSize)
        {
        rc = sfRead(offset, readSize);
        if (rc != OK)
            {
            return ERROR;
            }

        pSpt = (SPT_LAYOUT*) pRdBuf;
        if (pSpt->magic == SPT_MAGIC)
            {
            printf("Found SPT at offset 0x%08x\n", offset);
            break;
            }
        else
            {
            if (pSpt->magic != 0xffffffff)
                {
                printf("offset: 0x%08x magic: 0x%08x\n", offset, pSpt->magic);
                }
            }
        }

    sfShutdown();
    return OK;
    }


STATUS qspiFlashWrc(VOID)
    {
    STATUS rc = OK;
    UINT32* pBuf;
    UINT32 i;
    UINT32 offset = 0x0fff0000;

    rc = sfInit();
    if (rc != OK)
        {
        return ERROR;
        }

    pBuf = (UINT32*) pWrBuf;
    for (i = 0; (i < sectorSize / sizeof(UINT32)); i++)
        {
        pBuf[i] = offset;
        }

    printf("Erasing offset 0x%08x\n", offset);
    rc = sfEraseSector(offset);
    if (rc != OK)
        {
        return ERROR;
        }

    printf("Writing to offset 0x%08x\n", offset);
    rc = sfWrite(offset, sectorSize);
    if (rc != OK)
        {
        printf("sfWrite() failed\n");
        return ERROR;
        }

    printf("Reading from offset 0x%08x\n", offset);
    rc = sfRead(offset, sectorSize);
    if (rc != OK)
        {
        printf("sfRead() failed\n");
        return ERROR;
        }

    if (memcmp(pWrBuf, pRdBuf, sectorSize) != 0)
        {
        printf("Miscompare at offset 0x%08x\n", offset);
        return ERROR;
        }
    return OK;
    }

STATUS qspiFlashSectorsShow(VOID)
    {
    UINT32 offset;
    STATUS rc;
    BOOL isErased;

    rc = sfInit();
    if (rc != OK)
        {
        return ERROR;
        }

    for (offset = 0; offset < pFlashChip->uChipSize; offset += sectorSize)
        {
        rc = sfSectorIsErased(offset, &isErased);
        if (rc != OK)
            {
            return ERROR;
            }
        if (isErased)
            {
            printf("Sector at offset 0x%08x is erased\n", offset);
            }
        }
    return OK;
    }


LOCAL STATUS sfInit(VOID)
    {
    if (init)
        {
        return OK;
        }

    pFlashChip = (FLASH_CHIP*) vxbFlashChipFind("qspiflash",0);
    if (pFlashChip == NULL)
        {
        return ERROR;
        }
    printf("chip size = 0x%08llx, erase size = 0x%08x\n", pFlashChip->uChipSize, pFlashChip->uEraseSize);
    sectorSize = pFlashChip->uEraseSize;
    sectors = ((UINT32) pFlashChip->uChipSize) / sectorSize;
    pRdBuf = malloc(sectorSize);
    if (pRdBuf == NULL)
        {
        goto error;
        }
    pWrBuf = malloc(sectorSize);
    if (pWrBuf == NULL)
        {
        goto error;
        }
    init = TRUE;
    return OK;
error:
    sfShutdown();
    return ERROR;
    }

LOCAL VOID sfShutdown(VOID)
    {
    if (pRdBuf != NULL)
        {
        free(pRdBuf);
        }
    if (pWrBuf != NULL)
        {
        free(pWrBuf);
        }
    init = FALSE;
    }


LOCAL STATUS sfSectorIsErased
    (
    UINT32 offset,
    BOOL* isErased
    )
    {
    STATUS rc = OK;
    UINT32* pBuf;
    UINT32 i;

    if ((offset % sectorSize) != 0)
        {
        printf("Sector offset must be aligned to 0x%08x\n", sectorSize);
        return ERROR;
        }

    rc = sfRead(offset, sectorSize);
    if (rc != OK)
        {
        return ERROR;
        }

    pBuf = (UINT32*) pRdBuf;
    *isErased = TRUE;
    for (i = 0; i < (sectorSize / sizeof(UINT32)); i++)
        {
        if (pBuf[i] != 0xffffffff)
            {
            *isErased = FALSE;
            break;
            }
        }

    return OK;
    }

LOCAL STATUS sfRead
    (
    UINT32 offset,
    UINT32 count
    )
    {
    STATUS rc = OK;

    if (count == 0)
        {
        count = sectorSize;
        }

    if (count > sectorSize)
        {
        return ERROR;
        }

    memset(pRdBuf, 0, sectorSize);

    if (verbose)
        {
        printf("Reading %d bytes from offset 0x%08x\n", count, offset);
        }

    rc = (*pFlashChip->flashOPs.read)(pFlashChip, offset, 1, count, &pRdBuf, NULL);
    if (rc != OK)
        {
        printf("read() failed\n");
        return ERROR;
        }
    if (verbose)
        {
        printf("Read buffer at %p\n", pRdBuf);
        }
    return OK;
    }

LOCAL STATUS sfEraseSector
    (
    UINT32 offset
    )
    {
    STATUS rc = OK;

    if ((offset % sectorSize) != 0)
        {
        printf("Sector offset must be aligned to 0x%08x\n", sectorSize);
        return ERROR;
        }

    if (verbose)
        {
        printf("Erasing one sector at offset 0x%08x\n", offset);
        }
    rc = (*pFlashChip->flashOPs.blkErase)(pFlashChip, offset, 1);
    if (rc != OK)
        {
        printf("blkErase() failed\n");
        return ERROR;
        }

    return OK;
    }

LOCAL STATUS sfWrite
    (
    UINT32 offset,
    UINT32 count
    )
    {
    STATUS rc = OK;

    if (count == 0)
        {
        count = sectorSize;
        }

    if (count > sectorSize)
        {
        return ERROR;
        }

    if (verbose)
        {
        printf("Writing %d bytes to offset 0x%08x\n", count, offset);
        }

    rc = (*pFlashChip->flashOPs.write)(pFlashChip, offset, 1, count, &pWrBuf, NULL);
    if (rc != OK)
        {
        printf("read() failed\n");
        return ERROR;
        }
    return OK;
    }
