/* fsTest.c - file system test */

/* Copyright 1984-2011, 2024 Wind River Systems, Inc. */
#include "copyright_wrs.h"

#define FS_TEST_VERSION "02i"
 
/*
modification history
--------------------
02i,31jan24,tjf fixed errors/warnings seen in 23.09 LP64 (why now?!)
02h,13dec11,tjf added version to fsTestShow()
02g,13dec11,tjf fixed read rate bug in fsTestShow(), credit:Christian Didriksson
02f,08aug11,tjf added open/close overhead to data rates, cleaned up docs
02e,30nov10,tjf modified to include HRFS support - where permissions matter
02d,09dec06,tjf added read/write performance stats
02c,09dec06,tjf upgraded bytes added/removed stats to 64-bits
02b,19feb03,tjf tweaked to compile w/o warnings under MIPS T2.1.2
02a,25jan03,tjf added support to run multiple instances of fsTest()
01a,29jun01,tjf written by Ted Friedl, Wind River PS, Madison, WI
*/

/*
DESCRIPTION
This library provides the routines necessary to test a target
file system.

STARTING
From the shell, the user spawns (sp) fsTest() specifying a directory
on the file system to be tested, maximum file size to create and whether
to suppress verbose mode (boolean).  The maximum file size should always
be less than the bytes available on the file system -- roughly ten
percent (10%) is a good value if there is enough system memory.

For example, if (root) directory "/tffs" were empty and available,
and had 8 Mbytes of space available, fsTest() could be started as
follows:
.CS
    -> sp fsTest, "/tffs", 0x100000
.CE

fsTest() allocates twice the maximum file size of system memory
and then requires a small amount of memory to keep track of each
existing file.

OPERATION
fsTest() continuously loops and either creates a file with a
known pattern or randomly picks and removes a file.  After a
file creation or deletion, each existing file is read into
memory and checked.  All along, the create/remove/check
activity is displayed by the shell.  If an error occurs, the
test stops with proper notification.

fsTest() begins with a bias to create more files than it removes.
When the file system is full, the bias is reversed
(remove > create) and when the directory is void of files
created by fsTest(), the bias reverts back to its initial value
(create > remove).  Over the long run, this causes the file system
to be filled to near capacity, and then emptied in cycles.

STOPPING
A stop signal is sent to fsTest() by calling the routine
fsTestExit().  Before exiting, fsTest() checks all of the
existing files.  Upon exit, fsTest() makes no attempts to
remove those files.

STARTING AFTER STOPPING
fsTest() may be restarted after it is stopped without
compromising system memory.  It should be noted, however, that
the files created by the previous test run should be removed
before starting fsTest() again.

DEBUGGING ERRORS
fsTest() will exit for the following reasons:
.CS
    1. Invalid arguments
    2. Insufficient resources (system memory)
    3. Error opening an existing file
    4. Error reading an existing file
    5. Bad file data detected
    6. fsTestExit() was called
.CE

Cases 3, 4, and 5 indicate file system problems that probably
need to be investigated.  After an open or read error, the shell
display and the contents of the test directory should give some
insight as to what occured.  After bad file data is detected, the
global buffer pointers <pFsTestFillBuf> (known good data) and 
<pFsTestReadBuf> (read from the file) may be used to display an
image of the file in the area of the problem.

SHOWING STATISTICS
Important fsTest() statistics may be displayed while running or
after a test has stopped by calling fsTestShow(). fsTestShow()
displays the number of files and bytes added to the file system,
the number of files and bytes removed from the file system, and
the number of times the create/remove bias has changed.

Note that when fsTest() is stopped and started, the statistics of
the previous test may be cleared.

REMOVING FILES FROM A DIRECTORY
To avoid confusion, it is recommended that you start fsTest()
using an empty directory.  Note that when fsTest() is stopped
it usually leaves behind files that should be removed before
restarting.

Files created by fsTest() begin with the letter "f" and end with
a ".fst" extension.  If the directory under test is on the host,
fsTest-created files may be removed by using the "rm" (Unix) or
"delete" (DOS) commands with wildcards.  For example, on Unix:
.CS
    % rm /tmp/f*.fst
.CE

and from DOS:
.CS
    C:\> delete d:\tmp\f*.fst
.CE

TODO
This test was created in the era of 32-bit CPUs.  Module v02i was a
hurried upgrade made to satisfy the LP64 build environment.  This module
should be reviewed to ensure its conguence and quality operating in
32 and 64-bit environments.
*/

/* includes */

#include "vxWorks.h"
#include "stdioLib.h"
#include "stdlib.h"
#include "tickLib.h"
#include "string.h"
#include "sysLib.h"
#include "ioLib.h"
#include "fioLib.h"
#include "lstLib.h"
#include "usrLib.h"
#include "taskLib.h"
#include "usrFsLib.h"

/* defines */

#define FS_TEST_MAX_TESTS          8
#define FS_TEST_MAX_NAME_SIZE     16
#define FS_TEST_MAX_PATH_SIZE    128

/* typedefs */

typedef struct
    {
    NODE    node;
    char    pName[FS_TEST_MAX_NAME_SIZE];
    ssize_t size;
    int     pattern;
    } FS_TEST_FILE_DESC;

typedef struct
    {
    char *             pDirName;
    LIST *             pFileList;
    BOOL               shouldExit;
    unsigned           statFilesAdded;
    unsigned           statFilesRemoved;
    unsigned long long statBytesAdded;
    unsigned long long statBytesRemoved;
    unsigned long long statBytesRead;
    unsigned           statWriteTicks;
    unsigned           statReadTicks;
    unsigned           statBiasChanges;
    } FS_TEST_STAT_DESC;

/* globals (not in an structure array because they are easier to hand-parse!) */

TASK_ID fsTestTidArray[FS_TEST_MAX_TESTS];
char *  pFsTestFillBufArray[FS_TEST_MAX_TESTS];
char *  pFsTestReadBufArray[FS_TEST_MAX_TESTS];
ssize_t fsTestBufSizeArray[FS_TEST_MAX_TESTS];

/* locals */

LOCAL FS_TEST_STAT_DESC sd[FS_TEST_MAX_TESTS];
LOCAL int               numTests = 0;
LOCAL BOOL              globalVerbose = TRUE;

/* forward declarations */

/*******************************************************************************
*
* fsTestTidAdd - add task ID to fsTestTidArray[]
*
* This routine adds the task ID to fsTestTidArray[].
*
* RETURNS: Index of task, or ERROR if array is full.
*/

int fsTestTidAdd
    (
    TASK_ID tid
    )
    {
    int idx;

    for (idx = 0; idx < NELEMENTS(fsTestTidArray); idx++)
        {
        if (fsTestTidArray[idx] == 0)
            {
            fsTestTidArray[idx] = tid;
            numTests++;
            return (idx);
            }
        }

    return (ERROR);
    }

/*******************************************************************************
*
* fsTestTidDelete - delete task ID from fsTestTidArray[]
*
* This routine deletes the task ID from fsTestTidArray[].
*
* RETURNS: OK, or ERROR if task ID is not found.
*/

STATUS fsTestTidDelete
    (
    TASK_ID tid
    )
    {
    int idx;

    for (idx = 0; idx < NELEMENTS(fsTestTidArray); idx++)
        {
        if (fsTestTidArray[idx] == tid)
            {
            fsTestTidArray[idx] = 0;
            numTests--;
            return (OK);
            }
        }

    return (ERROR);
    }

/*******************************************************************************
*
* fsTestIndexGet - get array index
*
* This routine get the array index based on the task ID.
*
* RETURNS: Index of task, or ERROR if task is not found.
*/

int fsTestIndexGet
    (
    TASK_ID tid
    )
    {
    int idx;

    for (idx = 0; idx < NELEMENTS(fsTestTidArray); idx++)
        {
        if (fsTestTidArray[idx] == tid)
            return (idx);
        }

    return (ERROR);
    }

/*******************************************************************************
*
* fsTestCmp - compare one buffer to another
*
* This routine compares the first <nbytes> characters of <pBuf1>
* to <pBuf2> and returns the offset of the first miscompare.
*
* RETURNS: Offset of miscompare, or ERROR is buffers match.
*/

LOCAL int fsTestCmp
    (
    char *  pBuf1,    /* pointer to first buffer */
    char *  pBuf2,    /* pointer to second buffer */
    ssize_t nbytes    /* number of bytes to compare */
    )
    {
    int offset;

    for (offset = 0; offset < nbytes; offset++)
        {
        if (*pBuf1++ != *pBuf2++)
            return (offset);
        }

    return (ERROR);
    }

/*******************************************************************************
*
* fsTestBufFill - fill a buffer with a pattern
*
* This routine fills a buffer with an integer pattern specified
* by <pattern>.
*
* RETURNS: N/A
*/

LOCAL void fsTestBufFill
    (
    char *  pBuf,
    ssize_t nbytes,
    int     pattern
    ) 
    {
    char * pPattern = (char *)&pattern;

    while (nbytes >= sizeof (int))
        {
        *(int *)pBuf = pattern;

        pBuf += sizeof (int);
        nbytes -= sizeof (int);
        }

    while (nbytes > 0)
        {
        *pBuf = *pPattern;
        
        pBuf++;
        pPattern++;
        nbytes--;
        }
    }

/*******************************************************************************
*
* fsTestBufFree - free file buffers
*
* This routine manages the freeing of the file buffers.
*
* RETURNS: N/A
*/

LOCAL void fsTestBufFree
    (
    int idx
    )
    {
    if (pFsTestFillBufArray[idx] != (char *)NULL)
        free (pFsTestFillBufArray[idx]);
    if (pFsTestReadBufArray[idx] != (char *)NULL)
        free (pFsTestReadBufArray[idx]);

    pFsTestFillBufArray[idx] = (char *)NULL;
    pFsTestReadBufArray[idx] = (char *)NULL;
    fsTestBufSizeArray[idx] = 0;
    }

/*******************************************************************************
*
* fsTestDescFree - free file descriptors
*
* This routine manages the freeing the file descriptors.
*
* RETURNS: N/A
*/

LOCAL void fsTestDescFree
    (
    int idx
    )
    {
    FS_TEST_FILE_DESC * pDesc;

    pDesc = (FS_TEST_FILE_DESC *)lstFirst (sd[idx].pFileList);

    while (pDesc != (FS_TEST_FILE_DESC *)NULL)
        {
        lstDelete (sd[idx].pFileList, (NODE *)pDesc);

        free ((char *)pDesc);

        pDesc = (FS_TEST_FILE_DESC *)lstFirst (sd[idx].pFileList);
        }
    }

/*******************************************************************************
*
* fsTestShow - display statistics
*
* This routine displays the important statistics while running
* fsTest() or after fsTest() has stopped.
*
* RETURNS: OK.
*/

STATUS fsTestShow (void)
    {
    int     idx;
    TASK_ID tid;
    char    tidString[32];

    printf ("fsTest v%s statistics:\n", FS_TEST_VERSION);
   
    for (idx = 0; idx < NELEMENTS(fsTestTidArray); idx++)
        {
        if (sd[idx].pDirName == NULL)
            continue;

        if ((tid = fsTestTidArray[idx]) == 0)
            sprintf (tidString, "exited normally");
        else
            sprintf (tidString, "%p", (void *)tid);

        printf ("  Directory    : %s\n", sd[idx].pDirName);
        printf ("  Task ID      : %s\n", tidString);
        printf ("  Files added  : %d\n", sd[idx].statFilesAdded);
        printf ("  Files removed: %d\n", sd[idx].statFilesRemoved);
        printf ("  Bytes added  : %lld\n", sd[idx].statBytesAdded);
        printf ("  Bytes removed: %lld\n", sd[idx].statBytesRemoved);
        if (sd[idx].statWriteTicks != 0LL)
            printf ("  Write rate   : %lld bytes/sec\n",
                    (sd[idx].statBytesAdded * sysClkRateGet()) / 
                    sd[idx].statWriteTicks);
        if (sd[idx].statReadTicks != 0LL)
            printf ("  Read rate    : %lld bytes/sec\n",
                    (sd[idx].statBytesRead * sysClkRateGet()) / 
                    sd[idx].statReadTicks);
        printf ("  Bias changes : %d\n", sd[idx].statBiasChanges);
        printf ("\n");
        }

    return (OK);
    }

/*******************************************************************************
*
* fsTestExit - send stop signal to fsTest()
*
* This routine sends a stop signal to fsTest().
*
* RETURNS: OK.
*/

STATUS fsTestExit
    (
    TASK_ID exitTid
    )
    {
    int     idx;
    TASK_ID tid;

    for (idx = 0; idx < NELEMENTS(fsTestTidArray); idx++)
        {
        if ((tid = fsTestTidArray[idx]) == 0)
            continue;

        if ((exitTid == 0) || (exitTid == tid))
            {
            printf ("fsTestExit: signaling task 0x%x to stop\n", tid);
            sd[idx].shouldExit = TRUE;
            }
        }
    
    return (OK);
    }

/*******************************************************************************
*
* fsTestVerboseStop - stop verbose output
*
* This routine globally stops the verbosity of the tests that were
* started as verbose.
*
* RETURNS: OK.
*/

STATUS fsTestVerboseStop (void)
    {
    globalVerbose = FALSE;

    return (OK);
    }

/*******************************************************************************
*
* fsTestVerboseResume - resume verbose output
*
* This routine globally resumes the verbosity of the tests that were
* started as verbose.
*
* RETURNS: OK.
*/

STATUS fsTestVerboseResume (void)
    {
    globalVerbose = TRUE;

    return (OK);
    }

/*******************************************************************************
*
* fsTest - the file system test
*
* This routine tests a target file system by creating, deleting
* and verifying <maxFileSize>-sized files in the directory
* <pDirName>.
*
* Note: <maxFileSize> should alway be less than the bytes available
* on the file system -- roughly ten percent (10%) is a good value
* if there is enough system memory.
*
* RETURNS: OK, or ERROR if a failure is detected.
*/

STATUS fsTest
    (
    char *  pDirName,         /* clean directory on file system */
    ssize_t maxFileSize,      /* maximum file size in bytes */
    BOOL    supressVerbose    /* supress verbose mode */
    )
    {
    FS_TEST_FILE_DESC * pDesc;
    char                pFileName[FS_TEST_MAX_NAME_SIZE];
    char                pFullPath[FS_TEST_MAX_PATH_SIZE];
    BOOL                createBias = TRUE;
    BOOL                create;
    unsigned            count = 0;
    LIST                fileList;
    int                 fd;
    int                 nodeNum;
    ssize_t             size;
    int                 failedOffset;
    int                 idx;
    BOOL                verbose;
    ULONG               tickSave1;
    ULONG               tickSave2;
    
    if ((pDirName == (char *)NULL) || (maxFileSize <= 0) ||
        ((supressVerbose != TRUE) && (supressVerbose != FALSE)))
        {
        printf ("usage: sp fsTest, <dir1>, <max size>, <0|1>\n");
        return (ERROR);
        }

    idx = fsTestIndexGet (taskIdSelf());

    if (idx == ERROR)
        {
        idx = fsTestTidAdd (taskIdSelf());

        if (idx == ERROR)
            {
            printf ("fsTest: reached number of tests limit\n");
            return (ERROR);
            }
        }

    if (fsTestBufSizeArray[idx] != maxFileSize)
        {
        fsTestBufFree(idx);
    
        pFsTestFillBufArray[idx] = malloc (maxFileSize);
        pFsTestReadBufArray[idx] = malloc (maxFileSize);

        if ((pFsTestFillBufArray[idx] == (char *)NULL) ||
            (pFsTestReadBufArray[idx] == (char *)NULL))
            {
            printf ("fsTest: insufficient resources\n");
            fsTestBufFree(idx);
            fsTestTidDelete (taskIdSelf());
            return (ERROR);
            }

        fsTestBufSizeArray[idx] = maxFileSize;
        }

    verbose = !supressVerbose;
    
    sd[idx].pDirName = pDirName;

    bzero (pFsTestFillBufArray[idx], maxFileSize);
    bzero (pFsTestReadBufArray[idx], maxFileSize);

    sd[idx].pFileList = &fileList;
    lstInit (sd[idx].pFileList);

    sd[idx].statFilesAdded = 0;
    sd[idx].statFilesRemoved = 0;
    sd[idx].statBytesAdded = 0LL;
    sd[idx].statBytesRemoved = 0LL;
    sd[idx].statBytesRead = 0LL;
    sd[idx].statWriteTicks = 0;
    sd[idx].statReadTicks = 0;
    sd[idx].statBiasChanges = 0;

    srand (tickGet());

    sd[idx].shouldExit = FALSE;

    while (sd[idx].shouldExit == FALSE)
        {
        /* choose to create or remove a file */

        if (lstCount (sd[idx].pFileList) == 0)
            {
            create = TRUE;
            if (createBias == FALSE)
                {
                createBias = TRUE;
                sd[idx].statBiasChanges++;
                }
            }
        else if ((rand() % 3) == 0)
            create = !createBias;
        else
            create = createBias;

        /* perform create or remove action */

        if (create)
            {
            /* create the file name, size and contents */

            sprintf (pFileName, "f%x.fst", count);
            sprintf (pFullPath, "%s/%s", pDirName, pFileName);

            size = ((unsigned)rand() * (unsigned)rand()) % maxFileSize + 1;
 
            if (verbose && globalVerbose)
                printf ("fsTest: creating %s + %d bytes\n",
                        pFullPath, size);

            fsTestBufFill (pFsTestFillBufArray[idx], size, count);

            bzero (pFsTestFillBufArray[idx] + size, maxFileSize - size);

            /* write file */

            tickSave1 = tickGet();
            if ((fd = open (pFullPath, O_WRONLY | O_CREAT, 0777)) == ERROR)
                {
                if (verbose && globalVerbose)
                    printf ("fsTest: cannot create %s -- full?\n", pFullPath);
                if (createBias == TRUE)
                    {
                    createBias = FALSE;
                    sd[idx].statBiasChanges++;
                    }
                }
            else if (write (fd, pFsTestFillBufArray[idx], size) != size)
                {
                if (verbose && globalVerbose)
                    printf ("fsTest: cannot write %s -- full?\n", pFullPath);
                close (fd);
                rm (pFullPath);
                if (createBias == TRUE)
                    {
                    createBias = FALSE;
                    sd[idx].statBiasChanges++;
                    }
                }
            else
                {
                close (fd);
                tickSave2 = tickGet();

                /* create descriptor and add it to the list */

                pDesc =
                       (FS_TEST_FILE_DESC *)malloc (sizeof (FS_TEST_FILE_DESC));
                     
                if (pDesc == (FS_TEST_FILE_DESC *)NULL)
                    {
                    printf ("fsTest: insufficient resources\n");
                    fsTestBufFree(idx);
                    fsTestDescFree(idx);
                    fsTestTidDelete (taskIdSelf());
                    /* stop other tests so this problem may be diagnosed */
                    fsTestExit(0);
                    return (ERROR);
                    }

                strcpy (pDesc->pName, pFileName);
                pDesc->size = size;
                pDesc->pattern = count;

                lstAdd (sd[idx].pFileList, (NODE *)pDesc);

                sd[idx].statFilesAdded += 1;
                sd[idx].statBytesAdded += size;
                sd[idx].statWriteTicks += tickSave2 - tickSave1;
                }

            bzero (pFsTestFillBufArray[idx], size);
            }
        else
            {
            nodeNum = rand() % lstCount (sd[idx].pFileList) + 1;
            pDesc = (FS_TEST_FILE_DESC *)lstNth (sd[idx].pFileList, nodeNum);

            sprintf (pFullPath, "%s/%s", pDirName, pDesc->pName);

            if (verbose && globalVerbose)
                printf ("fsTest: removing %s - %d bytes\n",
                        pFullPath, pDesc->size);
 
            if (rm (pFullPath) != OK)
                {
                printf ("fsTest: ERROR removing %s\n", pFullPath);
                fsTestBufFree(idx);
                fsTestDescFree(idx);
                fsTestTidDelete (taskIdSelf());
                /* stop other tests so this problem may be diagnosed */
                fsTestExit(0);
                return (ERROR);
                }

            sd[idx].statFilesRemoved += 1;
            sd[idx].statBytesRemoved += pDesc->size;

            lstDelete (sd[idx].pFileList, (NODE *)pDesc);
            free ((char *)pDesc);
            }

        /* check all of the files */

        pDesc = (FS_TEST_FILE_DESC *)lstFirst (sd[idx].pFileList);

        while (pDesc != (FS_TEST_FILE_DESC *)NULL)
            {
            sprintf (pFullPath, "%s/%s", pDirName, pDesc->pName);

            if (verbose && globalVerbose)
                printf ("fsTest: checking %s\n", pFullPath);

            /* recreate what the file should look like */

            fsTestBufFill (pFsTestFillBufArray[idx], pDesc->size,
                           pDesc->pattern);

            /* read file into memory */

            tickSave1 = tickGet();
            fd = open (pFullPath, O_RDONLY, 0);

            if (fd == ERROR)
                {
                printf ("fsTest: ERROR opening %s\n", pFullPath);
                fsTestBufFree(idx);
                fsTestDescFree(idx);
                fsTestTidDelete (taskIdSelf());
                /* stop other tests so this problem may be diagnosed */
                fsTestExit(0);
                return (ERROR);
                }
            
            size = fioRead (fd, pFsTestReadBufArray[idx], maxFileSize);

            if (size != pDesc->size)
                {
                printf ("fsTest: ERROR reading %s\n", pFullPath);
                fsTestDescFree(idx);
                /* stop other tests so this problem may be diagnosed */
                fsTestExit(0);
                return (ERROR);
                }

            close (fd);
            tickSave2 = tickGet();

            sd[idx].statBytesRead += size;
            sd[idx].statReadTicks += tickSave2 - tickSave1;

            /* compare the two */

            if (bcmp (pFsTestFillBufArray[idx], pFsTestReadBufArray[idx],
                      size) != 0)
                {
                failedOffset = fsTestCmp (pFsTestFillBufArray[idx],
                                          pFsTestReadBufArray[idx], size);
                printf ("fsTest: ERROR bad data in %s at offset %d\n",
                        pFullPath, failedOffset);
                fsTestDescFree(idx);
                /* stop other tests so this problem may be diagnosed */
                fsTestExit(0);
                return (ERROR);
                }

            bzero (pFsTestFillBufArray[idx], size);
            bzero (pFsTestReadBufArray[idx], size);

            pDesc = (FS_TEST_FILE_DESC *)lstNext ((NODE *)pDesc);
            }

        if (verbose && globalVerbose && (numTests == 1))
            printf ("\n");
        count++;
        }

    fsTestBufFree(idx);
    fsTestDescFree(idx);
    fsTestTidDelete (taskIdSelf());

    printf ("fsTest(%s): exiting OK\n", pDirName);

    return (OK);
    }
